/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.benchmark;

import org.gillius.realdb.concept.ProgressReporter;
import org.gillius.realdb.concept.SimulationVehicleDataHandler;
import org.gillius.realdb.concept.VehicleDataSimulation;
import org.gillius.realdb.concept.data.DataHandlingException;
import org.gillius.realdb.concept.data.VehicleDataHandler;
import org.gillius.realdb.concept.data.VehicleDataParser;

import java.io.File;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.PrintStream;
import java.util.Arrays;
import java.util.Properties;
import java.util.regex.Pattern;

/**
 * The Benchmark class is the entry point for the RealDB benchmark program. A specific instance of
 * this application runs over only the "numRecords" and "sizeManagement" dimensions; leaving an
 * external process to iterate over the implementations dimension.
 *
 * @author Jason Winnebeck
 */
public class Benchmark {

	private static File inputFile;
	private static int  recordsLimit;
	private static int  recordsIncrement;

	private static ServerManager  server;
	private static BenchmarkStats stats;
	private static PrintStream    resultsTable;
	private static boolean        deleteOnEnd;

	private static BenchmarkManagerFactory bmf;

	private static File monitoredDir;
	private static PatternFileFilter monitoredPattern;

	public static void main( String[] args ) throws Exception {
		if ( args.length != 2 ) {
			System.out.println( "Two arguments required: <benchmark properties> <test properties>" );
			System.out.println( "Example: data/benchmark.properties data/myisam.properties" );
			System.out.println( "" );
			System.out.println( "Make sure you have run the proper DB instance setup." );
			System.exit( 1 );
		}

		Properties benchmarkProps = new Properties();
		benchmarkProps.load( new FileReader( args[0] ) );

		Properties testProps = new Properties();
		testProps.load( new FileReader( args[1] ) );

		String monitoredDisk = benchmarkProps.getProperty( "monitoredDisk" );
		inputFile = new File( benchmarkProps.getProperty( "inputFile" ) );
		recordsIncrement = Integer.parseInt( benchmarkProps.getProperty( "recordsIncrement" ) );
		if ( benchmarkProps.get( "recordsLimit" ) != null )
			recordsLimit = Integer.parseInt( benchmarkProps.getProperty( "recordsLimit" ) );
		else
			recordsLimit = Integer.MAX_VALUE;

		File resultsFile = new File( benchmarkProps.getProperty( "resultsFile" ) );
		boolean printHeader = !resultsFile.exists();
		resultsTable = new PrintStream( new FileOutputStream( resultsFile, true ) );
		if ( printHeader )
			resultsTable.println( BenchmarkResult.HEADER );

		server = new ServerManager( testProps.getProperty( "startDir" ),
		                            testProps.getProperty( "startCommand" ),
		                            testProps.getProperty( "endCommand" ) );

		monitoredDir     = new File( testProps.getProperty( "monitoredDir" ) ).getAbsoluteFile();
		monitoredPattern = getFilter( testProps.getProperty( "monitoredPattern" ) );
		stats = new BenchmarkStats( monitoredDisk, monitoredDir, monitoredPattern );


		bmf = new BenchmarkManagerFactory(
				Class.forName( testProps.getProperty( "benchmarkClass" ) ),
				testProps,
				Long.parseLong( benchmarkProps.getProperty( "maxAge" ) ) );

		deleteOnEnd = "true".equals( testProps.getProperty( "deleteOnEnd" ) );

		//sizeManagement dimension
		runBenchmarks( false );
		runBenchmarks( true );

		resultsTable.close();
	}

	/**
	 * Runs over the "numRecords" dimension
	 */
	private static void runBenchmarks( boolean sizeManagement ) throws Exception {
		int count;
		int maxRecords = 0;
		do {
			maxRecords += recordsIncrement;
			count = runSingleBenchmark( maxRecords, sizeManagement );
			if ( deleteOnEnd && monitoredDir != null && monitoredDir.exists() ) {
				File[] files = monitoredDir.listFiles( monitoredPattern );
				System.out.println( "Recursively deleting " + Arrays.toString( files ) );
				FileUtils.visitFiles( new FileDeleter(), files );
			}
		} while ( count >= maxRecords && maxRecords < recordsLimit );
	}

	private static int runSingleBenchmark( int maxRecords, boolean sizeManagement )
			throws Exception {
		BenchmarkManager db = bmf.create( sizeManagement );

		System.out.format( "Starting %s benchmark %s up to %d records%n",
		                   db.getImplementationName(),
		                   sizeManagement ? "(with size management)" : "",
		                   maxRecords );

		BenchmarkResult results = new BenchmarkResult();
		results.setSizeManagement( sizeManagement );
		results.setImplementation( db.getImplementationName() );
		stats.start();
		server.start();

		VehicleDataHandler writer = db.getHandler();

		stats.end( db.getUsedDatabaseSize() );
		results.setCreateTimeSec( stats.getDuration() );
		if ( stats.getDiskstats() != null ) {
			results.setCreateSectorReads( stats.getDiskstats().getSectorsRead() );
			results.setCreateSectorWrites( stats.getDiskstats().getSectorsWritten() );
		}

		System.out.println( "Starting and creation statistics" );
		stats.printResults( System.out );

		CountingVehicleDataHandler counter = new CountingVehicleDataHandler( writer );

		VehicleDataParser parser = VehicleDataSimulation.getParser( inputFile,
		                                                            new ProgressReporter( "bytes" ) );
		SimulationVehicleDataHandler sim = new SimulationVehicleDataHandler( counter );
		counter.setLimit( maxRecords );
		parser.setHandler( sim );

		stats.start();
		try {
			parser.parse();
		} catch ( DataHandlingException e ) {
			if ( !CountingVehicleDataHandler.EX_MESSAGE.equals( e.getMessage() ) )
				throw e;
		}

		db.finish();

		server.stop();

		stats.end( db.getUsedDatabaseSize() );

		results.setNumRecords( counter.getCount() );
		results.setLoadTimeSec( stats.getDuration() );
		results.setDatabaseSizeBytes( stats.getMonitoredFilesSize() );
		if ( stats.getDiskstats() != null ) {
			results.setLoadSectorReads( stats.getDiskstats().getSectorsRead() );
			results.setLoadSectorWrites( stats.getDiskstats().getSectorsWritten() );
			results.setLoadReadSec( stats.getDiskstats().getTotalReadMillis() / 1000.0 );
			results.setLoadWriteSec( stats.getDiskstats().getTotalWriteMillis() / 1000.0 );
		}

		if ( stats.getProcInfo() != null ) {
			results.setLoadUserModeTicks( stats.getProcInfo().getUserMode() +
			                              stats.getProcInfo().getChildrenUserMode() );
			results.setLoadKernelModeTicks( stats.getProcInfo().getKernelMode() +
			                                stats.getProcInfo().getChildrenKernelMode() );
		}

		resultsTable.println( results );

		System.out.println( "Number of events: " + counter.getCount() );
		stats.printResults( System.out );

		return counter.getCount();
	}

	private static PatternFileFilter getFilter( String pattern ) {
		if ( pattern == null )
			return null;
		else
			return new PatternFileFilter( Pattern.compile( pattern ) );
	}
}
