/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.benchmark;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileFilter;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintStream;

/**
 * BenchmarkStats provides a way to take monitor process (CPU) and system (disk) performance
 * metrics during a period on a Linux system, as well as clock time.
 *
 * @author Jason Winnebeck
 */
public class BenchmarkStats {
	private static final String PROC_STAT = "/proc/self/stat";

	private final String diskDevice;

	private long start;
	private long end;

	private LinuxProcInfo procInfo = null;
	private DiskstatsSample diskstats = null;

	private File monitoredDir;
	private FileFilter monitoredFileFilter;

	private long monitoredFilesSize;

	public BenchmarkStats( String diskDeviceName, File monitoredDir, FileFilter monitoredFileFilter ) {
		diskDevice = "/sys/block/" + diskDeviceName + "/stat";

		this.monitoredDir = monitoredDir;
		this.monitoredFileFilter = monitoredFileFilter;
	}

	public void start() throws IOException {
		start = System.currentTimeMillis();
		String raw = readFirstLine( PROC_STAT );
		if ( raw != null )
			procInfo = new LinuxProcInfo( raw );
		else
			System.err.println( "No CPU stats unless on Linux with proc mounted" );
		raw = readFirstLine( diskDevice );
		if ( raw != null )
			diskstats = new DiskstatsSample( raw );
		else
			System.err.println( "No disk stats unless on Linux with sysfs mounted; device: " + diskDevice );
	}

	public void end( Long manualSize ) throws IOException {
		end = System.currentTimeMillis();
		if ( procInfo != null )
			procInfo = procInfo.deltaTo( new LinuxProcInfo( readFirstLine( PROC_STAT ) ) );
		if ( diskstats != null )
			diskstats = diskstats.deltaTo( new DiskstatsSample( readFirstLine( diskDevice ) ) );

		if ( manualSize == null && monitoredDir != null && monitoredDir.exists() ) {
			DirectorySizer sizer = new DirectorySizer();
			FileUtils.visitFiles( sizer, monitoredDir.listFiles( monitoredFileFilter ) );
			monitoredFilesSize = sizer.getSize();
		} else {
			monitoredFilesSize = manualSize != null ? manualSize : -1;
		}
	}

	public void printResults( PrintStream out ) {
		out.println( "Duration: " + getDuration() + " seconds" );
		if ( procInfo != null )
			printResults( out, procInfo );
		if ( diskstats != null )
			printResults( out, diskDevice, diskstats );
		if ( monitoredFilesSize >= 0 )
			out.format( "Database size: %d bytes (%.4g MB)%n",
			            monitoredFilesSize, monitoredFilesSize / 1024.0 / 1024.0 );
	}

	public LinuxProcInfo getProcInfo() {
		return procInfo;
	}

	public DiskstatsSample getDiskstats() {
		return diskstats;
	}

	public double getDuration() {
		return ( end - start ) / 1000.0;
	}

	/**
	 * Total size of all monitored files, or -1 if there is no monitored directory or it does not
	 * exist.
	 */
	public long getMonitoredFilesSize() {
		return monitoredFilesSize;
	}

	/**
	 * Returns the first line of the file, or null if it does not exist or is empty.
	 */
	private static String readFirstLine( String path ) throws IOException {
		File f = new File( path );
		if ( !f.exists() ) {
			return null;
		} else {
			BufferedReader in = new BufferedReader( new FileReader( f ) );
			return in.readLine();
		}
	}

	private static void printResults( PrintStream out, LinuxProcInfo delta ) {
		out.println( "CPU statistics:" );

		long totalThis = delta.getUserMode() + delta.getKernelMode();
		long totalChildren = delta.getChildrenUserMode() + delta.getChildrenKernelMode();
		out.format( "  Total time this process: %d, children %d, total %d%n",
		                   totalThis, totalChildren, totalThis + totalChildren );
		out.format( "  Total user time: %d, kernel time: %d%n",
		                   delta.getUserMode() + delta.getChildrenUserMode(),
		                   delta.getKernelMode() + delta.getChildrenKernelMode() );
	}

	private static void printResults( PrintStream out, String device, DiskstatsSample delta ) {
		out.println( "Disk statistics for: " + device );

		out.format( "  Sectors read: %d, written: %d%n",
		                   delta.getSectorsRead(), delta.getSectorsWritten() );
		out.println( "  Details: " + delta );
	}
}
