/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.benchmark;

import java.util.Scanner;

/**
 * DiskstatsSample stores a single sample of disk statistics. This class is geared specifically for
 * the Linux kernels, and the information is based off of iostats.txt in the kernel source. On
 * kernels older than 2.6.25, there is a smaller set of 4 fields for partitions; this version of the
 * class only supports the later format which has all 11 fields, which match the disk devices.
 * <p/>
 * For a given device such as sda1, you can get the String you need by reading a single line from
 * /sys/block/sda/sda1/stat
 * <p/>
 * All subclasses must be Cloneable.
 * <p/>
 * iostats.txt was retrieved from
 * http://users.sosdg.org/~qiyong/lxr/source/Documentation/iostats.txt?v=2.6.30
 *
 * @author Jason Winnebeck
 */
public class DiskstatsSample implements Cloneable {
	private long readsCompleted; //1
	private long readsMerged; //2
	private long sectorsRead; //3
	private long totalReadMillis; //4
	private long writesCompleted; //5
	private long writesMerged; //6
	private long sectorsWritten; //7
	private long totalWriteMillis; //8
	private long currentIOs; //9
	private long totalIOMillis; //10
	private long weightedTotalIOMillis; //11

	public DiskstatsSample( String line ) {
		Scanner s = new Scanner( line );

		try {
			readsCompleted        = s.nextLong();
			readsMerged           = s.nextLong();
			sectorsRead           = s.nextLong();
			totalReadMillis       = s.nextLong();
			writesCompleted       = s.nextLong();
			writesMerged          = s.nextLong();
			sectorsWritten        = s.nextLong();
			totalWriteMillis      = s.nextLong();
			currentIOs            = s.nextLong();
			totalIOMillis         = s.nextLong();
			weightedTotalIOMillis = s.nextLong();

		} catch ( Exception e ) {
			throw new IllegalArgumentException( "Could not parse " + line + " as diskstats", e );
		}

		//In pre-2.6.25 there are only 4 fields, and after there are 11. But the first 4 of the 11
		//fields are not the same as the original 4 fields. Therefore there has been an example of the
		//number of fields determining the format; due to this I don't assume that if there are more
		//than 11 fields that the first 11 have the meaning they do now.
		if ( s.hasNextLong() ) {
			throw new IllegalArgumentException( line + " has extra, unknown fields; could signify a " +
			                                    "breaking change in format" );
		}
	}

	/**
	 * Returns a DiskstatsSample which holds the delta values of "other - this". This allows you to
	 * find out things like how many more reads were completed from a start to end of some sample
	 * period.
	 */
	public DiskstatsSample deltaTo( DiskstatsSample other ) {
		DiskstatsSample ret = other.doClone();

		ret.readsCompleted -= readsCompleted;
		ret.readsMerged -= readsMerged;
		ret.sectorsRead -= sectorsRead;
		ret.totalReadMillis -= totalReadMillis;
		ret.writesCompleted -= writesCompleted;
		ret.writesMerged -= writesMerged;
		ret.sectorsWritten -= sectorsWritten;
		ret.totalWriteMillis -= totalWriteMillis;
		ret.currentIOs -= currentIOs;
		ret.totalIOMillis -= totalIOMillis;
		ret.weightedTotalIOMillis -= weightedTotalIOMillis;

		return ret;
	}

	/**
	 * This is the total number of reads completed successfully.
	 */
	public long getReadsCompleted() {
		return readsCompleted;
	}

	/**
	 * Reads and writes which are adjacent to each other may be merged for efficiency.  Thus two 4K
	 * reads may become one 8K read before it is ultimately handed to the disk, and so it will be
	 * counted (and queued) as only one I/O.  This field lets you know how often this was done.
	 */
	public long getReadsMerged() {
		return readsMerged;
	}

	/**
	 * This is the total number of sectors read successfully.
	 */
	public long getSectorsRead() {
		return sectorsRead;
	}

	/**
	 * This is the total number of milliseconds spent by all reads (as measured from __make_request()
	 * to end_that_request_last()).
	 */
	public long getTotalReadMillis() {
		return totalReadMillis;
	}

	/**
	 * This is the total number of writes completed successfully.
	 */
	public long getWritesCompleted() {
		return writesCompleted;
	}

	/**
	 * Reads and writes which are adjacent to each other may be merged for efficiency.  Thus two 4K
	 * reads may become one 8K read before it is ultimately handed to the disk, and so it will be
	 * counted (and queued) as only one I/O.  This field lets you know how often this was done.
	 */
	public long getWritesMerged() {
		return writesMerged;
	}

	/**
	 * This is the total number of sectors written successfully.
	 */
	public long getSectorsWritten() {
		return sectorsWritten;
	}

	/**
	 * This is the total number of milliseconds spent by all writes (as measured from __make_request()
	 * to end_that_request_last()).
	 */
	public long getTotalWriteMillis() {
		return totalWriteMillis;
	}

	/**
	 * The only field that should go to zero. Incremented as requests are given to appropriate struct
	 * request_queue and decremented as they finish.
	 */
	public long getCurrentIOs() {
		return currentIOs;
	}

	/**
	 * This field is increases so long as field 9 is nonzero.
	 */
	public long getTotalIOMillis() {
		return totalIOMillis;
	}

	/**
	 * This field is incremented at each I/O start, I/O completion, I/O merge, or read of these stats
	 * by the number of I/Os in progress (field 9) times the number of milliseconds spent doing I/O
	 * since the last update of this field.  This can provide an easy measure of both I/O completion
	 * time and the backlog that may be accumulating.
	 */
	public long getWeightedTotalIOMillis() {
		return weightedTotalIOMillis;
	}

	@Override
	public boolean equals( Object o ) {
		if ( this == o ) return true;
		if ( o == null || getClass() != o.getClass() ) return false;

		DiskstatsSample that = (DiskstatsSample) o;

		if ( currentIOs != that.currentIOs ) return false;
		if ( readsCompleted != that.readsCompleted ) return false;
		if ( readsMerged != that.readsMerged ) return false;
		if ( sectorsRead != that.sectorsRead ) return false;
		if ( sectorsWritten != that.sectorsWritten ) return false;
		if ( totalIOMillis != that.totalIOMillis ) return false;
		if ( totalReadMillis != that.totalReadMillis ) return false;
		if ( totalWriteMillis != that.totalWriteMillis ) return false;
		if ( weightedTotalIOMillis != that.weightedTotalIOMillis ) return false;
		if ( writesCompleted != that.writesCompleted ) return false;
		if ( writesMerged != that.writesMerged ) return false;

		return true;
	}

	@Override
	public int hashCode() {
		int result = (int) ( readsCompleted ^ ( readsCompleted >>> 32 ) );
		result = 31 * result + (int) ( readsMerged ^ ( readsMerged >>> 32 ) );
		result = 31 * result + (int) ( sectorsRead ^ ( sectorsRead >>> 32 ) );
		result = 31 * result + (int) ( totalReadMillis ^ ( totalReadMillis >>> 32 ) );
		result = 31 * result + (int) ( writesCompleted ^ ( writesCompleted >>> 32 ) );
		result = 31 * result + (int) ( writesMerged ^ ( writesMerged >>> 32 ) );
		result = 31 * result + (int) ( sectorsWritten ^ ( sectorsWritten >>> 32 ) );
		result = 31 * result + (int) ( totalWriteMillis ^ ( totalWriteMillis >>> 32 ) );
		result = 31 * result + (int) ( currentIOs ^ ( currentIOs >>> 32 ) );
		result = 31 * result + (int) ( totalIOMillis ^ ( totalIOMillis >>> 32 ) );
		result = 31 * result + (int) ( weightedTotalIOMillis ^ ( weightedTotalIOMillis >>> 32 ) );
		return result;
	}

	@Override
	public String toString() {
		return "DiskstatsSample{" +
		       "readsCompleted=" + readsCompleted +
		       ", readsMerged=" + readsMerged +
		       ", sectorsRead=" + sectorsRead +
		       ", totalReadMillis=" + totalReadMillis +
		       ", writesCompleted=" + writesCompleted +
		       ", writesMerged=" + writesMerged +
		       ", sectorsWritten=" + sectorsWritten +
		       ", totalWriteMillis=" + totalWriteMillis +
		       ", currentIOs=" + currentIOs +
		       ", totalIOMillis=" + totalIOMillis +
		       ", weightedTotalIOMillis=" + weightedTotalIOMillis +
		       '}';
	}

	private DiskstatsSample doClone() {
		try {
			return (DiskstatsSample) clone();
		} catch ( CloneNotSupportedException e ) {
			throw new AssertionError( e );
		}
	}
}
