/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.benchmark;

import java.io.File;

/**
 * FileUtils
 *
 * @author Jason Winnebeck
 */
public class FileUtils {

	/**
	 * Visits each file root and recursively all files in directory roots. The files are visited in
	 * order of the "deepest" file first. That is, a/b/c is visited before a/b. This makes it possible
	 * for the visitor to implement recursive file deletions.
	 *
	 * @param handler non-null
	 * @param roots   list of 0 or more roots, can be files or directories
	 */
	public static void visitFiles( Visitor<File> handler, File... roots ) {
		for ( File root : roots ) {
			visitFiles( handler, root );
		}
	}

	/**
	 * Visits the root and all of its children. The files are visited in order of the "deepest" file
	 * first. That is, a/b/c is visited before a/b. This makes it possible for the visitor to implement
	 * recursive file deletions.
	 *
	 * @param handler non-null
	 * @param root    non-null
	 */
	public static void visitFiles( Visitor<File> handler, File root ) {
		if ( root.isDirectory() ) {
			visitFiles( handler, root.listFiles() );
		}
		handler.visit( root );
	}

	/**
	 * Deletes the given directory File recursively.
	 *
	 * @return true if the directory was deleted, false otherwise. When false is returned, as many
	 *         files that could have been deleted were deleted.
	 */
	public static boolean deleteRecursive( File dir ) {
		FileDeleter deleter = new FileDeleter();
		visitFiles( deleter, dir );
		return deleter.isAllDeleted();
	}
}
