/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.benchmark;

import org.gillius.realdb.concept.data.VehicleDataHandler;

import java.io.File;
import java.io.FileReader;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.SQLException;
import java.util.Properties;

import static java.util.Arrays.*;

/**
 * JDBCBenchmarkManager
 *
 * @author Jason Winnebeck
 */
public class JDBCBenchmarkManager implements BenchmarkManager {
	private String implName;
	private File createScript;
	private long maxAge;
	private ConnectionParameters params;

	private Connection conn;

	public JDBCBenchmarkManager() {
	}

	@Override
	public void init( Properties props ) {
		implName = props.getProperty( "implementationName", "JDBC" );
		createScript = new File( props.getProperty( "createScript" ) );
		if ( props.getProperty( "maxAge" ) != null )
			maxAge = Long.parseLong( props.getProperty( "maxAge" ) );
		else
			maxAge = -1;

		params = new ConnectionParameters();
		params.load( props );
	}

	@Override
	public String getImplementationName() {
		return implName;
	}

	@Override
	public VehicleDataHandler getHandler() throws Exception {
		conn = connect( params );
		SqlRunner.run( conn, new FileReader( createScript ) );
		conn.setCatalog( "realdb_benchmark" );

		JDBCVehicleDataHandler writer = new JDBCVehicleDataHandler( conn );
		SizeManagementStrategy sizeStrat;
		if ( maxAge < 0 ) {
			sizeStrat = new UnboundedSizeStrategy();
		} else {
			sizeStrat = new JDBCFixedTimeRangeStrategy( conn, asList( "FloatData", "BooleanData" ),
			                                            maxAge );
		}

		writer.setSizeManager( sizeStrat );

		return new BooleanRemoveDuplicatesVehicleDataHandler( writer );
	}

	@Override
	public Long getUsedDatabaseSize() {
		return null;
	}

	@Override
	public void finish() throws Exception {
		conn.close();
	}

	private static Connection connect( ConnectionParameters params )
			throws ClassNotFoundException, SQLException, InterruptedException {
		Class.forName( params.getClassName() );
		String jdbcUrl = "jdbc:" + params.getProtocol() + ":" + params.getUrl();
		Connection ret = null;
		while ( ret == null ) {
			try {
				ret = doConnect( params, jdbcUrl );
			} catch ( SQLException e ) {
				System.out.println( "Connect failed; trying again in 0.5 seconds: " + e.getMessage() );
				Thread.sleep( 500 );
			}
		}
		return ret;
	}

	private static Connection doConnect( ConnectionParameters params, String jdbcUrl )
			throws SQLException {
		return DriverManager.getConnection( jdbcUrl, params.getUser(), params.getPassword() );
	}
}
