/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.benchmark;

import org.gillius.realdb.concept.data.DataHandlingException;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.Collection;

/**
 * JDBCFixedTimeRangeStrategy deletes all records older than "maxAge" based on the most recently
 * inserted record. The delete occurs in all given tables. This class assumes that no records have
 * a negative timestamp. The size management is applied unconditionally on call to the manage
 * method.
 *
 * @author Jason Winnebeck
 */
public class JDBCFixedTimeRangeStrategy implements SizeManagementStrategy {
	private final long maxAge;

	private final PreparedStatement[] deleteStmts;

	private long maxTime = 0;

	public JDBCFixedTimeRangeStrategy( Connection conn, Collection<String> tables, long maxAge )
			throws SQLException {
		this.maxAge = maxAge;

		Statement stmt = conn.createStatement();
		try {
			for ( String table : tables ) {
				ResultSet rs = stmt.executeQuery( "SELECT MAX(time) FROM " + table );
				rs.next();
				maxTime = Math.max( maxTime, rs.getLong( 1 ) );
				rs.close();
			}
		} finally {
			stmt.close();
		}

		deleteStmts = new PreparedStatement[ tables.size() ];
		int idx = 0;
		for ( String table : tables ) {
			deleteStmts[idx] = conn.prepareStatement( "DELETE FROM " + table + " WHERE time < ?" );
			++idx;
		}
	}

	@Override
	public void onNewRecord( long time ) {
		if ( time > maxTime )
			maxTime = time;
	}

	@Override
	public void manage() throws DataHandlingException {
		long time = maxTime - maxAge;

		try {
			for ( PreparedStatement deleteStmt : deleteStmts ) {
				deleteStmt.setLong( 1, time );
				deleteStmt.executeUpdate();
			}
		} catch ( SQLException e ) {
			throw new DataHandlingException( e );
		}
	}
}
