/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.benchmark;

import org.gillius.realdb.concept.data.DataHandlingException;
import org.gillius.realdb.concept.data.VehicleDataHandler;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.SQLException;

/**
 * JDBCVehicleDataHandler is a {@link VehicleDataHandler} that writes JDBC batched prepared
 * statements.
 *
 * @author Jason Winnebeck
 */
public class JDBCVehicleDataHandler implements VehicleDataHandler {
	private final PreparedStatement floatData;
	private final PreparedStatement boolData;

	private SizeManagementStrategy sizeManager = null;

	private int floatBatched = 0;
	private int boolBatched = 0;

	public JDBCVehicleDataHandler( Connection conn ) throws SQLException {
		floatData = conn.prepareStatement(
				"INSERT INTO FloatData (streamId, time, value, discontinuity) VALUES (?, ?, ?, ? )" );
		boolData = conn.prepareStatement(
				"INSERT INTO BooleanData (streamId, time, value, discontinuity) VALUES (?, ?, ?, ? )" );
	}

	/**
	 * Sets a size manager, which will be called when the batch insert statements are executed.
	 *
	 * @param sizeManager nullable SizeManagementStrategy
	 */
	public void setSizeManager( SizeManagementStrategy sizeManager ) {
		this.sizeManager = sizeManager;
	}

	@Override
	public void startStream() throws DataHandlingException {
		floatBatched = 0;
		boolBatched = 0;
	}

	@Override
	public void dataRead( int stream, long time, float value, boolean discontinuity )
			throws DataHandlingException {
		try {
			floatData.setInt( 1, stream );
			floatData.setLong( 2, time );
			floatData.setFloat( 3, value );
			floatData.setBoolean( 4, discontinuity );
			floatData.addBatch();
			if ( sizeManager != null )
				sizeManager.onNewRecord( time );

			++floatBatched;
			if ( floatBatched == 1000 ) {
				writeFloat();
			}

		} catch ( SQLException e ) {
			throw new DataHandlingException( e );
		}
	}

	private void writeFloat() throws SQLException, DataHandlingException {
		floatData.executeBatch();
		floatBatched = 0;
		if ( sizeManager != null )
			sizeManager.manage();
	}

	@Override
	public void dataRead( int stream, long time, boolean value, boolean discontinuity )
			throws DataHandlingException {
		try {
			boolData.setInt( 1, stream );
			boolData.setLong( 2, time );
			boolData.setBoolean( 3, value );
			boolData.setBoolean( 4, discontinuity );
			boolData.addBatch();
			if ( sizeManager != null )
				sizeManager.onNewRecord( time );

			++boolBatched;
			if ( boolBatched == 1000 ) {
				writeBool();
			}

		} catch ( SQLException e ) {
			throw new DataHandlingException( e );
		}
	}

	private void writeBool() throws SQLException, DataHandlingException {
		boolData.executeBatch();
		boolBatched = 0;
		if ( sizeManager != null )
			sizeManager.manage();
	}

	@Override
	public void endStream() throws DataHandlingException {
		try {
			if ( floatBatched > 0 ) {
				writeFloat();
			}
			if ( boolBatched > 0 ) {
				writeBool();
			}
		} catch ( SQLException e ) {
			throw new DataHandlingException( e );
		}
	}
}
