/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.benchmark;

/**
 * LinuxProcInfo holds information read about a process in Linux. You can find out information about
 * arbitrary processes through /proc/[pid]/stat, or the current process by /proc/self/stat.
 * <p/>
 * The implementation of this class is based off of the proc(5) man page as of Ubuntu karmic.
 *
 * @author Jason Winnebeck
 */
public class LinuxProcInfo implements Cloneable {
	private final String raw;

	private final int processId; //1 (first field)

	private long userMode; //14
	private long kernelMode; //15

	private long childrenUserMode; //16
	private long childrenKernelMode; //17

	/**
	 * Construct a new LinuxProcInfo with the raw line from /proc/[pid]/stat. Only some of the values
	 * are parsed.
	 */
	public LinuxProcInfo( String raw ) {
		this.raw = raw;

		String[] parsed = raw.split( " " );
		processId = Integer.parseInt( parsed[0] );

		userMode = Long.parseLong( parsed[13] );
		kernelMode = Long.parseLong( parsed[14] );

		childrenUserMode = Long.parseLong( parsed[15] );
		childrenKernelMode = Long.parseLong( parsed[16] );
	}

	/**
	 * Returns the raw, unparsed proc stat line.
	 */
	public String getRaw() {
		return raw;
	}

	/**
	 * The process ID.
	 */
	public int getProcessId() {
		return processId;
	}

	/**
	 * Amount of time that this process has been scheduled in user mode,  measured  in  clock  ticks
	 * (divide by sysconf(_SC_CLK_TCK).  This includes guest time, guest_time (time spent running a
	 * virtual CPU,  see  below), so  that applications that are not aware of the guest time field do
	 * not lose that time from their calculations.
	 */
	public long getUserMode() {
		return userMode;
	}

	/**
	 * Amount of time that this process has been  scheduled  in  kernel  mode, measured in clock ticks
	 * (divide by sysconf(_SC_CLK_TCK).
	 */
	public long getKernelMode() {
		return kernelMode;
	}

	/**
	 * Amount of time that this process's waited-for children have been scheduled   in   user   mode,
	 *   measured   in   clock   ticks   (divide   by sysconf(_SC_CLK_TCK).   (See also times(2).)
	 * This includes guest time, cguest_time (time spent running a virtual CPU, see below).
	 */
	public long getChildrenUserMode() {
		return childrenUserMode;
	}

	/**
	 * Amount of time that this process's waited-for children have been scheduled   in   kernel
	 * mode,   measured   in   clock   ticks  (divide  by sysconf(_SC_CLK_TCK).
	 */
	public long getChildrenKernelMode() {
		return childrenKernelMode;
	}

	public LinuxProcInfo deltaTo( LinuxProcInfo other ) {
		if ( processId != other.processId )
			throw new IllegalArgumentException( "Different process ID" );

		LinuxProcInfo ret = other.doClone();

		ret.userMode -= userMode;
		ret.kernelMode -= kernelMode;
		ret.childrenUserMode -= childrenUserMode;
		ret.childrenKernelMode -= childrenKernelMode;

		return ret;
	}

	@Override
	public boolean equals( Object o ) {
		if ( this == o ) return true;
		if ( o == null || getClass() != o.getClass() ) return false;

		LinuxProcInfo that = (LinuxProcInfo) o;

		if ( childrenKernelMode != that.childrenKernelMode ) return false;
		if ( childrenUserMode != that.childrenUserMode ) return false;
		if ( kernelMode != that.kernelMode ) return false;
		if ( processId != that.processId ) return false;
		if ( userMode != that.userMode ) return false;
		if ( !raw.equals( that.raw ) ) return false;

		return true;
	}

	@Override
	public int hashCode() {
		int result = raw.hashCode();
		result = 31 * result + processId;
		result = 31 * result + (int) ( userMode ^ ( userMode >>> 32 ) );
		result = 31 * result + (int) ( kernelMode ^ ( kernelMode >>> 32 ) );
		result = 31 * result + (int) ( childrenUserMode ^ ( childrenUserMode >>> 32 ) );
		result = 31 * result + (int) ( childrenKernelMode ^ ( childrenKernelMode >>> 32 ) );
		return result;
	}

	@Override
	public String toString() {
		return "LinuxProcInfo{" +
		       "raw='" + raw + '\'' +
		       ", processId=" + processId +
		       ", userMode=" + userMode +
		       ", kernelMode=" + kernelMode +
		       ", childrenUserMode=" + childrenUserMode +
		       ", childrenKernelMode=" + childrenKernelMode +
		       '}';
	}

	private LinuxProcInfo doClone() {
		try {
			return (LinuxProcInfo) clone();
		} catch ( CloneNotSupportedException e ) {
			throw new AssertionError( e );
		}
	}
}
