/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.benchmark;

import org.gillius.realdb.concept.RecordSinkWritingHandler;
import org.gillius.realdb.concept.data.VehicleDataHandler;
import org.gillius.realdb.io.BlockFile;
import org.gillius.realdb.io.rdl.RDLFile;
import org.gillius.realdb.model.Database;
import org.gillius.realdb.tools.DatabaseBuilder;

import java.io.File;
import java.util.Properties;

/**
 * RealDBBenchmarkManager
 *
 * @author Jason Winnebeck
 */
public class RealDBBenchmarkManager implements BenchmarkManager {
	private static final String template =
			"SET blockSize     = 4096\n" +
			"SET fileSize      = %d\n" +
			"SET maxStreams    = 8\n" +
			"SET dataBlockSize = 4\n" +
			"\n" +
			"CREATE STREAM RPM WITH ID 190 {\n" +
			"  value float\n" +
			"}\n" +
			"CREATE STREAM FuelRate WITH ID 183 {\n" +
			"  value float\n" +
			"}\n" +
			"CREATE STREAM AccelPedalPosition WITH ID 91 {\n" +
			"  value float\n" +
			"}\n" +
			"CREATE STREAM BatteryVolts WITH ID 1682 {\n" +
			"  value float\n" +
			"}\n" +
			"CREATE STREAM FuelEcon WITH ID 184 {\n" +
			"  value float\n" +
			"}\n" +
			"CREATE STREAM Speed WITH ID 841 {\n" +
			"  value float\n" +
			"}\n" +
			"CREATE STREAM VehicleStatus WITH ID 1 {\n" +
			"  collecting boolean WITH CODEC StepAlgorithm\n" +
			"}\n" +
			"CREATE STREAM HardAccelEvent WITH ID 2 {\n" +
			"  active boolean WITH CODEC StepAlgorithm\n" +
			"}";

	private String dataFile;
	private boolean isDeviceFile;
	private long fileSize;

	private Database db;

	public RealDBBenchmarkManager() {
	}

	@Override
	public void init( Properties props ) {
		dataFile = props.getProperty( "dataFile" );
		isDeviceFile = "true".equals( props.getProperty( "isDeviceFile" ) );
		String sizeProp = "true".equals( props.getProperty( "sizeManagement" ) ) ?
		                  "managedSize" : "unmanagedSize";
		fileSize = Long.parseLong( props.getProperty( sizeProp ) );
	}

	@Override
	public String getImplementationName() {
		return isDeviceFile ? "RealDB Raw" : "RealDB File";
	}

	@Override
	public VehicleDataHandler getHandler() throws Exception {
		RDLFile rdl = RDLFile.load( String.format( template, fileSize ) );
		DatabaseBuilder builder = new DatabaseBuilder( rdl );
		if ( isDeviceFile ) {
			BlockFile bf = builder.createBlockFile( new File( dataFile ) );
			db = builder.createDatabase( bf );
		} else {
			db = builder.createDatabase( new File( dataFile ) );
		}
		return new RecordSinkWritingHandler( db );
	}

	@Override
	public Long getUsedDatabaseSize() {
		int freeBlocks = db.getTotalDataBlocks() - db.getDataBlocksUsed();
		long freeBytes = freeBlocks * db.getMetadata().getDataBlockSize() * db.getBlockSize();
		long totalBytes = db.getSizeInBlocks() * db.getBlockSize();
		return totalBytes - freeBytes;
	}

	@Override
	public void finish() throws Exception {
		db.close();
	}
}
