/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.benchmark;

import org.gillius.realdb.concept.IOUtils;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Arrays;

/**
 * ServerManager manages the starting and stopping of a server daemon process. The daemon's stdout
 * and stderr are redirected to those of this process.
 *
 * @author Jason Winnebeck
 */
public class ServerManager {
	private final String startDir;
	private final String startCommand;
	private final String endCommand;

	private Process daemonProc;

	public ServerManager( String startDir, String startCommand, String endCommand ) {
		this.startDir = startDir;
		this.startCommand = startCommand;
		this.endCommand = endCommand;
	}

	/**
	 * Starts the external process, if configured.
	 */
	public void start() throws IOException {
		if ( startCommand != null ) {
			daemonProc = runCommand( startCommand );
		}
	}

	/**
	 * Runs the end process command, if any, and waits for the external process created in {@link
	 * #start()}, if any.
	 */
	public void stop() throws IOException, InterruptedException {
		if ( endCommand != null ) {
			runCommand( endCommand ).waitFor();
		}

		if ( daemonProc != null ) {
			System.out.println( "Waiting for external process to end" );
			daemonProc.waitFor();
		}
	}

	private Process runCommand( String command ) throws IOException {
		String[] args = command.split( " " );
		File commandPath = new File( startDir, args[0] );
		args[0] = commandPath.getAbsolutePath();

		System.out.println( "Running " + Arrays.toString( args ) );
		ProcessBuilder proc = new ProcessBuilder( args );
		if ( startDir != null )
			proc.directory( new File( startDir ) );
		Process externalProc = proc.start();

		new Thread( new Copier( externalProc.getInputStream(), System.out ), "ext stdout" ).start();
		new Thread( new Copier( externalProc.getErrorStream(), System.err ), "ext stderr" ).start();

		return externalProc;
	}

	private static class Copier implements Runnable {
		private final InputStream in;
		private final OutputStream out;

		private Copier( InputStream in, OutputStream out ) {
			this.in = in;
			this.out = out;
		}

		@Override
		public void run() {
			try {
				IOUtils.copy( in, out );
			} catch ( IOException e ) {
				//ignore
			}
		}
	}

}
