/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.tools.gui;

import org.gillius.realdb.metadata.ElementType;
import org.gillius.realdb.metadata.RecordElement;
import org.gillius.realdb.model.algorithms.FlatElementInterval;
import org.gillius.realdb.model.data.DefaultStreamInterval;
import org.gillius.realdb.model.data.DiscontinuityInterval;
import org.gillius.realdb.model.data.ElementInterval;
import org.gillius.realdb.model.data.StreamInterval;
import org.jfree.chart.ChartFactory;
import org.jfree.chart.ChartPanel;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.data.xy.XYSeries;
import org.jfree.data.xy.XYSeriesCollection;

import javax.swing.*;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.Vector;

import static java.util.Arrays.*;

/**
 * ChartRecordsView displays the records in a StreamIntervalsModel in a line chart.
 * <p/>
 * TODO: fix bug around discontinuities when forceLinear is true
 *
 * @author Jason Winnebeck
 */
public class ChartRecordsView {
	private JPanel panel;
	private JComboBox elementsCombo;
	private JPanel chartPanel;

	private StreamIntervalsModel model;

	private boolean forceLinear = false;

	private ChangeListener listener = SwingProxy.create( ChangeListener.class, new ChangeListener() {
		@Override
		public void stateChanged( ChangeEvent e ) {
			updateChart();
		}
	} );

	public ChartRecordsView() {
		elementsCombo.addActionListener( new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent e ) {
				updateChart();
			}
		} );
	}

	public JPanel getPanel() {
		return panel;
	}

	public StreamIntervalsModel getModel() {
		return model;
	}

	public void setModel( StreamIntervalsModel model ) {
		if ( this.model != null ) {
			this.model.removeChangeListener( listener );
		}

		this.model = model;
		if ( model != null ) {
			model.addChangeListener( listener );
			RecordElement[] modelElements = model.getElements();
			Vector<RecordElementView> elements = new Vector<RecordElementView>( modelElements.length );
			for ( RecordElement element : modelElements ) {
				if ( Number.class.isAssignableFrom( element.getType().getTypeClass() ) )
					elements.add( new RecordElementView( element ) );
			}
			elementsCombo.setModel( new DefaultComboBoxModel( elements ) );
			updateChart();

		} else {
			elementsCombo.setModel( new DefaultComboBoxModel( new Object[0] ) );
		}
	}

	public boolean getForceLinear() {
		return forceLinear;
	}

	public void setForceLinear( boolean forceLinear ) {
		this.forceLinear = forceLinear;
		updateChart();
	}

	public static void main( String[] args ) {
		JFrame frame = new JFrame( "ChartRecordsView" );
		ChartRecordsView view = new ChartRecordsView();
		frame.setContentPane( view.panel );
		frame.setDefaultCloseOperation( JFrame.EXIT_ON_CLOSE );

		StreamIntervalsModel model = new StreamIntervalsModelImpl(
				new RecordElement[] {
						new RecordElement( 0, "value", null, ElementType.Float, false ),
						new RecordElement( 1, "other", null, ElementType.SInt16, false )
				},
				asList( interval( 100, 200, 1.0f, (short) 23 ),
				        interval( 200, 300, 3.0f, (short) 52 ),
				        new DiscontinuityInterval( 300, 400 ),
				        interval( 400, 500, 6.0f, (short) 60 ) )
		);
		view.setModel( model );
		view.setForceLinear( true );

		frame.pack();
		frame.setLocationByPlatform( true );
		frame.setVisible( true );
	}

	private static StreamInterval interval( long start, long end, float value, short other ) {
		return new DefaultStreamInterval( start, end,
		                                  new ElementInterval[] {
				                                  new FlatElementInterval( value, end - start ),
				                                  new FlatElementInterval( other, end - start )
		                                  } );
	}

	private void updateChart() {
		chartPanel.removeAll();

		if ( model == null || model.getRowCount() <= 0 ) {
			chartPanel.add( new JLabel( "No Data" ), BorderLayout.CENTER );

		} else if ( elementsCombo.getSelectedIndex() < 0 ) {
			chartPanel.add( new JLabel( "No Selection" ), BorderLayout.CENTER );

		} else {
			RecordElement element = ( (RecordElementView) elementsCombo.getSelectedItem() ).e;
			int idx = element.getPosition();
			XYSeries xy = new XYSeries( element.getName() );

			for ( int i = 0; i < model.getRowCount(); ++i ) {
				StreamInterval inter = model.getRow( i );
				long start = inter.getStartTime();
				if ( inter.isDiscontinuity() ) {
					xy.add( start, null );
				} else {
					xy.add( start, (Number) inter.getValueAt( start, idx ) );
					long end = inter.getEndTime();
					if ( !forceLinear && end > start + 1 ) {
						xy.add( end - 1, (Number) inter.getValueAt( end - 1, idx ) );
					}
				}
			}

			JFreeChart chart = ChartFactory.createXYLineChart(
					elementsCombo.getSelectedItem().toString(),
					"Time",
					null,
					new XYSeriesCollection( xy ),
					PlotOrientation.VERTICAL,
					false,
					false,
					false
			);
			chartPanel.add( new ChartPanel( chart ), BorderLayout.CENTER );
		}
		chartPanel.validate();
	}

	{
// GUI initializer generated by IntelliJ IDEA GUI Designer
// >>> IMPORTANT!! <<<
// DO NOT EDIT OR ADD ANY CODE HERE!
		$$$setupUI$$$();
	}

	/**
	 * Method generated by IntelliJ IDEA GUI Designer >>> IMPORTANT!! <<< DO NOT edit this method OR
	 * call it in your code!
	 *
	 * @noinspection ALL
	 */
	private void $$$setupUI$$$() {
		panel = new JPanel();
		panel.setLayout( new BorderLayout( 0, 0 ) );
		final JPanel panel1 = new JPanel();
		panel1.setLayout( new FlowLayout( FlowLayout.RIGHT, 5, 5 ) );
		panel.add( panel1, BorderLayout.NORTH );
		final JLabel label1 = new JLabel();
		label1.setText( "Element to graph:" );
		panel1.add( label1 );
		elementsCombo = new JComboBox();
		panel1.add( elementsCombo );
		chartPanel = new JPanel();
		chartPanel.setLayout( new BorderLayout( 0, 0 ) );
		panel.add( chartPanel, BorderLayout.CENTER );
		chartPanel.setBorder(
				BorderFactory.createTitledBorder( BorderFactory.createEmptyBorder( 5, 5, 5, 5 ), null ) );
	}

	/**
	 * @noinspection ALL
	 */
	public JComponent $$$getRootComponent$$$() {
		return panel;
	}

	private static class RecordElementView {
		public final RecordElement e;

		private RecordElementView( RecordElement e ) {
			this.e = e;
		}

		public String toString() {
			return e.getName() + " : " + e.getType();
		}
	}
}
