/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.tools.gui;

import org.gillius.realdb.BadFileException;
import org.gillius.realdb.model.DataStream;
import org.gillius.realdb.model.data.Record;
import org.gillius.realdb.model.data.StreamInterval;
import org.gillius.realdb.util.DBIterator;

import javax.swing.*;
import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;

/**
 * DataStreamDataViewer
 * <p/>
 * TODO: make a model to back this class; pull DB work out of event thread.
 *
 * @author Jason Winnebeck
 */
public class DataStreamDataViewer {
	private JPanel panel;
	private JFormattedTextField startTimeField;
	private JFormattedTextField limitField;
	private JLabel streamNameLabel;
	private JButton nextButton;
	private JTable table;
	private JButton graphButton;

	private RecordsTableModel records;

	private DataStream stream;

	public DataStreamDataViewer( DataStream stream ) {
		this.stream = stream;

		$$$setupUI$$$();
		streamNameLabel.setText( stream.getInfo().getName() );
		startTimeField.setValue( 0L );
		limitField.setValue( 1000 );

		records = new RecordsTableModel( stream.getInfo().getElements(),
		                                 Collections.<Record>emptyList() );
		table.setModel( records );

		nextButton.addActionListener( new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent e ) {
				if ( records.getRowCount() > 0 ) {
					Record r = records.getRow( records.getRowCount() - 1 );
					startTimeField.setValue( r.getTime() );
					load();
				}
			}
		} );

		PropertyChangeListener listener = new PropertyChangeListener() {
			@Override
			public void propertyChange( PropertyChangeEvent evt ) {
				load();
			}
		};

		startTimeField.addPropertyChangeListener( "value", listener );
		limitField.addPropertyChangeListener( "value", listener );

		load();
		graphButton.addActionListener( new ActionListener() {
			@Override
			public void actionPerformed( ActionEvent e ) {
				graph();
			}
		} );
	}

	private void graph() {
		long startTime = ( (Number) startTimeField.getValue() ).longValue();
		int limit = ( (Number) limitField.getValue() ).intValue();

		try {
			DBIterator<StreamInterval> iter = stream.readIntervals( startTime, Long.MAX_VALUE );

			ChartRecordsView view = new ChartRecordsView();
			view.setModel( new StreamIntervalsModelImpl( stream.getInfo().getElements(),
			                                             iterToList( iter, limit ) ) );

			JFrame frame = new JFrame( "Graphing " + stream.getInfo().getName() );
			frame.setDefaultCloseOperation( JFrame.DISPOSE_ON_CLOSE );
			frame.setContentPane( view.getPanel() );
			frame.pack();
			frame.setLocationByPlatform( true );
			frame.setVisible( true );

		} catch ( Exception e1 ) {
			e1.printStackTrace();
		}
	}

	public void makeAndShowFrame() {
		JFrame frame = new JFrame( "Viewing " + stream.getInfo().getName() );
		frame.setDefaultCloseOperation( JFrame.DISPOSE_ON_CLOSE );
		frame.setContentPane( panel );
		frame.pack();
		frame.setLocationByPlatform( true );
		frame.setVisible( true );
	}

	private void load() {
		long startTime = ( (Number) startTimeField.getValue() ).longValue();
		int limit = ( (Number) limitField.getValue() ).intValue();

		try {
			DBIterator<Record> iter = stream.readRecords( startTime, Long.MAX_VALUE );
			records.setRows( iterToList( iter, limit ) );

		} catch ( Exception e ) {
			e.printStackTrace();
			records.setRows( Collections.<Record>emptyList() );
		}
	}

	private static <T> ArrayList<T> iterToList( DBIterator<T> iter, int limit )
			throws IOException, BadFileException {
		ArrayList<T> data = new ArrayList<T>( limit );

		while ( iter.hasNext() && data.size() < limit ) {
			data.add( iter.next() );
		}
		return data;
	}

	/**
	 * Method generated by IntelliJ IDEA GUI Designer >>> IMPORTANT!! <<< DO NOT edit this method OR
	 * call it in your code!
	 *
	 * @noinspection ALL
	 */
	private void $$$setupUI$$$() {
		panel = new JPanel();
		panel.setLayout( new GridBagLayout() );
		streamNameLabel = new JLabel();
		streamNameLabel.setText( "Stream Name" );
		GridBagConstraints gbc;
		gbc = new GridBagConstraints();
		gbc.gridx = 0;
		gbc.gridy = 0;
		gbc.gridwidth = 4;
		gbc.anchor = GridBagConstraints.WEST;
		gbc.insets = new Insets( 8, 8, 8, 8 );
		panel.add( streamNameLabel, gbc );
		final JLabel label1 = new JLabel();
		label1.setText( "Starting Time" );
		gbc = new GridBagConstraints();
		gbc.gridx = 0;
		gbc.gridy = 1;
		gbc.anchor = GridBagConstraints.WEST;
		gbc.insets = new Insets( 8, 8, 8, 8 );
		panel.add( label1, gbc );
		startTimeField = new JFormattedTextField();
		gbc = new GridBagConstraints();
		gbc.gridx = 1;
		gbc.gridy = 1;
		gbc.weightx = 0.5;
		gbc.anchor = GridBagConstraints.WEST;
		gbc.fill = GridBagConstraints.HORIZONTAL;
		gbc.insets = new Insets( 8, 8, 8, 8 );
		panel.add( startTimeField, gbc );
		final JLabel label2 = new JLabel();
		label2.setText( "Limit" );
		gbc = new GridBagConstraints();
		gbc.gridx = 2;
		gbc.gridy = 1;
		gbc.anchor = GridBagConstraints.WEST;
		gbc.insets = new Insets( 8, 8, 8, 8 );
		panel.add( label2, gbc );
		limitField = new JFormattedTextField();
		gbc = new GridBagConstraints();
		gbc.gridx = 3;
		gbc.gridy = 1;
		gbc.weightx = 0.5;
		gbc.anchor = GridBagConstraints.WEST;
		gbc.fill = GridBagConstraints.HORIZONTAL;
		gbc.insets = new Insets( 8, 8, 8, 8 );
		panel.add( limitField, gbc );
		final JPanel panel1 = new JPanel();
		panel1.setLayout( new FlowLayout( FlowLayout.RIGHT, 8, 8 ) );
		gbc = new GridBagConstraints();
		gbc.gridx = 0;
		gbc.gridy = 3;
		gbc.gridwidth = 4;
		gbc.fill = GridBagConstraints.BOTH;
		panel.add( panel1, gbc );
		graphButton = new JButton();
		graphButton.setText( "Graph" );
		graphButton.setMnemonic( 'G' );
		graphButton.setDisplayedMnemonicIndex( 0 );
		panel1.add( graphButton );
		nextButton = new JButton();
		nextButton.setText( "Next" );
		nextButton.setMnemonic( 'N' );
		nextButton.setDisplayedMnemonicIndex( 0 );
		panel1.add( nextButton );
		final JScrollPane scrollPane1 = new JScrollPane();
		gbc = new GridBagConstraints();
		gbc.gridx = 0;
		gbc.gridy = 2;
		gbc.gridwidth = 4;
		gbc.weighty = 1.0;
		gbc.fill = GridBagConstraints.BOTH;
		gbc.insets = new Insets( 8, 8, 8, 8 );
		panel.add( scrollPane1, gbc );
		table = new JTable();
		scrollPane1.setViewportView( table );
	}

	/**
	 * @noinspection ALL
	 */
	public JComponent $$$getRootComponent$$$() {
		return panel;
	}
}
