/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.tools.gui;

import java.lang.reflect.InvocationHandler;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.util.Arrays;
import java.util.concurrent.Executor;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

/**
 * The SchedulingProxy schedules all calls on an interface to an Executor. Since the call could be
 * asynchronous, this can only proxy methods that return void. Any exceptions are output to stderr;
 * it is suggested that this proxy only proxy methods that do not throw to keep things clean.
 *
 * @author Jason Winnebeck
 */
public class SchedulingProxy implements InvocationHandler {
	private final Object target;
	private final Executor exec;

	public SchedulingProxy( Object target, Executor exec ) {
		this.target = target;
		this.exec = exec;
	}

	public Object invoke( Object proxy, final Method method, final Object[] args ) throws Throwable {
		if ( target == null )
			throw new NullPointerException();

		if ( method.getReturnType() != Void.TYPE ) {
			throw new UnsupportedOperationException(
					"SchedulingProxy can only proxy methods returning void; the method called was " +
					method + " with arguments " + Arrays.toString( args ) );
		}

		exec.execute( new Runnable() {
			public void run() {
				try {
					method.invoke( target, args );
				} catch ( IllegalAccessException e ) {
					e.printStackTrace();
				} catch ( InvocationTargetException e ) {
					e.getCause().printStackTrace();
				}
			}
		} );
		return null;
	}

	/**
	 * Creates a SchedulingProxy to proxy the given interface and delegate all calls to the target
	 * through a single-threaded executor. That executor's thread thread will time out after 1 second.
	 *
	 * @param clazz  interface to implement
	 * @param target target implementing that interface whose methods will be called in the Executor.
	 */
	public static <T> T create( Class<T> clazz, T target ) {
		ThreadPoolExecutor exec = new ThreadPoolExecutor(
				1, 1, 1000L, TimeUnit.MILLISECONDS, new LinkedBlockingQueue<Runnable>() );
		exec.allowCoreThreadTimeOut( true );

		return create( clazz, target, exec );
	}

	/**
	 * Creates a SchedulingProxy to proxy the given interface and delegate all calls to the target
	 * through an executor.
	 *
	 * @param clazz  interface to implement
	 * @param target target implementing that interface whose methods will be called in the Executor.
	 * @param exec   executor to use
	 */
	@SuppressWarnings( "unchecked" )
	public static <T> T create( Class<T> clazz, T target, Executor exec ) {
		SchedulingProxy handler = new SchedulingProxy( target, exec );
		return (T) Proxy.newProxyInstance( target.getClass().getClassLoader(),
		                                   new Class[] { clazz }, handler );
	}
}
