/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.tools.gui;

import javax.swing.*;
import java.beans.PropertyChangeListener;
import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.util.Arrays;

/**
 * The SwingProxy makes sure that all calls to an interface happen on the Swing event thread. If
 * the current thread is not the swing thread, it schedules the call to happen later, else it runs
 * the method immediately. Any exceptions thrown by the proxied method are simply printed to
 * standard error.
 *
 * @author Jason Winnebeck
 */
public class SwingProxy implements InvocationHandler {
	private final Object target;

	public SwingProxy( Object target ) {
		this.target = target;
	}

	public Object invoke( Object proxy, final Method method, final Object[] args ) throws Throwable {
		if ( target == null )
			throw new NullPointerException();

		if ( method.getReturnType() != Void.TYPE ) {
			if ( method.getName().equals( "equals" ) ) {
				return proxy == args[0];
			} else if ( method.getName().equals( "toString" ) ) {
				return "Proxy for " + target;
			}

			throw new UnsupportedOperationException(
					"SwingProxy can only proxy methods returning void; the method called was " +
					method + " with arguments " + Arrays.toString( args ) );
		}

		Runnable event = new Runnable() {
			public void run() {
				try {
					method.invoke( target, args );
				} catch ( Exception e ) {
					e.printStackTrace();
				}
			}
		};

		if ( SwingUtilities.isEventDispatchThread() ) {
			event.run();
		} else {
			SwingUtilities.invokeLater( event );
		}

		return null;
	}

	/**
	 * Creates a SwingProxy to proxy the given interface.
	 *
	 * @param clazz  interface to implement
	 * @param target target implementing that interface.
	 */
	public static <T> T create( Class<T> clazz, T target ) {
		SwingProxy handler = new SwingProxy( target );
		return (T) Proxy.newProxyInstance( target.getClass().getClassLoader(),
		                                   new Class[] { clazz }, handler );
	}

	/**
	 * Creates a SwingProxy to proxy PropertyChangeListener methods.
	 *
	 * @param target target implementing PropertyChangeListener
	 */
	public static PropertyChangeListener create( PropertyChangeListener target ) {
		return create( PropertyChangeListener.class, target );
	}
}
