/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.tools.cli;

import org.gillius.realdb.metadata.RecordElement;
import org.gillius.realdb.model.DataStream;
import org.gillius.realdb.model.Database;
import org.gillius.realdb.model.data.Discontinuity;
import org.gillius.realdb.model.data.GenericRecord;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.LineNumberReader;
import java.util.List;

import static java.util.Arrays.*;
import static java.util.Collections.*;

/**
 * BulkLoadCommand
 * TODO: document and make a lot more robust
 *
 * @author Jason Winnebeck
 */
public class BulkLoadCommand extends AbstractCommand {
	private static final List<Parameter> params = unmodifiableList( asList(
			new Parameter( "database", File.class, true, "Path to database" ),
			new Parameter( "stream", Integer.class, true, "Stream ID" ),
			new Parameter( "datafile", File.class, true, "File containing the tab-separated values" )
	) );

	public String getName() {
		return "bulkLoad";
	}

	public String getHelpText() {
		return "Bulk loads data into a stream from a tab-separated values file. The first row is " +
		       "the header row of element names. The first column must be named 'time' and " +
		       "contain the timestamp. The rest of the columns must be named after elements in the " +
		       "stream. Not all elements must be in the file, but missing elements must be nullable " +
		       "and will be set to null in each record. Rows representing discontinuities must have " +
		       "all elements with a value of '--'. Values may also be 'null' if allowed for that " +
		       "element. Blank lines and lines starting with # are allowed and are ignored.";
	}

	public List<Parameter> getParameters() {
		return params;
	}

	@Override
	protected void executeValidated( ExecutionContext context, Object[] params )
			throws CommandException {
		File    database = (File   ) params[0];
		Integer streamId = (Integer) params[1];
		File    dataFile = (File   ) params[2];

		try {
			Database db = Database.load( database );

			DataStream stream = db.getStreamsMap().get( streamId );
			if ( stream == null ) {
				throw new CommandException( "No such stream " + streamId );
			}
			RecordElement[] elements = stream.getInfo().getElements();

			LineNumberReader reader = new LineNumberReader( new FileReader( dataFile ) );

			String line = readLine( reader );
			String[] header = getHeader( line );
			int[] fileToElementMap = getColumnMapping( stream, header );

			line = readLine( reader );
			int numWritten = 0;
			while ( line != null ) {
				String[] cols = line.split( "\t" );
				if ( cols.length != header.length )
					throw new CommandException( "Invalid number of columns on line " + reader.getLineNumber() );

				long time = Converter.convert( Long.class, cols[0] );

				if ( cols[1].equals( "--" ) ) {
					for ( int i = 2; i < cols.length; ++i ) {
						if ( !cols[i].equals( "--" ) )
							throw new CommandException( "All columns must be '--' for a discontinuity on " +
							                            "line " + reader.getLineNumber() );
					}
					stream.writeRecord( new Discontinuity( time ) );

				} else {
					Object[] vals = new Object[header.length - 1];
					for ( int i = 1; i < cols.length; i++ ) {
						final int ePos = fileToElementMap[i - 1];
						vals[ePos] = Converter.convert( elements[ePos].getType().getTypeClass(), cols[i] );
					}

					stream.writeRecord( new GenericRecord( time, vals ) );
				}
				++numWritten;
				line = readLine( reader );
			}

			db.close();
			context.getOutput().println( "Wrote " + numWritten + " records into stream " + stream );

		} catch ( Exception e ) {
			throw new CommandException( "Error loading data from " + params[2] + " into " + params[0], e );
		}
	}

	/**
	 * Reads a line from the BufferedReader, skipping empty lines and comment lines.
	 */
	private static String readLine( BufferedReader reader ) throws IOException {
		String line = reader.readLine();
		while ( line != null && ( line.equals( "" ) || line.startsWith( "#" ) ) )
			line = reader.readLine();
		return line;
	}

	private static String[] getHeader( String line ) throws CommandException {
		if ( line == null ) {
			throw new CommandException( "Input file must at least have a header row" );
		}
		String[] header = line.split( "\t" );
		if ( header.length < 2 ) {
			throw new CommandException( "Input file must have at least 2 columns, but has " +
			                            header.length );
		}
		if ( !"time".equals( header[0] ) ) {
			throw new CommandException( "The first column of the input file must be 'time', but is " +
			                            header[0] );
		}
		return header;
	}

	private static int[] getColumnMapping( DataStream stream, String[] header )
			throws CommandException {
		int[] fileToElementMap = new int[ header.length - 1 ];
		for ( int i = 1; i < header.length; i++ ) {
			final RecordElement element = stream.getInfo().getElement( header[i] );
			if ( element == null ) {
				throw new CommandException( "Unknown element " + header[i] + " in stream " + stream );
			}
			fileToElementMap[i-1] = element.getPosition();
		}
		return fileToElementMap;
	}
}
