/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.tools.cli;

import org.gillius.Util.StringPrintWriter;

import java.util.List;

/**
 * CommandUtil has some utility methods to work with {@link Command}s.
 *
 * @author Jason Winnebeck
 */
public class CommandUtil {
	private CommandUtil() {}

	/**
	 * Validates that the parameters to be passed to {@link Command#execute} are valid, and if not
	 * throws a CommandException with a printable and user-friendly message text on the issues.
	 *
	 * @param c      non-null Command
	 * @param params non-null parameters to be passed
	 *
	 * @throws CommandException If the parameters to be passed are not valid in any way.
	 */
	public static void validateParameters( Command c, Object[] params ) throws CommandException {
		List<Parameter> cmdParams = c.getParameters();

		if ( cmdParams.size() != params.length ) {
			throw new CommandException( c + " requires " + cmdParams.size() + " parameters, but " +
			                            params.length + " were given." );
		}

		StringPrintWriter out = new StringPrintWriter();
		boolean errors = false;

		for ( int i = 0; i < params.length; i++ ) {
			Parameter cmdParam = cmdParams.get( i );
			if ( params[i] == null && cmdParam.isRequired() ) {
				out.println( "Parameter " + cmdParam + " is required but was not given." );
				errors = true;

			} else if ( params[i] != null && !cmdParam.getType().isAssignableFrom( params[i].getClass() ) ) {
				out.println( "A " + params[i].getClass().getSimpleName() + " was given to " + cmdParam +
				             ", but a " + cmdParam.getType().getSimpleName() + " is required." );
				errors = true;
			}
		}

		if ( errors )
			throw new CommandException( out.toString() );
	}

	/**
	 * Returns human-readable summary information on a Command in the returned array:
	 * <ol><li>Command Name</li>
	 *     <li>Brief description</li>
	 *     <li>Parameter list</li></ol>
	 */
	public static String[] getCommandSummary( Command c ) {
		StringBuilder ret = new StringBuilder();

		for ( Parameter param : c.getParameters() ) {
			ret.append( param.isRequired() ? " " : " [" );
			ret.append( param.getName() );
			ret.append( " : " );
			ret.append( param.getType().getSimpleName() );
			ret.append( param.isRequired() ? "" : "]" );
		}

		return new String[] { c.getName(), getBrief( c.getHelpText() ), ret.toString() };
	}

	public static String getFullHelpText( Command c ) {
		StringPrintWriter out = new StringPrintWriter();

		List<Parameter> params = c.getParameters();
		out.println( c.getName() + ": " + c.getHelpText() );
		out.println( c.getName() + " takes " + params.size() + " parameters:" );
		for ( Parameter param : params ) {
			out.println( "  " + param.getName() + " (" +
			             ( param.isRequired() ? "required " : "optional " ) +
			             param.getType().getSimpleName() + "): " + param.getHelpText() );
		}
		return out.toString();
	}

	private static String getBrief( String full ) {
		int idx = full.indexOf( '.' );
		if ( idx > 0 ) {
			return full.substring( 0, idx+1 );
		} else {
			return full;
		}
	}
}
