/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.tools.cli;

import org.gillius.realdb.metadata.DataStreamInfo;
import org.gillius.realdb.metadata.RecordElement;
import org.gillius.realdb.model.DataStream;
import org.gillius.realdb.model.Database;

import java.io.File;
import java.io.PrintStream;
import java.util.Collection;
import java.util.List;

import static java.util.Arrays.*;
import static java.util.Collections.*;

/**
 * DescribeCommand describes elements within a RealDB database.
 *
 * @author Jason Winnebeck
 */
public class DescribeCommand extends AbstractCommand {
	private static final List<Parameter> params = unmodifiableList( asList(
			new Parameter( "Database", File.class, true, "Path to database" )
	) );

	public String getName() {
		return "describe";
	}

	public String getHelpText() {
		return "Shows information about a database.";
	}

	public List<Parameter> getParameters() {
		return params;
	}

	@Override
	protected void executeValidated( ExecutionContext context, Object[] params )
			throws CommandException {
		PrintStream out = context.getOutput();

		File database = (File) params[0];

		try {
			Database db = Database.load( database );
			Collection<DataStream> streams = db.getStreams();

			out.println( "Database " + database.getCanonicalPath() + ":" );
			out.println( "  Block size     : " + db.getBlockSize() );
			out.println( "  Database size  : " + db.getSizeInBlocks() + " blocks (" +
		               ( db.getBlockSize() * db.getSizeInBlocks() ) + " bytes)" );
			out.println( "  Data block size: " + db.getMetadata().getDataBlockSize() );
			out.println( "Database has " + streams.size() + " streams:" );
			for ( DataStream stream : streams ) {
				DataStreamInfo info = stream.getInfo();
				RecordElement[] elements = info.getElements();
				out.println( info.getName() + " (" + info.getUserId() + "), with " + elements.length +
				             " elements:" );
				for ( RecordElement element : elements ) {
					out.println( "  " + element.getPosition() + " " + element.getName() + " (" +
					             element.getType() + ( element.isOptional() ? ", optional" : "" ) +
					             ") with codec " + element.getCodec().getClass().getSimpleName() );
				}
				out.println( "Stream contains records from times " + stream.getStartTime() + " to " +
				             stream.getEndTime() );
			}
			db.close();

		} catch ( Exception e ) {
			throw new CommandException( e );
		}
	}
}
