/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.tools.cli;

import java.io.PrintStream;
import java.util.List;
import java.util.SortedMap;

import static java.util.Arrays.*;
import static java.util.Collections.*;

/**
 * HelpCommand prints out a list of all commands or detailed help on a single command.
 *
 * @author Jason Winnebeck
 */
public class HelpCommand extends AbstractCommand {
	private static final List<Parameter> params = unmodifiableList( asList(
			new Parameter( "Command Name", String.class, false,
			               "Name of a command to get detailed help" ) ) );

	private final SortedMap<String, Command> commands;

	public HelpCommand( SortedMap<String, Command> commands ) {
		this.commands = commands;
	}

	public String getName() {
		return "help";
	}

	public String getHelpText() {
		return "Prints out a list of all commands or detailed help on a single command";
	}

	public List<Parameter> getParameters() {
		return params;
	}

	public void executeValidated( ExecutionContext context, Object[] params ) throws CommandException {
		PrintStream out = context.getOutput();

		if ( params[0] != null ) {
			//noinspection SuspiciousMethodCalls
			Command cmd = commands.get( params[0] );
			if ( cmd != null ) {
				out.println( CommandUtil.getFullHelpText( cmd ) );
			} else {
				context.getErrorOutput().println( "Unknown command " + params[0] );
			}

		} else {
			out.println( "List of commands: " );
			for ( Command command : commands.values() ) {
				out.format( "%-15s %s%n", (Object[]) CommandUtil.getCommandSummary( command ) );
			}
		}
	}
}
