/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.tools.cli;

import org.gillius.realdb.metadata.RecordElement;
import org.gillius.realdb.model.DataStream;
import org.gillius.realdb.model.Database;
import org.gillius.realdb.model.data.Record;
import org.gillius.realdb.util.DBIterator;

import java.io.File;
import java.io.PrintStream;
import java.util.Arrays;
import java.util.List;

import static java.util.Arrays.*;
import static java.util.Collections.*;

/**
 * ReadCommand reads all raw records or a range of raw records from a single stream.
 *
 * @author Jason Winnebeck
 */
public class ReadCommand extends AbstractCommand {
	private static final List<Parameter> params = unmodifiableList( asList(
	    new Parameter( "Database", File.class,    true,  "Path to database"               ),
	    new Parameter( "Stream",   Integer.class, true,  "Stream ID"                      ),
	    new Parameter( "Start",    Long.class,    false, "Starting timestamp (inclusive)" ),
	    new Parameter( "End",      Long.class,    false, "Ending timestamp (inclusive)"   )
	) );

	public String getName() {
		return "read";
	}

	public String getHelpText() {
		return "Reads all raw records or a range of raw records from a single stream and outputs a " +
		       "tab-delimited output with the first row as the column headers.";
	}

	public List<Parameter> getParameters() {
		return params;
	}

	@Override
	protected void executeValidated( ExecutionContext context, Object[] params )
			throws CommandException {
		PrintStream out = context.getOutput();
		File    database  = (File   ) params[0];
		Integer streamId  = (Integer) params[1];
		Long    startTime = (Long   ) params[2];
		Long    endTime   = (Long   ) params[3];

		if ( startTime == null )
			startTime = 0L;
		if ( endTime == null )
			endTime = Long.MAX_VALUE;

		try {
			Database db = Database.load( database );

			DataStream stream = db.getStreamsMap().get( streamId );
			if ( stream == null ) {
				throw new CommandException( "No such stream " + streamId );
			}
			RecordElement[] elements = stream.getInfo().getElements();

			String format = "%10s %10s";
			int col = 0;
			Object[] cols = new Object[ 2 + elements.length ];
			cols[ col++ ] = "stream";
			cols[ col++ ] = "time";
			for ( RecordElement element : elements ) {
				cols[ col++ ] = element.getName();
				format += " %10s";
			}
			format += "%n";
			out.format( format, cols );

			cols[0] = streamId;

			int numRecords = 0;
			DBIterator<Record> iter = stream.readRecords( startTime, endTime );
			while ( iter.hasNext() ) {
				Record r = iter.next();
				++numRecords;
				col = 1;
				if ( r.isDiscontinuity() ) {
					Arrays.fill( cols, 1, cols.length, "--" );
				} else {
					cols[ col++ ] = r.getTime();
					for ( RecordElement element : elements ) {
						cols[ col++ ] = r.getElement( element.getPosition() );
					}
				}
				out.format( format, cols );
			}
			out.println();
			out.println( "Read " + numRecords + " records." );

			db.close();

		} catch ( Exception e ) {
			throw new CommandException( "Error loading data from " + params[2] + " into " + params[0], e );
		}

	}
}
