/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.tools.cli;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.SortedMap;
import java.util.TreeMap;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * RealDBTool provides a command line interface to maintaining RealDB databases. Run the main method
 * to get help on valid commands.
 *
 * @author Jason Winnebeck
 */
public class RealDBTool {
	private static final Pattern argsRegex = Pattern.compile( "[^\\s\"']+|\"([^\"]*)\"|'([^']*)'" );

	public static void main( String[] args ) throws IOException {
		RealDBTool tool = new RealDBTool();
		ExecutionContext context = new DefaultExecutionContext( System.out, System.err );

		boolean anyErrors = false;

		if ( args.length > 0 ) {
			//Immediate execution
			anyErrors = tool.executeCommand( context, args );

		} else {
			//Interactive mode
			System.out.println( "RealDB Tool Interactive mode" );
			System.out.println( "This version is extremely experimental and the commands will change " +
			                    "in the next release." );
			System.out.println( "You can also invoke this program with arguments to run a single " +
			                    "command non-interactively." );
			System.out.println( "Use 'help' command to get a list of commands, 'exit' to end" );
			BufferedReader input = new BufferedReader( new InputStreamReader( System.in ) );

			String line = getNextCommand( input );
			while ( line != null ) {
				List<String> params = parseArgs( line );
				if ( params.isEmpty() ) {
					//ignore this line
					line = getNextCommand( input );

				} else if ( params.get( 0 ).equalsIgnoreCase( "exit" ) ) {
					line = null;

				} else {
					anyErrors |= tool.executeCommand( context, params.toArray( new String[ params.size() ] ) );
					line = getNextCommand( input );
				}
			}
		}

		System.exit( anyErrors ? 1 : 0 );
	}

	private static String getNextCommand( BufferedReader input ) throws IOException {
		System.out.print( "> " );
		System.out.flush();
		return input.readLine();
	}

	/**
	 * Parse the given line into arguments, which can contain quoted strings containing spaces.
	 * This code was found at
	 * http://stackoverflow.com/questions/366202/regex-for-splitting-a-string-using-space-when-not-surrounded-by-single-or-double/366532#366532
	 */
	private static List<String> parseArgs( String line ) {
		List<String> matchList = new ArrayList<String>();
		Matcher regexMatcher = argsRegex.matcher( line );
		while ( regexMatcher.find() ) {
			if ( regexMatcher.group( 1 ) != null ) {
				// Add double-quoted string without the quotes
				matchList.add( regexMatcher.group( 1 ) );
			} else if ( regexMatcher.group( 2 ) != null ) {
				// Add single-quoted string without the quotes
				matchList.add( regexMatcher.group( 2 ) );
			} else {
				// Add unquoted word
				matchList.add( regexMatcher.group() );
			}
		}

		return matchList;
	}

	private SortedMap<String, Command> commands = new TreeMap<String, Command>();

	public RealDBTool() {
		addCommand( new CreateDatabaseCommand() );
		addCommand( new BulkLoadCommand() );
		addCommand( new ReadCommand() );
		addCommand( new IntervalsCommand() );
		addCommand( new DescribeCommand() );
		addCommand( new SummaryCommand() );

		addCommand( new HelpCommand( Collections.unmodifiableSortedMap( commands ) ) );
	}

	public boolean executeCommand( ExecutionContext context, String... args ) {
		PrintStream err = context.getErrorOutput();
		Command c = commands.get( args[0] );

		if ( c == null ) {
			err.println( "Unknown command " + args[0] );
			return true;
		}

		Object[] parsedArgs;

		try {
			int numParams = args.length - 1;
			List<Parameter> cParams = c.getParameters();

			if ( numParams > cParams.size() ) {
				throw new CommandException( c.getName() + " requires " + cParams.size() + " parameters, but " +
				                            numParams + " were given." );
			}
			parsedArgs = new Object[ cParams.size() ];
			for ( int i = 0; i < parsedArgs.length && i < numParams; i++ ) {
				parsedArgs[i] = Converter.convert( cParams.get( i ).getType(), args[i+1] );
			}

			CommandUtil.validateParameters( c, parsedArgs );

		} catch ( CommandException e ) {
			err.println( "Invalid parameters for " + c.getName() );
			err.println( e.getMessage() );
			err.println();
			err.println( CommandUtil.getFullHelpText( c ) );
			return true;
		}

		try {
			c.execute( context, parsedArgs );
		} catch ( CommandException e ) {
			err.println( "Error while running command " + c.getName() + ":" );
			e.printStackTrace( err );
		}

		return false;
	}

	private void addCommand( Command c ) {
		commands.put( c.getName(), c );
	}
}
