/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.tools.cli;

import org.gillius.realdb.metadata.RecordElement;
import org.gillius.realdb.model.DataStream;
import org.gillius.realdb.model.Database;
import org.gillius.realdb.model.data.StreamInterval;
import org.gillius.realdb.util.DBIterator;

import java.io.File;
import java.io.PrintStream;
import java.util.Arrays;
import java.util.List;

import static java.util.Arrays.*;
import static java.util.Collections.*;

/**
 * SummaryCommand reads a sequence of intervals from a data stream and computes summary statistics
 * about a given element.
 *
 * @author Jason Winnebeck
 */
public class SummaryCommand extends AbstractCommand {
	private static final List<Parameter> params = unmodifiableList( asList(
	    new Parameter( "Database", File.class,    true,  "Path to database"                ),
	    new Parameter( "Stream",   Integer.class, true,  "Stream ID"                       ),
	    new Parameter( "Element",  String.class,  true,  "Element to read from the stream" ),
	    new Parameter( "Start",    Long.class,    false, "Starting timestamp (inclusive)"  ),
	    new Parameter( "End",      Long.class,    false, "Ending timestamp (inclusive)"    )
	) );

	public String getName() {
		return "summary";
	}

	public String getHelpText() {
		return "Reads intervals in a given time range on an element within a stream, and summarizes " +
		       "them into a single report. Ranges of discontinuity are treated as if they \"don't\" " +
		       "exist, meaning their time spans are not counted as part of the total weight. The " +
		       "summary of the average value is computed by weighted average on the time span.";
	}

	public List<Parameter> getParameters() {
		return params;
	}

	@Override
	protected void executeValidated( ExecutionContext context, Object[] params )
			throws CommandException {
		PrintStream out = context.getOutput();

		File    database    = (File   ) params[0];
		Integer streamId    = (Integer) params[1];
		String  elementName = (String ) params[2];
		Long    startTime   = (Long   ) params[3];
		Long    endTime     = (Long   ) params[4];

		if ( startTime == null )
			startTime = 0L;
		if ( endTime == null )
			endTime = Long.MAX_VALUE;

		try {
			Database db = Database.load( database );

			DataStream stream = db.getStreamsMap().get( streamId );
			if ( stream == null ) {
				throw new CommandException( "No such stream " + streamId );
			}
			RecordElement element = stream.getInfo().getElement( elementName );
			if ( element == null ) {
				throw new CommandException( "No such element " + elementName + " in stream " + streamId );
			}
			if ( !Number.class.isAssignableFrom( element.getType().getTypeClass() ) ) {
				throw new CommandException( "Element " + elementName + " in stream " + streamId + " must " +
				                            "be a number to be summarized" );
			}
			int ePos = element.getPosition();

			String format = "%10s %10s %10s %10s %10s %10s %10s %10s %10s%n";
			Object[] cols = new Object[] { "stream", "element", "count", "start", "end", "min", "avg",
			                               "max", "integral" };
			out.format( format, cols );

			cols[0] = streamId;
			cols[1] = elementName;

			int    count       = 0;
			Long   start       = null;
			long   end         = 0;
			double min         = Double.POSITIVE_INFINITY;
			double max         = Double.NEGATIVE_INFINITY;
			double sumAvgs     = 0.0;
			long   totalWeight = 0;
			double integral    = 0.0;

			DBIterator<StreamInterval> iter = stream.readIntervals( startTime, endTime );
			while ( iter.hasNext() ) {
				StreamInterval si = iter.next();
				++count;
				if ( start == null )
					start = si.getStartTime();
				end = si.getEndTime();

				if ( !si.isDiscontinuity() ) {
					min = min( min, si.getMinimumValue( ePos ) );
					max = max( max, si.getMaximumValue( ePos ) );

					if ( si.getIntegral( ePos ) == null ) {
						integral = Double.NaN;
					} else {
						integral += si.getIntegral( ePos ).doubleValue();
					}
					if ( si.getAverageValue( ePos ) != null ) {
						sumAvgs += ( (Number) si.getAverageValue( ePos ) ).doubleValue();
					}
					totalWeight += si.getTimespan();
				}
			}

			if ( count > 0 ) {
				cols[2] = count;
				cols[3] = start;
				cols[4] = end;
				cols[5] = min;
				cols[6] = sumAvgs / totalWeight;
				cols[7] = max;
				cols[8] = integral;

			} else {
				cols[2] = 0;
				Arrays.fill( cols, 3, cols.length, null );
			}
			out.format( format, cols );

		} catch ( Exception e ) {
			throw new CommandException( e );
		}
	}

	private static double min( double a, Object b ) {
		if ( b == null )
			return a;
		else
			return Math.min( a, ( (Number) b ).doubleValue() );
	}

	private static double max( double a, Object b ) {
		if ( b == null )
			return a;
		else
			return Math.max( a, ( (Number) b ).doubleValue() );
	}
}
