/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.concept;

import org.gillius.realdb.BadFileException;
import org.gillius.realdb.metadata.RecordElement;
import org.gillius.realdb.model.DataStream;
import org.gillius.realdb.model.Database;
import org.gillius.realdb.model.data.StreamInterval;
import org.gillius.realdb.tools.gui.DefaultBasicModel;
import org.gillius.realdb.tools.gui.SchedulingProxy;
import org.gillius.realdb.tools.gui.SelectionModel;
import org.gillius.realdb.tools.gui.SelectionModelImpl;
import org.gillius.realdb.tools.gui.StreamIntervalsModel;
import org.gillius.realdb.tools.gui.StreamIntervalsModelImpl;
import org.gillius.realdb.util.DBIterator;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.concurrent.Executor;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

/**
 * AnalysisModelImpl
 *
 * @author Jason Winnebeck
 */
public class AnalysisModelImpl extends DefaultBasicModel implements AnalysisModel {
	private static final int vehicleStatusId = 1;
	private static final int hardAccelEventId = 2;

	private final SelectionModel<StreamInterval> selectedIncidentModel =
			new SelectionModelImpl<StreamInterval>();
	private final SelectionModel<StreamView> selectedSourceModel =
			new SelectionModelImpl<StreamView>();

	private boolean loading = false;

	private IncidentsTableModel incidents;
	private StreamIntervalsModel chartData;
	private List<StreamView> chartDataSources;

	private Executor exec;

	public AnalysisModelImpl() {
		reset();

		ThreadPoolExecutor exec = new ThreadPoolExecutor(
				1, 1, 1000L, TimeUnit.MILLISECONDS, new LinkedBlockingQueue<Runnable>() );
		exec.allowCoreThreadTimeOut( true );
		this.exec = exec;

		PropertyChangeListener listener = new PropertyChangeListener() {
			@Override
			public void propertyChange( PropertyChangeEvent evt ) {
				StreamView selected = selectedSourceModel.getSelected();
				if ( selected != null )
					selectSource( selected.stream );
			}
		};

		listener = SchedulingProxy.create( PropertyChangeListener.class, listener, exec );

		selectedSourceModel.addPropertyChangeListener( "selected", listener );
		selectedIncidentModel.addPropertyChangeListener( "selected", listener );
	}

	@Override
	public SelectionModel<StreamInterval> getSelectedIncidentModel() {
		return selectedIncidentModel;
	}

	@Override
	@SuppressWarnings( "unchecked" )
	public SelectionModel<Object> getSelectedSourceModel() {
		//noinspection RedundantCast
		return (SelectionModel) selectedSourceModel;
	}

	public void setDatabase( final Database db ) {
		exec.execute( new Runnable() {
			public void run() {
				doSetDatabase( db );
			}
		} );
	}

	private void doSetDatabase( Database db ) {
		setLoading( true );
		reset();

		List<StreamView> streams = new ArrayList<StreamView>( db.getStreams().size() );
		for ( DataStream stream : db.getStreams() ) {
			int id = stream.getInfo().getUserId();
			if ( id != vehicleStatusId && id != hardAccelEventId ) {
				streams.add( new StreamView( stream ) );
			}
		}
		setChartDataSources( streams );

		try {
			DataStream incidentsStream = db.getStreamsMap().get( hardAccelEventId );
			ArrayList<StreamInterval> incidents = new ArrayList<StreamInterval>();
			DBIterator<StreamInterval> iter = incidentsStream.readIntervals();
			while ( iter.hasNext() ) {
				StreamInterval incident = iter.next();
				if ( !incident.isDiscontinuity() && incident.getAverageValue( 0 ) == Boolean.TRUE )
					incidents.add( incident );
			}
			setIncidents( new IncidentsTableModel( incidents ) );

		} catch ( Exception e ) {
			setError( "Error while loading incidents from the database", e );

		} finally {
			setLoading( false );
		}
	}

	private void selectSource( DataStream stream ) {
		StreamInterval incident = selectedIncidentModel.getSelected();
		if ( incident == null ) {
			setChartData( new StreamIntervalsModelImpl( new RecordElement[0],
			                                            Collections.<StreamInterval>emptyList() ) );
		} else {
			try {
				DBIterator<StreamInterval> iter = stream.readIntervals( incident.getStartTime(),
				                                                        incident.getEndTime() );

				setChartData( new StreamIntervalsModelImpl( stream.getInfo().getElements(),
				                                            iterToList( iter ) ) );
			} catch ( Exception e ) {
				setError( "Error loading chart data", e );
			}
		}
	}

	@Override
	public boolean isLoading() {
		return loading;
	}

	private synchronized void setLoading( boolean newVal ) {
		boolean oldVal = loading;
		loading = newVal;
		pcs.firePropertyChange( "loading", oldVal, newVal );
	}

	@Override
	public IncidentsTableModel getIncidents() {
		return incidents;
	}

	private synchronized void setIncidents( IncidentsTableModel newVal ) {
		IncidentsTableModel oldVal = incidents;
		incidents = newVal;
		pcs.firePropertyChange( "incidents", oldVal, newVal );
	}

	@Override
	public StreamIntervalsModel getChartData() {
		return chartData;
	}

	private synchronized void setChartData( StreamIntervalsModel newVal ) {
		StreamIntervalsModel oldVal = chartData;
		chartData = newVal;
		pcs.firePropertyChange( "chartData", oldVal, newVal );
	}

	@Override
	public List<Object> getChartDataSources() {
		return Collections.<Object>unmodifiableList( chartDataSources );
	}

	private synchronized void setChartDataSources( List<StreamView> newVal ) {
		List<StreamView> oldVal = chartDataSources;
		chartDataSources = newVal;
		pcs.firePropertyChange( "chartDataSources", oldVal, newVal );
	}

	private void reset() {
		setIncidents( new IncidentsTableModel( Collections.<StreamInterval>emptyList() ) );
		setChartData( new StreamIntervalsModelImpl( new RecordElement[0],
		                                            Collections.<StreamInterval>emptyList() ) );
		setChartDataSources( Collections.<StreamView>emptyList() );
	}

	private static <T> ArrayList<T> iterToList( DBIterator<T> iter )
			throws IOException, BadFileException {
		ArrayList<T> data = new ArrayList<T>();
		while ( iter.hasNext() ) {
			data.add( iter.next() );
		}
		return data;
	}

	private static class StreamView {
		public final DataStream stream;

		private StreamView( DataStream stream ) {
			this.stream = stream;
		}

		public String toString() {
			return stream.getInfo().getName() + " (" + stream.getInfo().getUserId() + ")";
		}
	}
}
