/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.concept;

import java.io.FilterInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.Timer;
import java.util.TimerTask;

/**
 * ProgressInputStream sends periodic events to a ProgressListener noting the progress of reading
 * the input stream. The class sends out events every second once the first byte is read and stops
 * once the stream is closed (sending out a final event of all progress).
 *
 * @author Jason Winnebeck
 */
public class ProgressInputStream extends FilterInputStream {

	private volatile int progress = 0;

	private final ProgressListener listener;

	private final Timer timer;
	private TimerTask task;

	public ProgressInputStream( InputStream in, ProgressListener listener ) {
		super( in );
		this.listener = listener;

		timer = new Timer( "ProgressInputStream", true );
		task = new TimerTask() {
			@Override
			public synchronized void run() {
				if ( progress != 0 )
					ProgressInputStream.this.listener.updatedProgress( progress );
			}
		};
		timer.scheduleAtFixedRate( task, 0, 1000 );
	}

	@Override
	public void close() throws IOException {
		timer.cancel();
		task.run(); //run it one last time
		super.close();
	}

	@Override
	public int read() throws IOException {
		int ret = super.read();
		if ( ret >= 0 )
			++progress;
		return ret;
	}

	@Override
	public int read( byte[] b ) throws IOException {
		return (int) measure( super.read( b ) );
	}

	@Override
	public int read( byte[] b, int off, int len ) throws IOException {
		return (int) measure( super.read( b, off, len ) );
	}

	@Override
	public long skip( long n ) throws IOException {
		return measure( super.skip( n ) );
	}

	@Override
	public void mark( int readlimit ) {
		throw new UnsupportedOperationException();
	}

	@Override
	public void reset() throws IOException {
		throw new UnsupportedOperationException();
	}

	@Override
	public boolean markSupported() {
		return false;
	}

	private long measure( long read ) {
		if ( read >= 0 )
			progress += read;
		return read;
	}
}
