/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.concept;

import org.gillius.realdb.concept.data.DataHandlingException;
import org.gillius.realdb.concept.data.VehicleDataHandler;
import org.gillius.realdb.model.DataStream;
import org.gillius.realdb.model.Database;
import org.gillius.realdb.model.data.Discontinuity;

import java.util.HashMap;
import java.util.Map;

/**
 * RecordSinkWritingHandler is an implementation of {@link VehicleDataHandler} that directly
 * transforms events into Records and then sends those records to {@link RecordSink}s.
 *
 * @author Jason Winnebeck
 */
public class RecordSinkWritingHandler implements VehicleDataHandler {

	private final Map<Integer, RecordSink> streams;

	public RecordSinkWritingHandler( Database db ) {
		streams = new HashMap<Integer, RecordSink>();
		for ( DataStream stream : db.getStreams() ) {
			streams.put( stream.getInfo().getUserId(),
			             new RecordSinkDataStreamAdaptor( stream ) );
		}
	}

	/**
	 * Constructs a RecordSinkWritingHandler with the sinks for output.
	 */
	public RecordSinkWritingHandler( Map<Integer, RecordSink> streams ) {
		this.streams = streams;
	}

	@Override
	public void startStream() throws DataHandlingException {
	}

	@Override
	public void dataRead( int stream, long time, float value, boolean discontinuity )
			throws DataHandlingException {
		writeData( stream, time, value, discontinuity );
	}

	@Override
	public void dataRead( int stream, long time, boolean value, boolean discontinuity )
			throws DataHandlingException {
		writeData( stream, time, value, discontinuity );
	}

	@Override
	public void endStream() throws DataHandlingException {
	}

	private void writeData( int stream, long time, Object value, boolean discontinuity )
			throws DataHandlingException {
		RecordSink sink = streams.get( stream );
		if ( sink == null )
			throw new DataHandlingException( "Unknown stream " + stream );

		try {
			if ( discontinuity ) {
				sink.writeRecord( new Discontinuity( time ) );
			} else {
				sink.writeRecord( new SingleFieldRecord<Object>( time, value ) );
			}
		} catch ( Exception e ) {
			throw new DataHandlingException( "Error writing to stream " + stream, e );
		}
	}
}
