/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.concept;

import org.gillius.Util.CheckUtils;
import org.gillius.realdb.concept.data.DataHandlingException;
import org.gillius.realdb.concept.data.VehicleDataHandler;

/**
 * SimulationVehicleDataHandler implements reading the simulation data streams and calculates the
 * VehicleStatus and HardAccelEvent streams off of the data. The passed-through events plus
 * additional events are sent to another {@link VehicleDataHandler}.
 * <p/>
 * This class assumes and uses the following streams:
 * <ul>
 * <li>Input: 91 - AccelPedalPosition with one field being a float of 0 to 100.0</li>
 * <li>Output: 1 - VehicleStatus, with one boolean field: set true when data is written and turned
 *     off then back on when a data gap larger than {@link #getDormantGap() dormantGap} is
 *     encountered.</li>
 * <li>Output: 2 - HardAccelEvent, with one boolean field: set true when AccelPedalPosition is
 *     above {@link #getHardAccelThreshold() hardAccelThreshold} for more than
 *     {@link #getHardAccelDuration() hardAccelDuration}, and set false any other time.</li>
 * <li>All inputs: forward to underlying handler</li>
 * </ul>
 *
 * Default property values: {@link #DEFAULT_DORMANT_GAP}, {@link #DEFAULT_HARD_ACCEL_DURATION},
 * {@link #DEFAULT_HARD_ACCEL_THRESHOLD}.
 *
 * @author Jason Winnebeck
 */
public class SimulationVehicleDataHandler implements VehicleDataHandler {
	public static final long  DEFAULT_DORMANT_GAP          = 60000;
	public static final long  DEFAULT_HARD_ACCEL_DURATION  = 10000;
	public static final float DEFAULT_HARD_ACCEL_THRESHOLD = 90.0f;

	public static final int vehicleStatus = 1;
	public static final int hardAccelEvent = 2;

	private final VehicleDataHandler handler;

	private long  dormantGap         = DEFAULT_DORMANT_GAP;
	private long  hardAccelDuration  = DEFAULT_HARD_ACCEL_DURATION;
	private float hardAccelThreshold = DEFAULT_HARD_ACCEL_THRESHOLD;

	//streaming processing state
	private long lastTime;
	private long hardAccelStart;
	private boolean inHardAccelEvent = false;

	/**
	 * Constructs a SimulationVehicleDataHandler for passing given and additional events to the
	 * handler.
	 * @param handler
	 */
	public SimulationVehicleDataHandler( VehicleDataHandler handler ) {
		CheckUtils.checkNull( handler, "handler" );
		this.handler = handler;
	}

	public long getDormantGap() {
		return dormantGap;
	}

	public void setDormantGap( long dormantGap ) {
		this.dormantGap = dormantGap;
	}

	public long getHardAccelDuration() {
		return hardAccelDuration;
	}

	public void setHardAccelDuration( long hardAccelDuration ) {
		this.hardAccelDuration = hardAccelDuration;
	}

	public float getHardAccelThreshold() {
		return hardAccelThreshold;
	}

	public void setHardAccelThreshold( float hardAccelThreshold ) {
		this.hardAccelThreshold = hardAccelThreshold;
	}

	@Override
	public void startStream() throws DataHandlingException {
		lastTime = -1;
		hardAccelStart = -1;
		handler.startStream();
	}

	@Override
	public void dataRead( int stream, long time, float value, boolean discontinuity )
			throws DataHandlingException {
		updateVehicleStatus( time );
		if ( stream == 91 ) {
			updateHardAccelEvent( time, value, discontinuity );
		}

		//passthrough the data
		handler.dataRead( stream, time, value, discontinuity );
	}

	@Override
	public void dataRead( int stream, long time, boolean value, boolean discontinuity )
			throws DataHandlingException {
		handler.dataRead( stream, time, value, discontinuity );
	}

	private void updateVehicleStatus( long time ) throws DataHandlingException {
		if ( lastTime < 0 ) {
			handler.dataRead( vehicleStatus, time, true, false );

		} else if ( time < lastTime ) {
			throw new DataHandlingException( "Out-of-order data" );

		} else if ( time - lastTime > dormantGap ) {
			handler.dataRead( vehicleStatus, lastTime+1, false, false );
			//Simulate reception of discontinuity data during non-collection time
			updateHardAccelEvent( lastTime+1, 0.0f, true );

			handler.dataRead( vehicleStatus, time, true, false );
		}
		lastTime = time;
	}

	private void updateHardAccelEvent( long time, float value, boolean discontinuity )
			throws DataHandlingException {
		if ( discontinuity ) {
			handler.dataRead( hardAccelEvent, time, false, true );
			hardAccelStart = -1;

		} else if ( hardAccelStart < 0 ) {
			handler.dataRead( hardAccelEvent, time, false, false );
			if ( value >= hardAccelThreshold ) {
				hardAccelStart = time;
			}

		} else if ( value >= hardAccelThreshold ) {
			//still above threshold
			if ( inHardAccelEvent ) {
				handler.dataRead( hardAccelEvent, time, true, false );
			} else {
				long candidateEventTime = hardAccelStart + hardAccelDuration;
				if ( candidateEventTime <= time ) {
					handler.dataRead( hardAccelEvent, candidateEventTime, true, false );
					if ( candidateEventTime != time ) {
						handler.dataRead( hardAccelEvent, time, true, false );
					}
					inHardAccelEvent = true;
				} else {
					//event still off
					handler.dataRead( hardAccelEvent, time, false, false );
				}
			}

		} else {
			//below threshold; reset count
			hardAccelStart = -1;
			inHardAccelEvent = false;
			handler.dataRead( hardAccelEvent, time, false, false );
		}
	}

	@Override
	public void endStream() throws DataHandlingException {
		if ( lastTime >= 0 ) {
			//We add +1 here in cast there's only 1 record.
			handler.dataRead( vehicleStatus, lastTime+1, false, false );
			//Simulate reception of discontinuity data during non-collection time
			updateHardAccelEvent( lastTime+1, 0.0f, true );
		}
		handler.endStream();
	}
}
