/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.concept.data;

import java.io.IOException;
import java.io.LineNumberReader;
import java.io.Reader;

/**
 * TextVehicleDataParser parses {@link VehicleDataHandler} events from a delimited text file
 * with four fields per line: stream, time, value, discontinuity. When discontinuity is true ("1"),
 * value must still be a valid float, but its value is ignored. It is suggested to put in the value
 * "0".
 * <p/>
 * The parser can only read a single time, since there is no way to rewind its input.
 *
 * @author Jason Winnebeck
 */
public class TextVehicleDataParser extends AbstractVehicleDataParser {
	private final LineNumberReader in;
	private final String delim;

	public TextVehicleDataParser( Reader reader, String delim ) {
		in = new LineNumberReader( reader );
		this.delim = delim;
	}

	@Override
	public void parse() throws IOException, DataHandlingException {
		if ( handler == null )
			throw new IllegalStateException( "No handler set" );

		handler.startStream();

		String line = null;
		try {
			while ( ( line = in.readLine() ) != null ) {
				String[] fields = line.split( delim, 5 );
				if ( fields.length != 4 ) {
					throw new DataHandlingException( String.format(
							"Invalid number of fields (%s) on line %s, text: %s",
							fields.length, in.getLineNumber(), line ) );
				}

				int streamId = Integer.parseInt( fields[0] );
				long time = Long.parseLong( fields[1] );
				boolean discontinuity;
				if ( "1".equals( fields[3] ) ) {
					discontinuity = true;
				} else if ( "0".equals( fields[3] ) ) {
					discontinuity = false;
				} else {
					throw new DataHandlingException( String.format(
							"Invalid discontinuity field (%s) on line %s, text: %s",
							fields[3], in.getLineNumber(), line ) );
				}
				float value = discontinuity ? 0.0f : Float.parseFloat( fields[2] );

				handler.dataRead( streamId, time, value, discontinuity );
			}

			handler.endStream();
			in.close();

		} catch ( NumberFormatException e ) {
			throw new DataHandlingException( String.format(
					"Invalid number format on line %s, text: %s",
					in.getLineNumber(), line ) );
		}
	}
}
