/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.concept;

import org.easymock.*;
import org.gillius.realdb.concept.data.VehicleDataHandler;
import org.junit.*;

import static org.easymock.EasyMock.*;

/**
 * TestSimulationVehicleDataHandler tests the {@link SimulationVehicleDataHandler}.
 *
 * @author Jason Winnebeck
 */
public class TestSimulationVehicleDataHandler extends EasyMockSupport {
	private VehicleDataHandler writer;

	private SimulationVehicleDataHandler handler;

	@Before
	public void setUp() throws Exception {
		writer = createNiceMock( "writer", VehicleDataHandler.class );
		handler = new SimulationVehicleDataHandler( writer );
	}

	@Test
	public void testStartStatus() throws Exception {
		resetAll();
		writer.startStream();
		expectBool( 1, 100, true );
		writer.endStream();
		replayAll();

		handler.startStream();
		handler.dataRead( 3, 100, 1.0f, false );
		handler.endStream();
		verifyAll();
	}

	@Test
	public void testEndStatus() throws Exception {
		resetAll();
		writer.startStream();
		expectBool( 1, 201, false );
		writer.endStream();
		replayAll();

		handler.startStream();
		handler.dataRead( 3, 100, 1.0f, false );
		handler.dataRead( 3, 200, 1.0f, false );
		handler.endStream();

		verifyAll();
	}

	@Test
	public void testDurationGap() throws Exception {
		resetAll();
		filter( 1 );
		writer.startStream();
		expectBool( 1, 175, true );  //start
		expectBool( 1, 201, false ); //gap
		expectBool( 1, 300, true );  //gap
		expectBool( 1, 331, false ); //end
		writer.endStream();
		replayAll();

		handler.setDormantGap( 50 );

		handler.startStream();
		handler.dataRead( 3, 175, 1.0f, false ); //start
		handler.dataRead( 3, 185, 1.0f, false );
		handler.dataRead( 3, 200, 1.0f, false );
		//gap
		handler.dataRead( 3, 300, 1.0f, false );
		handler.dataRead( 3, 325, 1.0f, false );
		handler.dataRead( 3, 330, 1.0f, false ); //end
		handler.endStream();

		verifyAll();
	}

	@Test
	public void testHardAccelTriggered() throws Exception {
		resetAll();
		filter( 2 );
		writer.startStream();
		expectBool( 2, 100, false );
		expectBool( 2, 105, false );
		expectBool( 2, 205, true );
		expectBool( 2, 300, true );
		replayAll();

		handler.setDormantGap( 10000 );
		handler.setHardAccelDuration( 100 );
		handler.setHardAccelThreshold( 0.5f );

		handler.startStream();
		handler.dataRead( 91, 100, 0.0f, false );
		handler.dataRead( 91, 105, 0.6f, false );
		handler.dataRead( 91, 300, 1.0f, false );

		verifyAll();
	}

	@Test
	public void testHardAccelTriggeredNoRepeatsAndEnds() throws Exception {
		resetAll();
		filter( 2 );
		writer.startStream();
		expectBool( 2, 100, false );
		expectBool( 2, 105, false );
		expectBool( 2, 205, true );
		expectBool( 2, 300, true );
		expectBool( 2, 310, true );
		writer.dataRead( 2, 311, false, true );
		writer.endStream();
		replayAll();

		handler.setDormantGap( 10000 );
		handler.setHardAccelDuration( 100 );
		handler.setHardAccelThreshold( 0.5f );

		handler.startStream();
		handler.dataRead( 91, 100, 0.0f, false );
		handler.dataRead( 91, 105, 0.6f, false );
		handler.dataRead( 91, 300, 1.0f, false );
		handler.dataRead( 91, 310, 1.0f, false );
		handler.endStream();

		verifyAll();
	}

	@Test
	public void testHardAccelDiscontinuityAtGap() throws Exception {
		testHardAccelTriggered();

		resetAll();
		filter( 2 );
		writer.dataRead( 2, 301, false, true ); //where collecting goes false
		replayAll();

		handler.setDormantGap( 100 );
		handler.dataRead( 3, 500, 0.0f, false );

		verifyAll();
	}

	@Test
	public void testHardAccelDiscontinuityAtEnd() throws Exception {
		testHardAccelTriggered();

		resetAll();
		filter( 2 );
		writer.dataRead( 2, 321, false, true ); //where collecting goes false
		writer.endStream();
		replayAll();

		handler.dataRead( 3, 320, 0.0f, false );
		handler.endStream();

		verifyAll();
	}

	@Test
	public void testHardAccelDiscontinuityAtDiscontinuity() throws Exception {
		testHardAccelTriggered();

		resetAll();
		filter( 2 );
		writer.dataRead( 2, 320, false, true );
		replayAll();

		handler.dataRead( 91, 320, 0.0f, true );

		verifyAll();
	}

	@Test
	public void testHardAccelDiscontinuityDoesNotTrigger() throws Exception {
		resetAll();
		filter( 2 );
		writer.startStream();
		expectBool( 2, 100, false );
		expectBool( 2, 105, false );
		writer.dataRead( 2, 300, false, true );
		expectBool( 2, 390, false );
		replayAll();

		handler.setDormantGap( 10000 );
		handler.setHardAccelDuration( 100 );
		handler.setHardAccelThreshold( 0.5f );

		handler.startStream();
		handler.dataRead( 91, 100, 0.0f, false );
		handler.dataRead( 91, 105, 0.6f, false );
		handler.dataRead( 91, 300, 0.0f, true );
		handler.dataRead( 91, 390, 1.0f, false );

		verifyAll();
	}

	@Test
	public void testHardAccelHighLowHighDoesNotTrigger() throws Exception {
		resetAll();
		filter( 2 );
		handler.startStream();
		expectBool( 2, 100, false );
		expectBool( 2, 105, false );
		expectBool( 2, 200, false );
		expectBool( 2, 220, false );
		replayAll();

		handler.setDormantGap( 10000 );
		handler.setHardAccelDuration( 100 );
		handler.setHardAccelThreshold( 0.5f );

		handler.startStream();
		handler.dataRead( 91, 100, 0.0f, false );
		handler.dataRead( 91, 105, 0.6f, false );
		handler.dataRead( 91, 200, 0.0f, false );
		handler.dataRead( 91, 220, 1.0f, false );

		verifyAll();
	}

	@Test
	public void testHardAccelLowDoesNotTrigger() throws Exception {
		resetAll();
		filter( 2 );
		handler.startStream();
		expectBool( 2, 100, false );
		expectBool( 2, 400, false );
		replayAll();

		handler.setDormantGap( 10000 );
		handler.setHardAccelDuration( 100 );
		handler.setHardAccelThreshold( 0.99f );

		handler.startStream();
		handler.dataRead( 91, 100, 0.95f, false );
		handler.dataRead( 91, 400, 0.96f, false );

		verifyAll();
	}

	private void filter( int stream ) throws Exception {
		resetToDefault( writer );
		writer.dataRead( not(eq(stream)), anyLong(), anyFloat(), anyBoolean() );
		expectLastCall().anyTimes();
		writer.dataRead( not(eq(stream)), anyLong(), anyBoolean(), anyBoolean() );
		expectLastCall().anyTimes();
	}

	private void expectFloat( int stream, long time, float value ) throws Exception {
		writer.dataRead( stream, time, value, false );
	}

	private void expectBool( int stream, long time, boolean value ) throws Exception {
		writer.dataRead( stream, time, value, false );
	}
}
