/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.Util;

import java.util.ArrayList;
import java.util.List;

import static java.util.Arrays.*;

/**
 * CircularLog is useful to provide the last 'n' log messages in a crash event, when extremely
 * verbose logging is desired, and outputting to stdout/err would cause problems due to the amount
 * of output.
 * <p/>
 * The intention of this class is that writeLog calls would be injected temporarily into code, and
 * then the debugger is used to view the actual log.
 *
 * @author Jason Winnebeck
 */
public class CircularLog {
	public static final int LOG_SIZE = 100;

	private static final String[] log = new String[ LOG_SIZE ];

	private static int currPos = 0;

	/**
	 * Writes a message to the log.
	 *
	 * @param s non-null String
	 */
	public static synchronized void writeLog( String s ) {
		log[currPos] = s;
		currPos = (currPos + 1) % LOG_SIZE;
	}

	/**
	 * Writes a message to the log using the {@link String#format(String, Object[])} method.
	 */
	public static void writeLog( String format, Object... args ) {
		writeLog( String.format( format, args ) );
	}

	/**
	 * Returns a copy of the last {@link #LOG_SIZE} messages as a list.
	 */
	public static synchronized List<String> getLastMessages() {
		ArrayList<String> ret = new ArrayList<String>( log.length );
		ret.addAll( asList( log ).subList( currPos, log.length ) );
		ret.addAll( asList( log ).subList( 0, currPos ) );
		return ret;
	}

	/**
	 * Returns {@link #getLastMessages()} but filtered such that only log entries containing the
	 * filter string appear are in the returned list.
	 */
	public static synchronized List<String> getFilteredMessages( String filter ) {
		ArrayList<String> ret = new ArrayList<String>( log.length );
		for ( String s : getLastMessages() ) {
			if ( s.contains( filter ) )
				ret.add( s );
		}
		return ret;
	}
}
