/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.experiments;

import org.gillius.realdb.io.BlockFile;
import org.gillius.realdb.io.RandomAccessBlockFile;
import org.gillius.realdb.util.ConsoleProgressMeter;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.RandomAccessFile;
import java.nio.ByteBuffer;
import java.util.Date;
import java.util.HashSet;
import java.util.Random;
import java.util.Set;
import java.util.zip.CRC32;
import java.util.zip.Checksum;

/**
 * CorruptionTest aids in the testing to see how a BlockFile can get corrupted, by writing data
 * to that file in an expected layout.
 *
 * @author Jason Winnebeck
 */
public class CorruptionTest {
	private static final String EOL = System.getProperty( "line.separator" );
	private static final int VALIDATE_SEG_SIZE = 262144;

	public static void main( String[] args ) throws IOException {
		if ( args.length != 3 ) {
			System.err.println( "Usage: java " + CorruptionTest.class + " <filename> <mode> <block_size>" );
			System.exit( 1 );
		}

		File f = new File( args[0] );
		String mode = args[1];
		int blockSize = Integer.parseInt( args[2] );
		System.out.println( "Using file " + f.getAbsolutePath() );
		System.out.println( "  with blocksize = " + blockSize + " bytes" );
		System.out.println( "  and mode " + mode );
		System.out.println();

		System.out.println( "Possible commands:" );
		System.out.println( "  allocate <num_blocks>" );
		System.out.println( "  clear" );
		System.out.println( "  write" );
		System.out.println( "  validate" );
		System.out.println( "  exit" );
		System.out.println();

		CorruptionTest ct = new CorruptionTest( f, mode, blockSize );
		ct.parseLoop();
	}

	private final String mode;
	private final File   f;
	private final int    blockSize;

	public CorruptionTest( File f, String mode, int blockSize ) {
		this.f         = f;
		this.mode      = mode;
		this.blockSize = blockSize;
	}

	public void parseLoop() throws IOException {
		BufferedReader in = new BufferedReader( new InputStreamReader( System.in ) );

		String nextLine = in.readLine();
		while ( nextLine != null ) {
			String[] line = nextLine.split( " " );
			long time = System.nanoTime();
			if ( line[0].equalsIgnoreCase( "allocate" ) )
				doAllocate( line );
			else if ( line[0].equalsIgnoreCase( "clear" ) )
				doClear();
			else if ( line[0].equalsIgnoreCase( "write" ) )
				doWrite();
			else if ( line[0].equalsIgnoreCase( "validate" ) )
				doValidate();
			else if ( line[0].equalsIgnoreCase( "exit" ) || line[0].equalsIgnoreCase( "quit" ) )
				System.exit( 0 );
			else
				System.err.println( "Unrecognized command " + nextLine );

			double diff = ( System.nanoTime() - time ) / 1000000000.0;
			System.out.println( "Completed " + nextLine + " in " + diff + " seconds" );
			nextLine = in.readLine();
		}
	}

	private void doAllocate( String[] line ) throws IOException {
		if ( line.length != 2 ) {
			System.err.println( "Usage: allocate <num_blocks>" );
			System.exit( 2 );
		}

		int numBlocks = Integer.parseInt( line[1] );

		RandomAccessFile raf = new RandomAccessFile( f, mode );
		raf.setLength( (long) blockSize * numBlocks );
		raf.close();
	}

	private void doClear() throws IOException {
		BlockFile bf = getBlockFile();
		byte[] data = new byte[ blockSize ];
		ByteBuffer buf = ByteBuffer.wrap( data, 0, 12 );
		Date fileCreation = new Date();
		System.out.println( "Creating file with timestamp of " + fileCreation );

		ConsoleProgressMeter meter = new ConsoleProgressMeter( bf.getSize(), 2, 20 );
		int i = 0;
		try {
			for ( i = 0; i < bf.getSize(); ++i ) {
				buf.rewind();
				buf.putInt( i );
				buf.putLong( fileCreation.getTime() );
				bf.writeBlock( i, data );
				meter.updateValue( i );
			}
		} catch ( IOException e ) {
			System.err.println( "Clear failed while writing block " + i );
			e.printStackTrace();
		}
		meter.finish();

		bf.close();
	}

	private void doWrite() throws IOException {
		BlockFile bf = getBlockFile();
		Checksum c = new CRC32();
		byte[] data = new byte[ blockSize ];
		ByteBuffer buf = ByteBuffer.wrap( data );
		Random rnd = new Random();
		Date fileCreation = new Date();
		System.out.println( "Creating file with timestamp of " + fileCreation );

		ConsoleProgressMeter meter = new ConsoleProgressMeter( bf.getSize(), 2, 20 );
		int i = -1;
		try {
			for ( i = 0; i < bf.getSize(); ++i ) {
				//fill data array with random bytes, but we won't use the first or last four
				rnd.nextBytes( data );

				//First 4 bytes is the block number
				buf.position( 0 );
				buf.putInt( i );
				buf.putLong( fileCreation.getTime() );

				//generate the checksum in the last 4 bytes
				c.reset();
				c.update( data, 0, blockSize - 4 );
				buf.position( buf.capacity() - 4 );
				buf.putInt( (int) c.getValue() );

				bf.writeBlock( i, data );
				meter.updateValue( i );
			}
		} catch ( IOException e ) {
			System.err.println( "Write failed while writing block " + i );
			throw e;
		}
		meter.finish();

		bf.close();
	}

	private void doValidate() throws IOException {
		BlockFile bf = getBlockFile();
		Checksum c = new CRC32();
		byte[] data = new byte[ blockSize ];
		ByteBuffer buf = ByteBuffer.wrap( data );
		ConsoleProgressMeter meter = new ConsoleProgressMeter( bf.getSize(), 2, 20 );

		int clearBlocks   = 0;
		int badBlocks     = 0;
		int goodBlocks    = 0;
		int wrongBlockNum = 0;

		StringBuilder errorReport = new StringBuilder();
		StringBuilder tsReport = new StringBuilder();
		Set<Date> timestamps = new HashSet<Date>();

		for ( int i=0; i < bf.getSize(); ++i ) {
			bf.readBlock( i, data );
			buf.rewind();
			int readBlockNum = buf.getInt();
			if ( readBlockNum != i ) {
				errorReport.append( "Block " );
				errorReport.append( i );
				errorReport.append( " says it is block number " );
				errorReport.append( readBlockNum );
				errorReport.append( EOL );
				++wrongBlockNum;
			}
			Date readBlockTime = new Date( buf.getLong() );
			if ( timestamps.add( readBlockTime ) ) {
				tsReport.append( "New timestamp " );
				tsReport.append( readBlockTime );
				tsReport.append( " found at block " );
				tsReport.append( i );
				tsReport.append( EOL );
			}

			//The block is a "clear" block if all but the first 4 bytes are clear
			boolean clear = true;
			while ( clear && buf.remaining() > 0 ) {
				if ( buf.get() != 0 )
					clear = false;
			}

			if ( clear )
				++clearBlocks;
			else {
				//check the CRC
				buf.position( buf.capacity() - 4 );
				int readCrc = buf.getInt();

				c.reset();
				c.update( data, 0, blockSize - 4 );
				int crc = (int) c.getValue();

				if ( readCrc == crc )
					++goodBlocks;
				else {
					errorReport.append( "Block " );
					errorReport.append( i );
					errorReport.append( " failed CRC check." );
					errorReport.append( EOL );
					++badBlocks;

					//Perform segment analysis, to see how far we wrote into a block
					buf.rewind();
					int b=0;
					while ( buf.hasRemaining() ) {
						int numB = 0;
						int lastNZ = 0;
						for ( int j=0; j < VALIDATE_SEG_SIZE && buf.hasRemaining(); ++j ) {
							if ( buf.get() != 0 ) {
								++numB;
								lastNZ = buf.position() - b * VALIDATE_SEG_SIZE;
							}
						}
						errorReport.append( "  Block " ).append( i );
						errorReport.append( ", segment " ).append( b ).append( " has " ).append( numB );
						errorReport.append( " non-zero bytes, with the last non-zero byte at " );
						errorReport.append( lastNZ ).append( '.' ).append( EOL );
						++b;
					}
				}
			}

			meter.updateValue( i );
		}

		meter.finish();
		System.out.format( "Blocks good=%s, bad_data=%s, bad_block_num=%s, clear=%s%n",
		                   goodBlocks, badBlocks, wrongBlockNum, clearBlocks );
		if ( tsReport.length() > 0 )
			System.out.print( tsReport );
		if ( errorReport.length() > 0 )
			System.out.print( errorReport );

		bf.close();
	}

	private BlockFile getBlockFile() throws IOException {
		if ( !f.exists() || f.isDirectory() ) {
			System.err.println( "Please allocate file before using it" );
		}
		RandomAccessFile raf = new RandomAccessFile( f, mode );
		return new RandomAccessBlockFile( raf, blockSize );
	}
}
