/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.io;

import org.gillius.Util.CircularLog;

import java.io.IOException;
import java.util.Random;

/**
 * AbstractFailingBlockFile is a base class for a class that can simulate a file failure at an
 * arbitrary time through the wrapped BlockFile. When a failure occurs, the block currently being
 * written is set to random data. It is assumed that the wrapped file itself does not fail.
 *
 * @author Jason Winnebeck
 */
public abstract class AbstractFailingBlockFile implements BlockFile {
	protected final BlockFile wrapped;
	protected final Random rng;

	/**
	 * Creates a new AbstractFailingBlockFile by wrapping a real BlockFile.
	 *
	 * @param wrapped non-null BlockFile
	 * @param random  non-null source of randomness, to generate consistent "random" data on failures.
	 */
	public AbstractFailingBlockFile( Random random, BlockFile wrapped ) {
		rng = random;
		this.wrapped = wrapped;
	}

	public int getSize() {
		return wrapped.getSize();
	}

	public int getBlockSize() {
		return wrapped.getBlockSize();
	}

	public void readBlock( int num, byte[] data ) throws IOException {
		if ( checkFailure() )
			throw makeFault();

		wrapped.readBlock( num, data );
	}

	public int readBlocks( int startBlock, byte[] data ) throws IOException {
		if ( checkFailure() )
			throw makeFault();

		return wrapped.readBlocks( startBlock, data );
	}

	public void writeBlock( int num, byte[] data ) throws IOException {
		if ( checkFailure() ) {
			data = new byte[ data.length ];
			rng.nextBytes( data );
			wrapped.writeBlock( num, data );
			throw makeFault();

		} else {
			wrapped.writeBlock( num, data );
		}
	}

	public int writeBlocks( int startBlock, byte[] data ) throws IOException {
		if ( checkFailure() ) {
			data = new byte[ data.length ];
			rng.nextBytes( data );
			wrapped.writeBlocks( startBlock, data );
			throw makeFault();

		} else {
			return wrapped.writeBlocks( startBlock, data );
		}
	}

	public int writeBlocks( int startBlock, byte[] data, int start, int len ) throws IOException {
		if ( checkFailure() ) {
			data = new byte[ len ];
			rng.nextBytes( data );
			wrapped.writeBlocks( startBlock, data, 0, len );
			throw makeFault();

		} else {
			return wrapped.writeBlocks( startBlock, data, start, len );
		}
	}

	public void close() throws IOException {
		if ( checkFailure() )
			throw makeFault();

		wrapped.close();
	}

	/**
	 * This method is called before every operation that can throw IOException. If it returns true,
	 * returns a failure. If it returns true while a write method is being called, random bytes are
	 * written instead of the requested bytes before throwing the exception.
	 */
	protected abstract boolean checkFailure();

	private static IOException makeFault() {
		CircularLog.writeLog( "Triggered fault" );
		return new IOException( "Triggered fault" );
	}
}
