/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.io;

import java.util.Arrays;

/**
 * BitField implements a fixed-size bit field that can be easily converted to a byte[] of the
 * smallest size. The byte array is stored "little-endian". The least significant bit (LSB) in the first
 * byte of the byte array is the LSB (bit 0) in the field, while the MSB in the last byte is the
 * MSB of the field. If the field's size is not evenly divisible by 8, it will be rounded up to the
 * nearest multiple of 8.
 * <p/>
 * This class differs from java.util.BitSet in that it is directly convertible to a byte[], but it
 * has much more limited functionality.
 *
 * @author Jason Winnebeck
 */
public class BitField {
	private final byte[] data;

	public BitField( int size ) {
		data = new byte[( size + 7 ) / 8];
	}

	/**
	 * Constructs a new BitField based on the given byte array, which is not copied.
	 *
	 * @param data non-null byte[] in the format described by the class documentation. This class
	 *             assumes ownership for the array; clone it if you want to copy.
	 */
	public BitField( byte[] data ) {
		this.data = data;
	}

	/**
	 * Returns the size of the BitField in bits.
	 */
	public int getSize() {
		return data.length * 8;
	}

	/**
	 * Returns the raw bit field values as a byte array, which must not be modified. The returned
	 * byte[] may be an intrusive view into the object.
	 *
	 * @return non-null read-only byte[] as described in the class documentation.
	 */
	public byte[] getData() {
		return data;
	}

	/**
	 * Gets the specified bit, with 0 being the least significant bit.
	 *
	 * @throws IndexOutOfBoundsException if index is &gt;= {@link #getSize()}.
	 */
	public boolean get( int index ) {
		//Don't need to check large case; let JVM do that
		if ( index < 0 )
			throw new IndexOutOfBoundsException( String.valueOf( index ) );

		int arrayIndex = index / 8;
		int bitIndex   = index % 8;
		int mask       = 1 << bitIndex;

		return ( data[arrayIndex] & mask ) != 0;
	}

	/**
	 * Sets the specified bit, with 0 being the least significant bit.
	 *
	 * @throws IndexOutOfBoundsException if index is &gt;= {@link #getSize()}.
	 */
	public void set( int index ) {
		//Don't need to check large case; let JVM do that
		if ( index < 0 )
			throw new IndexOutOfBoundsException( String.valueOf( index ) );

		int arrayIndex = index / 8;
		int bitIndex   = index % 8;
		int value      = 1 << bitIndex;

		data[arrayIndex] |= value;
	}

	/**
	 * Clears all bits.
	 */
	public void clear() {
		Arrays.fill( data, (byte) 0 );
	}
}
