/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.io;

import java.io.Closeable;
import java.io.EOFException;
import java.io.IOException;

/**
 * BlockFile is a low-level interface to perform synchronous writes to a file. The implementation
 * may be operating system and hardware-dependent to provide the synchronous write guarantee that
 * once a block is written, it can be reread even if a power loss or storage disconnect results
 * immediately after its return. There has been some research to see if and how this is possible in
 * a real system.
 * <p/>
 * Future extensions may include read/write methods with offset and length.
 * <p/>
 * BlockFile instances are usually not thread-safe.
 *
 * @author Jason Winnebeck
 */
public interface BlockFile extends Closeable {
	/**
	 * Returns the size of the file, in blocks.
	 */
	public int getSize();

	/**
	 * Returns the size of a block, in bytes.
	 */
	public int getBlockSize();

	/**
	 * Read a block of data into the given buffer, which is of the size returned by
	 * {@link #getBlockSize()}.
	 */
	public void readBlock( int num, byte[] data ) throws IOException;

	/**
	 * Reads one or more blocks from the file into the given buffer. The number of blocks read is
	 * the data array length divided by the block size, rounded up. If the data buffer does not have
	 * enough space to fully fit the last block, the content is truncated.
	 *
	 * @param startBlock   starting block address, from 0 to {@link #getSize()} - 1
	 * @param data         non-null data buffer that is fully populated
	 *
	 * @return the number of blocks actually read (includes the possibly truncated block)
	 *
	 * @throws IOException  if an error occurs reading the data
	 * @throws EOFException if there are not enough blocks to fill the buffer
	 */
	public int readBlocks( int startBlock, byte[] data ) throws IOException;

	/**
	 * Write a block of data from the given buffer, which is of the size returned by
	 * {@link #getBlockSize()}. When this method suceeds, the block has been safely committed to the
	 * disk. If it fails, then the block will have undefined content.
	 */
	public void writeBlock( int num, byte[] data ) throws IOException;

	/**
	 * Works identically to {@link #writeBlocks(int, byte[], int, int)} with start=0 and
	 * len=data.length.
	 *
	 * @param startBlock   starting block address, from 0 to {@link #getSize()} - 1
	 * @param data         non-null data block that is not too large (see main comment)
	 *
	 * @return The number of blocks actually written
	 *
	 * @throws IOException  if an error occurs while writing the data
	 * @throws EOFException if there are not enough blocks to write
	 */
	public int writeBlocks( int startBlock, byte[] data ) throws IOException;

	/**
	 * Write multiple blocks of data from the given buffer, which is of any size that will fit within
	 * the file. The block will fit if ({@link #getSize()} - startBlock) * {@link #getBlockSize()} is
	 * greater than the length of data. If the size of data is not an exact multiple of the block
	 * size, extra 0 bytes are padded at the end of the block. When this method suceeds, the block
	 * has been safely committed to the disk. If it fails, then all of the affected blocks have an
	 * undefined content.
	 * <p/>
	 * The primary purpose of this method is for performance, when writing a larger logical block
	 * than the minimum physical block size supported, when it is acceptable to entirely commit or
	 * entirely lose the logical block in one operation.
	 *
	 * @param startBlock   starting block address, from 0 to {@link #getSize()} - 1
	 * @param data         non-null data block that is not too large (see main comment)
	 *
	 * @return The number of blocks actually written
	 *
	 * @throws IOException  if an error occurs while writing the data
	 * @throws EOFException if there are not enough blocks to write
	 */
	public int writeBlocks( int startBlock, byte[] data, int start, int len ) throws IOException;
}
