/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.io;

import java.io.EOFException;
import java.io.IOException;

/**
 * ByteArrayBlockFile is an implementation of a BlockFile that is in-memory only, backed by a byte
 * array.
 * <p/>
 * TODO: figure out how to share a lot of this code with RandomAccessBlockFile
 *
 * @author Jason Winnebeck
 */
public class ByteArrayBlockFile implements BlockFile {
	private final byte[] d;
	private final int blockSize;
	private final int numBlocks;

	/**
	 * Constant buffer used for zero byte padding in writeBlocks.
	 */
	private final byte[] zeroBuf;

	public ByteArrayBlockFile( int blockSize, int numBlocks ) {
		d = new byte[ blockSize * numBlocks ];
		this.blockSize = blockSize;
		this.numBlocks = numBlocks;

		zeroBuf = new byte[ blockSize ];
	}

	/**
	 * Constructs a new ByteArrayBlockFile as a copy of the given source file.
	 *
	 * @param src non-null ByteArrayBlockFile
	 */
	public ByteArrayBlockFile( ByteArrayBlockFile src ) {
		d         = src.d.clone(); //We're not Cloneable because d is a final variable
		blockSize = src.blockSize;
		numBlocks = src.numBlocks;
		zeroBuf   = src.zeroBuf;   //this actually can be shared
	}

	public int getSize() {
		return numBlocks;
	}

	public int getBlockSize() {
		return blockSize;
	}

	public void readBlock( int num, byte[] data ) throws IOException {
		if ( num < 0 || num >= numBlocks ) {
			throw new IllegalArgumentException( "block " + num + " is not in the valid range of " +
			                                    "blocks from 0 to " + ( getSize() - 1 ) );
		}
		if ( data.length != blockSize )
			throw new IllegalArgumentException( "data block must be block size = " + blockSize );

		System.arraycopy( d, num * blockSize, data, 0, blockSize );
	}

	public int readBlocks( int startBlock, byte[] data ) throws IOException {
		if ( startBlock < 0 )
			throw new IllegalArgumentException( "startBlock must be non-negative" );
		if ( data.length > (numBlocks - startBlock) * blockSize ) {
			throw new EOFException( "Given buffer with " + data.length + " bytes is too large for " +
			                        "starting block " + startBlock );
		}

		int leftOver = data.length % blockSize;
		int numRead = data.length / blockSize;
		if ( leftOver > 0 )
			++numRead;

		System.arraycopy( d, startBlock * blockSize, data, 0, data.length );

		return numRead;
	}

	public void writeBlock( int num, byte[] data ) throws IOException {
		if ( num < 0 || num >= numBlocks ) {
			throw new IllegalArgumentException( "block " + num + " is not in the valid range of " +
			                                    "blocks from 0 to " + ( getSize() - 1 ) );
		}
		if ( data.length != blockSize )
			throw new IllegalArgumentException( "data block must be block size = " + blockSize );

		System.arraycopy( data, 0, d, num * blockSize, blockSize );
	}

	public int writeBlocks( int startBlock, byte[] data ) throws IOException {
		return writeBlocks( startBlock, data, 0, data.length );
	}

	public int writeBlocks( int startBlock, byte[] data, int start, int len ) throws IOException {
		if ( startBlock < 0 )
			throw new IllegalArgumentException( "startBlock must be non-negative" );
		if ( len > ( numBlocks - startBlock ) * blockSize ) {
			throw new EOFException( "Given buffer with " + data.length + " bytes will not fit at " +
			                        "starting block " + startBlock );
		}

		int leftOver   = len % blockSize;
		int numWritten = len / blockSize;
		if ( leftOver > 0 )
			++numWritten;

		int dStart = startBlock * blockSize;
		System.arraycopy( data, start, d, dStart, len );
		if ( leftOver > 0 )
			System.arraycopy( zeroBuf, 0, d, dStart + len, blockSize - leftOver ); //zero pad

		return numWritten;
	}

	public void close() throws IOException {
	}
}
