/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.io;

import org.gillius.Util.CheckUtils;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;
import java.nio.BufferOverflowException;
import java.nio.BufferUnderflowException;
import java.nio.ByteBuffer;

/**
 * ByteBufferDataOutput is a DataOutput and DataInput implementation that writes to a fixed-size
 * ByteBuffer. Each of the DataOutput methods delegate to the appropriate "put" or "get" method(s)
 * of ByteBuffer. Since the contained ByteBuffer is exposed through a get method, code using the
 * class should be aware of possible side-effects if ByteBuffer methods are called on it.
 * <p/>
 * Any BufferOverflowExceptions thrown by the ByteBuffer are wrapped as IOExceptions for DataOutput.
 * Any BufferUnderflowExceptions thrown are wrapped as EOFExceptions for DataInput.
 *
 * @author Jason Winnebeck
 */
public class ByteBufferDataIO implements DataOutput, DataInput {
	private final ByteBuffer buf;

	/**
	 * Creates a new ByteBufferDataOutput with a ByteBuffer with backing array of 'size' bytes, in
	 * big-endian byte ordering.
	 */
	public ByteBufferDataIO( int size ) {
		buf = ByteBuffer.allocate( size );
	}

	/**
	 * Constructs a new ByteBufferDataOutput with the given ByteBuffer. To follow the specification of
	 * DataOutput, the buffer's order should be big endian, but this class does not enforce that.
	 *
	 * @param buf non-null ByteBuffer
	 */
	public ByteBufferDataIO( ByteBuffer buf ) {
		CheckUtils.checkNull( buf, "buf" );
		this.buf = buf;
	}

	/**
	 * Constructs a new ByteBufferDataOutput with a ByteBuffer wrapped around the given array.
	 */
	public ByteBufferDataIO( byte[] data ) {
		CheckUtils.checkNull( data, "data" );
		buf = ByteBuffer.wrap( data );
	}

	public ByteBuffer getBuf() {
		return buf;
	}

	/**
	 * Returns a copy of all written data in the buffer (from 0 to the position).
	 */
	public byte[] getDataCopy() {
		byte[] ret = new byte[ buf.position() ];
		System.arraycopy( buf.array(), buf.arrayOffset(), ret, 0, buf.position() );
		return ret;
	}

	public void write( int b ) throws IOException {
		try {
			buf.put( (byte) b );
		} catch ( BufferOverflowException e ) {
			throw makeOverflow( e );
		}
	}

	public void write( byte b[] ) throws IOException {
		try {
			buf.put( b );
		} catch ( BufferOverflowException e ) {
			throw makeOverflow( e );
		}
	}

	public void write( byte b[], int off, int len ) throws IOException {
		try {
			buf.put( b, off, len );
		} catch ( BufferOverflowException e ) {
			throw makeOverflow( e );
		}
	}

	public void writeBoolean( boolean v ) throws IOException {
		try {
			buf.put( v ? (byte) 1 : (byte) 0 );
		} catch ( BufferOverflowException e ) {
			throw makeOverflow( e );
		}
	}

	public void writeByte( int v ) throws IOException {
		try {
			buf.put( (byte) v );
		} catch ( BufferOverflowException e ) {
			throw makeOverflow( e );
		}
	}

	public void writeShort( int v ) throws IOException {
		try {
			buf.putShort( (short) v );
		} catch ( BufferOverflowException e ) {
			throw makeOverflow( e );
		}
	}

	public void writeChar( int v ) throws IOException {
		try {
			buf.putChar( (char) v );
		} catch ( BufferOverflowException e ) {
			throw makeOverflow( e );
		}
	}

	public void writeInt( int v ) throws IOException {
		try {
			buf.putInt( v );
		} catch ( BufferOverflowException e ) {
			throw makeOverflow( e );
		}
	}

	public void writeLong( long v ) throws IOException {
		try {
			buf.putLong( v );
		} catch ( BufferOverflowException e ) {
			throw makeOverflow( e );
		}
	}

	public void writeFloat( float v ) throws IOException {
		try {
			buf.putFloat( v );
		} catch ( BufferOverflowException e ) {
			throw makeOverflow( e );
		}
	}

	public void writeDouble( double v ) throws IOException {
		try {
			buf.putDouble( v );
		} catch ( BufferOverflowException e ) {
			throw makeOverflow( e );
		}
	}

	public void writeBytes( String s ) throws IOException {
		try {
			for ( int i = 0; i < s.length(); ++i )
				buf.put( (byte) s.charAt( i ) );
		} catch ( BufferOverflowException e ) {
			throw makeOverflow( e );
		}
	}

	public void writeChars( String s ) throws IOException {
		try {
			for ( int i = 0; i < s.length(); ++i )
				buf.putChar( s.charAt( i ) );
		} catch ( BufferOverflowException e ) {
			throw makeOverflow( e );
		}
	}

	public void writeUTF( String s ) throws IOException {
		throw makeUnimplemented();
	}

	public void readFully( byte b[] ) throws IOException {
		try {
			buf.get( b );
		} catch ( BufferUnderflowException e ) {
			throw makeUnderflow( e );
		}
	}

	public void readFully( byte b[], int off, int len ) throws IOException {
		try {
			buf.get( b, off, len );
		} catch ( BufferUnderflowException e ) {
			throw makeUnderflow( e );
		}
	}

	public int skipBytes( int n ) throws IOException {
		//TODO: we could implement a partial skip rather than throwing exception if there aren't enough bytes
		buf.position( buf.position() + n );
		return n;
	}

	public boolean readBoolean() throws IOException {
		try {
			return buf.get() != 0;
		} catch ( BufferUnderflowException e ) {
			throw makeUnderflow( e );
		}
	}

	public byte readByte() throws IOException {
		try {
			return buf.get();
		} catch ( BufferUnderflowException e ) {
			throw makeUnderflow( e );
		}
	}

	public int readUnsignedByte() throws IOException {
		try {
			return buf.get() & 0xFF;
		} catch ( BufferUnderflowException e ) {
			throw makeUnderflow( e );
		}
	}

	public short readShort() throws IOException {
		try {
			return buf.getShort();
		} catch ( BufferUnderflowException e ) {
			throw makeUnderflow( e );
		}
	}

	public int readUnsignedShort() throws IOException {
		try {
			return buf.getShort() & 0xFFFF;
		} catch ( BufferUnderflowException e ) {
			throw makeUnderflow( e );
		}
	}

	public char readChar() throws IOException {
		try {
			return buf.getChar();
		} catch ( BufferUnderflowException e ) {
			throw makeUnderflow( e );
		}
	}

	public int readInt() throws IOException {
		try {
			return buf.getInt();
		} catch ( BufferUnderflowException e ) {
			throw makeUnderflow( e );
		}
	}

	public long readLong() throws IOException {
		try {
			return buf.getLong();
		} catch ( BufferUnderflowException e ) {
			throw makeUnderflow( e );
		}
	}

	public float readFloat() throws IOException {
		try {
			return buf.getFloat();
		} catch ( BufferUnderflowException e ) {
			throw makeUnderflow( e );
		}
	}

	public double readDouble() throws IOException {
		try {
			return buf.getDouble();
		} catch ( BufferUnderflowException e ) {
			throw makeUnderflow( e );
		}
	}

	public String readLine() throws IOException {
		throw makeUnimplemented();
	}

	public String readUTF() throws IOException {
		throw makeUnimplemented();
	}

	private static UnsupportedOperationException makeUnimplemented() {
		return new UnsupportedOperationException( "Not yet implemented; please implement" );
	}

	private static IOException makeOverflow( BufferOverflowException e ) {
		return new IOException( "Could not fit data in remaining buffer space", e );
	}

	private static IOException makeUnderflow( BufferUnderflowException e ) {
		return new IOException( "not enough data to be read", e );
	}
}
