/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.io;

import java.io.DataOutput;
import java.io.IOException;

/**
 * DataOutputWritable is an object that can be written to a DataOutput. This interface differs from
 * Serializable and Externizable in that DataOutputWritable objects are not identified in the
 * stream, so the type of the next object must be known. This is different if it is a
 * DataInputReadable, which can be loaded dynamically.
 * <p/>
 * Classes that implement DataOutputWritable should implement a method static public T read(
 * DataInput input ) throws IOException, which returns an instance that is the serialized form
 * written by the {@link #write(DataOutput)} method by reading it from a DataInput. Classes that
 * are DataOutputWritable, but not DataInputReadable, may marked final to prevent any possible
 * problems with serializing a subclass and trying to load the base version.
 *
 * @author Jason Winnebeck
 */
public interface DataOutputWritable {
	/**
	 * Writes this object to the given DataOutput. This method should not have any side-effects on this
	 * object as it may be used to discover sizes before writing.
	 *
	 * @param output non-null DataOutput
	 */
	public void write( DataOutput output ) throws IOException;
}
