/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.io;

import java.io.ByteArrayOutputStream;
import java.io.DataInput;
import java.io.DataOutput;
import java.io.DataOutputStream;
import java.io.IOException;

/**
 * DowUtil has some utilities for working with {@link DataOutputWritable} objects.
 *
 * @author Jason Winnebeck
 */
public class DowUtil {
	private DowUtil() {}

	/**
	 * Returns the size of the serialized DataOutputWritable object, as determined by counting the
	 * calls made by {@link DataOutputWritable#write(DataOutput)}.
	 *
	 * @param dow non-null DataOutputWritable
	 * @return length of byte array that would be returned by {@link #serialize(DataOutputWritable)}.
	 *
	 * @throws IOException if the object's write method throws. The DataOutput given to write will
	 *                     not cause an exception.
	 */
	public static int getSize( DataOutputWritable dow ) throws IOException {
		SizingDataOutput output = new SizingDataOutput();
		dow.write( output );
		return output.getSize();
	}

	/**
	 * Serializes the given DataOutputWritable to a byte array.
	 *
	 * @param dow non-null DataOutputWritable
	 * @return non-null byte[]
	 *
	 * @throws IOException if {@link DataOutputWritable#write(DataOutput)} throws
	 */
	public static byte[] serialize( DataOutputWritable dow ) throws IOException {
		ByteArrayOutputStream stream = new ByteArrayOutputStream();
		DataOutputStream output = new DataOutputStream( stream );
		dow.write( output );
		output.close();
		//Don't need to close ByteArrayOutputStream stream, per docs of DataOutputStream.close
		return stream.toByteArray();
	}

	/**
	 * Returns a Class represented by the next UTF string in the DataInput. The returned class is
	 * loaded, but not initialized. If the class cannot be found, IOException is thrown, a more
	 * appropriate exception when loading from the input. The class returned must be exactly base, or
	 * extend from/implement the class or interface 'base'.
	 *
	 * @param input non-null DataInput
	 * @param base  returned class must be assignable to base
	 *
	 * @return non-null Class&lt;? extends base&gt;
	 *
	 * @throws IOException if the class cannot be found or the string cannot be read
	 */
	@SuppressWarnings( { "unchecked" } )
	public static <T> Class<? extends T> readClass( DataInput input, Class<T> base )
			throws IOException {
		String className = input.readUTF();
		try {
			Class<?> c = Class.forName( className, false, DowUtil.class.getClassLoader() );
			if ( base.isAssignableFrom( c ) )
				return (Class<? extends T>) c;
			else
				throw new IOException( c + " does not extend/implement " + base );
		} catch ( ClassNotFoundException e ) {
			throw new IOException( e );
		}
	}

	public static <T extends DataInputReadable> T readObject( DataInput input, Class<T> base )
			throws IOException {
		Class<? extends T> clazz = readClass( input, base );
		T ret = null;
		try {
			ret = clazz.newInstance();
		} catch ( InstantiationException e ) {
			throw new IOException( clazz + " does not properly implement DataInputReadable (public no-args ctor)" );
		} catch ( IllegalAccessException e ) {
			throw new IOException( clazz + " does not properly implement DataInputReadable (public no-args ctor)" );
		}
		ret.read( input );
		return ret;
	}

	public static void writeObject( DataOutput output, DataInputReadable object ) throws IOException {
		output.writeUTF( object.getClass().getName() );
		object.write( output );
	}
}
