/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.io;

import org.gillius.Util.CheckUtils;
import org.gillius.realdb.metadata.ElementType;
import org.gillius.realdb.model.data.UnsignedByte;
import org.gillius.realdb.model.data.UnsignedInteger;
import org.gillius.realdb.model.data.UnsignedShort;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;
import java.util.HashMap;
import java.util.Map;

/**
 * ElementSerializerFactory creates {@link ElementSerializer} objects that work with a specific
 * {@link ElementType}.
 *
 * @author Jason Winnebeck
 */
public class ElementSerializerFactory {
	private static final Map<ElementType, ElementSerializer> serializers =
			new HashMap<ElementType, ElementSerializer>();

	static {
		serializers.put( ElementType.Boolean, new ElementSerializer() {
			public void writeValue( Object value, DataOutput out ) throws IOException {
				if ( value == null )
					out.writeBoolean( false );
				else
					out.writeBoolean( (Boolean) value );
			}

			public Object readValue( DataInput in ) throws IOException {
				return in.readBoolean();
			}

			public int getWrittenSize() {
				return 1;
			}
		} );
		serializers.put( ElementType.Double, new ElementSerializer() {
			public void writeValue( Object value, DataOutput out ) throws IOException {
				if ( value == null )
					out.writeDouble( 0.0 );
				else
					out.writeDouble( (Double) value );
			}

			public Object readValue( DataInput in ) throws IOException {
				return in.readDouble();
			}

			public int getWrittenSize() {
				return 8;
			}
		} );
		serializers.put( ElementType.Float, new ElementSerializer() {
			public void writeValue( Object value, DataOutput out ) throws IOException {
				if ( value == null )
					out.writeFloat( 0.0f );
				else
					out.writeFloat( (Float) value );
			}

			public Object readValue( DataInput in ) throws IOException {
				return in.readFloat();
			}

			public int getWrittenSize() {
				return 4;
			}
		} );
		serializers.put( ElementType.SInt8, new ElementSerializer() {
			public void writeValue( Object value, DataOutput out ) throws IOException {
				if ( value == null )
					out.writeByte( 0 );
				else
					out.writeByte( (Byte) value );
			}

			public Object readValue( DataInput in ) throws IOException {
				return in.readByte();
			}

			public int getWrittenSize() {
				return 1;
			}
		} );
		serializers.put( ElementType.SInt16, new ElementSerializer() {
			public void writeValue( Object value, DataOutput out ) throws IOException {
				if ( value == null )
					out.writeShort( 0 );
				else
					out.writeShort( (Short) value );
			}

			public Object readValue( DataInput in ) throws IOException {
				return in.readShort();
			}

			public int getWrittenSize() {
				return 2;
			}
		} );
		serializers.put( ElementType.SInt32, new ElementSerializer() {
			public void writeValue( Object value, DataOutput out ) throws IOException {
				if ( value == null )
					out.writeInt( 0 );
				else
					out.writeInt( (Integer) value );
			}

			public Object readValue( DataInput in ) throws IOException {
				return in.readInt();
			}

			public int getWrittenSize() {
				return 4;
			}
		} );
		serializers.put( ElementType.SInt64, new ElementSerializer() {
			public void writeValue( Object value, DataOutput out ) throws IOException {
				if ( value == null )
					out.writeLong( 0 );
				else
					out.writeLong( (Long) value );
			}

			public Object readValue( DataInput in ) throws IOException {
				return in.readLong();
			}

			public int getWrittenSize() {
				return 8;
			}
		} );
		serializers.put( ElementType.UInt8, new ElementSerializer() {
			public void writeValue( Object value, DataOutput out ) throws IOException {
				if ( value == null )
					out.writeByte( 0 );
				else
					out.writeByte( ( (UnsignedByte) value ).byteValue() );
			}

			public Object readValue( DataInput in ) throws IOException {
				return new UnsignedByte( in.readByte() );
			}

			public int getWrittenSize() {
				return 1;
			}
		} );
		serializers.put( ElementType.UInt16, new ElementSerializer() {
			public void writeValue( Object value, DataOutput out ) throws IOException {
				if ( value == null )
					out.writeShort( 0 );
				else
					out.writeShort( ( (UnsignedShort) value ).shortValue() );
			}

			public Object readValue( DataInput in ) throws IOException {
				return new UnsignedShort( in.readShort() );
			}

			public int getWrittenSize() {
				return 2;
			}
		} );
		serializers.put( ElementType.UInt32, new ElementSerializer() {
			public void writeValue( Object value, DataOutput out ) throws IOException {
				if ( value == null )
					out.writeInt( 0 );
				else
					out.writeInt( ( (UnsignedInteger) value ).intValue() );
			}

			public Object readValue( DataInput in ) throws IOException {
				return new UnsignedInteger( in.readInt() );
			}

			public int getWrittenSize() {
				return 4;
			}
		} );

		if ( serializers.size() != ElementType.values().length )
			throw new IllegalStateException( "There isn't a writer for every ElementType" );
	}

	/**
	 * Returns an DataOutputWriter that writes values for the given ElementType.
	 *
	 * @param type non-null ElementType
	 * @return non-null DataOutputWriter
	 */
	public static DataOutputWriter getWriter( ElementType type ) {
		CheckUtils.checkNull( type, "type" );
		//This shouldn't return null due to our assert made in the static constructor.
		return serializers.get( type );
	}

	/**
	 * Returns an DataInputReader that reads values for the given ElementType.
	 *
	 * @param type non-null ElementType
	 * @return non-null DataInputReader
	 */
	public static DataInputReader getReader( ElementType type ) {
		CheckUtils.checkNull( type, "type" );
		//This shouldn't return null due to our assert made in the static constructor.
		return serializers.get( type );
	}

	/**
	 * Returns an ElementSerializer that serializes values for the given ElementType.
	 *
	 * @param type non-null ElementType
	 * @return non-null ElementSerializer
	 */
	public static ElementSerializer getSerializer( ElementType type ) {
		CheckUtils.checkNull( type, "type" );
		//This shouldn't return null due to our assert made in the static constructor.
		return serializers.get( type );
	}
}
