/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.io;

import java.util.Random;

/**
 * FailingBlockFile wraps an existing BlockFile to produce controllable failures. When a failure
 * occurs, the block currently being written is set to random data. It is assumed that the wrapped
 * file itself does not fail.
 * <p/>
 * By default, the FailingBlockFile never fails any operations. By using methods such as
 * {@link #setFailureCount(int)}, we can say when to invoke a failure.
 *
 * @author Jason Winnebeck
 */
public class FailingBlockFile extends AbstractFailingBlockFile {

	private int count = -1;

	/**
	 * Creates a new FailingBlockFile by wrapping a real BlockFile.
	 *
	 * @param wrapped non-null BlockFile
	 * @param random  non-null source of randomness, to generate consistent "random" data on failures.
	 */
	public FailingBlockFile( BlockFile wrapped, Random random ) {
		super( random, wrapped );
	}

	/**
	 * Sets the failure count for the FailingBlockFile. Each time an operation occurs, the
	 * counter is decremented and checked to see if it is 0. If the counter is 0, then a failure
	 * occurs. Setting the count to 1 will cause the next read or write method to fail. Setting the
	 * count to 0 or a negative number will disable failures entirely. Once a failure occurs, it will
	 * not occur again until this method is called again.
	 * <p/>
	 * Methods not throwing IOException never fail.
	 */
	public void setFailureCount( int count ) {
		this.count = count;
	}

	protected boolean checkFailure() {
		if ( count > 0 ) {
			--count;
			return count == 0;

		} else {
			return false;
		}
	}
}
