/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.io;

import org.gillius.realdb.metadata.OptionalElementType;
import org.gillius.realdb.model.data.Discontinuity;
import org.gillius.realdb.model.data.GenericRecord;
import org.gillius.realdb.model.data.Record;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;
import java.util.Arrays;

/**
 * GenericRecordSerializer implements a generic Record serialization technique.
 *
 * @author Jason Winnebeck
 */
public class GenericRecordSerializer implements RecordSerializer {

	private static final int DISCONTINUITY_BIT = 0;
	private static final int FIRST_FIELD_BIT = 1;
	private static final int REQUIRED_ELEMENT = -1;

	/**
	 * Size of the bit field portion in bytes.
	 */
	private final int bitFieldSize;
	/**
	 * Size of the serialized record elements in bytes.
	 */
	private final int dataSize;
	/**
	 * Size of the entire record (including timestamp, bit field, and elements) in bytes.
	 */
	private final int recordSize;
	private final ElementSerializer[] serializers;
	private final int[] bitMapping;

	private final byte[] discontinuityData;

	/**
	 * Constructs a new GenericRecordSerializer to serialize GenericRecords that have the elements
	 * given.
	 *
	 * @param elements non-null array of ElementType with the types of each element in the
	 *                 GenericRecords to be read/written in the same order.
	 */
	public GenericRecordSerializer( OptionalElementType[] elements ) {
		serializers = new ElementSerializer[elements.length];
		bitMapping  = new int[elements.length];
		Arrays.fill( bitMapping, REQUIRED_ELEMENT );

		int elementsDataSize = 0;
		int bits = FIRST_FIELD_BIT; //need at least 1 bit for discontinuity
		for ( int i = 0; i < elements.length; i++ ) {
			if ( elements[i].isOptional() ) {
				bitMapping[i] = bits;
				++bits;
			}
			serializers[i] = ElementSerializerFactory.getSerializer( elements[i].getType() );
			elementsDataSize += serializers[i].getWrittenSize();
		}

		BitField bf = new BitField( bits );
		bitFieldSize = bf.getData().length;
		dataSize     = elementsDataSize;
		recordSize   = 8 + bitFieldSize + dataSize; //8 for the timestamp size

		//Create the discontinuityData field to speed up writing such records
		discontinuityData = new byte[ bitFieldSize + dataSize ];
		bf.set( DISCONTINUITY_BIT );
		System.arraycopy( bf.getData(), 0, discontinuityData, 0, bitFieldSize );
	}

	public int getWrittenSize() {
		return recordSize;
	}

	public void writeRecord( Record r, DataOutput out ) throws IOException {
		out.writeLong( r.getTime() ); //Required by interface spec

		if ( r.isDiscontinuity() ) {
			out.write( discontinuityData );

		} else {
			BitField bf = new BitField( new byte[bitFieldSize] );

			//Set the null bits for each null field
			for ( int i = 0; i < serializers.length; ++i ) {
				if ( r.getElement( i ) == null )
					bf.set( bitMapping[i] );
			}
			out.write( bf.getData() );

			//Write actual record elements (with placeholders for nulls)
			for ( int i = 0; i < serializers.length; ++i ) {
				serializers[i].writeValue( r.getElement( i ), out );
			}
		}
	}

	public Record readRecord( DataInput in ) throws IOException {
		long time = in.readLong();
		byte[] bfData = new byte[bitFieldSize];
		in.readFully( bfData );
		BitField bf = new BitField( bfData );

		if ( bf.get( DISCONTINUITY_BIT ) ) {
			in.skipBytes( dataSize );
			return new Discontinuity( time );

		} else {
			Object[] values = new Object[serializers.length];
			for ( int i = 0; i < serializers.length; ++i ) {
				if ( bitMapping[i] != REQUIRED_ELEMENT && bf.get( bitMapping[i] ) ) {
					values[i] = null;
					serializers[i].readValue( in ); //just to skip over those bytes
				} else {
					values[i] = serializers[i].readValue( in );
				}
			}

			return new GenericRecord( time, values );
		}
	}
}
