/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.io;

import java.io.IOException;

/**
 * ProfilingBlockFile wraps an existing BlockFile implementation to provide profiling information
 * such as the number of writes or reads. A call that throws an exception is counted as a call, but
 * not as blocks read/written. The close call is not counted as anything.
 *
 * @author Jason Winnebeck
 */
public class ProfilingBlockFile implements BlockFile {
	private final BlockFile wrapped;

	private int readCalls     = 0;
	private int writeCalls    = 0;
	private int readBlocks    = 0;
	private int writtenBlocks = 0;

	public ProfilingBlockFile( BlockFile wrapped ) {
		this.wrapped = wrapped;
	}

	/**
	 * Returns the number of read calls made.
	 */
	public int getReadCalls() {
		return readCalls;
	}

	/**
	 * Returns the number of write calls made.
	 */
	public int getWriteCalls() {
		return writeCalls;
	}

	/**
	 * Returns number of blocks read.
	 */
	public int getReadBlocks() {
		return readBlocks;
	}

	/**
	 * Returns number of blocks written.
	 */
	public int getWrittenBlocks() {
		return writtenBlocks;
	}

	public int getSize() {
		return wrapped.getSize();
	}

	public int getBlockSize() {
		return wrapped.getBlockSize();
	}

	public void readBlock( int num, byte[] data ) throws IOException {
		++readCalls;
		++readBlocks;
		wrapped.readBlock( num, data );
	}

	public int readBlocks( int startBlock, byte[] data ) throws IOException {
		++readCalls;
		int ret = wrapped.readBlocks( startBlock, data );
		readBlocks += ret;
		return ret;
	}

	public void writeBlock( int num, byte[] data ) throws IOException {
		++writeCalls;
		++writtenBlocks;
		wrapped.writeBlock( num, data );
	}

	public int writeBlocks( int startBlock, byte[] data ) throws IOException {
		++writeCalls;
		int ret = wrapped.writeBlocks( startBlock, data );
		writtenBlocks += ret;
		return ret;
	}

	public int writeBlocks( int startBlock, byte[] data, int start, int len ) throws IOException {
		++writeCalls;
		int ret = wrapped.writeBlocks( startBlock, data, start, len );
		writtenBlocks += ret;
		return ret;
	}

	public void close() throws IOException {
		wrapped.close();
	}
}
