/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.io;

import java.io.EOFException;
import java.io.IOException;
import java.io.RandomAccessFile;

/**
 * RandomAccessBlockFile implements the BlockFile interface through a given RandomAccessFile, which
 * should be opened in "rwd" or "rws" mode as appropriate. If the JVM and operating system do not
 * implement these modes properly, it is possible that RandomAccessBlockFile cannot completely make
 * the write behavior guarantees. On Windows XP and Linux using the Sun J2SE 1.6, this does appear
 * to work based on my research.
 * <p/>
 * An instance of this class cannot be shared between threads.
 * <p/>
 * TODO: this class may need an "offset" attribute to fully guarantee the writeBlock behavior
 *
 * @author Jason Winnebeck
 */
public class RandomAccessBlockFile implements BlockFile {
	private final RandomAccessFile file;
	private final int blockSize;

	/**
	 * Constant buffer used for zero byte padding in writeBlocks.
	 */
	private final byte[] zeroBuf;

	/**
	 * Cached size of the file, to prevent IOExceptions.
	 */
	private final int numBlocks;

	/**
	 * Constructs a new RandomAccessBlockFile to work with the given file and block size. The file
	 * must be preallocated, as the number of blocks is determined by the file length divided by the
	 * block size. The partial remaining block, if it exists, is not accessible from this BlockFile.
	 */
	public RandomAccessBlockFile( RandomAccessFile file, int blockSize ) throws IOException {
		this( file, blockSize, (int)( file.length() / blockSize ) );
	}

	/**
	 * Constructs a new RandomAccessBlockFile to work with the given parameters. The file must be
	 * preallocated, and large enough to hold 'numBlocks' blocks. Any leftover space will be
	 * unaccessible.
	 */
	public RandomAccessBlockFile( RandomAccessFile file, int blockSize, int numBlocks )
			throws IOException {
		int fileBlocks = (int) ( file.length() / blockSize );
		if ( numBlocks > fileBlocks ) {
			throw new IllegalArgumentException( "There is not enough space in the file for " +
			                                    numBlocks + " blocks." );
		}

		this.file = file;
		this.blockSize = blockSize;
		this.numBlocks = numBlocks;

		zeroBuf = new byte[blockSize];
	}

	public int getSize() {
		return numBlocks;
	}

	public void close() throws IOException {
		file.close();
	}

	public int getBlockSize() {
		return blockSize;
	}

	public void readBlock( int num, byte[] data ) throws IOException {
		if ( num < 0 || num >= numBlocks ) {
			throw new IllegalArgumentException( "block " + num + " is not in the valid range of " +
			                                    "blocks from 0 to " + ( getSize() - 1 ) );
		}
		if ( data.length != blockSize )
			throw new IllegalArgumentException( "data block must be block size = " + blockSize );

		file.seek( num * blockSize );
		file.readFully( data );
	}

	public int readBlocks( int startBlock, byte[] data ) throws IOException {
		if ( startBlock < 0 )
			throw new IllegalArgumentException( "startBlock must be non-negative" );
		if ( data.length > (numBlocks - startBlock) * blockSize ) {
			throw new EOFException( "Given buffer with " + data.length + " bytes is too large for " +
			                        "starting block " + startBlock );
		}

		int leftOver = data.length % blockSize;
		int numRead = data.length / blockSize;
		if ( leftOver > 0 )
			++numRead;

		file.seek( startBlock * blockSize );
		file.readFully( data );

		return numRead;
	}

	public void writeBlock( int num, byte[] data ) throws IOException {
		if ( num < 0 || num >= numBlocks ) {
			throw new IllegalArgumentException( "block " + num + " is not in the valid range of " +
			                                    "blocks from 0 to " + ( getSize() - 1 ) );
		}
		if ( data.length != blockSize )
			throw new IllegalArgumentException( "data block must be block size = " + blockSize );

		file.seek( num * blockSize );
		file.write( data );
	}

	public int writeBlocks( int startBlock, byte[] data ) throws IOException {
		return writeBlocks( startBlock, data, 0, data.length );
	}

	public int writeBlocks( int startBlock, byte[] data, int start, int len ) throws IOException {
		if ( startBlock < 0 )
			throw new IllegalArgumentException( "startBlock must be non-negative" );
		if ( len > ( numBlocks - startBlock ) * blockSize ) {
			throw new EOFException( "Given buffer with " + data.length + " bytes will not fit at " +
			                        "starting block " + startBlock );
		}

		int leftOver   = len % blockSize;
		int numWritten = len / blockSize;
		if ( leftOver > 0 )
			++numWritten;

		file.seek( startBlock * blockSize );
		file.write( data, start, len );
		if ( leftOver > 0 )
			file.write( zeroBuf, 0, ( blockSize - leftOver ) );

		return numWritten;
	}
}
