/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.io;

import java.io.DataOutput;
import java.io.IOException;
import java.io.UTFDataFormatException;

/**
 * SizingDataOutput is a DataOutput implementation that only counts bytes, and does not actually
 * store any written data. None of the methods in SizingDataOutput throw IOException except for
 * writeUTF, despite the interface in DataOutput. writeUTF will throw if the string is too large as
 * per the interface definition.
 *
 * @author Jason Winnebeck
 */
public class SizingDataOutput implements DataOutput {
	private int size = 0;

	/**
	 * Constructs a new SizingDataOutput with the size set to 0.
	 */
	public SizingDataOutput() {
	}

	/**
	 * Returns the number of bytes "written" since construction or last {@link #reset()}.
	 */
	public int getSize() {
		return size;
	}

	/**
	 * Resets the size to 0.
	 */
	public void reset() {
		size = 0;
	}

	public void write( int b ) {
		++size;
	}

	public void write( byte b[] ) {
		size += b.length;
	}

	public void write( byte b[], int off, int len ) {
		size += len;
	}

	public void writeBoolean( boolean v ) {
		++size;
	}

	public void writeByte( int v ) {
		++size;
	}

	public void writeShort( int v ) {
		size += 2;
	}

	public void writeChar( int v ) {
		size += 2;
	}

	public void writeInt( int v ) {
		size += 4;
	}

	public void writeLong( long v ) {
		size += 8;
	}

	public void writeFloat( float v ) {
		size += 4;
	}

	public void writeDouble( double v ) {
		size += 8;
	}

	public void writeBytes( String s ) {
		size += s.length();
	}

	public void writeChars( String s ) {
		size += s.length() * 2;
	}

	public void writeUTF( String s ) throws IOException {
		//The following is written per the documentation of DataOutput's writeUTF
		int strLen = 0;
		for ( int i = 0; i < s.length(); ++i ) {
			char c = s.charAt( i );
			if ( c <= 0x7f ) {
				++strLen;
			} else if ( c < 0x7ff ) {
				strLen += 2;
			} else {
				strLen += 4;
			}
		}
		if ( strLen > 65535 )
			throw new UTFDataFormatException( "String given is too large" );
		else
			size += 2 + strLen;
	}
}
