/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.io;

import java.io.IOException;
import java.util.Random;

/**
 * TriggeredFailingBlockFile is an AbstractFailingBlockFile implementation that allows a flag to be
 * set at any time to causes all further operations (except close) to fail. Like other BlockFile
 * implementations, TriggeredFailingBlockFile is generally not thread-safe. The one exception is
 * that the {@link #setFailureFlag()} may be called from a different thread.
 *
 * @author Jason Winnebeck
 */
public class TriggeredFailingBlockFile extends AbstractFailingBlockFile {
	private volatile boolean fail;

	/**
	 * Creates a new TriggeredFailingBlockFile by wrapping a real BlockFile. The failure flag is false
	 * by default.
	 *
	 * @param wrapped non-null BlockFile
	 * @param random  non-null source of randomness, to generate consistent "random" data on failures.
	 */
	public TriggeredFailingBlockFile( Random random, BlockFile wrapped ) {
		super( random, wrapped );
	}

	/**
	 * For a TriggeredFailingBlockFile, close never simulates an exception being thrown. It throws
	 * only if the underlying file throws when closing.
	 */
	public void close() throws IOException {
		wrapped.close();
	}

	public void setFailureFlag() {
		fail = true;
	}

	protected boolean checkFailure() {
		return fail;
	}
}
