/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.io.rdl;

import org.antlr.runtime.ANTLRFileStream;
import org.antlr.runtime.ANTLRStringStream;
import org.antlr.runtime.CommonTokenStream;
import org.antlr.runtime.RecognitionException;
import org.gillius.realdb.BadFileException;
import org.gillius.realdb.metadata.DataStreamInfo;
import org.gillius.realdb.metadata.ElementCodec;
import org.gillius.realdb.metadata.ElementType;
import org.gillius.realdb.metadata.RecordElement;
import org.gillius.realdb.model.algorithms.DeadbandAlgorithm;
import org.gillius.realdb.model.algorithms.SampledAlgorithm;
import org.gillius.realdb.model.algorithms.StepAlgorithm;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;

/**
 * RDLFile is a representation of an RDL file that the {@link RDBDefinitionParser} creates.
 * <p/>
 * The main purpose of this class is to also be a handler/builder for the ANTLR-generated parser, to
 * have the least amount of Java code in it as possible. As such, the design of this class is
 * specific to the design of the parser, so they are tightly-coupled. This makes RDLFile not as
 * generic as it could be; however, it converts into the metadata objects, which are loosely
 * coupled.
 * <p/>
 * TODO: Add better error handling and checking
 *
 * @author Jason Winnebeck
 */
public class RDLFile {
	private final Map<String, String> optionsMap = new HashMap<String, String>();
	private final List<DataStreamInfo> streamInfos = new ArrayList<DataStreamInfo>();

	public Map<String, String> getOptionsMap() {
		return optionsMap;
	}

	public List<DataStreamInfo> getStreamInfos() {
		return streamInfos;
	}

	public static RDLFile load( String s ) throws BadFileException {
		String errorMessage = "Could not load given RDL";
		RDBDefinitionLexer lexer = new RDBDefinitionLexer( new ANTLRStringStream( s ) );
		return doLoad( errorMessage, lexer );
	}

	public static RDLFile loadFile( File f ) throws IOException, BadFileException {
		String errorMessage = "Could not load RDL file at " + f.getAbsolutePath();
		RDBDefinitionLexer lexer = new RDBDefinitionLexer( new ANTLRFileStream( f.getPath() ) );
		return doLoad( errorMessage, lexer );
	}

	private static RDLFile doLoad( String errorMessage, RDBDefinitionLexer lexer )
			throws BadFileException {
		CommonTokenStream tokens = new CommonTokenStream( lexer );
		RDBDefinitionParser parser = new RDBDefinitionParser( tokens );

		try {
			RDLFile ret = parser.statements();
			if ( !ret.errors.isEmpty() ) {
				throw new BadFileException( "Encountered errors when loading RDL file: " + ret.errors );
			}
			return ret;

		} catch ( RecognitionException e ) {
			throw new BadFileException( errorMessage, e );

		} catch ( RDLFileException e ) {
			throw new BadFileException( e );
		}
	}

	//Builder portion of this class
	private static final Map<String, Class<? extends ElementCodec>> CODECS =
			new HashMap<String, Class<? extends ElementCodec>>();
	private static final Map<String, ElementType> TYPE_MAP =
			new HashMap<String, ElementType>();

	static {
		CODECS.put( "StepAlgorithm", StepAlgorithm.class );
		CODECS.put( "SampledAlgorithm", SampledAlgorithm.class );
		CODECS.put( "DeadbandAlgorithm", DeadbandAlgorithm.class );

		for ( ElementType type : ElementType.values() ) {
			TYPE_MAP.put( type.name().toUpperCase( Locale.ENGLISH ), type );
		}
	}

	private final List<String> errors = new ArrayList<String>();
	private List<RecordElement> elements = new ArrayList<RecordElement>();
	private Map<String, String> codecParams = new HashMap<String, String>();

	void reportError( String error ) {
		errors.add( error );
	}

	void addOption( String name, String value ) {
		//TODO: check for duplicate keys
		optionsMap.put( name, value );
	}

	void addCodecParam( String name, String value ) {
		codecParams.put( name, value );
	}

	@SuppressWarnings( { "unchecked" } )
	void addElement( String name, String typeName, boolean isNull, String codec ) {
		try {
			ElementType type = TYPE_MAP.get( typeName.toUpperCase( Locale.ENGLISH ) );
			Class<? extends ElementCodec> c = null;
			if ( codec == null )
				c = SampledAlgorithm.class;

			if ( c == null )
				c = CODECS.get( codec );

			if ( c == null ) {
				//try to resolve it as a classname
				Class<?> temp = Class.forName( codec );
				if ( ElementCodec.class.isAssignableFrom( temp ) )
					c = (Class<? extends ElementCodec>) temp;
				else
					throw new RDLFileException( "Codec " + codec + " is not a valid ElementCodec (does not implement ElementCodec)" );
			}

			if ( c != null ) {
				ElementCodec codecInstance = c.newInstance();
				codecInstance.setParams( codecParams );
				elements.add( new RecordElement( elements.size(), name, codecInstance, type, isNull ) );
			}

			codecParams.clear();

		} catch ( ClassNotFoundException e ) {
			throw new RDLFileException( "Codec " + codec + " is not a valid ElementCodec", e );
		} catch ( IllegalAccessException e ) {
			throw new RDLFileException( "Codec " + codec + " is not a valid ElementCodec (IllegalAccessException)", e );
		} catch ( InstantiationException e ) {
			throw new RDLFileException( "Could not create codec " + codec, e );
		}
	}

	void addStream( String name, String id ) {
		RecordElement[] elementsArray = elements.toArray( new RecordElement[elements.size()] );
		streamInfos.add( new DataStreamInfo( Integer.valueOf( id ), name, elementsArray ) );
		elements.clear();
	}

	@SuppressWarnings( { "UncheckedExceptionClass" } )
	private static class RDLFileException extends RuntimeException {
		private RDLFileException( String message ) {
			super( message );
		}

		private RDLFileException( String message, Throwable cause ) {
			super( message, cause );
		}
	}
}
