/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.metadata;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

/**
 * AbstractElementCodec provides functionaly to better implement the ElementCodec interface. It
 * provides a method to check the parameters to {@link #setParams(Map)} and it also implements the
 * read/write methods suitable for a stateless ElementCodec that takes no parameters.
 *
 * @author Jason Winnebeck
 */
public abstract class AbstractElementCodec implements ElementCodec {
	public void getRequiredParams( Set<String> params ) {
	}

	public void getOptionalParams( Set<String> params ) {
	}

	/**
	 * Checks the given parameters to make sure that all required parameters exist, and that no
	 * unrecognized parameters are present.
	 *
	 * @param params non-null parameter map
	 *
	 * @throws IllegalArgumentException If a required parameter is not provided, or if a provided
	 *                                  parameter is not recognized.
	 */
	public void checkParams( Map<String, String> params ) throws IllegalArgumentException {
		Set<String> required = new HashSet<String>();
		getRequiredParams( required );
		required.removeAll( params.keySet() );
		if ( !required.isEmpty() ) {
			throw new IllegalArgumentException( "Parameters " + required + " are required for " +
			                                    getName() );
		}

		Set<String> allAcceptableParams = new HashSet<String>();
		getRequiredParams( allAcceptableParams );
		getOptionalParams( allAcceptableParams );

		Set<String> providedParams = new HashSet<String>( params.keySet() );
		providedParams.removeAll( allAcceptableParams );
		if ( !providedParams.isEmpty() ) {
			throw new IllegalArgumentException( "Parameters " + providedParams + " are unknown to " +
			                                    getName() );
		}
	}

	/**
	 * The default implementation for setParams simply calls {@link #checkParams(Map)}.
	 */
	public void setParams( Map<String, String> params ) throws IllegalArgumentException {
		checkParams( params );
	}

	/**
	 * Returns the name of this ElementCodec to use in exception messages. By default, returns the
	 * name of the class for this object.
	 */
	protected String getName() {
		return getClass().getName();
	}

	public void read( DataInput input ) throws IOException {
	}

	public void write( DataOutput output ) throws IOException {
	}
}
