/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.metadata;

import org.gillius.Util.CheckUtils;
import org.gillius.Util.StringPrintWriter;
import org.gillius.realdb.io.DataOutputWritable;
import org.gillius.realdb.model.data.Record;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;
import java.util.Arrays;

/**
 * DataStreamInfo describes information about a DataStream. A new DataStream could be constructed
 * from the combination of this info and other database elements such as its index.
 *
 * @author Jason Winnebeck
 */
public final class DataStreamInfo implements DataOutputWritable {
	private final int userId;

	private final String name;

	private final RecordElement[] elements;

	public DataStreamInfo( int userId, String name, RecordElement[] elements ) {
		CheckUtils.checkNull( name,        "name"        );
		CheckUtils.checkNull( elements,    "elements"    );

		this.userId      = userId;
		this.name        = name;
		this.elements    = elements;
	}

	public int getUserId() {
		return userId;
	}

	public String getName() {
		return name;
	}

	/**
	 * Returns an array of all of the elements in this stream, which must not be modified.
	 */
	public RecordElement[] getElements() {
		return elements;
	}

	/**
	 * Returns the RecordElement with the given name, or null if there is no such element.
	 */
	public RecordElement getElement( String name ) {
		for ( RecordElement element : elements ) {
			if ( element.getName().equals( name ) )
				return element;
		}
		return null;
	}
	/**
	 * Checks if the given Record is proper for this stream. Generates an exception describing the
	 * problem(s) with the Record, if it is not proper. This method checks that the record has the
	 * appropriate number of elements, that they are of the right type, and that only nullable
	 * elements are null.
	 *
	 * @param r non-null Record
	 *
	 * @throws IllegalArgumentException if Record is null or is not proper for this stream.
	 */
	public void checkRecord( Record r ) throws IllegalArgumentException {
		CheckUtils.checkNull( r, "r" );

		if ( !r.isDiscontinuity() ) {
			if ( r.getNumElements() != elements.length ) {
				throw new IllegalArgumentException( "Given record has " + r.getNumElements() +
				                                    " elements, but " + elements.length + " are expected." );
			}

			StringPrintWriter out = new StringPrintWriter();
			for ( RecordElement element : elements ) {
				final int ePos = element.getPosition();
				if ( r.getElement( ePos ) == null ) {
					if ( !element.isOptional() )
						out.println( eToS( element ) + " is not optional, but null was given" );
				} else {
					final Class<?> rClass = r.getElement( ePos ).getClass();
					if ( !element.getType().getTypeClass().isAssignableFrom( rClass ) )
						out.println( rClass + " is not compatible with " + eToS( element ) );
				}
			}
			if ( out.wasUsed() ) {
				throw new IllegalArgumentException( String.format( "%s is not compatible with %s:%n%s",
			                                                     r, this, out ) );
			}
		}
	}

	public void write( DataOutput output ) throws IOException {
		output.writeInt( userId );
		output.writeUTF( name );
		output.writeInt( elements.length );
		for ( RecordElement element : elements ) {
			element.write( output );
		}
	}

	public static DataStreamInfo read( DataInput input ) throws IOException {
		int    userId      = input.readInt();
		String name        = input.readUTF();
		int    numElements = input.readInt();
		RecordElement[] elements = new RecordElement[ numElements ];
		for ( int i = 0; i < numElements; ++i ) {
			elements[i] = RecordElement.read( i, input );
		}

		return new DataStreamInfo( userId, name, elements );
	}

	public boolean equals( Object o ) {
		if ( this == o ) return true;
		if ( o == null || getClass() != o.getClass() ) return false;

		DataStreamInfo that = (DataStreamInfo) o;

		if ( userId != that.userId ) return false;
		if ( !Arrays.equals( elements, that.elements ) ) return false;
		if ( !name.equals( that.name ) ) return false;

		return true;
	}

	public int hashCode() {
		int result;
		result = userId;
		result = 31 * result + name.hashCode();
		result = 31 * result + Arrays.hashCode( elements );
		return result;
	}

	public String toString() {
		return "DataStreamInfo{" +
		       "userId=" + userId +
		       ", name='" + name + '\'' +
		       ", elements=" + ( elements == null ? null : Arrays.asList( elements ) ) +
		       '}';
	}

	private static String eToS( RecordElement element ) {
		return "Element " + element.getPosition() + " <" + element.getName() + " : " +
		       element.getType() + ">";
	}
}
