/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.metadata;

import org.gillius.realdb.io.DataInputReadable;
import org.gillius.realdb.model.ElementCompressor;
import org.gillius.realdb.model.ElementReconstructor;

import java.io.DataInput;
import java.util.Map;
import java.util.Set;

/**
 * ElementCodec contains the information to create ElementCompressor and ElementReconstructor
 * objects to work on a RecordElement of a given type. An ElementCodec can be configured via two
 * provided methods: {@link #setParams(Map)} and {@link #read(DataInput)}. The ElementCodec must
 * be configured before it is used.
 *
 * @author Jason Winnebeck
 */
public interface ElementCodec extends DataInputReadable {
	/**
	 * Adds to the given Set all of the parameters required to configure this codec.
	 * <p/>
	 * Implementation note: the Set is a parameter rather than a return to support inheritance by
	 * calling super's method.
	 *
	 * @param params non-null Set
	 */
	public void getRequiredParams( Set<String> params );

	/**
	 * Adds to the given Set all of the non-required parameters that can be used to configure this
	 * codec.
	 * <p/>
	 * Implementation note: the Set is a parameter rather than a return to support inheritance by
	 * calling super's method.
	 *
	 * @param params non-null Set
	 */
	public void getOptionalParams( Set<String> params );

	/**
	 * Completely configures this ElementCodec with the given parameters.
	 *
	 * @param params non-null Map with entires at least for all keys returned from
	 *               {@link #getRequiredParams(Set)}.
	 *
	 * @throws IllegalArgumentException If a required parameter is not provided, if a provided
	 *                                  parameter is invalid, or if a provided parameter is not
	 *                                  recognized by this method.
	 */
	public void setParams( Map<String, String> params ) throws IllegalArgumentException;

	/**
	 * Returns an ElementCompressor implementation that will handle the given type.
	 */
	public ElementCompressor getCompressor( ElementType type );

	/**
	 * Returns an ElementReconstructor implementation that will handle the given type.
	 */
	public ElementReconstructor getReconstructor( ElementType type );
}
