/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.metadata;

import org.gillius.realdb.model.data.UnsignedByte;
import org.gillius.realdb.model.data.UnsignedInteger;
import org.gillius.realdb.model.data.UnsignedShort;

/**
 * ElementType represents one of the possible data types for a record element.
 *
 * @author Jason Winnebeck
 */
public enum ElementType {
	//signed integer types
	SInt8( Byte.class, (byte)0 ),
	SInt16( Short.class, (byte)1 ),
	SInt32( Integer.class, (byte)2 ),
	SInt64( Long.class, (byte)3 ),

	//unsigned integer types
	UInt8( UnsignedByte.class, (byte)4 ),
	UInt16( UnsignedShort.class, (byte)5 ),
	UInt32( UnsignedInteger.class, (byte)6 ),

	//real-number types
	Float( Float.class, (byte)7 ),
	Double( Double.class, (byte)8 ),

	//discrete value types
	Boolean( Boolean.class, (byte)9 );

	private static final ElementType[] byCode;

	static {
		byCode = new ElementType[ values().length ];
		for ( ElementType type : values() ) {
			if ( byCode[ type.getDbCode() ] != null )
				throw new IllegalStateException( "Duplcate ElementType codes" );
			//ArrayOutOfBounds means the numbers aren't packed...
			byCode[ type.getDbCode() ] = type;
		}
	}

	public static ElementType getByDbCode( byte code ) {
		if ( code < 0 || code >= byCode.length )
			throw new IllegalArgumentException( "No ElementType with DB code " + code );
		return byCode[code];
	}

	private final Class<?> typeClass;
	private final byte  dbCode;

	private ElementType( Class<?> typeClass, byte dbCode ) {
		this.typeClass = typeClass;
		this.dbCode    = dbCode;
	}

	public Class<?> getTypeClass() {
		return typeClass;
	}

	public byte getDbCode() {
		return dbCode;
	}
}
