/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.metadata;

import org.gillius.realdb.BadFileException;
import org.gillius.realdb.io.BlockFile;

import java.io.IOException;
import java.io.InputStream;
import java.nio.ByteBuffer;
import java.util.Arrays;

/**
 * FileHeader is a header for a RealDB database file.
 *
 * @author Jason Winnebeck
 */
public class FileHeader {
	/**
	 * The file magic value, the string "RLDB" encoded in ASCII. Every RealDB database starts with
	 * this value.
	 */
	public static final byte[] FILE_MAGIC = new byte[] { 82, 69, 65, 76 };

	/**
	 * Returns the most current header version known to this version of RealDB.
	 */
	public static final int CURRENT_HEADER_VER = 2;

	/**
	 * Returns the most current database version known to this version of RealDB.
	 */
	public static final int CURRENT_DB_VER = 3;

	private final int headerVersion;
	private final int databaseVersion;
	private final int blockSize;
	private final int numBlocks;

	/**
	 * Returns the serialized size of the file header on disk. Includes MAGIC, plus the 4 fields.
	 */
	private static final int SIZE = 4 + 4 * 4;

	/**
	 * Constructs a FileHeader by explicitly specifying all fields.
	 */
	public FileHeader( int headerVersion, int databaseVersion, int blockSize, int numBlocks ) {
		try {
			checkBlockSize( blockSize );
			checkNumBlocks( numBlocks );
		} catch ( BadFileException e ) {
			throw new IllegalArgumentException( e );
		}

		this.headerVersion   = headerVersion;
		this.databaseVersion = databaseVersion;
		this.blockSize       = blockSize;
		this.numBlocks       = numBlocks;
	}

	/**
	 * Constructs a FileHeader by assuming defaults for the version fields.
	 */
	public FileHeader( int blockSize, int numBlocks ) {
		try {
			checkBlockSize( blockSize );
			checkNumBlocks( numBlocks );
		} catch ( BadFileException e ) {
			throw new IllegalArgumentException( e );
		}

		headerVersion = CURRENT_HEADER_VER;
		databaseVersion = CURRENT_DB_VER;
		this.blockSize = blockSize;
		this.numBlocks = numBlocks;
	}

	public int getHeaderVersion() {
		return headerVersion;
	}

	public int getDatabaseVersion() {
		return databaseVersion;
	}

	public int getBlockSize() {
		return blockSize;
	}

	public int getNumBlocks() {
		return numBlocks;
	}

	/**
	 * Writes this header out to a file. The blockSize field and the block size of f must match. The
	 * first block of the file is written.
	 *
	 * @param f non-null BlockFile
	 *
	 * @throws IOException if the block cannot be written
	 */
	public void writeHeader( BlockFile f ) throws IOException {
		if ( blockSize != f.getBlockSize() ) {
			throw new IllegalArgumentException( "Header block size of " + blockSize + " does not match" +
			                                    "BlockFile block size of " + f.getSize() );
		}

		ByteBuffer buf = ByteBuffer.allocate( SIZE );
		buf.put( FILE_MAGIC );
		buf.putInt( headerVersion );
		buf.putInt( databaseVersion );
		buf.putInt( blockSize );
		buf.putInt( numBlocks );
		f.writeBlocks( 0, buf.array() );
	}

	/**
	 * Parses a RealDB FileHeader from the given InputStream. The returned FileHeader will have been
	 * checked for compatible version numbers.
	 *
	 * @throws IOException if an error occurs getting data from the InputStream
	 * @throws BadFileException if enough bytes are read, but does not contain supported values, such
	 *                          as a newer/older version supported
	 */
	public static FileHeader getHeader( InputStream is ) throws IOException, BadFileException {
		byte[] buf = new byte[ SIZE ];
		int pos = 0;
		while ( pos < SIZE ) {
			int read = is.read( buf, pos, buf.length - pos );
			if ( read < 0 )
				throw makeError( "not enough data" );
			pos += read;
		}

		return parse( buf );
	}

	/**
	 * Parses a RealDB FileHeader from the given BlockFile, and verifies the block size against the
	 * BlockFile and the version numbers.
	 *
	 * @throws IOException if an error occurs getting data from the BlockFile
	 * @throws BadFileException if enough bytes are read, but does not contain supported values, such
	 *                          as a newer/older version supported
	 */
	public static FileHeader verify( BlockFile bf ) throws IOException, BadFileException {
		int blockSize = bf.getBlockSize();
		checkBlockSize( blockSize );

		byte[] data = new byte[ SIZE ];
		bf.readBlocks( 0, data );
		FileHeader ret = parse( data );
		if ( ret.getBlockSize() != blockSize )
			throw makeError( "Block size " + ret.getBlockSize() + " in header does not match block size " +
	                     blockSize + " of given BlockFile" );
		if ( ret.getNumBlocks() != bf.getSize() )
			throw makeError( "Number of blocks " + ret.getNumBlocks() + " in header does not match " +
	                     "number " + bf.getSize() + " in given BlockFile" );

		return ret;
	}

	private static FileHeader parse( byte[] buf ) throws BadFileException {
		ByteBuffer bb = ByteBuffer.wrap( buf );
		byte[] magic = new byte[4];
		bb.get( magic );
		if ( !Arrays.equals( magic, FILE_MAGIC ) )
			throw makeError( "file magic does not match; not a RealDB type file" );

		int headerVersion   = bb.getInt();

		if ( headerVersion != CURRENT_HEADER_VER ) {
			throw makeError( "Header version " + headerVersion + " is not the " +
	                     "version (" + CURRENT_HEADER_VER + ") for this version of RealDB" );
		}
		int databaseVersion = bb.getInt();
		int blockSize       = bb.getInt();
		int numBlocks       = bb.getInt();

		if ( databaseVersion != CURRENT_DB_VER ) {
			throw makeError( "Database version " + headerVersion + " is not the " +
			                 "recognized version (" + CURRENT_DB_VER + ") for this version of RealDB" );
		}
		checkBlockSize( blockSize );
		checkNumBlocks( numBlocks );

		return new FileHeader( headerVersion, databaseVersion, blockSize, numBlocks );
	}

	private static void checkBlockSize( int blockSize ) throws BadFileException {
		if ( blockSize < SIZE ) {
			throw makeError( "Block size " + blockSize + " specified, but minimum of " + SIZE + " is " +
		                   "supported" );
		}
	}

	private static void checkNumBlocks( int numBlocks ) throws BadFileException {
		if ( numBlocks < 1 ) {
			throw makeError( "Number of blocks in the file is a negative number or 0" );
		}
	}

	private static BadFileException makeError( String msg ) {
		return new BadFileException( "Given data does not have a valid RealDB header: " + msg );
	}
}
