/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.metadata;

import org.gillius.Util.CheckUtils;
import org.gillius.realdb.io.BlockFile;
import org.gillius.realdb.io.DataOutputWritable;
import org.gillius.realdb.io.SizingDataOutput;

import java.io.ByteArrayInputStream;
import java.io.DataInputStream;
import java.io.DataOutput;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

/**
 * TODO: rewrite later
 * MetadataSection represents a typically static metadata section of the database. This section
 * contains information on the data streams and their formats. During normal execution of RealDB
 * (reading and writing data stream data), this section does not change.
 * <p/>
 * The section can contain some slack space when written to allow the number of the data streams
 * to grow without having to rebuild the database.
 * <p/>
 * TODO: consider storing database time unit (milliseconds?) and epoch start.
 *
 * @author Jason Winnebeck
 */
public class MetadataSection implements DataOutputWritable {
	private final int dataBlockSize;

	private final List<DataStreamInfo> streamInfos;
	private final int maxStreams;

	/**
	 * The size variable is lazily initialized. It's a cached variable, and is not used in equals
	 * and hashCode (as it is dervied from the other final variables).
	 */
	private transient int size = -1;

	/**
	 * Constructs a new MetadataSection, not from an existing database. Sets the initial
	 * data block size and a list of DataStreams. If the maxStreams is greater than streams.size(),
	 * this allows for a slack space to be able to add more streams without rebuilding the database.
	 *
	 * @param dataBlockSize size of a data block, in low-level blocks
	 * @param streamInfos   non-null List of 1 to maxStreams DataStream
	 * @param maxStreams    number of streams there could be without rebuilding DB.
	 */
	public MetadataSection( int dataBlockSize, List<DataStreamInfo> streamInfos, int maxStreams ) {
		CheckUtils.checkNull( streamInfos, "streamInfos" );

		this.dataBlockSize = dataBlockSize;
		this.streamInfos   = new ArrayList<DataStreamInfo>( streamInfos );
		this.maxStreams    = maxStreams;

		if ( streamInfos.size() > maxStreams )
				throw new IllegalArgumentException( "Too many streams are given to fit within maxStreams" );
	}

	@SuppressWarnings( { "unchecked" } )
	public MetadataSection( BlockFile file, int startBlock ) throws IOException {
		CheckUtils.checkNull( file, "file" );

		byte[] raw = new byte[4];
		file.readBlocks( startBlock, raw );
		DataInputStream input = new DataInputStream( new ByteArrayInputStream( raw ) );
		int sectionSize = input.readInt();

		raw = new byte[ sectionSize ];
		size = file.readBlocks( startBlock, raw );
		input = new DataInputStream( new ByteArrayInputStream( raw ) );
		input.readInt(); //skip sectionSize, which we just read
		maxStreams = input.readInt();
		dataBlockSize = input.readInt();


		int numStreams = input.readInt();
		streamInfos = new ArrayList<DataStreamInfo>( numStreams );
		for ( int i = 0; i < numStreams; ++i ) {
			streamInfos.add( DataStreamInfo.read( input ) );
		}
	}

	/**
	 * Returns the size of a data block, in the number of low-level file blocks (not bytes).
	 */
	public int getDataBlockSize() {
		return dataBlockSize;
	}

	/**
	 * Return the size of the section, in bytes.
	 */
	public int getSizeBytes() {
		if ( size < 0 ) {
			SizingDataOutput out = new SizingDataOutput();
			try {
				writeAllButSize( out );
			} catch ( IOException e ) {
				//Due to the objects being used, we know that we shouldn't be getting an IOException, unless
				//one of the strings is larger than 65535 bytes in UTF8 encoding.
				throw new IllegalStateException( e );
			}
			size = out.getSize() + 4; //add 4 for size header
		}

		return size;
	}

	/**
	 * Returns the stream information stored in this MetadataSection. The returned list may not be
	 * modified.
	 */
	public List<DataStreamInfo> getStreams() {
		return streamInfos;
	}

	/**
	 * Returns the maximum number of streams that can be stored in the index.
	 */
	public int getMaxStreams() {
		return maxStreams;
	}

	public void write( DataOutput output ) throws IOException {
		output.writeInt( getSizeBytes() );
		writeAllButSize( output );
	}

	private void writeAllButSize( DataOutput output ) throws IOException {
		output.writeInt( maxStreams );
		output.writeInt( dataBlockSize );
		output.writeInt( streamInfos.size() );
		for ( DataStreamInfo info : streamInfos ) {
			info.write( output );
		}
	}

	public boolean equals( Object o ) {
		if ( this == o ) return true;
		if ( o == null || getClass() != o.getClass() ) return false;

		MetadataSection that = (MetadataSection) o;

		if ( dataBlockSize != that.dataBlockSize ) return false;
		if ( maxStreams != that.maxStreams ) return false;
		if ( !streamInfos.equals( that.streamInfos ) ) return false;

		return true;
	}

	public int hashCode() {
		int result;
		result = dataBlockSize;
		result = 31 * result + streamInfos.hashCode();
		result = 31 * result + maxStreams;
		return result;
	}
}
