/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.metadata;

import org.gillius.realdb.io.DataOutputWritable;
import org.gillius.realdb.io.DowUtil;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;

/**
 * RecordElement describes a single element (equivalent to column in relational DB) in a Record.
 *
 * @author Jason Winnebeck
 */
public final class RecordElement implements DataOutputWritable, OptionalElementType {
	private final int          position;
	private final String       name;
	private final ElementCodec codec;
	private final ElementType  type;
	private final boolean      optional;

	public RecordElement( int position, String name, ElementCodec codec, ElementType type,
	                      boolean optional ) {
		this.position = position;
		this.name     = name;
		this.codec    = codec;
		this.type     = type;
		this.optional = optional;
	}

	public int getPosition() {
		return position;
	}

	public String getName() {
		return name;
	}

	public ElementCodec getCodec() {
		return codec;
	}

	public ElementType getType() {
		return type;
	}

	public boolean isOptional() {
		return optional;
	}

	public void write( DataOutput output ) throws IOException {
		output.writeUTF( name );
		DowUtil.writeObject( output, codec );
		output.writeByte( type.getDbCode() );
		output.writeBoolean( optional );
	}

	public static RecordElement read( int position, DataInput input ) throws IOException {
		String name = input.readUTF();
		ElementCodec codec = DowUtil.readObject( input, ElementCodec.class );
		ElementType type = ElementType.getByDbCode( input.readByte() );
		boolean optional = input.readBoolean();

		return new RecordElement( position, name, codec, type, optional );
	}

	@Override
	public boolean equals( Object o ) {
		if ( this == o ) return true;
		if ( o == null || getClass() != o.getClass() ) return false;

		RecordElement that = (RecordElement) o;

		if ( optional != that.optional ) return false;
		if ( position != that.position ) return false;
		if ( !codec.equals( that.codec ) ) return false;
		if ( !name.equals( that.name ) ) return false;
		if ( type != that.type ) return false;

		return true;
	}

	@Override
	public int hashCode() {
		int result = position;
		result = 31 * result + name.hashCode();
		result = 31 * result + codec.hashCode();
		result = 31 * result + type.hashCode();
		result = 31 * result + ( optional ? 1 : 0 );
		return result;
	}

	public String toString() {
		return "RecordElement{" +
		       "position=" + position +
		       ", name='" + name + '\'' +
		       ", codec=" + codec +
		       ", type=" + type +
		       ", optional=" + optional +
		       '}';
	}
}
