/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.model;

import org.gillius.realdb.BadFileException;
import org.gillius.realdb.io.Formattable;
import org.gillius.realdb.model.pool.TransactionLog;

import java.io.Flushable;
import java.io.IOException;

/**
 * BlockPool
 * <p/>
 * TODO: document
 *
 * @author Jason Winnebeck
 */
public interface BlockPool extends Formattable, Flushable {
	/**
	 * Opens the TransactionLog given when this object was constructed. Any blocks in open
	 * transactions become reclaimable blocks.
	 * <p/>
	 * The pool must be opened before calling any other methods besides {@link #format()}.
	 */
	void open() throws IOException, BadFileException;

	TransactionLog getLog();

	/**
	 * Returns true if there are any free blocks in the pool.
	 */
	public boolean hasFreeBlocks();

	/**
	 * Returns the total number of data blocks managed by this pool.
	 */
	int getNumDataBlocks();

	/**
	 * Returns the number of data blocks in use.
	 */
	int getDataBlocksUsed();

	/**
	 * If {@link #hasFreeBlocks()}, then returns a new transaction with the free block.
	 *
	 * @return a BlockTransaction with the free block and a single action to flush the transaction
	 *         log.
	 *
	 * @throws IOException if there is an implicit log flush and it fails.
	 * @throws IllegalStateException if {@link #hasFreeBlocks()} is false.
	 */
	BlockTransaction getFreeBlock() throws IOException;

	/**
	 * Signals an intention to remove the indicated block from the indicated index immediately after
	 * calling this method.
	 *
	 * @param block       block number to remove
	 * @param sourceIndex source index ID
	 * @param seqNum      index's head sequence number (used in recovery to verify action)
	 * @param secondBlockStartTime start time of the second block in the index that will be passed to
	 *                    {@link DataIndex#removeFirstBlock}; needed for replay.
	 *
	 * @return a BlockTransaction with the removed block and a single action to flush the transaction
	 *         log.
	 *
	 * @throws IOException if there is an implicit log flush and it fails.
	 */
	BlockTransaction removeBlock( int block, int sourceIndex, int seqNum, long secondBlockStartTime )
			throws IOException;

	/**
	 * Adds a block obtained from {@link #getFreeBlock} or {@link #removeBlock}. This method must be
	 * called after writing the data block to the new state and before adding the block to the index.
	 * This method finalizes the transaction; the only operation remaining is to commit it.
	 *
	 * @param tx          transaction returned from {@link #getFreeBlock} or {@link #removeBlock}
	 * @param lastBlock   current last block in the index, used only for extra verification in
	 *                    recovery, or -1 if the index is empty.
	 * @param destination index destination ID
	 * @param seqNum      index tail sequence number on disk (needed for recovery)
	 */
	void addBlock( BlockTransaction tx, int lastBlock, int destination, int seqNum )
			throws IOException;
}
