/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.model;

import org.gillius.Util.CheckUtils;
import org.gillius.Util.MathUtils;
import org.gillius.realdb.BadFileException;
import org.gillius.realdb.io.BlockFile;
import org.gillius.realdb.io.ByteBufferDataIO;
import org.gillius.realdb.io.DataInputReader;
import org.gillius.realdb.io.DataOutputWritable;
import org.gillius.realdb.io.Formattable;
import org.gillius.realdb.io.SizingDataOutput;
import org.gillius.realdb.util.DBIterator;

import java.io.Flushable;
import java.io.IOException;
import java.io.PrintStream;
import java.nio.ByteBuffer;
import java.util.NoSuchElementException;

/**
 * CircularQueue implements a small, circular buffer of {@link DataOutputWritable} objects. Objects
 * are inserted at the "head" end and the oldest objects are in the "tail" end.
 * <p/>
 * CircularQueue saves some overhead by not implementing a header to keep track of the head and
 * tail, so it scans all of its blocks to find the head and tail. Therefore, the definition of
 * "small" is based on the time it takes to scan the whole log being acceptable. This technique
 * trades faster write speeds for slower startup time.
 * <p/>
 * This code relies on a 32-bit sequence number, and assumes that there will be fewer than (2^31 -
 * number of blocks) writes before the queue wraps around.
 *
 * @author Jason Winnebeck
 */
public class CircularQueue implements Flushable, Formattable {
	/**
	 * Constant overhead (regardless of size) of this queue on disk, in blocks.
	 */
	private static final int overhead = 1;

	private final CircularQueueBlockIO io;
	private final int logicalSize;

	/**
	 * Holds the actual data content of the head.
	 */
	private transient final ByteBufferDataIO buf;
	private transient final SizingDataOutput sizer;

	private final CircularQueueState state;

	/**
	 * newData is set true whenever new data is added, and false whenever flushed.
	 */
	private boolean newData = false;

	/**
	 * @param io          non-null
	 * @param state       initial queue state
	 * @param maxDowSize  maximum size of queued object
	 * @param startData   nullable starting data for the header block
	 */
	private CircularQueue( CircularQueueBlockIO io, CircularQueueState state, int maxDowSize,
	                       byte[] startData )
			throws IOException {
		this.io = io;

		//checkSizes has already been called...
		logicalSize = io.getSize() - overhead;

		if ( state.tail < 0 || state.tail >= logicalSize )
			throw new IllegalArgumentException( "bad tail: " + state.tail );
		if ( state.head < 0 || state.head >= logicalSize )
			throw new IllegalArgumentException( "bad head: " + state.head );
		if ( maxDowSize <= 0 )
		  throw new IllegalArgumentException( "maxDowSize must be positive: " + maxDowSize );

		this.state = state;

		if ( io.getBlockDataSize() < maxDowSize )
			throw new IllegalArgumentException( "BlockFile block is too small" );

		buf = new ByteBufferDataIO( io.getBlockDataSize() );
		sizer = new SizingDataOutput();

		if ( startData != null ) {
			buf.write( startData );
		}
	}

	/**
	 * Creates a newly formatted CircularQueue in the BlockFile.
	 *
	 * @param bf          non-null
	 * @param startBlock  within bf bounds
	 * @param size        startBlock + logicalSize within bf bounds
	 * @param maxDowSize  maximum size of queued object
	 */
	public static CircularQueue createNew( BlockFile bf, int startBlock, int size, int maxDowSize )
			throws IOException {
		checkSizes( bf, startBlock, size );
		CircularQueueBlockIO io = new DefaultCircularQueueBlockIO( bf, startBlock, size );
		CircularQueue ret = new CircularQueue( io, new CircularQueueState(), maxDowSize, null );
		ret.format();
		return ret;
	}

	/**
	 * Loads a previously created CircularQueue from the BlockFile.
	 *
	 * @param bf          non-null
	 * @param startBlock  within bf bounds
	 * @param size        startBlock + logicalSize within bf bounds
	 * @param maxDowSize  maximum size of queued object
	 *
	 * @throws IOException      error reading from the file
	 * @throws BadFileException if method can read from the file, but data is corrupted
	 */
	public static CircularQueue load( final BlockFile bf, final int startBlock, final int size,
	                                  int maxDowSize )
			throws IOException, BadFileException {
		checkSizes( bf, startBlock, size );

		CircularQueueBlockIO io = new DefaultCircularQueueBlockIO( bf, startBlock, size );

		CircularQueueState state = loadState( io );

		byte[] startData;
		if ( !state.altHead )
			startData = io.getData( io.getAltHeadIndex() );
		else
			startData = io.getData( state.head );

		return new CircularQueue( io, state, maxDowSize, startData );
	}

	private static CircularQueueState loadState( CircularQueueBlockIO io )
			throws IOException, BadFileException {
		CircularQueueState ret = null;

		for ( int i = 0; i < io.getSize(); ++i ) {
			try {
				CircularQueueState next = io.getState( i );
				if ( ret == null || next.nextSeqNum > ret.nextSeqNum )
					ret = next;

			} catch ( Exception e ) {
				//ignore this block
			}
		}

		if ( ret == null ) {
			throw new BadFileException( "Did not find any valid blocks in the queue, but even a " +
			                            "freshly formatted queue has one." );
		} else {
			//Check if the tail is bad, which is possible on a rotation scenario
			if ( ret.head != ret.tail ) {
				try {
					io.getState( ret.tail );
				} catch ( Exception e ) {
					ret.tail = ( ret.tail + 1 ) % ( io.getSize() - overhead );
				}
			}
			++ret.nextSeqNum;
			return ret;
		}
	}

	/**
	 * Given the block size in bytes and the total amount of data to hold in bytes, returns the
	 * minimum size in blocks the queue will need to be able to hold at least 'size' bytes at all
	 * times, and with each block having a header of up to 'maxHeaderSize' bytes and each item in the
	 * queue being up to maxItemSize.
	 *
	 * @throws IllegalArgumentException if there is not enough room to store the largest item.
	 */
	public static int getMinSize( int size, int maxHeaderSize, int maxItemSize, int blockSize ) {
		int bytesPerBlock = blockSize - CircularQueueBlockIO.blockOverhead - maxHeaderSize;
		if ( bytesPerBlock < maxItemSize )
			throw new IllegalArgumentException( "blocks are too small for the items" );

		int itemsPerBlock = bytesPerBlock / maxItemSize;
		int effectiveBytesPerBlock = itemsPerBlock * maxItemSize;
		int blocks = MathUtils.divideRoundUp( size, effectiveBytesPerBlock ) + 1;
		//+1 because head could contain as low as 0 items in a flush
		return blocks + overhead;
	}

	public void format() throws IOException {
		io.format();
		state.reset();
		io.writeBlock( 0, state, new byte[0], 0, 0 );
		buf.getBuf().position( 0 );
		newData = false;
	}

	/**
	 * Returns the number of bytes available in each block for the added elements. The file's block
	 * size minus this property gives you the per-block overhead of the queue.
	 */
	public int getUsableBytesPerBlock() {
		return io.getBlockDataSize();
	}

	/**
	 * Returns the number of blocks totally claimed by overhead.
	 */
	@SuppressWarnings( { "MethodMayBeStatic" } )
	public int getOverheadBlocks() {
		return overhead;
	}

	/**
	 * Returns the total size of the queue in blocks.
	 */
	public int getSize() {
		return logicalSize + overhead;
	}

	/**
	 * Returns the total number of usable bytes in the queue for the serialized added elements.
	 */
	public int getTotalUsableBytes() {
		return ( getSize() - getOverheadBlocks() ) * getUsableBytesPerBlock();
	}

	/**
	 * Adds the given element to the queue.
	 */
	public void add( DataOutputWritable dow ) throws IOException {
		add( dow, null );
	}

	/**
	 * Adds the given element to the queue, using the given block header if a new block is
	 * started. If the header is null, no header is written. This can be used to ensure that a
	 * given element is always in the queue at least once, providing a form of persistence. The very
	 * first add to the queue does count as "starting a new block."
	 */
	public void add( DataOutputWritable dow, DataOutputWritable blockHeader ) throws IOException {
		if ( blockHeader != null && buf.getBuf().position() == 0 ) {
			safeWrite( blockHeader );
		}
		sizer.reset();
		dow.write( sizer );
		int dowSize = sizer.getSize();

		if ( buf.getBuf().remaining() < dowSize ) {
			writeAndAdvanceHead();
			if ( blockHeader != null )
				safeWrite( blockHeader );
		}
		//Check here separately because of block header
		if ( buf.getBuf().remaining() < dowSize ) {
			throw new IllegalArgumentException( "Given writable is larger than the block" );
		}

		safeWrite( dow );
	}

	public void flush() throws IOException {
		flush( false );
	}

	public void flush( boolean force ) throws IOException {
		if ( !newData && !force )
			return;

		int b = state.altHead ? io.getAltHeadIndex() : state.head;
		io.writeBlock( b, state, buf.getBuf().array(), buf.getBuf().arrayOffset(),
		               buf.getBuf().position() );
		newData = false;

		state.altHead = !state.altHead;
		++state.nextSeqNum;
	}

	/**
	 * Returns a {@link DBIterator} to read all of the data in this queue, using the given reader to
	 * deserialize the elements. The iterator is invalidated if the queue is modified.
	 */
	public <T> DBIterator<T> iterator( DataInputReader<T> reader ) {
		return new Iterator<T>( reader );
	}

	/**
	 * Outputs a raw output of the data in the file blocks underlying this queue.
	 */
	public <T> void printForensics( DataInputReader<T> reader, PrintStream out ) {
		out.format( "%s%n", state );

		for( int i = 0; i < io.getSize(); ++i ) {
			if ( i == io.getAltHeadIndex() ) {
				out.print( "AltHead Block " + logicalSize );
			} else {
				out.print( "Block " + i );
			}

			try {
				ByteBufferDataIO buf = new ByteBufferDataIO( io.getData( i ) );
				CircularQueueState blockState = io.getState( i );
				out.println( ", seq " + blockState.nextSeqNum + ", state " + blockState );
				printBufferForensics( reader, out, buf );

			} catch ( Exception e ) {
				out.println( " -- BAD: " + e.getMessage() );
			}
		}

		out.println( "Buffered Head:" );
		ByteBufferDataIO buf = new ByteBufferDataIO( (ByteBuffer) this.buf.getBuf().duplicate().flip() );
		printBufferForensics( reader, out, buf );
	}

	private static <T> void printBufferForensics( DataInputReader<T> reader, PrintStream out,
	                                              ByteBufferDataIO buf ) {
		out.print( " Content: " );
		try {
			while( buf.getBuf().hasRemaining() ) {
				out.print( reader.readValue( buf ) + " " );
			}
			out.println();
		} catch ( IOException e ) {
			out.println( " -- BAD: " + e.getMessage() );
		}
	}

	/**
	 * Writes the DOW to the buffer, reverting the position on exception.
	 */
	private void safeWrite( DataOutputWritable dow ) throws IOException {
		int savedPos = buf.getBuf().position();

		try {
			dow.write( buf );
			newData = true;
			savedPos = -1;

		} finally {
			if ( savedPos >= 0 )
				buf.getBuf().position( savedPos );
		}
	}

	/**
	 * Writes and advances the header.
	 */
	private void writeAndAdvanceHead() throws IOException {
		flush();

		//We need to make sure the "real" head block is written before advancing. This might require a
		//forced "double" write because we can't safely overwrite our old block.
		if ( !state.altHead )
			flush( true );
		state.altHead = false; //next write should be to the "real" block

		state.head = ( state.head + 1 ) % logicalSize;
		if ( state.head == state.tail )
			state.tail = ( state.tail + 1 ) % logicalSize; //we overwrote the tail...

		buf.getBuf().clear();
	}

	private int getLength() {
		if ( state.head >= state.tail )
			return state.head - state.tail + 1;
		else //head < tail
			return logicalSize - state.tail + state.head + 1;
	}

	/**
	 * Loads data from the block (starting at tail = 0 and head = length - 1) into the given buffer.
	 * The buffer must have a backing array with remaining bytes equal to or greater than
	 * {@link #getUsableBytesPerBlock()}.
	 */
	private void loadBlockDataTailRelativeIndex( int tailRelativeIdx, ByteBuffer dest )
			throws IOException, BadFileException {
		int length = getLength();
		if ( tailRelativeIdx < 0 || tailRelativeIdx >= length ) {
			throw new IndexOutOfBoundsException();

		} else if ( tailRelativeIdx == length - 1 ) {
			dest.put( (ByteBuffer) buf.getBuf().duplicate().flip() );

		} else {
			int rawIdx = ( state.tail + tailRelativeIdx ) % logicalSize;

			dest.put( io.getData( rawIdx ) );
		}
	}

	private static void checkSizes( BlockFile bf, int startBlock, int size ) {
		CheckUtils.checkNull( bf, "bf" );
		if ( startBlock < 0 || startBlock >= bf.getSize() )
			throw new IllegalArgumentException( "bad startBlock: " + startBlock );
		if ( size < overhead )
		  throw new IllegalArgumentException( "size must be >= " + overhead + ": " + size );
		if ( startBlock + size > bf.getSize() )
			throw new IllegalArgumentException( "size extends past end of file: " + size );
	}

	private class Iterator<T> implements DBIterator<T> {
		private final DataInputReader<T> reader;
		private final int length;
		private final ByteBuffer raw;
		private final ByteBufferDataIO buf;

		private int nextIdx;

		private Iterator( DataInputReader<T> reader ) {
			this.reader = reader;

			length = getLength();
			raw = ByteBuffer.allocate( io.getBlockDataSize() );
			buf = new ByteBufferDataIO( raw );
			raw.limit( 0 );

			nextIdx = 0;
		}

		public boolean hasNext() throws IOException, BadFileException {
			if( !raw.hasRemaining() ) {
				if ( nextIdx < length ) {
					raw.clear();
					loadBlockDataTailRelativeIndex( nextIdx, raw );
					raw.flip();
					++nextIdx;
				}
			}
			return raw.hasRemaining();
		}

		public T next() throws IOException, BadFileException {
			if ( !hasNext() )
				throw new NoSuchElementException();

			int savedPos = raw.position();
			try {
				T ret = reader.readValue( buf );
				savedPos = -1;
				return ret;

			} finally {
				if ( savedPos >= 0 )
					raw.position( savedPos );
			}
		}
	}
}
