/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.model;

import org.gillius.realdb.io.DataOutputWritable;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;

/**
 * CircularQueueState holds the current CircularQueue state. The persistent parts of the state
 * when written as a {@link DataOutputWritable} are the head and tail pointers. The altHead can be
 * determined on loading on whether or not the serialized state is found in the alternate head
 * block or not, and the nextSeqNum is the sequence number of the block it is found within.
 *
 * @author Jason Winnebeck
 */
class CircularQueueState implements DataOutputWritable {
	/**
	 * Constructs a new CircularQueueState for a freshly created queue.
	 */
	CircularQueueState() {
		reset();
	}

	/**
	 * Constructs a new CircularQueueState explicitly.
	 *
	 * @param tail        [0, logicalSize)
	 * @param head        [0, logicalSize)
	 * @param altHead     true if next write should be to alternative head
	 * @param nextSeqNum  next sequence number to write
	 */
	CircularQueueState( int tail, int head, boolean altHead, int nextSeqNum ) {
		this.tail = tail;
		this.head = head;
		this.altHead = altHead;
		this.nextSeqNum = nextSeqNum;
	}

	/**
	 * Loads the serialized form of the CircularQueueState from the DataInput.
	 */
	CircularQueueState( DataInput in, boolean altHead, int nextSeqNum ) throws IOException {
		head = in.readInt();
		tail = in.readInt();
		this.altHead = altHead;
		this.nextSeqNum = nextSeqNum;
	}

	/**
	 * Returns the written size of the CircularQueueState.
	 */
	public static int getWrittenSize() {
		return 4 + 4;
	}

	/**
	 * Tail index, relative to queue's start.
	 */
	public int tail;
	/**
	 * Head index, relative to queue's start.
	 */
	public int head;

	/**
	 * True if the next write should be to the alternative head.
	 */
	public boolean altHead;

	/**
	 * Sequence number for the next block written.
	 */
	public int nextSeqNum;

	/**
	 * Resets the state as if the queue were freshly created.
	 */
	public void reset() {
		tail = head = 0;
		altHead = false;
		nextSeqNum = 0;
	}

	@Override
	public void write( DataOutput output ) throws IOException {
		output.writeInt( head );
		output.writeInt( tail );
	}

	@Override
	public String toString() {
		return "CircularQueueState{" +
		       "tail=" + tail +
		       ", head=" + head +
		       ", altHead=" + altHead +
		       ", nextSeqNum=" + nextSeqNum +
		       '}';
	}

	@Override
	public boolean equals( Object o ) {
		if ( this == o ) return true;
		if ( o == null || getClass() != o.getClass() ) return false;

		CircularQueueState that = (CircularQueueState) o;

		if ( altHead != that.altHead ) return false;
		if ( head != that.head ) return false;
		if ( nextSeqNum != that.nextSeqNum ) return false;
		if ( tail != that.tail ) return false;

		return true;
	}

	@Override
	public int hashCode() {
		int result = tail;
		result = 31 * result + head;
		result = 31 * result + ( altHead ? 1 : 0 );
		result = 31 * result + nextSeqNum;
		return result;
	}
}
