/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.model;

import org.gillius.realdb.BadFileException;
import org.gillius.realdb.io.ByteBufferDataIO;
import org.gillius.realdb.io.RecordReader;
import org.gillius.realdb.io.RecordWriter;
import org.gillius.realdb.model.data.Record;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.util.Arrays;

/**
 * DataBlock represents a block in the {@link DataSection}. It provides the implementation of
 * managing the block's header and being able to write records. This class is not thread-safe.
 *
 * @author Jason Winnebeck
 */
public class DataBlock {
	private final int blockNum;

	private int recordSize;
	private int totalSize;
	//TODO: Consider CRC

	//TODO: Can we get rid of these in the serialized version?
	//We could actually do this knowing that the first 8 bytes of every record is the timestamp.
	//Although, we need to keep into consideration the DataBlockSummaryReader's needs...
	private long startTime = -1L;
	private long endTime   = -1L;

	private ByteBufferDataIO data;

	public static final int HEADER_SIZE = 8 + 8 + 2;

	//Redundant variables cached for performance
	private int numRecords; //seriously, getNumRecords was top method in hotspot profiling!
	private int maxRecords;

	/**
	 * Returns the size of the header in a DataBlock, the overhead that is not used for record
	 * storage in the allocated size.
	 */
	public static int getHeaderSize() {
		return HEADER_SIZE;
	}

	/**
	 * Constructs a new DataBlock that can write fixed-length records to a fixed-length array. Only
	 * full records can be written in a data block, so there can be unused slack space at the end.
	 * <p/>
	 * Currently there can only be up to 65535 records in a block, so the size and recordSize
	 * should be set appropriately.
	 *
	 * @param blockNum   data block index
	 * @param recordSize size of the records to be written
	 * @param size       total size of the data block, in bytes
	 */
	public DataBlock( int blockNum, int recordSize, int size ) {
		maxRecords = ( size - HEADER_SIZE ) / recordSize;
		if ( maxRecords == 0 )
			throw new IllegalArgumentException( "Size must be at least large enough to hold one record plus the header." );
		if ( maxRecords > 65535 )
			throw new IllegalArgumentException( "Currently there can only be up to 65535 records in a data block." );

		this.blockNum   = blockNum;
		this.recordSize = recordSize;
		this.totalSize  = size;
		numRecords      = 0;

		data = new ByteBufferDataIO( size );
		data.getBuf().position( HEADER_SIZE );
	}

	/**
	 * Loads a written DataBlock, based on a given byte array generated from {@link #getRawData()},
	 * plus any number of random bytes.
	 *
	 * @param data       non-null byte array of serialized DataBlock
	 * @param blockNum   data block index
	 * @param recordSize size of the records to be written
	 * @param size       total size of the data block, in bytes
	 */
	public static DataBlock load( byte[] data, int blockNum, int recordSize, int size )
			throws BadFileException {
		DataBlock  ret = new DataBlock( blockNum, recordSize, size );
		ByteBuffer buf = ByteBuffer.wrap( data, 0, HEADER_SIZE );

		ret.startTime  = buf.getLong();
		ret.endTime    = buf.getLong();
		int numRecords = buf.getShort() & 0xFFFF;

		try {
			ret.data.write( data, HEADER_SIZE, numRecords * recordSize );
		} catch ( IOException e ) {
			throw new BadFileException( "numRecords is not correct size", e );
		}

		ret.calculateNumRecords();

		return ret;
	}

	/**
	 * Given the first {@link #HEADER_SIZE} bytes of the DataBlock's data, returns a DefaultRange.
	 */
	public static DefaultRange loadBlockRange( byte[] data ) {
		ByteBuffer buf = ByteBuffer.wrap( data, 0, HEADER_SIZE );

		return new DefaultRange( buf.getLong(), buf.getLong() );
	}

	/**
	 * Returns the data block index number.
	 */
	public int getBlockNum() {
		return blockNum;
	}

	/**
	 * Returns size size of a record in the block, in bytes. This value came from the class
	 * constructor.
	 */
	public int getRecordSize() {
		return recordSize;
	}

	/**
	 * Returns the total size of this data block, in bytes.
	 */
	public int getTotalSize() {
		return totalSize;
	}

	/**
	 * Returns the number of records written to this DataBlock so far.
	 */
	public int getNumRecords() {
		return numRecords;
	}

	/**
	 * Returns the raw data in this DataBlock, whose value could change as a result of calls to this
	 * object. Therefore, the return from this method should not be cached. Only the used bytes are
	 * returned, so the returned array could be smaller than {@link #getTotalSize()}.
	 */
	public byte[] getRawData() {
		ByteBuffer buf = data.getBuf();
		buf.putLong( 0, startTime );
		buf.putLong( 8, endTime );
		buf.putShort( 16, (short) getNumRecords() );

		final int offset = buf.arrayOffset();
		return Arrays.copyOfRange( buf.array(), offset, offset + buf.position() );
	}

	/**
	 * Returns the time of the first record written to this block, or -1 if there are no records.
	 */
	public long getStartTime() {
		return startTime;
	}

	/**
	 * Returns the time of the last record written to this block, or -1 if there are no records.
	 */
	public long getEndTime() {
		return endTime;
	}

	/**
	 * Returns true if there is space to write another record in this block.
	 */
	public boolean canWriteRecord() {
		return numRecords < maxRecords;
	}

	/**
	 * Uses the given RecordWriter to write the Record r to this DataBlock. The caller must ensure that
	 * there is space to write a record, by calling {@link #canWriteRecord()}. The caller also must
	 * make sure that the given writer can write records of the expected size.
	 *
	 * @param r Record to write
	 * @param w non-null RecordWriter
	 *
	 * @throws IOException              if writing fails; this block is unusable if this is thrown.
	 * @throws IllegalStateException    if there is not space for a record to be written
	 * @throws IllegalArgumentException if the RecordWriter writes too few or too many bytes. If
	 *                                  getWrittenSize does not match record size, then no harm is
	 *                                  done, but if a different number of bytes is written, then the
	 *                                  write is reverted and this method had to effect.
	 */
	public void writeRecord( Record r, RecordWriter w ) throws IOException {
		int startSize = getSize();
		int endSize   = startSize + recordSize;

		if ( endSize > totalSize )
			throw new IllegalStateException( "No space to write another record" );
		if ( w.getWrittenSize() != recordSize ) {
			throw new IllegalArgumentException( "Given RecordCompressor writes " + w.getWrittenSize() +
			                                    ", but " + recordSize + " expected." );
		}
		if ( r.getTime() <= endTime )
			throw new IllegalArgumentException( "Given record is before previous record (out-of-order)" );

		w.writeRecord( r, data );
		if ( getSize() != endSize ) {
			data.getBuf().position( startSize ); //revert the damage
			throw new IllegalArgumentException(
					"The given Record and compressor wrote " + ( getSize() - startSize ) +
					" instead of the expected " + recordSize + " bytes." );
		}

		if ( startTime < 0 )
			startTime = r.getTime();
		endTime = r.getTime();
		++numRecords;
	}

	/**
	 * Finds the index of the record immediately before or exactly at the given time. Put another way,
	 * finds the index of the record with the largest timestamp less than or equal to the given time.
	 * If the timestamp is before the first record, the index of the first record is given.
	 *
	 * @throws IllegalStateException if there are no records in this DataBlock
	 */
	public int findRecord( long time ) {
		if ( numRecords == 0 )
			throw new IllegalStateException( "findRecord called on a block with no records" );

		final ByteBuffer buf = data.getBuf();

		//The below algorithm is adapted from Arrays.binarySearch.
		int low  = 0;
		int high = numRecords - 1;

		while ( low <= high ) {
			int mid = ( low + high ) >>> 1;
			long midVal = buf.getLong( HEADER_SIZE + mid * recordSize );

			if ( midVal < time )
				low = mid + 1;
			else if ( midVal > time )
				high = mid - 1;
			else
				return mid; //exact timestamp found
		}
		return low > 0 ? low - 1 : low; //exact timestamp not found; return the point below (or 0)
	}

	public Record readRecord( int recordNum, RecordReader reader ) throws IOException {
		if ( reader.getWrittenSize() != recordSize )
			throw new IllegalArgumentException( "Reconstructor's read size is different from DataBlock record size" );
		if ( recordNum < 0 || recordNum >= getNumRecords() )
			throw new IllegalArgumentException( "recordNum too large or negative" );

		//I wanted this to be memory allocation-free for speed purposes, but it does make the code a
		//touch awkward, and there may be the potential for side-effects if a read can change something
		//besides the position and limit.
		final ByteBuffer buf = data.getBuf();
		final int savedPos   = buf.position();
		final int savedLimit = buf.limit();
		final int startIndex = HEADER_SIZE + recordNum * recordSize;

		buf.limit( startIndex + recordSize );
		buf.position( startIndex );
		Record ret = reader.readRecord( data );

		buf.limit( savedLimit );
		buf.position( savedPos );

		return ret;
	}

	private void calculateNumRecords() {
		numRecords = ( getSize() - HEADER_SIZE ) / recordSize;
	}

	/**
	 * Returns the total number of bytes used in the data buffer so far.
	 */
	private int getSize() {
		return data.getBuf().position();
	}

	public boolean equals( Object o ) {
		if ( this == o ) return true;
		if ( o == null || getClass() != o.getClass() ) return false;

		DataBlock dataBlock = (DataBlock) o;

		if ( blockNum   != dataBlock.blockNum  ) return false;
		if ( endTime    != dataBlock.endTime   ) return false;
		if ( recordSize != dataBlock.recordSize) return false;
		if ( startTime  != dataBlock.startTime ) return false;
		if ( totalSize  != dataBlock.totalSize ) return false;
		return Arrays.equals( getRawData(), dataBlock.getRawData() );
	}

	public int hashCode() {
		int result;
		result = blockNum;
		result = 31 * result + recordSize;
		result = 31 * result + totalSize;
		result = 31 * result + (int) ( startTime ^ ( startTime >>> 32 ) );
		result = 31 * result + (int) ( endTime ^ ( endTime >>> 32 ) );
		result = 31 * result + Arrays.hashCode( getRawData() );
		return result;
	}
}
