/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.model;

import org.gillius.Util.MathUtils;
import org.gillius.realdb.BadFileException;
import org.gillius.realdb.io.BlockFile;
import org.gillius.realdb.io.DataOutputWritable;
import org.gillius.realdb.io.DowUtil;
import org.gillius.realdb.io.Formattable;
import org.gillius.realdb.util.DBIterator;

import java.io.Flushable;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.NoSuchElementException;

import static java.lang.Math.*;

/**
 * DataIndex represents a block list index for a data stream in a database.
 * <p/>
 * The design of the DataIndex is meant so that if the underlying BlockFile fails during any single
 * writeBlock(s) operation, the DataIndex can be reloaded into a state that is equal to the last
 * flush, or a state following some or all operations performed since the last flush. Therefore,
 * there is an upper limit on data loss based on the frequency of flushing (internally a partial or
 * full flush results when a buffer fills up).
 * <p/>
 * Currently, the order of operations is not necessarily preserved in the reloaded DataIndex after
 * a failure, but this could be ensured by inserting forced flushes, at a later time if needed.
 *
 * <h2>Fail-safe Implementation</h2>
 *
 * (Only notes for now)
 * Alternate writing for header, head, and tail blocks.
 * Always update tail index after updating tail block, same for head.
 * Need to read all 3 blocks if head=tail.
 * In some cases, a forced write is needed to duplicate a block (but with a newer seq num).
 * Main number one rule: the most recently written block for head or tail must always be pointed to
 * by the header, and it cannot be overwritten in a following write (hence alternate writing).
 * If head==tail, then use only ALT_HEAD.
 *
 * @author Jason Winnebeck
 */
public class DataIndex implements Flushable, Formattable, DataStreamIndex {
	/**
	 * The constant number of blocks that any DataIndex will have. Note this is not the same as the
	 * minimum size, since there must be some index blocks.
	 */
	private static final int NUM_OVERHEAD_BLOCKS = 4;

	/**
	 * The smallest a DataIndex can be in blocks.
	 */
	public static final int MIN_SIZE = NUM_OVERHEAD_BLOCKS + 2; //need two IndexBlocks

	/**
	 * The minimum size that a low-level data block can be, in bytes. Note that a DataIndex operating
	 * on such a file would be drastically inefficient.
	 */
	public static final int MIN_BLOCK_SIZE = max( DataIndexHeader.getSize(),
	                                                   IndexBlock.getSize( 1 ) );

	//TODO: public static final int MAX_DATA_BLOCKS = 2^30?; //based on seq num size and writes/blocks

	/**
	 * The first header block.
	 */
	private static final int H1 = 0;
	/**
	 * Alternate copy of header block.
	 */
	private static final int H2 = H1 + 1; //1
	/**
	 * Alternate copy of index head block.
	 */
	private static final int ALT_HEAD = H2 + 1; //2
	/**
	 * Alternate copy of index tail block.
	 */
	private static final int ALT_TAIL = ALT_HEAD + 1; //3
	/**
	 * The first index block in the DataIndex section.
	 */
	private static final int FIRST_IB = ALT_TAIL + 1; //4

	private final BlockFile file;

	/**
	 * The low-level starting block in the actual BlockFile.
	 */
	private final int startBlock;

	/**
	 * The number of index blocks. The total size is this variable plus NUM_OVERHEAD_BLOCKS.
	 */
	private final int capacity;

	/**
	 * Next sequence number for a written block.
	 */
	private int nextSeqNum;

	/**
	 * Current state of the head, or both in the head==tail case.
	 */
	private IndexBlockFactory headFactory;

	/**
	 * Current state of the tail.
	 */
	private IndexBlockFactory tailFactory;

	/**
	 * Current state of the head pointer in the DataIndexHeader physically on disk.
	 */
	private int currHead;
	/**
	 * Current state of the tail pointer in the DataIndexHeader physically on disk.
	 */
	private int currTail;
	/**
	 * Current state of this header flag in the DataIndexHeader physically on disk.
	 */
	private boolean headAdvanced;
	/**
	 * Current sequence number of the most recent head physically on disk.
	 */
	private int headSeqNum;
	/**
	 * Current sequence number of the most recent tail physically on disk.
	 */
	private int tailSeqNum;

	/**
	 * True if the next block to be written for the header block is the H2 block.
	 */
	private boolean altHeader;
	/**
	 * True if the next block to be written for the head block is the ALT_HEAD block.
	 */
	private boolean altHead;
	/**
	 * True if the next block to be written for the tail block is the ALT_TAIL block.
	 */
	private boolean altTail;
	/**
	 * True if on the next write the next block should be used for the tail. This flag can be set only
	 * if there is room for another tail.
	 */
	private boolean advanceTail;

	private final int indiciesPerBlock;

	/**
	 * List of actions to perform before flushing the head (transaction prerequisites from remove
	 * block).
	 */
	private final ArrayList<Flushable> headFlushables = new ArrayList<Flushable>();
	/**
	 * List of actions to perform before flushing the tail (transaction prerequisites from add block).
	 */
	private final ArrayList<Flushable> tailFlushables = new ArrayList<Flushable>();

	/**
	 * Lazy-initialized, shared, object to flush the head to use with transactions.
	 */
	private HeadFlusher headFlusher = null;

	/**
	 * Lazy-initialized, shared, object to flush the tail to use with transactions.
	 */
	private TailFlusher tailFlusher = null;

	/**
	 * Constructs a new DataIndex, whose data begins at absolute file block 'startBlock' and continues
	 * for 'size' blocks. Whatever data is in "file" is ignored. It is suggested that a format
	 * operation takes place if the file is newly created.
	 *
	 * @param file       non-null BlockFile
	 * @param startBlock Starting block number for this index
	 * @param size       Size of this index in blocks
	 */
	private DataIndex( BlockFile file, int startBlock, int size ) {
		if ( file.getBlockSize() < MIN_BLOCK_SIZE )
			throw new IllegalArgumentException( "Minimum block size supported is " + MIN_BLOCK_SIZE );
		if ( size < MIN_SIZE )
			throw new IllegalArgumentException( "A DataIndex must be at least " + MIN_SIZE + " blocks." );

		this.file        = file;
		this.startBlock  = startBlock;
		this.capacity    = size - NUM_OVERHEAD_BLOCKS;
		indiciesPerBlock = IndexBlock.getMaxDataBlocks( this.file.getBlockSize() );

		//other fields must be set by format or load call.
	}

	/**
	 * Constructs a new DataIndex on the given file to start at the given startBlock and take up
	 * 'size' blocks.
	 *
	 * @param file       non-null BlockFile
	 * @param startBlock Starting block number for this index
	 * @param size       Size of this index in blocks
	 *
	 * @return non-null new DataIndex backed by a newly formatted section of a BlockFile.
	 *
	 * @throws IOException If an error occurs formatting the BlockFile. If this happens, the BlockFile
	 *                     should be considered to be in an inconsistent state between startBlock and
	 *                     startBlock + size (exclusive). It can be recovered by attempting this
	 *                     operation again.
	 */
	public static DataIndex createNew( BlockFile file, int startBlock, int size ) throws IOException {
		DataIndex ret = new DataIndex( file, startBlock, size );
		ret.format();
		return ret;
	}

	/**
	 * Creates a DataIndex from an existing BlockFile, starting at the given startBlock and going for
	 * "size" blocks. The BlockFile is read to discover the current state of the index.
	 *
	 * @param file       non-null BlockFile containing an active database
	 * @param startBlock Starting block number for this index
	 * @param size       Size of this index in blocks
	 *
	 * @return non-null DataIndex
	 *
	 * @throws IOException if an error occurs reading the BlockFile
	 * @throws BadFileException if the file can be read but the data is corrupted
	 */
	public static DataIndex load( BlockFile file, int startBlock, int size )
			throws IOException, BadFileException {
		DataIndex ret = new DataIndex( file, startBlock, size );
		ret.load();
		return ret;
	}

	/**
	 * Sequence number of the tail physically on disk.
	 */
	public int getTailSeqNum() {
		return tailSeqNum;
	}

	/**
	 * Sequence number of the head physically on disk.
	 */
	public int getHeadSeqNum() {
		return headSeqNum;
	}

	public long getStartTime() {
		if ( headFactory.getNumIndices() > 0 )
			return headFactory.getStartTime();
		else
			return -1;
	}

	public long getEndTime() {
		IndexBlockFactory factory = ( tailFactory != null ) ? tailFactory : headFactory;
		if ( factory.getNumIndices() > 0 )
			return factory.getEndTime();
		else
			return -1;
	}

	public int getFirstBlock() {
		if ( headFactory.getNumIndices() > 0 )
			return headFactory.getFirstIndex();
		else
			return -1;
	}

	/**
	 * Return the second data block index for this data stream, or -1 if there is not a second block.
	 * This method is useful for the {@link #removeFirstBlock(long, Transaction)} method, which needs
	 * the time of the first record in the second block. Unlike getFirstBlock, this method might throw
	 * an exception if it needs to read a second block.
	 */
	public int getSecondBlock() throws IOException, BadFileException {
		if ( headFactory.getNumIndices() > 1 ) {
			return headFactory.getSecondIndex();

		} else if ( getNumIndexBlocks() > 1 ) {
			//We know there is a second index if there is more than 1 block (the head)
			DBIterator<Integer> iter = iterator();
			iter.next();
			return iter.next();

		} else {
			return -1;
		}
	}

	@Override
	public int getLastBlock() {
		IndexBlockFactory factory = ( tailFactory != null ) ? tailFactory : headFactory;
		if ( factory.getNumIndices() > 0 )
			return factory.getLastIndex();
		else
			return -1;
	}

	public DBIterator<Integer> iterator( DataBlockSummaryReader summaryReader, long startTime,
	                                     long endTime ) throws IOException, BadFileException {
		if ( headFactory.getNumIndices() == 0 ) {
			//Then there is not a single block index in this DataIndex, so there's nothing to iterate.
			return iterator();
		}
		DataIndexReader iReader = new DataIndexReader();
		int iStart = RangeReaderUtil.findRange( iReader, startTime );
		int iEnd   = RangeReaderUtil.findRange( iReader, endTime );

		//Now we need to find the actual DataBlock within the IndexBlock. The code below would fail
		//if it is given a block with no indices, so we checked for that above.
		int dStart = RangeReaderUtil.findRange(
				new DataBlockRangeReader( summaryReader, iReader.getIndexBlock( iStart ).getDataBlocks() ),
				startTime );
		int dEnd   = RangeReaderUtil.findRange(
				new DataBlockRangeReader( summaryReader, iReader.getIndexBlock( iEnd ).getDataBlocks() ),
				endTime );

		return new DataIndexIterator( new DataIndexReader(), iStart, dStart, iEnd, dEnd );
	}

	public DBIterator<Integer> iterator() {
		return new DataIndexIterator( new DataIndexReader() );
	}

	/**
	 * Reads the entire DataIndex and returns a cached result of all of the data indices in order.
	 * This method scales linearly with data index size and could take a long time if there are many
	 * blocks to read, so use {@link #iterator()} for a streaming method if there is concern about
	 * hitting any memory limit.
	 *
	 * @return non-null new List of 0 or more non-null Integer.
	 */
	public List<Integer> getAllIndicies() throws IOException, BadFileException {
		ArrayList<Integer> ret = new ArrayList<Integer>( getNumIndexBlocks() * indiciesPerBlock );
		DBIterator<Integer> iter = iterator();
		while ( iter.hasNext() ) {
			ret.add( iter.next() );
		}
		return ret;
	}

	/**
	 * Removes the first block as being allocated to this data stream.
	 * <p/>
	 * A note about transactions: Since the head and tail could be in the same physical block, it is
	 * NOT supported to have more than one remove and/or add operation within the same Transaction.
	 * Attempting to do so will result in an infinite recursion as it is impossible to flush head
	 * before tail in this case. A possible workaround is to force a flush of the index after calling
	 * this method and before calling another add/remove in the same transaction.
	 *
	 * @param secondBlockStartTime the time of the first data point in the second block, if there is
	 *                             one. If there isn't a second block, this parameter is ignored. If
	 *                             that time is known (because it's the start of the next index
	 *                             block), then it is checked for equality with this parameter. If
	 *                             IllegalArgumentException is thrown because of this, the operation
	 *                             still took effect, and the index is still in a good state (the
	 *                             bad parameter was ignored). This is arguably a bug, but treat this
	 *                             like an assert failure.
	 * @param tx           nullable Transaction that, if set, adds an action to flush this operation.
	 *                     The Flushable returned from that method is added internally to a list of
	 *                     actions to perform if this operation is flushed implicitly or explicitly.
	 *
	 * @throws IOException            If this operation caused a write that failed.
	 * @throws NoSuchElementException If there are no data block indices in this index.
	 */
	public void removeFirstBlock( long secondBlockStartTime, Transaction tx ) throws IOException {
		if ( headFactory.getNumIndices() == 0 )
			throw new NoSuchElementException( "No data block indices to remove" );

		headFactory.removeFirstDataBlock( secondBlockStartTime );
		if ( tx != null ) {
			if ( headFlusher == null )
				headFlusher = new HeadFlusher();

			headFlushables.add( tx.addAction( headFlusher ) );
		}
		if ( headFactory.getNumIndices() == 0 ) {
			//If we just emptied the head, we need to flush and move the head pointer if there is a block
			//after.
			boolean advanceHead = currHead != currTail || advanceTail;
			flushHead( advanceHead );
			//If we moved the head, the secondBlockStartTime should match our current head
			if ( advanceHead && headFactory.getStartTime() != secondBlockStartTime )
				throw new IllegalArgumentException( "secondBlockStartTime does not match start next head" );
		}
	}

	/**
	 * Adds the given data block number to the end of the list of allocated blocks for this index.
	 * The startTime values must be increasing on each call to this method (you must add data blocks
	 * in order of time and there must not be any overlaps).
	 * <p/>
	 * A note about transactions: Since the head and tail could be in the same physical block, it is
	 * NOT supported to have more than one remove and/or add operation within the same Transaction.
	 * Attempting to do so will result in an infinite recursion as it is impossible to flush head
	 * before tail in this case. A possible workaround is to force a flush of the index after calling
	 * this method and before calling another add/remove in the same transaction.
	 *
	 * @param dataBlockNum data block number
	 * @param startTime    timestamp of the first Record in the data block
	 * @param endTime      timestamp of the last Record in the data block, >= startTime
	 * @param tx           nullable Transaction that, if set, adds an action to flush this operation.
	 *                     The Flushable returned from that method is added internally to a list of
	 *                     actions to perform if this operation is flushed implicitly or explicitly.
	 */
	public void addDataBlock( int dataBlockNum, long startTime, long endTime, Transaction tx )
			throws IOException {
		//Is the tail distinct from the head currently?
		IndexBlockFactory factory = ( tailFactory != null ) ? tailFactory : headFactory;

		if ( !factory.hasSpace() ) {
			if ( getNumIndexBlocks() == capacity )
				throw new IOException( "No more space available for another data block index" );

			flushTail();
			ensureRealTailFlushed(); //might need to duplicate...

			//Check to see if tail is distinct from head. If not, make a new one.
			if ( tailFactory == null ) {
				assert( currHead == currTail );
				tailFactory = new IndexBlockFactory( indiciesPerBlock, headFactory.getEndTime() );
				factory = tailFactory;
			} else {
				tailFactory.startNextBlock();
			}
			advanceTail = true;
		}

		//factory handles start/endTime checking
		factory.addDataBlock( dataBlockNum, startTime, endTime );
		if ( tx != null ) {
			if ( tailFlusher == null )
				tailFlusher = new TailFlusher();

			tailFlushables.add( tx.addAction( tailFlusher ) );
		}
	}

	/**
	 * Updates the last data block added via {@link #addDataBlock(int, long, long, Transaction)}.
	 *
	 * @param dataBlockNum data block number, exactly equal to the last call of addDataBlock
	 * @param startTime    data block start time, exactly equal to the last call of addDataBlock
	 * @param endTime      new endTime, greater than or equal to the last call of addDataBlock
	 */
	public void updateDataBlock( int dataBlockNum, long startTime, long endTime ) {
		//TODO: this method is currently unused; consider removal
		tailFactory.updateDataBlock( dataBlockNum, startTime, endTime );
	}

	public void flush() throws IOException {
		//TODO: do an actual flush only if needed
		flushHead( false );
		flushTail();
	}

	public void format() throws IOException {
		//Set all of the blocks to be invalid. Remember, this operation should be restartable but does
		//not need to be fail-safe.
		byte[] zeros = new byte[ file.getBlockSize() ];
		//H1 will be written later
		file.writeBlock( startBlock + H2, zeros );
		file.writeBlock( startBlock + ALT_HEAD, zeros );
		file.writeBlock( startBlock + ALT_TAIL, zeros );
		//FIRST_IB will be written later

		//Write initial header and head/tail block
		nextSeqNum = 0;
		writeBlock( H1,       new DataIndexHeader( nextSeqNum, 0, false, 0 ) );
		writeBlock( FIRST_IB, new IndexBlock( nextSeqNum ) );
		headSeqNum = tailSeqNum = nextSeqNum - 1;

		currHead  = 0;
		currTail  = 0;
		altHeader = true;  //since we just wrote the primary version
		altHead   = true;  //since we just wrote the primary version
		altTail   = false; //ignored on the first advanceTail, but we say false to be consistent

		//start at 1 instead of 0 because of the first "empty" block that we just wrote
		headFactory = new IndexBlockFactory( indiciesPerBlock, -1 );
		tailFactory = null;
	}

	/**
	 * Returns the size in blocks required to hold an index capable of tracking 'numDataBlocks'.
	 *
	 * @param numDataBlocks the maximum size allowed for this index in data blocks
	 * @param blockSize     the size of a file block, in bytes
	 */
	public static int getIndexSize( int numDataBlocks, int blockSize ) {
		if ( blockSize < MIN_BLOCK_SIZE )
			throw new IllegalArgumentException( "Minimum block size supported is " + MIN_BLOCK_SIZE );

		int indiciesPerBlock = IndexBlock.getMaxDataBlocks( blockSize );
		//Plus 1 because of wasted space in head and tail
		int blocksNeeded = MathUtils.divideRoundUp( numDataBlocks, indiciesPerBlock ) + 1;

		return NUM_OVERHEAD_BLOCKS + blocksNeeded;
	}

	public boolean equals( Object o ) {
		if ( this == o ) return true;
		if ( o == null || getClass() != o.getClass() ) return false;

		DataIndex dataIndex = (DataIndex) o;

		if ( advanceTail      != dataIndex.advanceTail     ) return false;
		if ( altHead          != dataIndex.altHead         ) return false;
		if ( altHeader        != dataIndex.altHeader       ) return false;
		if ( altTail          != dataIndex.altTail         ) return false;
		if ( capacity         != dataIndex.capacity        ) return false;
		if ( currHead         != dataIndex.currHead        ) return false;
		if ( currTail         != dataIndex.currTail        ) return false;
		if ( headAdvanced     != dataIndex.headAdvanced    ) return false;
		if ( indiciesPerBlock != dataIndex.indiciesPerBlock) return false;
		if ( nextSeqNum       != dataIndex.nextSeqNum      ) return false;
		if ( startBlock       != dataIndex.startBlock      ) return false;
		if ( !file.equals( dataIndex.file )                ) return false;
		if ( headFactory != null ? !headFactory.equals( dataIndex.headFactory ) :
		     dataIndex.headFactory != null ) return false;
		if ( headFlushables != null ? !headFlushables.equals( dataIndex.headFlushables ) :
		     dataIndex.headFlushables != null ) return false;
		if ( headFlusher != null ? !headFlusher.equals( dataIndex.headFlusher ) :
		     dataIndex.headFlusher != null ) return false;
		if ( tailFactory != null ? !tailFactory.equals( dataIndex.tailFactory ) :
		     dataIndex.tailFactory != null ) return false;
		if ( tailFlushables != null ? !tailFlushables.equals( dataIndex.tailFlushables ) :
		     dataIndex.tailFlushables != null ) return false;
		if ( tailFlusher != null ? !tailFlusher.equals( dataIndex.tailFlusher ) :
		     dataIndex.tailFlusher != null ) return false;

		return true;
	}

	public int hashCode() {
		int result;
		result = file.hashCode();
		result = 31 * result + startBlock;
		result = 31 * result + capacity;
		result = 31 * result + nextSeqNum;
		result = 31 * result + ( headFactory != null ? headFactory.hashCode() : 0 );
		result = 31 * result + ( tailFactory != null ? tailFactory.hashCode() : 0 );
		result = 31 * result + currHead;
		result = 31 * result + currTail;
		result = 31 * result + ( headAdvanced ? 1 : 0 );
		result = 31 * result + ( altHeader ? 1 : 0 );
		result = 31 * result + ( altHead ? 1 : 0 );
		result = 31 * result + ( altTail ? 1 : 0 );
		result = 31 * result + ( advanceTail ? 1 : 0 );
		result = 31 * result + indiciesPerBlock;
		result = 31 * result + ( headFlushables != null ? headFlushables.hashCode() : 0 );
		result = 31 * result + ( tailFlushables != null ? tailFlushables.hashCode() : 0 );
		result = 31 * result + ( headFlusher != null ? headFlusher.hashCode() : 0 );
		result = 31 * result + ( tailFlusher != null ? tailFlusher.hashCode() : 0 );
		return result;
	}

	/**
	 * Writes out either an IndexBlock or DataIndexHeader to the file and increments nextSeqNum, only
	 * if writing was successful.
	 *
	 * @param idx   block index relative to start of DataIndex
	 * @param block block to write
	 */
	private void writeBlock( int idx, DataOutputWritable block ) throws IOException {
		byte[] data = DowUtil.serialize( block );
		if ( data.length > file.getBlockSize() )
			throw new AssertionError( "Block is too large to fit" );

		file.writeBlocks( startBlock + idx, data );
		writeLog( "Wrote block at absolute index %d", startBlock + idx );
		++nextSeqNum;
	}

	private void writeHeader( int nextHead, boolean nextHeadAdvanced, int nextTail )
			throws IOException {
		DataIndexHeader h = new DataIndexHeader( nextSeqNum, nextHead, nextHeadAdvanced, nextTail );
		writeLog( "About to write header: idx=%d, currHead=%d, nextTail=%d",
		          (altHeader ? H2 : H1), currHead, nextTail );
		writeBlock( altHeader ? H2 : H1, h );
		altHeader = !altHeader;
//		CircularLog.writeLog( "Index %d: wrote header %s", startBlock, h );
		headAdvanced = nextHeadAdvanced;
		currHead     = nextHead;
		currTail     = nextTail;
	}

	private void writeLog( String format, Object... args ) {
//		Object[] newArgs = new Object[ args.length + 1 ];
//		newArgs[0] = startBlock;
//		System.arraycopy( args, 0, newArgs, 1, args.length );
//		CircularLog.writeLog( "Index %d: " + format, newArgs );
	}

	/**
	 * Flushes the head. Every statement in this method is carefully ordered to make it exception and
	 * crash safe, except in the case where a following index block cannot be read. If this occurs,
	 * then there is unexpected corruption (there is no repair operation yet), and currently this
	 * throws IllegalStateException to indicate that we cannot continue.
	 *
	 * @param advanceHead If we are moving up the head pointer. This cannot be true if head==tail.
	 */
	private void flushHead( boolean advanceHead ) throws IOException {
		flushAll( headFlushables );
		//TODO: can we optimize out this write in some cases?
		//Ensure that the tail is realized
		if ( advanceTail )
			flushTail();

		if ( advanceHead ) {
			if ( currHead == currTail )
				throw new AssertionError( "advanceHead set when head==tail" );

			int nextHead = ( currHead + 1 ) % capacity;
			if ( nextHead == currTail ) {
				//Merge the tail and head

				//Ensure that tail is in main position
				ensureRealTailFlushed();
				writeHeader( nextHead, false, currTail );
				headFactory = tailFactory;
				headSeqNum = tailSeqNum;
				tailFactory = null;

			} else {
				//Move up the header and repopulate headFactory
				writeHeader( nextHead, true, currTail );
				IndexBlock next;
				try {
					next = loadRawIndexBlock( nextHead );
				} catch ( BadFileException e ) {
					throw new IllegalStateException( "Error advancing head; couldn't read next block. " +
					                                 "DataIndex corrupted", e );
				}
				headFactory = new IndexBlockFactory( indiciesPerBlock, next);
				headSeqNum = next.getSeqNum();
			}
			altHead = true;

		} else {
			//If head==tail, then we're also flushing the tail, so we need to clear those flushables
			if ( currHead == currTail ) {
				flushAll( tailFlushables );
				tailFlusher = null;
			}

			int idx = altHead ? ALT_HEAD : FIRST_IB + currHead;
			writeLog( "Writing head index block at %d", idx );
			writeBlock( idx, headFactory.makeBlock( nextSeqNum ) );
			headSeqNum = nextSeqNum - 1;
			if ( currHead == currTail )
				tailSeqNum = headSeqNum;

			//After we wrote the block, we want to make to "re-enable" ALT_HEAD on reloading
			if ( headAdvanced ) {
				//If we just advanced the head, the only safe write would have been to ALT_HEAD.
				if ( !altHead )
					throw new AssertionError( "altHead false when headAdvanced is true" );

				writeHeader( currHead, false, currTail );
			}
			altHead = !altHead;
		}

		//Effectively notify the existing headFlusher that it no longer needs to flush
		headFlusher = null;
	}

	/**
	 * Flushes all of the Flushable objects from the given List and removes them one at a time, from
	 * the end. This is done so that if any Flushable throws an IOException, it and all previous
	 * Flushable objects are NOT removed. Iteration is done backwards since this is the most
	 * efficient way to clear an ArrayList, as it prevents copying.
	 */
	private static void flushAll( ArrayList<? extends Flushable> flushables ) throws IOException {
		while ( flushables.size() > 0 ) {
			int lastIdx = flushables.size() - 1;
			flushables.get( lastIdx ).flush();
			flushables.remove( lastIdx );
		}
	}

	/**
	 * Flushes the tail. Every statement in this method is carefully ordered to make it exception and
	 * crash safe. Relevant rules:
	 * <p/>
	 * Write next tail block before moving the header. If tail block write fails; it never happened.
	 * If header write fails, also tail block was never written.
	 */
	private void flushTail() throws IOException {
		flushAll( tailFlushables );
		int nextTail = currTail;

		if ( currHead == currTail && !advanceTail ) {
			flushHead( false );

		} else {
			if ( advanceTail ) {
				altTail = false;
				nextTail = ( currTail + 1 ) % capacity;
			}
			int idx = altTail ? ALT_TAIL : FIRST_IB + nextTail;
			writeLog( "Writing tail block at %d", idx );
			writeBlock( idx, tailFactory.makeBlock( nextSeqNum ) );
			tailSeqNum = nextSeqNum - 1;
			altTail = !altTail;

			if ( advanceTail ) {
				//TODO: why is advanceHead set false (rather than unchanged?)
				writeHeader( currHead, false, nextTail );
				advanceTail = false;
			}
		}

		//Effectively notify the existing tailFlusher that it no longer needs to flush
		tailFlusher = null;
	}

	/**
	 * If the latest copy of the tail is in ALT_TAIL, flush the latest data to the "real" tail block.
	 */
	private void ensureRealTailFlushed() throws IOException {
		if ( !altTail )
			flushTail();
	}

	/**
	 * Returns the number of index blocks currently in use. The {@link #capacity} is the most that
	 * this number can be.
	 */
	private int getNumIndexBlocks() {
		//If advanceTail is set then we have another block not yet flushed
		int modifier = advanceTail ? 1 : 0;

		if ( currTail > currHead )
			return modifier + currTail - currHead + 1;
		else if ( currTail == currHead )
			return modifier + 1;
		else
			return modifier + capacity - currHead + currTail + 1; //end plus beginning
	}

	private void load() throws IOException, BadFileException {
		//Pick the most recent header that passes checksum
		byte[] firstHeader  = new byte[DataIndexHeader.getSize()];
		byte[] secondHeader = new byte[DataIndexHeader.getSize()];
		file.readBlocks( startBlock + H1, firstHeader );
		file.readBlocks( startBlock + H2, secondHeader );

		int[] bestIndex = new int[1];
		DataIndexHeader header = SequencedBlockUtil.loadBestBlock( DataIndexHeader.loader, bestIndex,
		                                                           firstHeader, secondHeader );

		//If the firstHeader was the best block, then we want to write to alternate header block next
		altHeader = ( bestIndex[0] == 0 );

		currHead     = header.getHead();
		currTail     = header.getTail();
		headAdvanced = header.isHeadAdvanced();

		IndexBlock head;
		if ( !headAdvanced ) {
			head    = loadHeadBlock( bestIndex );
			altHead = ( bestIndex[0] == 0 );

		} else {
			//If the head was just advanced, then we ignore the ALT_HEAD block.
			head    = loadRawIndexBlock( currHead );
			altHead = true;
		}
		headFactory = new IndexBlockFactory( indiciesPerBlock, head );
		headSeqNum = head.getSeqNum();

		if ( currHead == currTail ) {
			nextSeqNum  = head.getSeqNum() + 1;
			tailSeqNum = head.getSeqNum(); //tail is also the head
			tailFactory = null;
			altTail     = false; //although this really doesn't matter, but to be consistent.

		} else {
			IndexBlock tail = loadTailBlock( bestIndex );
			tailSeqNum = tail.getSeqNum();
			altTail = ( bestIndex[0] == 0 );
			nextSeqNum  = max( max( head.getSeqNum(), tail.getSeqNum() ), header.getSeqNum() ) + 1;
			tailFactory = new IndexBlockFactory( indiciesPerBlock, tail );
		}
	}

	/**
	 * Loads an IndexBlock from a given raw offset, not by logical block number (where 0 would be the
	 * head).
	 */
	private IndexBlock loadRawIndexBlock( int idx ) throws IOException, BadFileException {
		byte[] buf = new byte[ file.getBlockSize() ];
		file.readBlock( startBlock + FIRST_IB + idx, buf );
		return new IndexBlock( buf );
	}

	/**
	 * Returns the best head IndexBlock based on 'currHead' only.
	 *
	 * @param best out parameter; if block loaded, set best[0] to 0 if primary block is best, else
	 *             set to 1 if alternate block is best.
	 * @return non-null IndexBlock
	 */
	private IndexBlock loadHeadBlock( int[] best ) throws IOException, BadFileException {
		return getBestIndexBlock( best, currHead + FIRST_IB, ALT_HEAD );
	}

	/**
	 * Loads the best tail IndexBlock based on 'currTail' only.
	 *
	 * @param best out parameter; if block loaded, set best[0] to 0 if primary block is best, else
	 *             set to 1 if alternate block is best.
	 * @return non-null IndexBlock
	 */
	private IndexBlock loadTailBlock( int[] best ) throws IOException, BadFileException {
		return getBestIndexBlock( best, currTail + FIRST_IB, ALT_TAIL );
	}

	private IndexBlock getBestIndexBlock( int[] best, int... blocks )
			throws IOException, BadFileException {
		byte[][] data = new byte[blocks.length][];
		for ( int i = 0; i < data.length; ++i ) {
			data[i] = new byte[ file.getBlockSize() ];
			file.readBlock( startBlock + blocks[i], data[i] );
		}

		return SequencedBlockUtil.loadBestBlock( IndexBlock.loader, best, data );
	}

	private class DataIndexReader implements DataStreamIndexReader {
		final int    numBlocks;
		final byte[] buf;

		public DataIndexReader() {
			//We cache this to keep from recalculating it each time
			numBlocks = getNumIndexBlocks();
			buf       = new byte[ file.getBlockSize() ];
		}

		public int getNumRanges() {
			return numBlocks;
		}

		/**
		 * "Smart" method of getting index blocks AFTER DataIndex is loaded, by taking advantage of
		 * the cached head or tail. The sequence number of blocks returned from this method should be
		 * ignored.
		 */
		public IndexBlock getIndexBlock( int idx ) throws IOException, BadFileException {
			if ( idx < 0 || idx >= numBlocks )
				throw new IndexOutOfBoundsException( "" + idx );

			if ( idx == 0 ) {
				//Use the cached head (or head/tail)
				return headFactory.makeBlock( 0 );

			} else if ( idx == numBlocks - 1 ) {
				//Use the cached tail. We know there is a distinct tail because we checked head first
				return tailFactory.makeBlock( 0 );

			} else {
				//middle block -- need to load it from the file
				int ibIdx = startBlock + FIRST_IB + ( currHead + idx ) % capacity;
				file.readBlock( ibIdx, buf );
				return new IndexBlock( buf );
			}
		}

		public Range getRange( int idx ) throws IOException, BadFileException {
			return getIndexBlock( idx );
		}
	}

	private class TailFlusher implements Flushable {
		public void flush() throws IOException {
			//If we are still active, then do the flush
			if ( tailFlusher == this )
				flushTail();
		}
	}

	private class HeadFlusher implements Flushable {
		public void flush() throws IOException {
			//If we are still active, then do the flush
			if ( headFlusher == this )
				flushHead( false );
		}
	}
}
