/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.model;

import org.gillius.realdb.BadFileException;

import java.io.DataOutput;
import java.io.IOException;
import java.nio.ByteBuffer;

/**
 * DataIndexHeader represents a header block in a stream's {@link DataIndex}.
*
* @author Jason Winnebeck
*/
public class DataIndexHeader implements SequencedBlock {
	/**
	 * A global static final thread-safe instance of a SequencedBlockLoader to load DataIndexHeaders.
	 */
	public static final SequencedBlockLoader<DataIndexHeader> loader =
			new SequencedBlockLoader<DataIndexHeader>() {
				public DataIndexHeader load( byte[] serialized ) throws BadFileException {
					return new DataIndexHeader( serialized );
				}
			};

	private final int seqNum;
	private final int head;
	private final int tail;

	/**
	 * True if the head pointer is newly advanced and the alternate head block (backup) should NOT
	 * be considered.
	 */
	private final boolean headAdvanced;

	/**
	 * Returns the size of the index header.
	 */
	public static int getSize() {
		return 17; //3 fields of 4 bytes + headAdvancedByte + checksum
	}

	/**
	 * Constructs a new DataIndexHeader with the given information.
	 *
	 * @param seqNum       block sequence number
	 * @param head         index of the head block
	 * @param headAdvanced true if backup head block should not be considered.
	 * @param tail         index of the tail block
	 */
	public DataIndexHeader( int seqNum, int head, boolean headAdvanced, int tail ) {
		this.seqNum       = seqNum;
		this.head         = head;
		this.tail         = tail;
		this.headAdvanced = headAdvanced;
	}

	/**
	 * Constructs a new DataIndexHeader by reading it from the given byte array.
	 *
	 * @param data non-null byte[]
	 *
	 * @throws BadFileException if the data integrity of the block is not correct
	 */
	public DataIndexHeader( byte[] data ) throws BadFileException {
		ByteBuffer buf = ByteBuffer.wrap( data );
		seqNum = buf.getInt();
		head   = buf.getInt();
		tail   = buf.getInt();
		byte headAdvancedByte = buf.get();
		if ( headAdvancedByte != 0 && headAdvancedByte != 1 )
			throw new BadFileException( "Invalid valid for headAdvanced: " + headAdvancedByte );
		headAdvanced = headAdvancedByte != 0;

		int check = buf.getInt();
		if ( tail - head + seqNum + headAdvancedByte + 1 != check )
			throw new BadFileException( "DataIndexHeader does not pass checksum" );
	}

	public int getSeqNum() {
		return seqNum;
	}

	public int getHead() {
		return head;
	}

	public int getTail() {
		return tail;
	}

	public boolean isHeadAdvanced() {
		return headAdvanced;
	}

	public void write( DataOutput output ) throws IOException {
		output.writeInt( seqNum );
		output.writeInt( head );
		output.writeInt( tail );
		byte headAdvancedByte = headAdvanced ? (byte) 1 : 0;
		output.write( headAdvancedByte );
		//calculate a simple, fast "checksum", should protect against most random data and all 0's
		output.writeInt( tail - head + seqNum + headAdvancedByte + 1 );
	}

	public boolean equals( Object o ) {
		if ( this == o ) return true;
		if ( o == null || getClass() != o.getClass() ) return false;

		DataIndexHeader that = (DataIndexHeader) o;

		if ( head != that.head ) return false;
		if ( headAdvanced != that.headAdvanced ) return false;
		if ( seqNum != that.seqNum ) return false;
		if ( tail != that.tail ) return false;

		return true;
	}

	public int hashCode() {
		int result;
		result = seqNum;
		result = 31 * result + head;
		result = 31 * result + tail;
		result = 31 * result + ( headAdvanced ? 1 : 0 );
		return result;
	}

	public String toString() {
		return "DataIndexHeader{" +
		       "seqNum=" + seqNum +
		       ", head=" + head +
		       ", tail=" + tail +
		       ", headAdvanced=" + headAdvanced +
		       '}';
	}
}
