/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.model;

import org.gillius.realdb.BadFileException;
import org.gillius.realdb.util.DBIterator;

import java.io.IOException;
import java.util.NoSuchElementException;

/**
 * DataIndexIterator implements a generic method to iterate over data block indices of a
 * {@link DataStreamIndex} by using a {@link DataStreamIndexReader}.
 *
 * @author Jason Winnebeck
 */
public class DataIndexIterator implements DBIterator<Integer> {
	final DataStreamIndexReader reader;

	final int endBlock;
	final int endIndex;

	int   currBlock = 0;
	int   currIndex = 0;
	int[] data      = null;

	/**
	 * Constructs a DataIndexIterator that iterates over all data block indices.
	 *
	 * @param reader non-null DataStreamIndexReader
	 */
	public DataIndexIterator( DataStreamIndexReader reader ) {
		this.reader = reader;
		endBlock    = reader.getNumRanges() - 1;
		endIndex    = Integer.MAX_VALUE;
	}

	/**
	 * Constructs a DataIndexIterator that starts at the given position in the starting index block,
	 * and will proceed no further than the given endIndex position of the ending index block
	 *
	 * @param reader     non-null DataStreamIndexReader
	 * @param startBlock starting index block number
	 * @param startIndex starting position in that block
	 * @param endBlock   index block number of the last block to iterate, can be larger than the number
	 *                   of total blocks (like Integer.MAX_VALUE), if so iterates to the end.
	 * @param endIndex   position to iterate to (inclusive), can be larger than the number of data
	 *                   indices (like Integer.MAX_VALUE) to iterate to the end.
	 *
	 * @throws IOException If there is an error reading from the DataStreamIndexReader.
	 * @throws BadFileException If the database can be read but is corrupted
	 */
	public DataIndexIterator( DataStreamIndexReader reader, int startBlock, int startIndex,
	                          int endBlock, int endIndex )
			throws IOException, BadFileException {
		int numBlocks = reader.getNumRanges();
		if ( endBlock >= numBlocks )
			endBlock = numBlocks - 1;
		if ( startBlock < 0 || startBlock > endBlock || startBlock >= numBlocks )
			throw new IllegalArgumentException( "0 < startBlock < endBlock && startBlock < numBlocks" );
		if ( endIndex < 0 )
			throw new IllegalArgumentException( "endIndex < 0" );

		this.reader    = reader;
		this.currBlock = startBlock;
		this.currIndex = startIndex;
		this.endBlock  = endBlock;
		this.endIndex  = endIndex;

		hasNext(); //pre-load it now just to check the index
		if ( startIndex >= data.length )
			throw new IllegalArgumentException( "startIndex >= data.length" );
	}

	public boolean hasNext() throws IOException, BadFileException {
		if ( data == null ) {
			if ( currBlock <= endBlock ) {
				data = reader.getIndexBlock( currBlock ).getDataBlocks();
			} //else, leave data null
		}

		return data != null && data.length > 0;
	}

	public Integer next() throws IOException, BadFileException {
		if ( !hasNext() )
			throw new NoSuchElementException();

		int ret = data[currIndex];
		++currIndex;
		if ( currIndex >= data.length || ( currIndex == endIndex && currIndex >= endIndex ) ) {
			//Move to the next block and reset our cached data
			data = null;
			currIndex = 0;
			++currBlock;
		}

		return ret;
	}
}
