/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.model;

import org.gillius.Util.CheckUtils;
import org.gillius.realdb.BadFileException;
import org.gillius.realdb.io.BlockFile;
import org.gillius.realdb.io.Formattable;

import java.io.Flushable;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

/**
 * DataIndexSection represents a dynamic section of the database that keeps track of free and
 * allocated blocks. The primary purpose is to provide a linear view of the data stream when
 * reading, and to provide information to assist in block allocation/recycling.
 *
 * @author Jason Winnebeck
 */
public class DataIndexSection implements Formattable, Flushable {
	private final List<DataIndex> indices;

	/**
	 * Size of this section, in blocks
	 */
	private final int size;

	/**
	 * Manual construction of a DataIndexSection; Normally the {@link #createNew} and {@link #load}
	 * factory methods would be used.
	 */
	public DataIndexSection( List<DataIndex> indices, int size ) {
		CheckUtils.checkNull( indices, "indices" );

		this.indices = indices;
		this.size    = size;
	}

	/**
	 * Creates and formats a new DataIndexSection in the given BlockFile.
	 *
	 * @param file           non-null BlockFile
	 * @param startBlock     absolute file starting block for this section
	 * @param streams        number of data streams to hold
	 * @param numDataBlocks  number of data blocks in the DataSection of the Database
	 * @param numIndexBlocks size of this section, in blocks, must be large enough
	 *
	 * @return non-null DataIndexSection
	 */
	public static DataIndexSection createNew( BlockFile file, int startBlock, int streams,
	                                          int numDataBlocks, int numIndexBlocks )
			throws IOException {
		try {
			return new DataIndexSection( file, startBlock, streams, numDataBlocks, numIndexBlocks, false );
		} catch ( BadFileException e ) {
			throw new RuntimeException( "BadFileException cannot occur on creation", e );
		}
	}

	/**
	 * Loads a DataIndexSection in the given BlockFile.
	 *
	 * @param file           non-null BlockFile
	 * @param startBlock     absolute file starting block for this section
	 * @param streams        number of data streams to hold
	 * @param numDataBlocks  number of data blocks in the DataSection of the Database
	 * @param numIndexBlocks size of this section, in blocks, must be large enough
	 *
	 * @return non-null DataIndexSection
	 */
	public static DataIndexSection load( BlockFile file, int startBlock, int streams,
	                                     int numDataBlocks, int numIndexBlocks )
			throws IOException, BadFileException {
		return new DataIndexSection( file, startBlock, streams, numDataBlocks, numIndexBlocks, true );
	}

	/**
	 * Returns the minimum size ('numIndexBlocks') parameter for a DataIndexSection with the given
	 * parameters.
	 *
	 * @param streams       number of data streams to hold
	 * @param numDataBlocks number of data blocks in the DataSection of the Database
	 * @param blockSize     size of a low-level file block in bytes
	 */
	public static int getMinimumSize( int streams, int numDataBlocks, int blockSize ) {
		return DataIndex.getIndexSize( numDataBlocks, blockSize ) * streams;
	}

	/**
	 * Constructs a new DataIndexSection.
	 *
	 * @param file           non-null BlockFile to load from
	 * @param startBlock     absolute file starting block for this section
	 * @param streams        number of data streams to hold
	 * @param numDataBlocks  number of data blocks in the DataSection of the Database
	 * @param numIndexBlocks size of this section, in blocks, must be large enough
	 * @param load           true if loading, false if creating
	 *
	 * @throws IOException if there is an error reading/writing to the BlockFile
	 * @throws BadFileException if load is true and the file is corrupted
	 */
	private DataIndexSection( BlockFile file, int startBlock, int streams, int numDataBlocks,
	                          int numIndexBlocks, boolean load ) throws IOException, BadFileException {
		int indexSize = DataIndex.getIndexSize( numDataBlocks, file.getBlockSize() );
		int usedSize = indexSize * streams;

		size = numIndexBlocks;
		if ( size < usedSize ) {
			throw new IllegalArgumentException( "The size given is not large enough to hold indicies " +
			                                    "for " + streams + " streams with " + numDataBlocks +
			                                    " data blocks." );
		}

		List<DataIndex> indices = new ArrayList<DataIndex>( streams );
		for ( int i = 0; i < streams; ++i ) {
			if ( load )
				indices.add( DataIndex.load( file, startBlock, indexSize ) );
			else
				indices.add( DataIndex.createNew( file, startBlock, indexSize ) );
			startBlock += indexSize;
		}
		this.indices = Collections.unmodifiableList( indices );
	}

	/**
	 * Returns the size of this section, in blocks.
	 */
	public int getSize() {
		return size;
	}

	/**
	 * Returns a list of all indices in this section. The returned list must not be modified (although
	 * any methods on the indices themselves can be called).
	 */
	public List<DataIndex> getIndices() {
		return indices;
	}

	/**
	 * Returns a specific DataIndex in this section.
	 */
	public DataIndex getIndex( int indexNum ) {
		return indices.get( indexNum );
	}

	/**
	 * Returns the number of DataIndex objects in this section.
	 */
	public int numIndices() {
		return indices.size();
	}

	public void format() throws IOException {
		for ( DataIndex index : indices ) {
			index.format();
		}
	}

	public void flush() throws IOException {
		for ( DataIndex index : indices ) {
			index.flush();
		}
	}

	public boolean equals( Object o ) {
		if ( this == o ) return true;
		if ( o == null || getClass() != o.getClass() ) return false;

		DataIndexSection that = (DataIndexSection) o;

		if ( size != that.size ) return false;
		if ( !indices.equals( that.indices ) ) return false;

		return true;
	}

	public int hashCode() {
		int result;
		result = indices.hashCode();
		result = 31 * result + size;
		return result;
	}
}
