/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.model;

import org.gillius.realdb.BadFileException;
import org.gillius.realdb.io.BlockFile;

import java.io.IOException;

/**
 * DataSection is the last and (usually) the largest section of the database, and contains the
 * actual data records for the data streams.
 *
 * @author Jason Winnebeck
 */
public class DataSection {
	/**
	 * Backing database file.
	 */
	public final BlockFile file;

	/**
	 * Starting block index, an absolute low-level file block index.
	 */
	private final int startBlock;

	/**
	 * Size of a data block in low-level file blocks.
	 */
	private final int blockSize;

	/**
	 * Number of data blocks.
	 */
	private final int numBlocks;

	/**
	 * Constructs a new DataSection.
	 *
	 * @param file       non-null BlockFile
	 * @param startBlock starting physical block in the BlockFile
	 * @param blockSize  size of a logical data block, in physical blocks
	 * @param numBlocks  number of logical data blocks
	 */
	public DataSection( BlockFile file, int startBlock, int blockSize, int numBlocks ) {
		if ( blockSize < 0 )
			throw new IllegalArgumentException( "blockSize must be positive" );
		if ( numBlocks < 0 )
			throw new IllegalArgumentException( "numBlocks must be positive" );
		if ( startBlock < 0 )
			throw new IllegalArgumentException( "startBlock must be positive" );

		int physicalSize = blockSize * numBlocks;
		if ( startBlock + physicalSize > file.getSize() )
			throw new IllegalArgumentException( "section is too big to fit in the file" );

		this.file       = file;
		this.startBlock = startBlock;
		this.blockSize  = blockSize;
		this.numBlocks  = numBlocks;
	}

	/**
	 * Returns the absolute low-level file block index for the requested data block index.
	 */
	public int getDataBlockStart( int dataBlock ) {
		return startBlock + blockSize * dataBlock;
	}

	/**
	 * Returns the size of a data block in low-level file blocks.
	 */
	public int getBlockSize() {
		return blockSize;
	}

	/**
	 * Returns the size of a data block in bytes.
	 */
	public int getBlockSizeBytes() {
		return blockSize * file.getBlockSize();
	}

	/**
	 * Returns the number of logical data blocks in this section.
	 */
	public int getNumBlocks() {
		return numBlocks;
	}

	public void writeBlock( DataBlock block ) throws IOException {
		if ( block.getTotalSize() != getBlockSizeBytes() )
			throw new IllegalArgumentException( "Given DataBlock is not of the proper size" );
		int blockNum = block.getBlockNum();
		if ( blockNum < 0 || blockNum >= numBlocks )
			throw new IllegalArgumentException( "Invalid block number, " + blockNum );

		file.writeBlocks( getDataBlockStart( blockNum ), block.getRawData() );
	}

	/**
	 * Reads a block directly from the BlockFile.
	 * @param blockNum   index of logical data block
	 * @param recordSize size of a record within that block
	 *
	 * @return non-null DataBlock
	 *
	 * @throws IOException if the block could not be read
	 * @throws BadFileException if the block loaded was corrupted
	 */
	public DataBlock readBlock( int blockNum, int recordSize ) throws IOException, BadFileException {
		if ( blockNum >= numBlocks )
			throw new IllegalArgumentException( "Given block number " + blockNum + " is too large" );

		//TODO: optimize; share buffer
		byte[] buf = new byte[ getBlockSizeBytes() ];
		file.readBlocks( getDataBlockStart( blockNum ), buf );
		return DataBlock.load( buf, blockNum, recordSize, buf.length );
	}

	/**
	 * Reads summary information about a block directly from the BlockFile.
	 *
	 * @param blockNum index of logical data block
	 *
	 * @return non-null DefaultRange
	 *
	 * @throws IOException      if the block could not be read
	 * @throws BadFileException if the block loaded was corrupted
	 */
	public DefaultRange readBlockRange( int blockNum ) throws IOException, BadFileException {
		if ( blockNum >= numBlocks )
			throw new IllegalArgumentException( "Given block number " + blockNum + " is too large" );

		byte[] buf = new byte[ DataBlock.HEADER_SIZE ];
		file.readBlocks( getDataBlockStart( blockNum ), buf );
		return DataBlock.loadBlockRange( buf );
	}
}
