/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.model;

import org.gillius.Util.Factory;
import org.gillius.realdb.BadFileException;
import org.gillius.realdb.io.GenericRecordSerializer;
import org.gillius.realdb.io.RecordSerializer;
import org.gillius.realdb.metadata.DataStreamInfo;
import org.gillius.realdb.metadata.RecordElement;
import org.gillius.realdb.model.data.DefaultStreamReconstructorFactory;
import org.gillius.realdb.model.data.Record;
import org.gillius.realdb.model.data.RecordCompressor;
import org.gillius.realdb.model.data.StreamInterval;
import org.gillius.realdb.util.DBIterable;
import org.gillius.realdb.util.DBIterator;

import java.io.Closeable;
import java.io.Flushable;
import java.io.IOException;
import java.util.NoSuchElementException;

/**
 * A DataStream is an ordered sequence of timestamped {@link Record}s within the database.
 *
 * @author Jason Winnebeck
 */
public class DataStream implements Flushable, Closeable {
	private final DataStreamInfo   info;
	private final DataStreamIndex  index;
	private final DataBlockManager manager;

	private final RecordSerializer serializer;
	private final RecordCompressor compressor;
	private final Factory<? extends StreamReconstructor> recon;

	private long startTime;
	private long endTime;

	private DataBlock currBlock = null;

	/**
	 * Constructs a new DataStream. The proper way to obtain a DataStream from a database is by
	 * getting them through the {@link Database#getStreams()} method.
	 *
	 * @param info    non-null DataStreamInfo describing this DataStream
	 * @param index   non-null DataStreamIndex managing the index for this DataStream
	 * @param manager non-null DataBlockManager used as a source of data blocks
	 */
	public DataStream( DataStreamInfo info, DataStreamIndex index, DataBlockManager manager ) {
		this.info    = info;
		this.index   = index;
		this.manager = manager;

		RecordElement[] recordElements = info.getElements();

		//Remember to pick the right impl when we support custom serializers (for speed)
		serializer = new GenericRecordSerializer( recordElements );
		compressor = new RecordCompressor( recordElements );
		recon      = new DefaultStreamReconstructorFactory( recordElements );

		startTime = index.getStartTime();
		endTime = index.getEndTime();
	}

	/**
	 * Returns the DataStreamInfo that describes this DataStream.
	 */
	public DataStreamInfo getInfo() {
		return info;
	}

	/**
	 * Writes the given record to the end of this stream.
	 *
	 * @param r non-null Record with a time greater than all previously written records.
	 *
	 * @return true if a Record was physically written to the database, false if the record was removed
	 *         due to the stream compressor.
	 *
	 * @throws IOException              If there is an error writing to the database
	 * @throws BadFileException         If the database is corrupted
	 * @throws IllegalArgumentException If the time for this record is before the last record written
	 *                                  (out-of-order write), or if {@link #checkRecord} would throw
	 */
	public boolean writeRecord( Record r ) throws IOException, BadFileException {
		checkRecord( r );

		Record toWrite = compressor.handleRecord( r );

		if ( toWrite != null ) {
			if ( currBlock == null ) {
				currBlock = manager.getNextBlock( serializer.getWrittenSize() );

			} else if ( !currBlock.canWriteRecord() ) {
				manager.closeBlock();
				currBlock = null; //for possible exception-safety later
				currBlock = manager.getNextBlock( serializer.getWrittenSize() );
			}

			//TODO: handle the case where the block is unusable if writeRecord throws
			currBlock.writeRecord( r, serializer );

			if ( startTime < 0 )
				startTime = r.getTime();
			endTime = r.getTime();

			return true;

		} else {
			return false;
		}
	}

	/**
	 * Checks if the given Record is proper for this DataStream, by calling
	 * {@link DataStreamInfo#checkRecord(Record)}.
	 *
	 * @param r non-null Record
	 *
	 * @throws IllegalArgumentException if Record is null or is not proper for this stream.
	 */
	public void checkRecord( Record r ) throws IllegalArgumentException {
		info.checkRecord( r );
		if ( r.getTime() <= getEndTime() )
			throw new IllegalArgumentException( "Out-of-order data at " + r.getTime() );
	}

	/**
	 * Returns the timestamp of the first record in this DataStream -- the minimum time for all data in
	 * the stream. If there are no data points, -1 is returned.
	 */
	public long getStartTime() {
		return startTime;
	}

	/**
	 * Returns the timestamp of the last record in this DataStream -- the maximum time for all data in
	 * the stream. If there are no data points, -1 is returned.
	 */
	public long getEndTime() {
		return endTime;
	}

	/**
	 * Returns a read-only Iterator over all of the Records contained in this DataStream. The
	 * records are not read from any cache, and any {@link #writeRecord(Record)} or {@link #flush()}
	 * operation invalidates the returned Iterator, which may not be fail-fast to detect this
	 * condition.
	 * <p/>
	 * Each element returned from the iterator is non-null.
	 */
	public DBIterator<Record> readRecords() {
		return new RecordIterator();
	}

	/**
	 * Returns a DBIterable that returns iterators from {@link #readRecords()}.
	 */
	public DBIterable<Record> records() {
		return new DBIterable<Record>() {
			public DBIterator<Record> iterator() {
				return readRecords();
			}
		};
	}

	/**
	 * Returns a read-only Iterator over Records in this DataStream between the given times. The
	 * records are not read from any cache, and any {@link #writeRecord(Record)} or {@link #flush()}
	 * operation invalidates the returned Iterator, which may not be fail-fast to detect this
	 * condition.
	 * <p/>
	 * Each element returned from the iterator is non-null.
	 *
	 * @param startTime iteration start time (inclusive), or 0 to start at the beginning
	 * @param endTime   iteration end time (inclusive), or Long.MAX_VALUE to go to the end
	 *
	 * @return iterator of Records, where the first record's time is the latest record in the stream
	 *         &lt;= startTime, or the first record in the stream.
	 *
	 * @throws IOException      if there is an error accessing the database while searching
	 * @throws BadFileException if the database can be accessed, but it is irreparably corrupted.
	 */
	public DBIterator<Record> readRecords( long startTime, long endTime )
			throws IOException, BadFileException {
		return new RecordIterator( startTime, endTime );
	}

	/**
	 * Returns a DBIterable that returns iterators from {@link #readRecords(long, long)}.
	 */
	public DBIterable<Record> records( final long startTime, final long endTime ) {
		return new DBIterable<Record>() {
			public DBIterator<Record> iterator() throws IOException, BadFileException {
				return readRecords( startTime, endTime );
			}
		};
	}

	/**
	 * Returns a read-only Iterator over all of the StreamIntervals contained in this DataStream. Any
	 * operation that would invalidate a Record iterator like the one returned from {@link
	 * #readRecords()} also invalidates this iterator. The last interval in a DataStream will have the
	 * same start and end time (a zero duration), since the next Record is not yet known.
	 */
	public DBIterator<StreamInterval> readIntervals() {
		return new IntervalIterator( new RecordIterator(), recon.create(), Long.MAX_VALUE );
	}

	/**
	 * Returns a DBIterable that returns iterators from {@link #readIntervals()}.
	 */
	public DBIterable<StreamInterval> intervals() {
		return new DBIterable<StreamInterval>() {
			public DBIterator<StreamInterval> iterator() {
				return readIntervals();
			}
		};
	}

	/**
	 * Returns a read-only Iterator over a range of StreamIntervals contained in this DataStream.
	 * Any operation that would invalidate a Record iterator like the one returned from
	 * {@link #readRecords()} also invalidates this iterator. The start time of first interval
	 * returned may be before startTime, if the interval contains startTime. The end time of the last
	 * interval returned will be after endTime, if that interval does not end exactly on the given
	 * endTime. The last interval in a DataStream will have the same start and end time (a zero
	 * duration), since the next Record is not yet known.
	 *
	 * @param startTime iteration start time (inclusive), or 0 to start at the beginning
	 * @param endTime   iteration end time (inclusive), or Long.MAX_VALUE to go to the end
	 *
	 * @return iterator of StreamInterval, where the first interval's start time is the latest record
	 *         in the stream &lt;= startTime, or the first record in the stream.
	 *
	 * @throws IOException      if there is an error accessing the database while searching
	 * @throws BadFileException if the database can be accessed, but it is irreparably corrupted.
	 */
	public DBIterator<StreamInterval> readIntervals( long startTime, long endTime )
			throws IOException, BadFileException {
		return new IntervalIterator( new RecordIterator( startTime, Long.MAX_VALUE ),
		                             recon.create(), endTime );
	}

	/**
	 * Returns a DBIterable that returns iterators from {@link #readIntervals(long, long)}.
	 */
	public DBIterable<StreamInterval> intervals( final long startTime, final long endTime ) {
		return new DBIterable<StreamInterval>() {
			public DBIterator<StreamInterval> iterator() throws IOException, BadFileException {
				return readIntervals( startTime, endTime );
			}
		};
	}

	/**
	 * A debugging method to read the data block indices in order from this stream's index.
	 * <p/>
	 * Each element returned from the iterator is non-null.
	 */
	public DBIterator<Integer> readDataBlockIndices() {
		return index.iterator();
	}

	/**
	 * Commits all written data to the database. This could cause a partial data block to be written,
	 * possibly wasting space in the current data block.
	 */
	public void flush() throws IOException {
		if ( currBlock != null ) {
			manager.closeBlock();
			currBlock = null;
		}
	}

	/**
	 * Closes this DataStream by flushing and closing out this data stream. Normal users should call
	 * {@link Database#close()} instead. Once a DataStream is closed it cannot be used anymore.
	 * <p/>
	 * Currently the closed state is not enforced, so this works similar to a flush.
	 */
	public void close() throws IOException {
		flush();
	}

	public String toString() {
		return info.getName() + " (ID " + info.getUserId() + ")";
	}

	private class RecordIterator implements DBIterator<Record> {
		private final DBIterator<Integer> indexIter;
		private final Long endTime;

		private DataBlock block = null;
		private int recordNum  = 0;

		private Record nextRecord = null;

		private RecordIterator() {
			indexIter = index.iterator();
			endTime   = null;
		}

		private RecordIterator( long startTime, long endTime )
				throws IOException, BadFileException {
			this.endTime = endTime;

			indexIter = index.iterator( manager, startTime, endTime );
			if ( indexIter.hasNext() ) {
				block = manager.readBlock( indexIter.next(), serializer.getWrittenSize() );
				recordNum = block.findRecord( startTime );

			} else if ( currBlock != null ) {
				//The only records we have are the unflushed ones, so look there for the starting record
				block = currBlock;
				recordNum = currBlock.findRecord( startTime );
			}
		}

		public boolean hasNext() throws IOException, BadFileException {
			if ( nextRecord != null )
				return true;

			if ( block == null || block.getNumRecords() <= recordNum ) {
				if ( indexIter.hasNext() ) {
					block = manager.readBlock( indexIter.next(), serializer.getWrittenSize() );
					recordNum = 0;

				} else if ( block != currBlock && currBlock != null ) {
					//We've got some unflushed records to read back
					block = currBlock;
					recordNum = 0;

				} else {
					//No more blocks
					return false;
				}
			}

			nextRecord = block.readRecord( recordNum, serializer );

			//If we are told to stop at an endTime, check for that
			if ( endTime != null && nextRecord.getTime() > endTime ) {
				//Terminate early
				nextRecord = null;
				return false;

			} else {
				//This record is OK, so increment and return
				++recordNum;
				return true;
			}
		}

		public Record next() throws IOException, BadFileException {
			if ( !hasNext() )
				throw new NoSuchElementException();

			Record ret = nextRecord;
			nextRecord = null;
			return ret;
		}
	}
}
