/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.model;

import org.gillius.realdb.BadFileException;
import org.gillius.realdb.util.DBIterable;
import org.gillius.realdb.util.DBIterator;

import java.io.IOException;

/**
 * DataStreamIndex is an interface to a stream's data index to support reading. A DataStreamIndex
 * is Iterable over the data block indices stored in the index.
 *
 * @author Jason Winnebeck
 */
public interface DataStreamIndex extends DBIterable<Integer> {
	/**
	 * Returns the timestamp of the first data point in the first data block in this index -- the
	 * minimum time for all data in the stream. If there are no data points, -1 is returned.
	 */
	long getStartTime();

	/**
	 * Returns the timestamp of the last data point in the last data block in this index -- the
	 * maximum time for all data in the stream. If there are no data points, -1 is returned.
	 */
	long getEndTime();

	/**
	 * Returns the number of the first data block for this data stream, or -1 if there are no blocks
	 * in this index.
	 */
	int getFirstBlock();

	/**
	 * Return the second data block index for this data stream, or -1 if there is not a second block.
	 */
	int getSecondBlock() throws IOException, BadFileException;

	/**
	 * Returns the last data block index for this data stream, or -1 if there are no blocks in this
	 * index.
	 */
	int getLastBlock();

	/**
	 * Returns an iterator over a range of data blocks that contain at least one record within the
	 * given time bounds. The first index returned will be at a DataBlock that contains at least one
	 * record &gt;= startTime, and the last index returned is either the last DataBlock in the index,
	 * or the last DataBlock containing a record &lt;= endTime.
	 *
	 * @param summaryReader non-null reader to get data block summaries (their time ranges)
	 * @param startTime     start time, inclusive, or 0 for unbounded start
	 * @param endTime       end time, inclusive, or Long.MAX_VALUE for unbounded end
	 */
	public DBIterator<Integer> iterator( DataBlockSummaryReader summaryReader, long startTime,
	                                     long endTime ) throws IOException, BadFileException;

	/**
	 * Returns an iterator over all of the data blocks in this DataIndex. Any modification of this
	 * index invalidates all iterators returned from this method. The returned iterator is not
	 * fail-fast -- it won't throw ConcurrentModificationException if there has been a modification.
	 * The returned iterator also does not support the remove operation.
	 */
	public DBIterator<Integer> iterator();
}
