/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.model;

import org.gillius.Util.IOUtils;
import org.gillius.Util.MathUtils;
import org.gillius.realdb.BadFileException;
import org.gillius.realdb.io.BlockFile;
import org.gillius.realdb.io.DowUtil;
import org.gillius.realdb.io.Formattable;
import org.gillius.realdb.io.ProfilingBlockFile;
import org.gillius.realdb.io.RandomAccessBlockFile;
import org.gillius.realdb.metadata.DataStreamInfo;
import org.gillius.realdb.metadata.FileHeader;
import org.gillius.realdb.metadata.MetadataSection;
import org.gillius.realdb.model.pool.AddIndexTxEntry;
import org.gillius.realdb.model.pool.BlockPoolFactory;
import org.gillius.realdb.model.pool.IndexTxEntry;
import org.gillius.realdb.model.pool.RemoveIndexTxEntry;
import org.gillius.realdb.model.pool.TxEntry;
import org.gillius.realdb.tools.DatabaseBuilder;
import org.gillius.realdb.util.DBIterator;

import java.io.Closeable;
import java.io.File;
import java.io.FileInputStream;
import java.io.Flushable;
import java.io.IOException;
import java.io.InputStream;
import java.io.RandomAccessFile;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.Map;

import static org.gillius.Util.MathUtils.*;

/**
 * The Database class is the main class to be used in RealDB. A Database holds a set of DataStream
 * objects for reading and writing data. The static factory methods or the {@link DatabaseBuilder}
 * can be used to obtain a Database instance.
 *
 * @author Jason Winnebeck
 */
public class Database implements Formattable, Flushable, Closeable {
	private final BlockFile        file;

	private final MetadataSection  metadata;
	private final BlockPool        pool;
	private final DataIndexSection index;
	private final DataSection      data;

	private Map<Integer, DataStream> streams = new HashMap<Integer, DataStream>();

	/**
	 * Constructs a new Database with all of the sections already loaded and configured.
	 *
	 * @param file nullable BlockFile, used only in the close method.
	 */
	public Database( BlockFile file, MetadataSection metadata, BlockPool pool,
	                 DataIndexSection index, DataSection data ) {
		this.file     = file;
		this.metadata = metadata;
		this.pool     = pool;
		this.index    = index;
		this.data     = data;

		createDataStreams();
	}

	/**
	 * Creates or loads a database.
	 *
	 * @param bf              BlockFile where DB will be created or loaded
	 * @param metadataSection Constructed (loaded or created) MetadataSection
	 * @param create          true if creating a new Database (will format)
	 *
	 * @throws IOException if an error occurs accessing the file
	 * @throws BadFileException if loading and the database is corrupted
	 */
	private Database( BlockFile bf, MetadataSection metadataSection, boolean create )
			throws IOException, BadFileException {
		metadata = metadataSection;
		this.file = bf;

		//1 block for the file header
		//Metadata size div file block size gives blocks that the metadata takes up
		int startBlock = 1 + MathUtils.divideRoundUp( metadata.getSizeBytes(), file.getBlockSize() );

		//Now, we need to determine the size of the index and data sections that maximizes the size of
		//the data section.

		//First generation algorithm, start with the largest data size and keep going down until we can
		//fit. Originally we had a real math algorithm here, but it required knowledge of the exact
		//details of the sections' implementations. Instead, it is better to deduce it. The second
		//generation algorithm should use binary search. The third generation could add a heuristic.
		final int streams = metadataSection.getMaxStreams();
		final int dBS     = metadata.getDataBlockSize();
		final int physicalDataBlocks = file.getSize() - startBlock;

		int logicalDataBlocks = physicalDataBlocks / dBS;
		int otherBlocks       = physicalDataBlocks % dBS;
		while ( logicalDataBlocks > 0 &&
		        getPoolAndIndexSize( streams, logicalDataBlocks, file.getBlockSize() ) >= otherBlocks ) {
			otherBlocks += dBS;
			--logicalDataBlocks;
		}
		if ( logicalDataBlocks <= 0 )
			throw new IllegalArgumentException( "It is impossible to make a database so small" );

		int indexBlocks = DataIndexSection.getMinimumSize( streams, logicalDataBlocks, file.getBlockSize() );
		int maxOpenTx = metadataSection.getMaxStreams(); //assume at most 1 open Tx per stream
		int poolBlocks  = BlockPoolFactory.getMinSize( file.getBlockSize(), maxOpenTx );
		if ( create ) {
			pool = BlockPoolFactory.createNew( file, startBlock, poolBlocks, logicalDataBlocks );
			startBlock += poolBlocks;
			//TODO: create/load section with actual number of streams, not all reserved?
			index = DataIndexSection.createNew( file, startBlock, streams, logicalDataBlocks, indexBlocks );
			startBlock += indexBlocks;
		} else {
			pool  = BlockPoolFactory.load( file, startBlock, poolBlocks, logicalDataBlocks );
			startBlock += poolBlocks;
			index = DataIndexSection.load( file, startBlock, streams, logicalDataBlocks, indexBlocks );
			startBlock += indexBlocks;
		}
		pool.open();

		data = new DataSection( file, startBlock, dBS, logicalDataBlocks );

		doRecovery();
		createDataStreams();
	}

	@SuppressWarnings( { "CastConflictsWithInstanceof" } )
	private void doRecovery() throws IOException, BadFileException {
		DBIterator<TxEntry> iter = pool.getLog().iterator();

		while ( iter.hasNext() ) {
			TxEntry tx = iter.next();
			if ( tx instanceof IndexTxEntry ) {
				IndexTxEntry indexTxEntry = (IndexTxEntry) tx;
				DataIndex ix = index.getIndex( indexTxEntry.getIndex() );
				int seqNum = indexTxEntry.getSeqNum();
				switch( indexTxEntry.getType() ) {
					case Add:
						if ( ix.getTailSeqNum() == seqNum ) {
							if ( ix.getLastBlock() != ((AddIndexTxEntry)tx).getLastBlock() ) {
								throw new BadFileException( "Index not in expected state, Tx: " + indexTxEntry +
								                            " index tail: " + ix.getLastBlock() );
							}
							//TODO: why don't we replay the transaction here?
							//It seems to be a bug that we don't; the block would not marked as free
							//but it's not in any index, so this block is lost forever.
							//TODO: check that not replaying really causes a block leak
							//If we really implement this, I think what has to be done is load the data
							//block to find the start/end times then add it. Or at minimum the simpler
							//idea is to just notify the TransactionLog that the block is really free.
						}
						break;

					case Remove:
						if ( ix.getHeadSeqNum() == seqNum ) {
							if ( ix.getFirstBlock() != indexTxEntry.getBlock() ) {
								throw new BadFileException( "Index not in expected state, Tx: " + indexTxEntry +
								                            " index head: " + ix.getFirstBlock() );
							}
							ix.removeFirstBlock( ((RemoveIndexTxEntry)tx).getNextBlockStartTime(), null );
						}
						break;

					default:
						throw new AssertionError( "Invalid IndexTxEntry type " + indexTxEntry.getType() );
				}
			}
		}
	}

	private static int getPoolAndIndexSize( int streams, int logicalDataBlocks, int blockSize ) {
		return DataIndexSection.getMinimumSize( streams, logicalDataBlocks, blockSize ) +
	         BlockPoolFactory.getMinSize( blockSize, streams );
	}

	public MetadataSection getMetadata() {
		return metadata;
	}

	/**
	 * Returns a mapping of section names to their sizes in blocks. All of the sections are entirely
	 * accounted for and none overlap. Therefore, the values should add up to the file's size in
	 * blocks except for any "slack" space (odd unused blocks at the end)
	 */
	public Map<String, Integer> getSectionSizes() {
		LinkedHashMap<String, Integer> ret = new LinkedHashMap<String, Integer>();

		ret.put( "Header", 1 );
		ret.put( "Metadata", divideRoundUp( metadata.getSizeBytes(), file.getBlockSize() ) );
		ret.put( "BlockPool", BlockPoolFactory.getMinSize( file.getBlockSize(),
		                                                   metadata.getMaxStreams() ) );
		ret.put( "Indexes", index.getSize() );
		ret.put( "Data", data.getNumBlocks() * data.getBlockSize() );

		int totalSize = 0;
		for ( Integer i : ret.values() ) {
			totalSize += i;
		}
		ret.put( "Slack", file.getSize() - totalSize );

		return ret;
	}

	/**
	 * Returns the number of data blocks used. A data block normally consists of multiple file blocks,
	 * so this is not the same value as from {@link #getSectionSizes()}.
	 */
	public int getDataBlocksUsed() {
		return pool.getDataBlocksUsed();
	}

	/**
	 * Returns the total number of data blocks. A data block normally consists of multiple file blocks,
	 * so this is not the same value as from {@link #getSectionSizes()}.
	 */
	public int getTotalDataBlocks() {
		return pool.getNumDataBlocks();
	}

	/**
	 * Returns the block size, in bytes, of the underlying storage.
	 */
	public int getBlockSize() {
		return file.getBlockSize();
	}

	/**
	 * Returns the size of the database, in blocks.
	 */
	public int getSizeInBlocks() {
		return file.getSize();
	}

	/**
	 * Returns all of the DataStreams in this database.
	 */
	public Collection<DataStream> getStreams() {
		return streams.values();
	}

	/**
	 * Returns a map of DataStream user ID to DataStream.
	 */
	public Map<Integer, DataStream> getStreamsMap() {
		return streams;
	}

	/**
	 * Clears out the database entirely by reformatting the index and data sections. The file header
	 * and metadata sections are unchanged. If a format operation fails, the underlying file is an
	 * undefined state. You can try to format it again, but that is the only allowed operation on such
	 * a file.
	 *
	 * @throws IOException if an error occurs writing the database
	 */
	public void format() throws IOException {
		index.format();
		pool.format();
	}

	/**
	 * Flushes the Database, but keeps it open for further operations. Flushing a database will commit
	 * all changes to disk. If this operation succeeds, any data written can be recovered if a failure
	 * occurs later. If this operation fails, then some or all data could be lost in the database
	 * since the last flush.
	 */
	public void flush() throws IOException {
		for ( DataStream stream : streams.values() ) {
			stream.flush();
		}
		pool.flush();
		index.flush();
	}

	/**
	 * Closes the database by flushing all buffers and closing any open resources. A close operation is
	 * performed on the underlying file, if there is one. Once a Database is closed, it and nothing
	 * that comes from it should be used again, but the Database may be reloaded with {@link #load}
	 * later.
	 * <p/>
	 * If an IOException is thrown while trying to close the database, the underlying file, if there
	 * is one, is forcefully closed to release this resource. In this case not all streams and
	 * buffers were flushed properly.
	 */
	public void close() throws IOException {
		try {
			for ( DataStream stream : streams.values() ) {
				stream.close();
			}
			pool.flush();
			index.flush();

		} finally {
			IOUtils.safeClose( file );
		}
	}

	/**
	 * Forces the underlying BlockFile to be closed immediately without flushing any data remaining
	 * to be written.
	 * </p>
	 * TODO: remove this method when RealDB code is exception-safe.
	 */
	public void forceClose() {
		IOUtils.safeClose( file );
	}

	/**
	 * If this Database is enabled for profiling, returns a {@link DatabaseStatistics}, else returns
	 * null.
	 */
	public DatabaseStatistics getProfilingStatistics() {
		if ( file instanceof ProfilingBlockFile ) {
			ProfilingBlockFile pbf = (ProfilingBlockFile) file;

			return new DatabaseStatistics( pbf.getReadCalls(), pbf.getWriteCalls(),
			                               pbf.getReadBlocks(), pbf.getWrittenBlocks() );
		} else {
			return null;
		}
	}

	/**
	 * Creates a new Database and formats the given BlockFile with it. This includes writing of the
	 * file header and serialization of the given MetadataSection.
	 *
	 * @param bf       BlockFile where DB will be created or loaded
	 * @param metadata Constructed MetadataSection
	 *
	 * @return non-null new Database
	 *
	 * @throws IOException if there is an error writing to the file. If this occurs the file may have
	 *                     been partially written.
	 */
	public static Database createNew( BlockFile bf, MetadataSection metadata )
			throws IOException {
		FileHeader header = new FileHeader( bf.getBlockSize(), bf.getSize() );
		header.writeHeader( bf );
		bf.writeBlocks( 1, DowUtil.serialize( metadata ) );

		try {
			return new Database( bf, metadata, true );
		} catch ( BadFileException e ) {
			throw new IOException( "BadFileException not expected on creation", e );
		}
	}

	/**
	 * Loads a database from the given File.
	 *
	 * @param f non-null existing File
	 *
	 * @return non-null Database
	 *
	 * @throws IOException      if an error occurs accessing the file
	 * @throws BadFileException if the file is read, but the data is unrecoverably corrupted
	 */
	public static Database load( File f ) throws IOException, BadFileException {
		return load( f, false );
	}

	/**
	 * Loads a database from the given File.
	 *
	 * @param f non-null existing File
	 *
	 * @param profiling if true, enables collection of profiling statistics (has some overhead).
	 * @return non-null Database
	 *
	 * @throws IOException      if an error occurs accessing the file
	 * @throws BadFileException if the file is read, but the data is unrecoverably corrupted
	 */
	public static Database load( File f, boolean profiling ) throws IOException, BadFileException {
		InputStream is = new FileInputStream( f );
		FileHeader header = FileHeader.getHeader( is );
		//version numbers are checked, if it is loaded.
		is.close();

		RandomAccessFile raf = new RandomAccessFile( f, "rwd" );
		BlockFile file = new RandomAccessBlockFile( raf, header.getBlockSize(), header.getNumBlocks() );

		if ( profiling ) {
			file = new ProfilingBlockFile( file );
		}

		return secondStageLoad( file );
	}

	/**
	 * Loads a database from a properly constructed BlockFile. The file header is in the first block
	 * and its block size is checked against the given BlockFile size and the versions are checked.
	 *
	 * @param bf non-null BlockFile
	 *
	 * @return new Database attached to the given BlockFile.
	 *
	 * @throws IOException      if an error occurs accessing the file
	 * @throws BadFileException if the file is read, but the data is unrecoverably corrupted
	 */
	public static Database load( BlockFile bf ) throws IOException, BadFileException {
		FileHeader.verify( bf );
		return secondStageLoad( bf );
	}

	/**
	 * Performs the load of the database, after the header has already been parsed and/or verified.
	 */
	private static Database secondStageLoad( BlockFile bf ) throws IOException, BadFileException {
		MetadataSection metadata = new MetadataSection( bf, 1 );

		return new Database( bf, metadata, false );
	}

	/**
	 * Once the sections are set up, allows for creating all of the data streams.
	 */
	private void createDataStreams() {
		for ( int i = 0; i < metadata.getStreams().size(); i++ ) {
			DataStreamInfo streamInfo = metadata.getStreams().get( i );
			DataBlockManager manager = new DefaultDataBlockManager( index, i, pool, data );
			DataStream stream = new DataStream( streamInfo, index.getIndex( i ), manager );
			if ( streams.put( streamInfo.getUserId(), stream ) != null ) {
				throw new IllegalStateException( "There is more than one stream with ID " +
				                                 streamInfo.getUserId() );
			}
		}

		streams = Collections.unmodifiableMap( streams );
	}
}
