/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.model;

import org.gillius.realdb.BadFileException;
import org.gillius.realdb.io.BlockFile;
import org.gillius.realdb.io.ByteBufferDataIO;

import java.io.IOException;
import java.nio.ByteBuffer;

/**
* DefaultCircularQueueBlockIO
*
* @author Jason Winnebeck
*/
class DefaultCircularQueueBlockIO implements CircularQueueBlockIO {

	private final ByteBufferDataIO buf;
	private final BlockFile bf;
	private final int startBlock;
	private final int size;

	public DefaultCircularQueueBlockIO( BlockFile bf, int startBlock, int size ) {
		this.bf = bf;
		this.startBlock = startBlock;
		this.size = size;
		buf = new ByteBufferDataIO( bf.getBlockSize() );

		if ( getBlockDataSize() <= 0 ) {
			throw new IllegalArgumentException( "Block size is too small" );
		}
	}

	@Override
	public int getSize() {
		return size;
	}

	@Override
	public int getBlockDataSize() {
		return bf.getBlockSize() - blockOverhead;
	}

	@Override
	public int getAltHeadIndex() {
		return size - 1;
	}

	@Override
	public CircularQueueState getState( int block ) throws IOException, BadFileException {
		if ( block < 0 || block >= size )
			throw new IndexOutOfBoundsException( String.valueOf( block ) );

		bf.readBlock( block + startBlock, buf.getBuf().array() );
		RawSequencedBlock rsb = new RawSequencedBlock( buf.getBuf().array() );
		int nextSeqNum = rsb.getSeqNum();
		boolean altHead = ( block != getAltHeadIndex() );
		return new CircularQueueState( new ByteBufferDataIO( rsb.getRawData() ),
		                               altHead, nextSeqNum );
	}

	@Override
	public byte[] getData( int block ) throws IOException, BadFileException {
		if ( block < 0 || block >= size )
			throw new IndexOutOfBoundsException( String.valueOf( block ) );

		bf.readBlock( block + startBlock, buf.getBuf().array() );
		RawSequencedBlock rsb = new RawSequencedBlock( buf.getBuf().array() );
		ByteBuffer data = ByteBuffer.wrap( rsb.getRawData() );
		data.position( CircularQueueState.getWrittenSize() );
		byte[] ret = new byte[ data.remaining() ];
		data.get( ret );
		return ret;
	}

	@Override
	public void writeBlock( int block, CircularQueueState state, byte[] data, int offset, int length )
			throws IOException {
		if ( block < 0 || block >= size )
			throw new IndexOutOfBoundsException( String.valueOf( block ) );

		ByteBuffer raw = buf.getBuf();
		raw.position( 0 );
		raw.limit( CircularQueueState.getWrittenSize() );
		state.write( buf );
		raw.limit( raw.capacity() );
		raw.put( data, offset, length );

		byte[] packed = new byte[ raw.position() ];
		((ByteBuffer)raw.flip()).get( packed );

		RawSequencedBlock rsb = new RawSequencedBlock( state.nextSeqNum, packed );
		raw.clear();
		rsb.write( buf );
		bf.writeBlock( block + startBlock, raw.array() );
	}

	@Override
	public void format() throws IOException {
		byte[] zeros = new byte[ bf.getBlockSize() ];
		for ( int i = 0; i < size; ++ i ) {
			bf.writeBlock( i + startBlock, zeros );
		}
	}
}
