/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.model;

import org.gillius.realdb.BadFileException;

import java.io.Flushable;
import java.io.IOException;

import static org.gillius.Util.CheckUtils.*;

/**
 * DefaultDataBlockManager implements the full data space management algorithm for RealDB, to
 * allocate new blocks and delete the oldest blocks when the database fills up.
 *
 * @author Jason Winnebeck
 */
public class DefaultDataBlockManager implements DataBlockManager {
	private final DataIndexSection indices;
	private final DataIndex        index;
	private final int              indexId;
	private final BlockPool        pool;
	private final DataSection      data;

	private DataBlock        currBlock = null;
	private BlockTransaction currTx    = null;

	/**
	 * Constructs a new DefaultDataBlockManager for a particular stream in a database.
	 *
	 * @param indices non-null DataIndexSection
	 * @param stream  stream index inside of the DataIndexSection
	 * @param pool    non-null BlockPool
	 * @param data    non-null DataSection
	 */
	public DefaultDataBlockManager( DataIndexSection indices, int stream, BlockPool pool,
	                                DataSection data ) {
		checkNull( indices, "indices" );
		checkNull( pool, "pool" );
		checkNull( data, "data" );

		this.indices = indices;
		index        = indices.getIndex( stream );
		indexId      = stream;
		this.pool    = pool;
		this.data    = data;
	}

	public DataBlock getNextBlock( int recordSize ) throws IOException, BadFileException {
		if ( currBlock != null )
			throw new IllegalStateException( "Must close existing block" );

		if ( pool.hasFreeBlocks() ) {
			currTx = pool.getFreeBlock();
			currBlock = new DataBlock( currTx.getBlockNum(), recordSize, data.getBlockSizeBytes() );

		} else {
			//Find the oldest data block in the database and reclaim it for our own use.
			int oldest = findOldestIndex();

			if ( oldest < 0 )
				throw new BadFileException( "No stream has any data, yet the database is full" );

			currTx = releaseFirstBlock( oldest, recordSize );
			currBlock = new DataBlock( currTx.getBlockNum(), recordSize, data.getBlockSizeBytes() );
		}

		return currBlock;
	}

	public void closeBlock() throws IOException {
		if ( currBlock == null )
			throw new IllegalStateException( "No open block to close" );

		//TODO: Complete failure of the logging concept; we have to flush
		//TODO: the following isn't exception-safe if commit or addDataBlock fail
		currTx.addAction( new DataBlockWriter( data, currBlock ) ).flush();
		pool.addBlock( currTx, index.getLastBlock(), indexId, index.getTailSeqNum() );
		index.addDataBlock( currBlock.getBlockNum(), currBlock.getStartTime(), currBlock.getEndTime(),
		                    currTx );
		//TODO: Apparently we still need the commit; need to find out how to eliminate, so we can do more than one tx at a time
		currTx.commit();
		currBlock = null;
		currTx    = null;
	}

	public DataBlock readBlock( int blockNum, int recordSize ) throws IOException, BadFileException {
		return data.readBlock( blockNum, recordSize );
	}

	public DefaultRange readBlockRange( int blockNum ) throws IOException, BadFileException {
		return data.readBlockRange( blockNum );
	}

	/**
	 * Returns the DataIndex in the DataIndexSection with the oldest record, or -1 if there is no
	 * such index. And yes, this method may even return our index -- and this is a valid situation.
	 */
	private int findOldestIndex() {
		long minTime = Long.MAX_VALUE;
		int oldest = -1;
		for ( int i = 0; i < indices.getIndices().size(); i++ ) {
			long start = indices.getIndices().get( i ).getStartTime();
			if ( start >= 0 && ( oldest < 0 || start < minTime ) ) {
				oldest = i;
				minTime = start;
			}
		}
		return oldest;
	}

	private BlockTransaction releaseFirstBlock( int oldest, int recordSize )
			throws IOException, BadFileException {
		DataIndex oldestIndex = indices.getIndex( oldest );
		int nextBlock = oldestIndex.getFirstBlock();

		long newStartTime = 0; //ignored if there isn't a second block
		int secondBlock = oldestIndex.getSecondBlock();
		if ( secondBlock >= 0 ) {
			newStartTime = readBlock( secondBlock, recordSize ).getStartTime();
		}

		BlockTransaction ret = pool.removeBlock( nextBlock, oldest, oldestIndex.getHeadSeqNum(),
		                                         newStartTime );
		oldestIndex.removeFirstBlock( newStartTime, ret );
		//TODO: need to solve the flush problem like the commit above
		if ( oldest == indexId ) {
			//To prevent recursion, we have to force a flush before we write the data block. We do it
			//here because it is easiest; although we leave open a little more time where we could lose
			//the first data block from this stream.
			oldestIndex.flush();
		}

		return ret;
	}

	private static class DataBlockWriter implements Flushable {
		private final DataSection dataSection;
		private final DataBlock   block;

		private DataBlockWriter( DataSection dataSection, DataBlock block ) {
			this.dataSection = dataSection;
			this.block = block;
		}

		public void flush() throws IOException {
			dataSection.writeBlock( block );
		}
	}
}
