/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.model;

/**
 * An ElementCompressor is specialized for a given ElementType and handles whether or not a
 * RecordElement should be written, and how it is written when it is. The compressors hold a state
 * of the last record written.
 * <p/>
 * An ElementCompressor does not need to be responsible for recording a null value; a null value
 * always is a significant change that causes a write and is kept in a separate area.
 *
 * @author Jason Winnebeck
 */
public interface ElementCompressor {
	/**
	 * Returns the number of previously written Records needed to make assessments in {@link
	 * #handleNewValue}.
	 */
	public int getLookBehindDesired();

	/**
	 * Returns the number of future Records beyond the record of interest to make assessments in {@link
	 * #handleNewValue}.
	 */
	public int getLookAheadDesired();

	/**
	 * Gives the ElementCompressor the values around the current record of interest and asks if that
	 * record should be written. This method is not guaranteed to be called for every record, if
	 * the determination to write a Record can be made without consulting this compressor. However,
	 * all written values will be given to {@link #valueWritten}
	 * <p/>
	 * The size of the given arrays is at least 2 + {@link #getLookBehindDesired()} + {@link
	 * #getLookAheadDesired()}, and the value of currentIndex will be at least the value of {@link
	 * #getLookBehindDesired()}. At the start and end of a stream's values, the values array may
	 * contain nulls (the corresponding timestamps will be negative). However, at least always
	 * values[currentIndex] will be defined.
	 *
	 * @param times        timestamps array corresponding directly to the values array, which must not
	 *                     be modified.
	 * @param values       sampled values of the element at the given timestamps, which must not be
	 *                     modified.
	 * @param currentIndex index of the sample that is the start of the ElementInterval.
	 *
	 * @return true if the record at currentIndex should be written.
	 */
	public boolean handleNewValue( long[] times, Object[] values, int currentIndex );

	/**
	 * Tells the compressor that the given value was written for this element.
	 * <p/>
	 * This method will be needed for the more complex (greater than first-order) algorithms.
	 *
	 * @param time  time of the value
	 * @param value nullable value, same as was given earlier to handleNewValue.
	 */
	public void valueWritten( long time, Object value );
}
