/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.model;

import org.gillius.Util.CheckUtils;
import org.gillius.realdb.BadFileException;

import java.io.DataOutput;
import java.io.IOException;
import java.nio.BufferUnderflowException;
import java.nio.ByteBuffer;
import java.util.Arrays;

/**
 * IndexBlock represents a single index data block within a stream's {@link DataIndex}. A large
 * sequence number (32 bit) is used so we can recognize that the active block written longest ago is
 * still recognized as being older than the newest block.
 *
 * @author Jason Winnebeck
 */
public class IndexBlock implements SequencedBlock, Range {
	/**
	 * A global static final thread-safe instance of a SequencedBlockLoader to load IndexBlocks.
	 */
	public static SequencedBlockLoader<IndexBlock> loader =
			new SequencedBlockLoader<IndexBlock>() {
				public IndexBlock load( byte[] serialized ) throws BadFileException {
					return new IndexBlock( serialized );
				}
			};

	/**
	 * seq, start, end, dataBlocksSize, checksum
	 */
	private static final int HEADER_SIZE = 4 + 8 + 8 + 2 + 4;

	private final int   seqNum;
	private final long  startTime;
	private final long  endTime;
	private final int[] dataBlocks;

	/**
	 * Returns the size, in bytes, of the smallest possible IndexBlock, the empty block.
	 */
	public static int getMinSize() {
		return getSize( 0 );
	}

	/**
	 * Returns the size, in bytes, of the IndexBlock that holds 'numBlocks' data block indexes.
	 */
	public static int getSize( int numBlocks ) {
		return HEADER_SIZE + numBlocks * 4;
	}

	/**
	 * Given a block size in bytes, returns the number of data block indicies that this IndexBlock
	 * can hold.
	 *
	 * @param blockSize a block size in bytes, greater than or equal to getMinSize().
	 *
	 * @return maximum size of the dataBlocks array to fit within 'blockSize' bytes
	 */
	public static int getMaxDataBlocks( int blockSize ) {
		return ( blockSize - HEADER_SIZE ) / 4;
	}

	/**
	 * Creates an "empty" IndexBlock with no data blocks, and start and end times are 0.
	 *
	 * @param seqNum sequence number for the IndexBlock
	 */
	public IndexBlock( int seqNum ) {
		this.seqNum = seqNum;
		startTime = endTime = -1;
		dataBlocks = new int[0];
	}

	/**
	 * Creates an IndexBlock with the given data blocks.
	 *
	 * @param seqNum     block's sequence number
	 * @param startTime  time of the first data point in the first data block
	 * @param endTime    time of the last data point in the last data block
	 * @param dataBlocks array of 1 or more data blocks.
	 */
	public IndexBlock( int seqNum, long startTime, long endTime, int[] dataBlocks ) {
		CheckUtils.checkNull( dataBlocks, "dataBlocks" );
		if ( endTime < startTime )
			throw new IllegalArgumentException( "endTime must be at or after startTime" );
		if ( dataBlocks.length == 0 || dataBlocks.length > 65535 )
			throw new IllegalArgumentException( "dataBlocks must have length between 1 and 65535, inclusive" );

		this.seqNum     = seqNum;
		this.startTime  = startTime;
		this.endTime    = endTime;
		this.dataBlocks = dataBlocks;
	}

	/**
	 * Constructs a new IndexBlock by reading it from the given byte array.
	 *
	 * @param data non-null byte[]
	 *
	 * @throws BadFileException if the data integrity of the block is not correct
	 */
	public IndexBlock( byte[] data ) throws BadFileException {
		ByteBuffer buf = ByteBuffer.wrap( data );
		try {
			seqNum    = buf.getInt();
			startTime = buf.getLong();
			endTime   = buf.getLong();
			if ( endTime < startTime )
				throw new BadFileException( "Data corruption: endTime before startTime" );

			int dataBlocksSize = buf.getShort() & 0xFFFF;
			if ( dataBlocksSize < 1 && ( startTime != -1 || endTime != -1 ) )
				throw new BadFileException( "Data corruption: empty IndexBlock must have -1 start and end time" );
			dataBlocks = new int[ dataBlocksSize ];
			int sum = 1;
			for ( int i = 0; i < dataBlocks.length; ++i ) {
				int next = buf.getInt();
				sum += next;
				dataBlocks[i] = next;
			}

			//TODO: replace later with a real checksum
			int check = buf.getInt();
			if ( sum != check )
				throw new BadFileException( "Data corruption: checksum is not correct" );

		} catch ( BufferUnderflowException e ) {
			//This can happen when the block is corrupted, especially when dataBlocks is read to be an
			//excessively large number.
			throw new BadFileException( "Data corruption: buffer too small (likely corrupted dataBlocks field)" );
		}
	}

	public int getSeqNum() {
		return seqNum;
	}

	/**
	 * Returns the timestamp of the first data point in the first data block
	 */
	public long getStartTime() {
		return startTime;
	}

	public long getEndTime() {
		return endTime;
	}

	public int[] getDataBlocks() {
		return dataBlocks;
	}

	public void write( DataOutput output ) throws IOException {
		output.writeInt( seqNum );
		output.writeLong( startTime );
		output.writeLong( endTime );
		output.writeShort( dataBlocks.length );
		int sum = 1;
		for ( int blockNum : dataBlocks ) {
			sum += blockNum;
			output.writeInt( blockNum );
		}
		output.writeInt( sum );
	}

	public boolean equals( Object o ) {
		if ( this == o ) return true;
		if ( o == null || getClass() != o.getClass() ) return false;

		IndexBlock that = (IndexBlock) o;

		if ( endTime != that.endTime ) return false;
		if ( seqNum != that.seqNum ) return false;
		if ( startTime != that.startTime ) return false;
		if ( !Arrays.equals( dataBlocks, that.dataBlocks ) ) return false;

		return true;
	}

	public int hashCode() {
		int result;
		result = seqNum;
		result = 31 * result + (int) ( startTime ^ ( startTime >>> 32 ) );
		result = 31 * result + (int) ( endTime ^ ( endTime >>> 32 ) );
		result = 31 * result + Arrays.hashCode( dataBlocks );
		return result;
	}

	public String toString() {
		return "IndexBlock{" +
		       "seqNum=" + seqNum +
		       ", startTime=" + startTime +
		       ", endTime=" + endTime +
		       ", dataBlocks=" + Arrays.toString( dataBlocks ) +
		       '}';
	}
}
