/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.model;

import java.util.Arrays;
import java.util.NoSuchElementException;

/**
 * IndexBlockFactory holds the state to iteratively build up {@link IndexBlock}s.
 *
 * @author Jason Winnebeck
 */
public class IndexBlockFactory {
	private final int maxIndices;
	private final int[] blocks;

	private long  startTime;
	private long  endTime;
	private int   numBlocks;

	/**
	 * Builds a new IndexBlockFactory.
	 *
	 * @param maxIndices     the number of data blocks the IndexBlock can hold
	 * @param minStartTime   The startTime parameter to the first {@link #addDataBlock}
	 *                       call must be greater than this number.
	 */
	public IndexBlockFactory( int maxIndices, long minStartTime ) {
		this( maxIndices, 0, minStartTime, null );
	}

	/**
	 * Builds a new IndexBlockFactory, "resuming" an in-progress block.
	 *
	 * @param maxIndices     the number of data blocks the IndexBlock can hold
	 * @param blockIndices   nullable array with length &lt;= 'maxBlocks' with starting data block
	 *                       indices.
	 * @param startTime      timestamp of the first data in the first data block
	 * @param endTime        timestamp of the data data in the last data block
	 */
	public IndexBlockFactory( int maxIndices, long startTime, long endTime,
	                          int[] blockIndices ) {
		this.maxIndices = maxIndices;
		this.startTime = startTime;
		this.endTime   = endTime;

		blocks    = new int[ maxIndices ];
		numBlocks = 0;
		if ( blockIndices != null ) {
			//Let the arraycopy method throw the exception for us if blockIndices is too large
			System.arraycopy( blockIndices, 0, blocks, 0, blockIndices.length );
			numBlocks = blockIndices.length;
		}
	}

	/**
	 * Constructs a new IndexBlockFactory to "resume" writing to a given IndexBlock.
	 *
	 * @param maxIndices     the number of data blocks the IndexBlock can hold
	 * @param block          IndexBlock to resume, wihch has &lt;= maxBlocks indices in it
	 */
	public IndexBlockFactory( int maxIndices, IndexBlock block ) {
		this( maxIndices, block.getStartTime(), block.getEndTime(), block.getDataBlocks() );
	}

	/**
	 * Returns the maxmimum number of data block indices an IndexBlock created by this factory can
	 * hold.
	 */
	public int getMaxIndices() {
		return maxIndices;
	}

	/**
	 * The start time for the first data block in this index block, or -1 if there are none.
	 */
	public long getStartTime() {
		if ( numBlocks == 0 )
			return -1;
		else
			return startTime;
	}

	/**
	 * Returns the 'endTime' from the last {@link #addDataBlock(int, long, long)} call, or the
	 * 'minStartTime' from the constructor if it has never been called.
	 */
	public long getEndTime() {
		return endTime;
	}

	/**
	 * Return the number of data block indices currently in the IndexBlock under construction.
	 */
	public int getNumIndices() {
		return numBlocks;
	}

	/**
	 * Returns the data index of the first block, or throws IndexOutOfBoundsException if there isn't
	 * one.
	 */
	public int getFirstIndex() {
		if ( numBlocks < 1 )
			throw new IndexOutOfBoundsException();

		return blocks[0];
	}

	/**
	 * Returns the data index of the second block, or throws IndexOutOfBoundsException if there isn't
	 * one.
	 */
	public int getSecondIndex() {
		if ( numBlocks < 2 )
			throw new IndexOutOfBoundsException();

		return blocks[1];
	}

	/**
	 * Returns the data index of the last block, or throws IndexOutOfBoundsException if there isn't
	 * one.
	 */
	public int getLastIndex() {
		if ( numBlocks < 1 )
			throw new IndexOutOfBoundsException();

		return blocks[ numBlocks - 1 ];
	}

	/**
	 * Returns an array of the indices currently stored by the factory.
	 */
	public int[] getIndices() {
		int[] blocksCopy = new int[ numBlocks ];
		System.arraycopy( blocks, 0, blocksCopy, 0, numBlocks );
		return blocksCopy;
	}

	/**
	 * Returns true if {@link #addDataBlock} can safely be called on this factory, which is if
	 * {@link #getNumIndices()} < {@link #getMaxIndices()}.
	 */
	public boolean hasSpace() {
		return numBlocks < maxIndices;
	}

	/**
	 * Adds the given data block number to the end of the list of data blocks. Can be called only if
	 * {@link #hasSpace()} returns true. Over a given factory instance, the startTime values must be
	 * greater than the greatest endTime (or minStartTime from constructor) ever passed to this
	 * factory.
	 *
	 * @param dataBlockNum data block number
	 * @param startTime    timestamp of the first Record in the data block, > last endTime
	 * @param endTime      timestamp of the last Record in the data block, >= startTime
	 */
	public void addDataBlock( int dataBlockNum, long startTime, long endTime ) {
		if ( !hasSpace() )
			throw new IllegalStateException( "No space left to add a block" );
		if ( endTime < startTime )
			throw new IllegalArgumentException( "endTime before startTime" );
		if ( startTime <= this.endTime )
			throw new IllegalArgumentException( "startTime at or before endTime from previous call" );

		if ( numBlocks == 0 )
			this.startTime = startTime;

		blocks[ numBlocks ] = dataBlockNum;
		++numBlocks;
		this.endTime = endTime;
	}

	/**
	 * Updates the last data block added via {@link #addDataBlock(int, long, long)}.
	 *
	 * @param dataBlockNum data block number, exactly equal to the last call of addDataBlock
	 * @param startTime    data block start time, exactly equal to the last call of addDataBlock
	 * @param endTime      new endTime, greater than or equal to the last call of addDataBlock
	 */
	public void updateDataBlock( int dataBlockNum, long startTime, long endTime ) {
		if ( numBlocks == 0 )
			throw new IllegalStateException( "Trying to update non-existant data block" );
		if ( blocks[numBlocks - 1] != dataBlockNum )
			throw new IllegalArgumentException( "Trying to update a different data block than last added" );
		if ( this.startTime != startTime )
			throw new IllegalArgumentException( "Trying to update a different data block than last added" );
		if ( this.endTime > endTime )
			throw new IllegalArgumentException( "Trying to update endTime backwards" );

		this.endTime = endTime;
	}

	/**
	 * Removes the first data block index stored in this factory. The endTime property of this factory
	 * is unchanged by this method, even if the factory becomes empty, so new data blocks must still
	 * be increasing in time.
	 *
	 * @param startTime a new value for factory's startTime that must be less than or equal to
	 *                  endTime and should be the startTime for the data block following the first,
	 *                  which should be greater than the first's startTime. Only the condition with
	 *                  endTime is checked for now. If removing the last block, this parameter is
	 *                  ignored.
	 *
	 * @throws IllegalArgumentException if startTime &gt; endTime
	 * @throws NoSuchElementException   if numBlocks is zero
	 */
	public void removeFirstDataBlock( long startTime ) {
		if ( numBlocks > 1 && startTime > endTime )
			throw new IllegalArgumentException( "startTime > endTime" );
		if ( numBlocks == 0 )
			throw new NoSuchElementException( "No blocks left to remove" );

		//TODO: optimize this by having a circular buffer
		System.arraycopy( blocks, 1, blocks, 0, numBlocks - 1 );
		this.startTime = ( numBlocks == 1 ) ? 0 : startTime;
		--numBlocks;
	}

	/**
	 * Starts a new block by resetting the data block list. The startTime of any data blocks must be
	 * greater than the endTime of the last block (the one existing before this call).
	 */
	public void startNextBlock() {
		//start time and end time remain the same -- they won't be used until the first block is added
		//in which case we do want endTime to be present for checking.
		numBlocks = 0;
	}

	/**
	 * Constructs an IndexBlock with the current factory's state. If any exception is thrown, the
	 * sequence is unchanged. If no blocks have been added, an empty IndexBlock is returned.
	 */
	public IndexBlock makeBlock( int seqNum ) {
		if ( numBlocks == 0 )
			return new IndexBlock( seqNum );
		else
			return new IndexBlock( seqNum, startTime, endTime, getIndices() );
	}

	public boolean equals( Object o ) {
		if ( this == o ) return true;
		if ( o == null || getClass() != o.getClass() ) return false;

		IndexBlockFactory that = (IndexBlockFactory) o;

		if ( endTime != that.endTime ) return false;
		if ( maxIndices != that.maxIndices ) return false;
		if ( numBlocks != that.numBlocks ) return false;
		//startTime only matters if numBlocks > 0
		if ( numBlocks > 0 && startTime != that.startTime ) return false;
		if ( !Arrays.equals( blocks, that.blocks ) ) return false;

		return true;
	}

	public int hashCode() {
		int result;
		result = maxIndices;
		result = 31 * result + (int) ( startTime ^ ( startTime >>> 32 ) );
		result = 31 * result + (int) ( endTime ^ ( endTime >>> 32 ) );
		result = 31 * result + Arrays.hashCode( blocks );
		result = 31 * result + numBlocks;
		return result;
	}
}
