/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.model;

import org.gillius.realdb.BadFileException;
import org.gillius.realdb.model.data.Record;
import org.gillius.realdb.model.data.StreamInterval;
import org.gillius.realdb.util.DBIterator;

import java.io.IOException;
import java.util.NoSuchElementException;

/**
 * IntervalIterator passes a sequence of {@link Record}s from an {@link DBIterator} through a
 * {@link StreamReconstructor} to generate a sequence of {@link StreamInterval}s.
 *
 * @author Jason Winnebeck
 */
public class IntervalIterator implements DBIterator<StreamInterval> {
	private final DBIterator<Record> iter;
	private final StreamReconstructor recon;

	private final long endTime;

	private boolean        started     = false;
	private boolean        ended       = false;
	private StreamInterval next        = null;
	private long           lastEndTime = -1;

	/**
	 * @param iter    A DBIterator that iterates records to the end of the DataStream
	 * @param recon   StreamReconstructor to construct StreamIntervals from Records
	 */
	public IntervalIterator( DBIterator<Record> iter, StreamReconstructor recon ) {
		this( iter, recon, Long.MAX_VALUE );
	}

	/**
	 * @param iter    A DBIterator that iterates records to the end of the DataStream
	 * @param recon   StreamReconstructor to construct StreamIntervals from Records
	 * @param endTime end time of the range to reconstruct, can be Long.MAX_VALUE for all intervals
	 */
	public IntervalIterator( DBIterator<Record> iter, StreamReconstructor recon, long endTime ) {
		this.iter    = iter;
		this.recon   = recon;
		this.endTime = endTime;
	}

	public boolean hasNext() throws IOException, BadFileException {
		if ( next != null ) {
			return true;

		} else if ( ended ) {
			return false;

		} else if ( !started ) {
			if ( !iter.hasNext() ) {
				//there are no records at all
				ended = true;
				return false;
			}

			Record r = iter.next();
			if ( r.getTime() > endTime ) {
				//The first interval will start after our endTime, so there are no intervals to iterate
				ended = true;
				return false;
			}

			recon.startRange( r );
			lastEndTime = r.getTime(); //need this for "single" Record situation, for finishRecords.
			started = true;
		}

		//While we have records left, push them into the reconstructor
		while ( next == null && iter.hasNext() ) {
			next = recon.handleRecord( iter.next() );
		}
		if ( next == null ) {
			//If we still don't have a next, then we ran out of records, so we do a finish.
			next = recon.finishRecords( lastEndTime );
		}
		if ( next != null ) {
			lastEndTime = next.getEndTime();
			ended = ( lastEndTime > endTime );
			return true;

		} else {
			//Reached the end the intervals without hitting our endTime
			return false;
		}
	}

	public StreamInterval next() throws IOException, BadFileException {
		if ( !hasNext() )
			throw new NoSuchElementException();

		StreamInterval ret = next;
		next = null;
		return ret;
	}
}
