/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.model;

import org.gillius.realdb.BadFileException;

import java.io.IOException;

/**
 * RangeReaderUtil contains utility methods working with {@link RangeReader}s.
 *
 * @author Jason Winnebeck
 */
public class RangeReaderUtil {
	private RangeReaderUtil() {
	}

	/**
	 * Given a sorted, non-overlapping set of Ranges from a RangeReader, performs a binary search
	 * for the range that either contains the given time. If no range contains the given time, return
	 * the range immediately before the time, or the first range if before all ranges.
	 *
	 * @throws IllegalArgumentException if the RangeReader has no ranges
	 */
	public static int findRange( RangeReader reader, long time ) throws IOException, BadFileException {
		//The below algorithm is adapted from Arrays.binarySearch.
		int low = 0;
		int high = reader.getNumRanges() - 1;

		if ( high < 0 )
			throw new IllegalArgumentException( "The given RangeReader must have some ranges" );

		while ( low <= high ) {
			int mid = ( low + high ) >>> 1;
			Range midVal = reader.getRange( mid );

			if ( midVal.getStartTime() < time ) {
				if ( midVal.getEndTime() > time )
					return mid; //record is in this block and isn't the first in it.
				else
					low = mid + 1;
			} else if ( midVal.getStartTime() > time ) {
				//record is entirely before the middle block
				high = mid - 1;
			} else {
				//The record is exactly the first in the block, but we really want to get at the record
				//immediately before, unless it is the first record or before is false.
				return mid;
			}
		}
		return low > 0 ? low - 1 : low; //exact timestamp not found; return the point below (or 0)
	}
}
