/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.model;

import org.gillius.Util.CheckUtils;
import org.gillius.realdb.BadFileException;

import java.io.DataOutput;
import java.io.IOException;
import java.nio.BufferUnderflowException;
import java.nio.ByteBuffer;
import java.util.Arrays;
import java.util.zip.CRC32;
import java.util.zip.Checksum;

/**
 * The RawSequencedBlock holds a byte array of raw data with a sequence number. The data is
 * checksummed when written and checked on reading.
 *
 * @author Jason Winnebeck
 */
public class RawSequencedBlock implements SequencedBlock {
	public static final SequencedBlockLoader<RawSequencedBlock> loader =
			new SequencedBlockLoader<RawSequencedBlock>() {
				public RawSequencedBlock load( byte[] serialized ) throws BadFileException {
					return new RawSequencedBlock( serialized );
				}
			};

	/**
	 * Returns the minimum size of the RawSequencedBlock (its header size). The final size of the
	 * block will be this value (the header) plus the length of the rawData array.
	 */
	public static int getMinSize() {
		return 4 + 2 + 4; //seqNum + rawData.length + checksum
	}

	private final int    seqNum;
	private final byte[] rawData;

	/**
	 * Precomputed checksum, not used in equals and hashCode since it is directly derived from
	 * rawData.
	 */
	private final int    checksum;

	public RawSequencedBlock( int seqNum, byte[] rawData ) {
		CheckUtils.checkNull( rawData, "rawData" );
		if ( rawData.length > 65535 )
			throw new UnsupportedOperationException( "max supported data array length is 65535" );

		this.seqNum  = seqNum;
		this.rawData = rawData;

		checksum = getChecksum( rawData );
	}

	/**
	 * Reads a RawSequencedBlock from the given serialized form data.
	 */
	public RawSequencedBlock( byte[] data ) throws BadFileException {
		ByteBuffer buf = ByteBuffer.wrap( data );

		try {
			seqNum  = buf.getInt();
			int len = buf.getShort() & 0xFFFF;
			rawData = new byte[ len ];
			buf.get( rawData );
			checksum = buf.getInt();
		} catch ( BufferUnderflowException e ) {
			throw new BadFileException( "Corrupted block", e );
		}

		int expectedCrc = getChecksum( rawData );

		if ( expectedCrc != checksum )
			throw new BadFileException( "Checksum does not match" );
	}

	public int getSeqNum() {
		return seqNum;
	}

	public byte[] getRawData() {
		return rawData;
	}

	public void write( DataOutput output ) throws IOException {
		output.writeInt( seqNum );
		output.writeShort( rawData.length );
		output.write( rawData );
		output.writeInt( checksum );
	}

	public boolean equals( Object o ) {
		if ( this == o ) return true;
		if ( o == null || getClass() != o.getClass() ) return false;

		RawSequencedBlock that = (RawSequencedBlock) o;

		if ( seqNum != that.seqNum ) return false;
		if ( !Arrays.equals( rawData, that.rawData ) ) return false;

		return true;
	}

	public int hashCode() {
		int result;
		result = seqNum;
		result = 31 * result + Arrays.hashCode( rawData );
		return result;
	}

	private static int getChecksum( byte[] rawData ) {
		Checksum chk = new CRC32();
		chk.update( rawData, 0, rawData.length );

		return (int) chk.getValue() + 1; //add 1 so that an "all zeros" block is not "good"
	}
}
