/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.model;

import org.gillius.realdb.BadFileException;
import org.gillius.realdb.io.BlockFile;
import org.gillius.realdb.io.ByteBufferDataIO;
import org.gillius.realdb.io.Formattable;

import java.io.IOException;
import java.nio.ByteBuffer;

/**
 * ReliableSequencedBlock persists a SequencedBlock in a fail-safe manner such that every write to
 * the block either succeeds entirely, or if it fails, it reverts back to its previous state. This
 * is achieved by writing two alternating copies of the block in the BlockFile.
 *
 * @author Jason Winnebeck
 */
public class ReliableSequencedBlock<T extends SequencedBlock> implements Formattable {
	private final BlockFile file;
	private final int       firstBlock;
	private final int       secondBlock;

	private final ByteBufferDataIO writer;

	private T       block;
	private int     nextSeqNum;
	private boolean onPrimary;

	/**
	 * Creates a new ReliableSequencedBlock by ignoring existing data on disk and formatting both
	 * allocated spaces to all 0's, based on 'size'. Per the contract of {@link SequencedBlock}, data
	 * consisting of all NUL bytes must not be a valid serialized form on disk, and this class uses
	 * that assumption to perform the format.
	 * <p/>
	 * This means that a newly created/formatted ReliableSequencedBlock will throw an exception when
	 * reloading it, unless {@link #writeBlock(SequencedBlock)} is called.
	 *
	 * @param file        non-null BlockFile
	 * @param firstBlock  block index of start of primary space
	 * @param secondBlock block index of start of secondary space
	 * @param size        maximum size of serialized block, in bytes
	 */
	public static <T extends SequencedBlock> ReliableSequencedBlock<T> createNew(
			BlockFile file, int firstBlock, int secondBlock, int size ) throws IOException {
		ReliableSequencedBlock<T> ret =
				new ReliableSequencedBlock<T>( file, firstBlock, secondBlock, size );
		ret.format();
		return ret;
	}

	/**
	 * Creates a new ReliableSequencedBlock by loading the most recent and valid block.
	 *
	 * @param file        non-null BlockFile
	 * @param firstBlock  block index of start of primary space
	 * @param secondBlock block index of start of secondary space
	 * @param size        maximum size of serialized block, in bytes
	 * @param loader      loader to deserialize the data
	 *
	 * @throws IOException      if the file could not be read
	 * @throws BadFileException if neither block is valid
	 */
	public static <T extends SequencedBlock> ReliableSequencedBlock<T> load(
			BlockFile file, int firstBlock, int secondBlock, int size, SequencedBlockLoader<T> loader )
			throws IOException, BadFileException {
		ReliableSequencedBlock<T> ret =
				new ReliableSequencedBlock<T>( file, firstBlock, secondBlock, size );
		byte[] first  = new byte[size];
		byte[] second = new byte[size];
		file.readBlocks( firstBlock,  first  );
		file.readBlocks( secondBlock, second );
		int[] idx = new int[1];

		ret.block      = SequencedBlockUtil.loadBestBlock( loader, idx, first, second );
		ret.nextSeqNum = ret.block.getSeqNum() + 1;
		ret.onPrimary  = idx[0] != 0;

		return ret;
	}

	private ReliableSequencedBlock( BlockFile file, int firstBlock, int secondBlock, int size ) {
		this.file        = file;
		this.firstBlock  = firstBlock;
		this.secondBlock = secondBlock;

		writer = new ByteBufferDataIO( size );
	}

	/**
	 * Returns the cached block; either that which was last written, or the block loaded if just
	 * loaded, or null if just created. If this ReliableSequencedBlock was created by load, or if
	 * {@link #writeBlock} has ever been called, this method will never return null.
	 */
	public T getBlock() {
		return block;
	}

	/**
	 * Returns the sequence number for the next block to be written.
	 */
	public int getNextSeqNum() {
		return nextSeqNum;
	}

	/**
	 * Formats both primary and secondary block space by writing all NUL bytes. Per the contract of
	 * {@link SequencedBlock}, data consisting of all NUL bytes must not be a valid serialized form on
	 * disk, and this class uses that assumption to perform the format.
	 */
	public void format() throws IOException {
		int maxBytes = writer.getBuf().rewind().capacity();
		byte[] zeros = new byte[ maxBytes ];

		file.writeBlocks( firstBlock, zeros );
		file.writeBlocks( secondBlock, zeros );
	}

	/**
	 * Writes the given block to the file. If an exception occurs, there is NO change in state of
	 * this object, however, one or both of the portions of the underlying file may be corrupted.
	 */
	public void writeBlock( T block ) throws IOException {
		if ( block.getSeqNum() != nextSeqNum )
			throw new IllegalArgumentException( "Block to be written had unexpected sequence number" );

		int dest = onPrimary ? firstBlock : secondBlock;

		ByteBuffer buf = writer.getBuf();
		buf.rewind();
		block.write( writer );

		file.writeBlocks( dest, buf.array(), buf.arrayOffset(), buf.position() );
		this.block = block;
		onPrimary  = !onPrimary;
		++nextSeqNum;
	}
}
