/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.model;

import org.gillius.realdb.BadFileException;

/**
 * SequencedBlockUtil provides utility functions useful with {@link SequencedBlock}s.
 *
 * @author Jason Winnebeck
 */
class SequencedBlockUtil {
	/**
	 * Given a SequencedBlockLoader and a list of blocks, returns the one with the highest sequence
	 * number. Note there is some subtleties with comparing the sequence numbers since they wrap
	 * around. In this implementation, the first successfully loaded block becomes the "best" block,
	 * then the blocks are compared in order by keeping best if (best.seqNum - next.seqNum >= 0). This
	 * works perfectly if the space between the oldest and newest block is not more than 2^31. If a
	 * block has the same sequence number as the best block, then the first block loaded is the best.
	 *
	 * @param loader      non-null loader to use to load the blocks
	 * @param blockLoaded if not null and non-zero length, the first index in the array is set to the
	 *                    index of the best block if function returns.
	 * @param blocks      list of 1 or more blocks to try to load
	 *
	 * @return the best block loaded
	 *
	 * @throws BadFileException if none of the blocks can be loaded. The exception's cause is the
	 *                          exception thrown by the loader trying to load the last block.
	 */
	public static <T extends SequencedBlock> T loadBestBlock( SequencedBlockLoader<T> loader,
	                                                          int[] blockLoaded,
	                                                          byte[]... blocks )
			throws BadFileException {
		T best = null;
		T next = null;
		BadFileException lastError = null;

		for ( int i = 0; i < blocks.length; i++ ) {
			try {
				next = loader.load( blocks[i] );

				if ( best == null || next.getSeqNum() - best.getSeqNum() > 0 ) {
					best = next;
					if ( blockLoaded != null && blockLoaded.length > 0 )
						blockLoaded[0] = i;
				}
			} catch ( BadFileException e ) {
				lastError = e;
			}
		}

		if ( best != null )
			return best;
		else
			throw new BadFileException( "Could not load any blocks", lastError );
	}
}
