/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.model;

import org.gillius.realdb.model.data.Record;
import org.gillius.realdb.model.data.StreamInterval;

/**
 * A StreamReconstructor takes a sequence of {@link Record}s and builds a sequence of {@link
 * StreamInterval} objects from it.
 *
 * @author Jason Winnebeck
 */
public interface StreamReconstructor {
	/**
	 * Starts a new stream reconstruction on a range of records that starts with the given Record.
	 * All subsequent calls to {@link #handleRecord(Record)} should be Records in ascending time
	 * order.
	 *
	 * @param start non-null Record
	 */
	public void startRange( Record start );

	/**
	 * Handles the next record from the raw stream. Sometimes more records ahead are needed and null
	 * will be returned from a finite number of calls. Once this buffer is filled, all subsequent
	 * calls should return new StreamInterval objects.
	 *
	 * @param r non-null Record in ascending time order
	 *
	 * @return non-null StreamInterval
	 */
	public StreamInterval handleRecord( Record r );

	/**
	 * After all of the Records have been exhausted, there may will be intervals remaining to be
	 * output if there is a buffered delay. This method is called to denote the end of the time range
	 * to be reconstructed. Continue calling this method with the same endTime until it returns null
	 * to obtain all of the remaining intervals (or, the last interval's endTime will be the time
	 * given, so the caller could also check for this condition).
	 *
	 * @param endTime the end of the time range to be reconstructed, which is greater than the time of
	 *                the last Record given to {@link #handleRecord(Record)}.
	 *
	 * @return nullable StreamInterval.
	 */
	public StreamInterval finishRecords( long endTime );
}
