/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.model;

import org.gillius.realdb.model.data.Record;

/**
 * StreamTransformer provides an interface to something that handles a stream of Records and outputs
 * another stream of Records that has the same number or fewer Records. It is also possible for the
 * StreamTransformer to delay the output until a later input is received -- a buffer delay.
 *
 * @author Jason Winnebeck
 */
public interface StreamTransformer {
	/**
	 * Resets this StreamTransformer to the initial state. If this transformer has any buffers, they
	 * will be cleared.
	 */
	public void reset();

	/**
	 * Handles the record by possibly sending out a new record, or it may drop the record by returning
	 * null. Since this is the most speed-sensitive part of RealDB, there is no additional checking
	 * here that the Record given matches the expected type.
	 *
	 * @param r non-null Record whose type matches the elements expected.
	 *
	 * @return An output Record, or null if there is no additional record to output based on the
	 *         input.
	 */
	public Record handleRecord( Record r );

	/**
	 * After all of the Records have been given to the transformer, there may still be records left
	 * to output. This method should be called in a loop until null is returned to ensure that all
	 * Records have been obtained.
	 * <p/>
	 * After calling this method, {@link #reset} has to be called before the next
	 * {@link #handleRecord}.
	 *
	 * @return An output Record, or null if there is no additional records to output
	 */
	public Record flushRecords();
}
