/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.model;

import java.io.Flushable;
import java.io.IOException;

/**
 * A Transaction represents a ordered series of actions that need special attention if any failures
 * occur while it is open. The first action of a transaction is to open it, and the last action is
 * to close it (commit it). RealDB does not support the concept of a rollback.
 *
 * @author Jason Winnebeck
 */
public interface Transaction {
	/**
	 * Adds an ordered action to this Transaction. Returns a Flushable that must be called before
	 * actually performing (or "flushing") the action given. Typically, the returned flushable flushes
	 * all previous actions, ensuring that all operations occur in the order given. Due to this
	 * design, the flush methods involved should be reentrant in that they can handle being flushed
	 * multiple times.
	 * <p/>
	 * The first call to this method returns a non-null Flushable that has no effect, so it can be
	 * discarded.
	 *
	 * @param action non-null Flushable
	 * @return non-null Flushable
	 */
	public Flushable addAction( Flushable action );

	/**
	 * Flushes all actions in this transaction, then closes it. No further actions can be added to the
	 * Transaction once it is committed.
	 *
	 * @throws IOException if any flush operation throws. If a transaction fails, the database needs
	 *                     to be reloaded, so that the transaction can be recovered.
	 */
	public void commit() throws IOException;
}
