/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.model.algorithms;

import org.gillius.realdb.metadata.AbstractElementCodec;
import org.gillius.realdb.metadata.ElementType;
import org.gillius.realdb.model.ElementCompressor;
import org.gillius.realdb.model.ElementReconstructor;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;
import java.util.Map;
import java.util.Set;

/**
 * DeadbandAlgorithm implements an ElementCodec that suppresses the output of values if the value has
 * not changed enough from the last recorded point.
 *
 * @author Jason Winnebeck
 */
public class DeadbandAlgorithm extends AbstractElementCodec {
	private double deadband;

	public DeadbandAlgorithm() {
	}

	public DeadbandAlgorithm( double deadband ) {
		this.deadband = deadband;
	}

	@Override
	public void getRequiredParams( Set<String> params ) {
		super.getRequiredParams( params );
		params.add( "deadband" );
	}

	@Override
	public void setParams( Map<String, String> params ) throws IllegalArgumentException {
		super.setParams( params );

		deadband = Double.parseDouble( params.get( "deadband" ) );
	}

	public ElementCompressor getCompressor( ElementType type ) {
		return new Compressor();
	}

	public ElementReconstructor getReconstructor( ElementType type ) {
		return FlatElementReconstructor.instance;
	}

	public void read( DataInput input ) throws IOException {
		deadband = input.readDouble();
	}

	public void write( DataOutput output ) throws IOException {
		output.writeDouble( deadband );
	}

	public boolean equals( Object o ) {
		if ( this == o ) return true;
		if ( o == null || getClass() != o.getClass() ) return false;

		DeadbandAlgorithm that = (DeadbandAlgorithm) o;

		if ( Double.compare( that.deadband, deadband ) != 0 ) return false;

		return true;
	}

	public int hashCode() {
		long temp = deadband != +0.0d ? Double.doubleToLongBits( deadband ) : 0L;
		return (int) ( temp ^ ( temp >>> 32 ) );
	}

	public String toString() {
		return "DeadbandAlgorithm with deadband " + deadband;
	}

	private class Compressor implements ElementCompressor {
		public int getLookBehindDesired() {
			return 1;
		}

		public int getLookAheadDesired() {
			return 0;
		}

		public boolean handleNewValue( long[] times, Object[] values, int currentIndex ) {
			final Object rawLastValue = values[ currentIndex - 1 ];
			final Object value        = values[ currentIndex ];

			final double lastValue = rawLastValue == null ? Double.NaN :
			                         ( (Number)rawLastValue ).doubleValue();

			final double nextVal;
			if ( value == null )
				nextVal = Double.NaN;
			else
				nextVal = ((Number)value).doubleValue();

			//This first check should handle double NaN, double Infinity, and exact equality
			if ( Double.compare( lastValue, nextVal ) == 0 ) {
				return false;
			} else {
				if ( Double.isNaN( lastValue ) || Double.isNaN( nextVal ) )
					return true;

				return ( nextVal - lastValue ) >= deadband;
			}
		}

		public void valueWritten( long time, Object value ) {
		}
	}
}