/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.model.data;

import org.gillius.realdb.model.ElementReconstructor;
import org.gillius.realdb.model.StreamReconstructor;

/**
 * DefaultStreamReconstructor rebuilds records written with a RecordCompressor. You can get
 * instances of this class via the {@link DefaultStreamReconstructorFactory}.
 *
 * @see RecordCompressor
 *
 * @author Jason Winnebeck
 */
public class DefaultStreamReconstructor implements StreamReconstructor {
	private final ElementReconstructor[] recons;

	private final RecordBuffer buf;

	/**
	 * If true, we are clearing out the buffer or aren't in a range
	 */
	private boolean finishing = true;

	/**
	 * Constructor meant to be called from the factory, which has calculated the values.
	 */
	DefaultStreamReconstructor( ElementReconstructor[] recons, int pivot, int bufferLen ) {
		this.recons = recons;
		buf = new RecordBuffer( bufferLen, recons.length, pivot );
	}

	public void startRange( Record start ) {
		buf.clear();
		buf.rotate( start );
		finishing = false;
	}

	public StreamInterval handleRecord( Record r ) {
		if ( finishing )
			throw new IllegalStateException( "Invalid to call handleRecord; not in a range" );

		buf.rotate( r );
		if ( buf.getTimes()[ buf.getPivot() ] < 0 ) {
			//We haven't filled our lookahead buffer yet
			return null;

		} else {
			return makeInterval();
		}
	}

	public StreamInterval finishRecords( long endTime ) {
		if ( !finishing ) {
			finishing = true;
			buf.rotateEmpty( endTime );
		} else {
			buf.rotateEmpty( -1 );
		}
		if ( buf.getTimes()[ buf.getPivot() + 1 ] < 0 ) {
			return null;
		} else {
			return makeInterval();
		}
	}

	private StreamInterval makeInterval() {
		final long[]    t       = buf.getTimes();
		final int       pivot   = buf.getPivot();
		final boolean[] discons = buf.getDiscontinuities();

		if ( discons[pivot] ) {
			return new DiscontinuityInterval( t[pivot], t[pivot+1] );

		} else {
			final Object[][] data = buf.getData();
			ElementInterval[] intervals = new ElementInterval[ recons.length ];
			for ( int i = 0; i < recons.length; ++i ) {
				intervals[i] = recons[i].getInterval( t, data[i], pivot );
			}
			return new DefaultStreamInterval( t[pivot], t[pivot + 1], intervals );
		}
	}
}
