/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.model.data;

import java.util.Arrays;

/**
 * GenericRecord implements a generic (dynamically-typed) Record that acts like an array of Object.
 * Such a record is never discontinuous.
 * <p/>
 * At some point, the plan is to have RealDB allow to have concrete instances of Record with named
 * class members instead of just this simple Object array. These classes are envisioned to be
 * auto-generated from a tool that reads the RDL file.
 *
 * @author Jason Winnebeck
 */
public class GenericRecord implements Record {
	private final long     time;
	private final Object[] values;

	/**
	 * Constructs a new GenericRecord with a timestamp and 0 or more elements. <b>For speed purposes,
	 * the elements array is not cloned and should never be modified after passing it into this
	 * constructor.</b> The clone has been found to be a bottle neck of up to 33% of total CPU load
	 * for straight reading and writing.
	 *
	 * @param time     non-negative timestamp, as defined in {@link #getTime()}
	 * @param elements non-null array of 0 or more Object (elements in this record)
	 */
	public GenericRecord( long time, Object... elements ) {
		if ( time < 0 )
			throw new IllegalArgumentException( "time cannot be negative" );

		this.time   = time;
		this.values = elements;
	}

	public long getTime() {
		return time;
	}

	public boolean isDiscontinuity() {
		return false;
	}

	public int getNumElements() {
		return values.length;
	}

	public Object getElement( int index ) {
		return values[index];
	}

	public String toString() {
		return "GenericRecord, time " + time + ", data " + Arrays.toString( values );
	}

	public boolean equals( Object o ) {
		if ( this == o ) return true;
		if ( o == null || getClass() != o.getClass() ) return false;

		GenericRecord that = (GenericRecord) o;

		if ( time != that.time ) return false;
		// Probably incorrect - comparing Object[] arrays with Arrays.equals
		if ( !Arrays.deepEquals( values, that.values ) ) return false;

		return true;
	}

	public int hashCode() {
		int result;
		result = (int) ( time ^ ( time >>> 32 ) );
		result = 31 * result + Arrays.deepHashCode( values );
		return result;
	}
}
