/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.model.data;

/**
 * Record represents a low-level sample of data within the RealDB system. In some cases it
 * represents the next value to be presented to the stream, in others it is the actual value to be
 * written to or just read from the database.
 * <p/>
 * A Record in an of itself is not inherently associated with any DataStream, so there has to be
 * an awareness outside of this object alone to know what DataStream to which the Record applies,
 * to know the types and meaning of the values.
 * <p/>
 * Record objects must be immutable; meaning that they cannot change their value. This also implies
 * that they should be thread-safe.
 *
 * @author Jason Winnebeck
 */
public interface Record {
	/**
	 * Returns the timestamp of this record, expressed as a time in some unit since some epoch.
	 * The application that uses RealDB defines the unit and epoch, but it is almost always
	 * milliseconds since Jan 1, 1970 GMT, which is the same as the result returned by
	 * {@link System#currentTimeMillis()}.
	 *
	 * @return non-negative timestamp
	 */
	public long getTime();

	/**
	 * Returns true if this Record represents the beginning of a discontinuity in the data stream.
	 * Such a record will have 0 elements.
	 */
	public boolean isDiscontinuity();

	/**
	 * Returns the total number of values stored in this Record, not including the timestamp.
	 */
	public int getNumElements();

	/**
	 * Returns the value at the given index.
	 *
	 * @param index value from 0 to {@link #getNumElements()}-1.
	 *
	 * @return nullable Object
	 *
	 * @throws IndexOutOfBoundsException if index is out of the allowed range
	 */
	public Object getElement( int index );
}
