/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.model.data;

import java.util.Arrays;

/**
 * RecordBuffer stores a view of "future" and "past" snapshots of Record data in a raw format. The
 * 'times' array stores the timestamps for the records. The 'data' array stores the look-ahead and
 * look-behind buffer with each element getting its own row. The 'pivot' is the "current" record
 * index, so the current Record's timestamp is times[pivot] and the current value for the second
 * element is data[1][pivot]. The initial values in the times array is -1 (an invalid timestamp),
 * and nulls in the data array.
 *
 * @author Jason Winnebeck
 */
public class RecordBuffer {
	private long[] times;
	private Object[][] data;
	private boolean[] discon;

	private final int pivot;

	/**
	 * Constructs a new RecordBuffer with the given number of records and elements in each record.
	 *
	 * @param length   buffer length, at least 1
	 * @param elements number of elements, at least 1
	 * @param pivot    "current record" pointer, less than length
	 */
	public RecordBuffer( int length, int elements, int pivot ) {
		times  = new long[ length ];
		data   = new Object[elements][length];
		discon = new boolean[ length ];

		this.pivot = pivot;

		Arrays.fill( times, -1 );
	}

	public void clear() {
		Arrays.fill( times, -1 );
		for ( Object[] elementData : data ) {
			Arrays.fill( elementData, null );
		}
	}

	public long[] getTimes() {
		return times;
	}

	public Object[][] getData() {
		return data;
	}

	public boolean[] getDiscontinuities() {
		return discon;
	}

	public int getPivot() {
		return pivot;
	}

	/**
	 * Returns the total number of records in the buffer.
	 */
	public int getSize() {
		return times.length;
	}

	/**
	 * Returns the number of records behind the pivot.
	 */
	public int getLookBehind() {
		return pivot;
	}

	/**
	 * Returns the number of records after the pivot.
	 */
	public int getLookAhead() {
		return times.length - pivot - 1;
	}

	/**
	 * Rotates a record with all null element values in the front with the given time.
	 */
	public void rotateEmpty( long time ) {
		System.arraycopy( times,  1, times,  0, times.length  - 1 );
		System.arraycopy( discon, 1, discon, 0, discon.length - 1 );
		times[ times.length - 1 ] = time;
		discon[ times.length - 1 ] = false;

		for ( Object[] values : data ) {
			System.arraycopy( values, 1, values, 0, values.length - 1 );
			values[ values.length - 1 ] = null;
		}
	}

	/**
	 * Rotates the buffer by pushing everything back and inserting the new record at the front.
	 */
	public void rotate( Record r ) {
		System.arraycopy( times,  1, times,  0, times.length  - 1 );
		System.arraycopy( discon, 1, discon, 0, discon.length - 1 );
		times[ times.length - 1 ] = r.getTime();
		discon[ times.length - 1 ] = r.isDiscontinuity();

		for ( int i = 0, dataLength = data.length; i < dataLength; i++ ) {
			Object[] values = data[i];
			System.arraycopy( values, 1, values, 0, values.length - 1 );
			values[ values.length - 1 ] = r.isDiscontinuity() ? null : r.getElement( i );
		}
	}

	/**
	 * Removes the record at the pivot point and pulls back all records in the look-ahead. This is
	 * similar to a {@link #rotateEmpty(long)} call with -1 but leaves the records behind the pivot
	 * unchanged.
	 */
	public void dropPivot() {
		final int lookAhead = getLookAhead();

		//Since the times and values sets are unconditional, no operation is needed if there is no
		//lookahead at all.
		if ( lookAhead > 0 ) {
			System.arraycopy( times,  pivot + 1, times,  pivot, lookAhead );
			System.arraycopy( discon, pivot + 1, discon, pivot, lookAhead );
			times[ times.length - 1 ] = -1;
			discon[ times.length - 1 ] = false;

			for ( Object[] values : data ) {
				System.arraycopy( values, pivot + 1, values, pivot, lookAhead );
				values[ values.length - 1 ] = null;
			}
		}
	}

	/**
	 * Fills in the most recent Record.
	 */
	public void fillLast( Record r ) {
		times[ times.length - 1 ] = r.getTime();
		discon[ times.length - 1 ] = r.isDiscontinuity();

		for ( int i = 0, dataLength = data.length; i < dataLength; i++ ) {
			Object[] values = data[i];
			values[ values.length - 1 ] = r.isDiscontinuity() ? null : r.getElement( i );
		}
	}
}
