/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.model.data;

import org.gillius.realdb.metadata.RecordElement;
import org.gillius.realdb.model.ElementCompressor;
import org.gillius.realdb.model.StreamTransformer;

/**
 * RecordCompressor handles the set of {@link ElementCompressor}s needed to compress all of the
 * elements of a Record for a DataStream. It transforms a stream by eliminating records from it
 * based on the ElementCompressor operations.
 *
 * @author Jason Winnebeck
 * @see DefaultStreamReconstructor
 */
public class RecordCompressor implements StreamTransformer {
	private final ElementCompressor[] compressors;

	private final PreservingRecordBuffer buf;

	/**
	 * Constructs a RecordCompressor to compress Records consisting of the given elements. The
	 * forceNext property starts out as true, so that the first call to {@link #handleRecord(Record)}
	 * will output the record given.
	 *
	 * @param elements non-null RecordElement array
	 */
	public RecordCompressor( RecordElement[] elements ) {
		compressors = new ElementCompressor[ elements.length ];

		int maxBehind = 1; //RecordCompressor itself needs at least 1 look-behind
		int maxAhead  = 0;

		for ( int i = 0; i < elements.length; i++ ) {
			RecordElement element = elements[i];
			compressors[i] = element.getCodec().getCompressor( element.getType() );
			maxBehind = Math.max( maxBehind, compressors[i].getLookBehindDesired() );
			maxAhead  = Math.max( maxAhead,  compressors[i].getLookAheadDesired () );
		}

		int bufferLen = maxBehind + 1 + maxAhead;
		int pivot     = maxBehind;

		buf = new PreservingRecordBuffer( bufferLen, compressors.length, pivot );
	}

	public void reset() {
		buf.clear();
	}

	public Record handleRecord( Record r ) {
		buf.fillLast( r );

		if ( buf.getTimes()[ buf.getPivot() ] < 0 ) {
			//We're not filled up yet
			buf.rotateEmpty( -1 );
			return null;

		} else {
			return doCompression();
		}
	}

	private Record doCompression() {
		final long[]     t      = buf.getTimes();
		final int        pivot  = buf.getPivot();
		final Object[][] d      = buf.getData();
		final boolean[]  discon = buf.getDiscontinuities();

		//If this is the first record, or we are transitioning between discontinuity states
		boolean doWrite = ( ( t[pivot-1] == -1 ) || ( discon[pivot-1] != discon[pivot] ) );

		for ( int i = 0; !doWrite && i < compressors.length; i++ ) {
			doWrite = compressors[i].handleNewValue( t, d[i], pivot );
		}

		if ( doWrite ) {
			final Record ret = buf.getRecords()[pivot];
			for ( int i = 0; i < compressors.length; ++i ) {
				compressors[i].valueWritten( t[pivot], d[i][pivot] );
			}
			buf.rotateEmpty( -1 );
			return ret;

		} else {
			buf.dropPivot();
			return null;
		}
	}

	public Record flushRecords() {
		if ( buf.getTimes()[ buf.getPivot() ] < 0 ) {
			return null;
		} else {
			return doCompression();
		}
	}
}
