/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.model.data;

import org.gillius.realdb.model.Range;

/**
 * StreamInterval represents data within a given interval on a DataStream.
 *
 * @author Jason Winnebeck
 */
public interface StreamInterval extends Range {

	/**
	 * Returns true if this interval represents a discontinuity in the data stream.
	 * Such an interval will have 0 elements.
	 */
	public boolean isDiscontinuity();

	/**
	 * Returns the starting time (inclusive) of this interval.
	 */
	public long getStartTime();

	/**
	 * Returns the end time (exclusive) of this interval.
	 */
	public long getEndTime();

	/**
	 * Returns {@link #getEndTime()} - {@link #getStartTime()}.
	 */
	public long getTimespan();

	/**
	 * Returns the total number of elements (fields) in this interval, not including the timestamp.
	 */
	public int getNumElements();

	/**
	 * Samples the value at the given time between startTime (inclusive) and endTime (exclusive).
	 * The result of calling this method with a time less than startTime or greater than or equal to
	 * endTime is undefined. It may throw IllegalArgumentException, but is not guaranteed to.
	 *
	 * @throws IndexOutOfBoundsException if element is &lt; 0 or &gt;= {@link #getNumElements()}.
	 */
	public Object getValueAt( long time, int element );

	/**
	 * Returns the average value for the element over this interval's time range. For non-numeric
	 * types, such as Boolean, this method returns the exact value.
	 *
	 * @throws IndexOutOfBoundsException if element is &lt; 0 or &gt;= {@link #getNumElements()}.
	 */
	public Object getAverageValue( int element );

	/**
	 * Returns the maximum value for the element over this interval's time range. For non-numeric
	 * types, such as Boolean, this method returns the exact value.
	 *
	 * @throws IndexOutOfBoundsException if element is &lt; 0 or &gt;= {@link #getNumElements()}.
	 */
	public Object getMaximumValue( int element );

	/**
	 * Returns the minimum value for the element over this interval's time range. For non-numeric
	 * types, such as Boolean, this method returns the exact value.
	 *
	 * @throws IndexOutOfBoundsException if element is &lt; 0 or &gt;= {@link #getNumElements()}.
	 */
	public Object getMinimumValue( int element );

	/**
	 * Returns the result of the definite integral between this interval's start and end times. For
	 * non-numeric ElementTypes (such as Boolean), or for ElementCodecs that do not represent
	 * continuous ranges (such as SampledAlgorithm) this method always returns null. For numeric
	 * ElementTypes, this method always returns a Number object of the appropriate type.
	 *
	 * @throws IndexOutOfBoundsException if element is &lt; 0 or &gt;= {@link #getNumElements()}.
	 */
	public Number getIntegral( int element );
}
