/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.model.pool;

import org.gillius.realdb.BadFileException;
import org.gillius.realdb.io.BlockFile;
import org.gillius.realdb.model.CircularQueue;

import java.io.IOException;
import java.util.Collections;

import static java.util.Arrays.*;

/**
 * BlockPoolFactory contains methods to load and create {@link DefaultBlockPool}s and the associated
 * {@link CircularQueue} and {@link TransactionLog}.
 *
 * @author Jason Winnebeck
 */
public class BlockPoolFactory {
	private static final int maxItemSize = Collections.max( asList(
			FreeStateTxEntry.getWrittenSize(),
			AddIndexTxEntry.getWrittenSize(),
			RemoveIndexTxEntry.getWrittenSize(),
			RemoveFreeTxEntry.getWrittenSize() ) );

	public static int getMinSize( int blockSize, int maxOpenTx ) {
		int maxHeaderSize = FreeStateTxEntry.getWrittenSize();
		int totalSize = maxItemSize * maxOpenTx;

		return CircularQueue.getMinSize( totalSize, maxHeaderSize, maxItemSize, blockSize );
	}

	/**
	 * Creates and formats a new {@link DefaultBlockPool}.
	 *
	 * @param bf         non-null backing BlockFile
	 * @param startBlock starting block in BlockFile
	 * @param size       size in blocks (get from {@link #getMinSize}
	 * @param numDataBlocks number of data blocks in the pool
	 */
	public static DefaultBlockPool createNew( BlockFile bf, int startBlock, int size, int numDataBlocks )
			throws IOException {
		CircularQueue q = CircularQueue.createNew( bf, startBlock, size, maxItemSize );
		return makePool( numDataBlocks, q );
	}

	/**
	 * Loads an existing {@link DefaultBlockPool} from the given BlockFile.
	 *
	 * @param bf         non-null backing BlockFile
	 * @param startBlock starting block in BlockFile
	 * @param size       size in blocks (get from {@link #getMinSize}
	 * @param numDataBlocks number of data blocks in the pool
	 */
	public static DefaultBlockPool load( BlockFile bf, int startBlock, int size, int numDataBlocks )
		throws IOException, BadFileException {
		CircularQueue q = CircularQueue.load( bf, startBlock, size, maxItemSize );
		return makePool( numDataBlocks, q );
	}

	private static DefaultBlockPool makePool( int numDataBlocks, CircularQueue q )
			throws IOException {
		TransactionLog log = new TransactionLog( q );
		DefaultBlockPool ret = new DefaultBlockPool( log, numDataBlocks );
		return ret;
	}
}
