/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.model.pool;

import org.gillius.realdb.BadFileException;
import org.gillius.realdb.io.DataOutputWritable;
import org.gillius.realdb.model.BlockPool;
import org.gillius.realdb.model.BlockTransaction;

import java.io.DataOutput;
import java.io.IOException;
import java.util.ArrayDeque;
import java.util.HashSet;
import java.util.Queue;
import java.util.Set;
import java.util.TreeSet;

/**
 * This implementation always chooses reclaimable blocks before unused blocks, and always chooses
 * the lowest number block in {@link #getFreeBlock}.
 * <p/>
 * The pool must be opened before calling any other methods besides {@link #format()}.
 *
 * @author Jason Winnebeck
 */
public class DefaultBlockPool implements BlockPool {
	private final TransactionLog log;
	private final int numDataBlocks;

	private int nextUnusedBlock;
	private Queue<Integer> freeBlocks;

	/**
	 * Creates a DefaultBlockPool. The given TransactionLog will have its header writer set.
	 * <p/>
	 * The pool must be opened before calling any other methods besides {@link #format()}.
	 *
	 * @param log non-null TransactionLog
	 * @param numDataBlocks the number of data blocks in the pool
	 */
	public DefaultBlockPool( TransactionLog log, int numDataBlocks ) {
		this.log = log;
		this.numDataBlocks = numDataBlocks;

		log.setHeaderWriter( new DataOutputWritable() {
			@Override
			public void write( DataOutput output ) throws IOException {
				FreeStateTxEntry.write( output, nextUnusedBlock );
			}
		} );
	}

	@Override
	public void open() throws IOException, BadFileException {
		log.loadOpenTransactions( new NextUnusedBlockCalculator() );

		freeBlocks = new ArrayDeque<Integer>( new TreeSet<Integer>( log.getOpenTransactions() ) );
	}

	private class NextUnusedBlockCalculator implements TxEntryHandler {
		@Override
		public void handleEntry( TxEntry entry ) {
			if ( entry instanceof FreeStateTxEntry ) {
				FreeStateTxEntry freeEntry = (FreeStateTxEntry) entry;
				nextUnusedBlock = freeEntry.getNextFreeBlock();

			} else if ( entry instanceof RemoveFreeTxEntry ) {
				RemoveFreeTxEntry removeEntry = (RemoveFreeTxEntry) entry;
				if ( removeEntry.getBlock() >= nextUnusedBlock )
					nextUnusedBlock = removeEntry.getBlock() + 1;
			}
		}
	}

	@Override
	public TransactionLog getLog() {
		return log;
	}

	@Override
	public int getNumDataBlocks() {
		return numDataBlocks;
	}

	@Override
	public int getDataBlocksUsed() {
		return nextUnusedBlock;
	}

	public int getNextUnusedBlock() {
		return nextUnusedBlock;
	}

	/**
	 * Returns a set of blocks that were used at one point but no longer and will be reclaimed
	 * first when getting free blocks.
	 */
	public Set<Integer> getReclaimableBlocks() {
		return new HashSet<Integer>( freeBlocks );
	}

	public boolean isOpened() {
		return freeBlocks != null;
	}

	private void checkOpened() {
		if ( !isOpened() )
			throw new IllegalStateException( getClass().getSimpleName() + " not yet opened" );
	}

	@Override
	public boolean hasFreeBlocks() {
		checkOpened();
		return nextUnusedBlock < numDataBlocks || !freeBlocks.isEmpty();
	}

	@Override
	public BlockTransaction getFreeBlock() throws IOException {
		int block;
		if ( !freeBlocks.isEmpty() ) {
			block = freeBlocks.remove();
			return new BT( block );

		} else if ( nextUnusedBlock < numDataBlocks ) {
			block = nextUnusedBlock;
			log.startTransaction( new RemoveFreeTxEntry( block ) );
			BT ret = new BT( block );
			ret.addAction( log );
			++nextUnusedBlock;
			return ret;

		} else {
			throw new IllegalStateException( "No free blocks in the pool" );
		}
	}

	@Override
	public BlockTransaction removeBlock( int block, int sourceIndex, int seqNum,
	                                    long secondBlockStartTime ) throws IOException {
		log.startTransaction( new RemoveIndexTxEntry( block, sourceIndex, seqNum, secondBlockStartTime ) );
		BT ret = new BT( block );
		ret.addAction( log );
		return ret;
	}

	@Override
	public void addBlock( BlockTransaction tx, int lastBlock, int destination, int seqNum )
			throws IOException {
		log.endTransaction( new AddIndexTxEntry( tx.getBlockNum(), destination, seqNum, lastBlock ) );
		tx.addAction( log );
	}

	@Override
	public void flush() throws IOException {
		log.flush();
	}

	public void format() throws IOException {
		log.format();
		freeBlocks = new ArrayDeque<Integer>( 0 );
		nextUnusedBlock = 0;
	}

	private static class BT extends DefaultTransaction implements BlockTransaction {
		private final int blockNum;

		private BT( int blockNum ) {
			this.blockNum = blockNum;
		}

		@Override
		public int getBlockNum() {
			return blockNum;
		}
	}
}
