/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.model.pool;

import org.gillius.realdb.model.Transaction;

import java.io.Flushable;
import java.io.IOException;
import java.util.ArrayList;

/**
 * DefaultTransaction provides a default implementation that ensures that all actions are performed
 * in order, only once, and entirely on {@link #commit}.
 *
 * @author Jason Winnebeck
 */
public class DefaultTransaction implements Transaction {
	private final ArrayList<Flushable> actions;

	private int     nextToFlush = 0;
	private boolean committed   = false;

	public DefaultTransaction() {
		actions = new ArrayList<Flushable>();
	}

	public Flushable addAction( Flushable action ) {
		if ( committed )
			throw new IllegalStateException( "Transaction is already committed" );

		actions.add( action );
		return new Flusher( actions.size() - 1 ); //to flush everything before this action
	}

	public void commit() throws IOException {
		if ( !committed ) {
			flushTo( actions.size() );
			committed = true;
		}
	}

	private void flushTo( int idx ) throws IOException {
		//TODO: check if we are in flushTo already with a lower idx; this likely leads to infinite loop
		if ( idx > actions.size() )
			throw new IllegalArgumentException( "Not enough actions to flush to given index" );

		while ( nextToFlush < idx ) {
			actions.get( nextToFlush ).flush();
			++nextToFlush;
		}
	}

	private class Flusher implements Flushable {
		private final int idx;

		private Flusher( int idx ) {
			this.idx = idx;
		}

		public void flush() throws IOException {
			flushTo( idx );
		}
	}
}