/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.model.pool;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;

/**
 * FreeStateTxEntry represents a transaction log entry containing the current state of the
 * next free block counter in the block pool.
 *
 * @author Jason Winnebeck
 */
public class FreeStateTxEntry implements TxEntry {
	private final int nextFreeBlock;

	public FreeStateTxEntry( int nextFreeBlock ) {
		this.nextFreeBlock = nextFreeBlock;
	}

	/**
	 * Given a DataInput where the type byte was already read, reads a new FreeStateTxEntry.
	 */
	public static FreeStateTxEntry read( DataInput in ) throws IOException {
		return new FreeStateTxEntry( in.readInt() );
	}

	public static int getWrittenSize() {
		return 1 + 4;
	}

	@Override
	public TxEntryType getType() {
		return TxEntryType.FreeState;
	}

	public int getNextFreeBlock() {
		return nextFreeBlock;
	}

	@Override
	public void write( DataOutput output ) throws IOException {
		write( output, nextFreeBlock );
	}

	/**
	 * Performs the same operation as the {@link #write(DataOutput)} method, but without creating an
	 * actual object, to reduce memory allocations.
	 */
	public static void write( DataOutput output, int nextFreeBlock ) throws IOException {
		output.write( TxEntryType.FreeState.getExternalCode() );
		output.writeInt( nextFreeBlock );
	}

	@Override
	public String toString() {
		return "FreeStateTxEntry{" +
		       "nextFreeBlock=" + nextFreeBlock +
		       '}';
	}

	@Override
	public boolean equals( Object o ) {
		if ( this == o ) return true;
		if ( o == null || getClass() != o.getClass() ) return false;

		FreeStateTxEntry that = (FreeStateTxEntry) o;

		if ( nextFreeBlock != that.nextFreeBlock ) return false;

		return true;
	}

	@Override
	public int hashCode() {
		return nextFreeBlock;
	}
}
