/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.model.pool;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;

/**
 * IndexTxEntry holds an add or remove block transaction log entry.
 *
 * @author Jason Winnebeck
 */
public abstract class IndexTxEntry implements BlockTxEntry {
	private final int block;
	private final int index;
	private final int seqNum;

	/**
	 * See get methods for information about each property.
	 */
	public IndexTxEntry( int block, int index, int seqNum ) {
		this.block  = block;
		this.index  = index;
		this.seqNum = seqNum;
	}

	/**
	 * Given a DataInput where the type was already read, reads the state of this object.
	 */
	protected IndexTxEntry( DataInput in ) throws IOException {
		block = in.readInt();
		index = in.readInt();
		seqNum = in.readInt();
	}

	protected static int getWrittenSize() {
		return 1 + 4 + 4 + 4;
	}

	/**
	 * Returns the ID of the affected block. Also serves as a transaction ID.
	 */
	@Override
	public int getBlock() {
		return block;
	}

	/**
	 * Returns the ID of the affected index.
	 */
	public int getIndex() {
		return index;
	}

	/**
	 * Returns the sequence number of the currently-on-disk head (remove) or tail (add) index block.
	 * If the sequence number is higher, we can conclude the transaction was committed in the index.
	 */
	public int getSeqNum() {
		return seqNum;
	}

	@Override
	public void write( DataOutput output ) throws IOException {
		output.write( getType().getExternalCode() );
		output.writeInt( block );
		output.writeInt( index );
		output.writeInt( seqNum );
	}

	@Override
	public String toString() {
		return "IndexTxEntry{" +
		       "type=" + getType() +
		       ", block=" + block +
		       ", index=" + index +
		       ", seqNum=" + seqNum +
		       '}';
	}

	@Override
	public boolean equals( Object o ) {
		if ( this == o ) return true;
		if ( o == null || getClass() != o.getClass() ) return false;

		IndexTxEntry that = (IndexTxEntry) o;

		if ( block != that.block ) return false;
		if ( index != that.index ) return false;
		if ( seqNum != that.seqNum ) return false;

		return true;
	}

	@Override
	public int hashCode() {
		int result = block;
		result = 31 * result + index;
		result = 31 * result + seqNum;
		return result;
	}
}
