/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.model.pool;

import org.gillius.Util.CheckUtils;
import org.gillius.realdb.BadFileException;
import org.gillius.realdb.io.DataOutputWritable;
import org.gillius.realdb.io.Formattable;
import org.gillius.realdb.model.CircularQueue;
import org.gillius.realdb.util.DBIterator;

import java.io.Flushable;
import java.io.IOException;
import java.util.HashSet;
import java.util.Set;

/**
 * TransactionLog holds a rotating log of {@link TxEntry}. The given header writer is passed into
 * every {@link CircularQueue#add} method called by this class.
 *
 * @author Jason Winnebeck
 */
public class TransactionLog implements Flushable, Formattable {
	private final CircularQueue q;
	private final TxEntryReader reader;

	private DataOutputWritable headerWriter;

	//TODO: consider a Map to assert that no openTx is overwritten by queue rotation
	//Right now we assume this doesn't happen by sizing the queue to the largest number of possible
	//outstanding transactions.
	private final Set<Integer> openTx = new HashSet<Integer>();

	public TransactionLog( CircularQueue q, DataOutputWritable headerWriter ) {
		this.q = q;
		this.headerWriter = headerWriter;

		reader = new TxEntryReader();
	}

	public TransactionLog( CircularQueue q ) {
		this( q, null );
	}

	public DataOutputWritable getHeaderWriter() {
		return headerWriter;
	}

	public void setHeaderWriter( DataOutputWritable headerWriter ) {
		this.headerWriter = headerWriter;
	}

	/**
	 * (Re)scans the queue for any open transactions; should be run before starting or ending any
	 * transactions.
	 */
	public void loadOpenTransactions() throws IOException, BadFileException {
		loadOpenTransactions( null );
	}

	/**
	 * (Re)scans the queue for any open transactions; should be run before starting or ending any
	 * transactions.
	 *
	 * @param handler optional TxEntryHandler to pass all TxEntry objects found in the log.
	 */
	public void loadOpenTransactions( TxEntryHandler handler ) throws IOException, BadFileException {
		openTx.clear();

		DBIterator<TxEntry> iter = iterator();
		while ( iter.hasNext() ) {
			TxEntry entry = iter.next();
			if ( entry.getType() == TxEntryType.Remove ||
					 entry.getType() == TxEntryType.RemoveFree ) {
				int id = ((BlockTxEntry) entry).getBlock();
				if ( !openTx.add( id ) ) {
					throw new BadFileException( "Transaction " + id + " is open more than once" );
				}

			} else if ( entry.getType() == TxEntryType.Add ) {
				int id = ((IndexTxEntry) entry).getBlock();
				//We could see "double adds" on recovery if we reclaimed a block, or if the log was
				//truncated.
				openTx.remove( id );
			}
			if ( handler != null )
				handler.handleEntry( entry );
		}
	}

	public DBIterator<TxEntry> iterator() {
		return q.iterator( reader );
	}

	public CircularQueue getQueue() {
		return q;
	}

	public TxEntryReader getQueueReader() {
		return reader;
	}

	/**
	 * Returns all open transactions. For performance, this is an intrusive, non-cloned view and
	 * should not be modified or used after calling any other methods.
	 */
	public Set<Integer> getOpenTransactions() {
		return openTx;
	}

	public void startTransaction( RemoveFreeTxEntry tx ) throws IOException {
		startTransaction( tx, tx.getBlock() );
	}

	public void startTransaction( RemoveIndexTxEntry tx ) throws IOException {
		startTransaction( tx, tx.getBlock() );
	}

	private void startTransaction( TxEntry tx, int id ) throws IOException {
		CheckUtils.checkNull( tx, "tx" );
		if ( openTx.contains( id ) )
			throw new IllegalStateException( "Transaction " + id + " already started" );

		q.add( tx, headerWriter );
		openTx.add( id );
	}

	public void endTransaction( AddIndexTxEntry tx ) throws IOException {
		CheckUtils.checkNull( tx, "tx" );
		int id = tx.getBlock();
		if ( !openTx.contains( id ) )
			throw new IllegalStateException( "Transaction " + id + " not started" );

		q.add( tx, headerWriter );
		openTx.remove( id );
	}

	@Override
	public void flush() throws IOException {
		q.flush();
	}

	@Override
	public void format() throws IOException {
		q.format();
	}
}
