/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.tools;

import org.gillius.Util.CheckUtils;
import org.gillius.realdb.BadFileException;
import org.gillius.realdb.io.BlockFile;
import org.gillius.realdb.io.ByteArrayBlockFile;
import org.gillius.realdb.io.RandomAccessBlockFile;
import org.gillius.realdb.io.rdl.RDLFile;
import org.gillius.realdb.metadata.MetadataSection;
import org.gillius.realdb.model.Database;

import java.io.File;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

/**
 * DatabaseBuilder assists in the construction of a new Database and BlockFile from scratch.
 * <p/>
 * TODO: consider moving class to same package as Database due to the tight coupling
 *
 * @author Jason Winnebeck
 */
public class DatabaseBuilder {
	private int fileSize = -1;
	private int blockSize = -1;

	private MetadataSection metadata = null;

	/**
	 * Utility method to create an in-memory Database from the RDLFile given.
	 *
	 * @param rdl non-null RDLFile
	 *
	 * @return non-null Database
	 *
	 * @throws BadFileException if the RDLFile does not describe a valid database configuration
	 */
	public static Database createMemoryDatabase( RDLFile rdl ) throws BadFileException {
		DatabaseBuilder builder = new DatabaseBuilder( rdl );
		BlockFile file;
		file = new ByteArrayBlockFile( builder.getBlockSize(),
		                               builder.getFileSize() / builder.getBlockSize() );

		try {
			return builder.createDatabase( file );
		} catch ( IOException e ) {
			throw new AssertionError( "IOException not expected for in-memory DB" );
		}
	}

	/**
	 * Creates a default database builder with no parameters set.
	 */
	public DatabaseBuilder() {
	}

	/**
	 * Creates a DatabaseBuilder with all parameters set as from the RDLFile.
	 *
	 * @param rdl non-null RDLFile
	 *
	 * @throws BadFileException if not all parameters are defined properly
	 */
	public DatabaseBuilder( RDLFile rdl ) throws BadFileException {
		Map<String, String> options = rdl.getOptionsMap();
		Set<String> requiredOptions = new HashSet<String>(
				Arrays.asList( "fileSize", "blockSize", "dataBlockSize", "maxStreams" )
		);
		requiredOptions.removeAll( options.keySet() );
		if ( requiredOptions.size() > 0 )
			throw new BadFileException( "Options " + requiredOptions + " not specified in given RDL file" );

		fileSize  = parseRDLInt( "fileSize",  options );
		blockSize = parseRDLInt( "blockSize", options );

		int dataBlockSize = parseRDLInt( "dataBlockSize", options );
		int maxStreams    = parseRDLInt( "maxStreams",    options );

		metadata = new MetadataSection( dataBlockSize, rdl.getStreamInfos(), maxStreams );

		StringBuilder errors = new StringBuilder();
		if ( !validate( errors ) )
			throw new BadFileException( "RDL file settings are not valid: " + errors );
	}

	/**
	 * Returns the set file size of the database to be built, or -1 if it has not yet been set.
	 */
	public int getFileSize() {
		return fileSize;
	}

	/**
	 * Returns the block size of the database to be built, or -1 if it has not yet been set.
	 */
	public int getBlockSize() {
		return blockSize;
	}

	/**
	 * Returns the MetadataSection of the database to be built, or null if it has not yet been set.
	 */
	public MetadataSection getMetadata() {
		return metadata;
	}

	public void setFileSize( int fileSize ) {
		if ( fileSize < 0 )
		  throw new IllegalArgumentException( "Parameter fileSize must be positive or zero, but " + fileSize + " was given." );
		this.fileSize = fileSize;
	}

	public void setBlockSize( int blockSize ) {
		if ( blockSize < 0 )
		  throw new IllegalArgumentException( "Parameter blockSize must be positive or zero, but " + blockSize + " was given." );
		this.blockSize = blockSize;
	}

	public void setMetadata( MetadataSection metadata ) {
		CheckUtils.checkNull( metadata, "metadata" );
		this.metadata = metadata;
	}

	/**
	 * Validates that the DatabaseBuilder can build a working database. This means that the block
	 * size and file size are set appropriately so that there are enough blocks to store a minimal
	 * database, as well as any other criteria that may be required in the future.
	 *
	 * @param errors if not null, appends errors followed by a newline (\n), to the builder.
	 *
	 * @return true if the DatabaseBuilder can build a valid database
	 */
	public boolean validate( StringBuilder errors ) {
		boolean ret = true;
		if ( fileSize < 0 ) {
			ret = false;
			if ( errors != null )
				errors.append( "fileSize must be positive: " ).append( fileSize ).append( '\n' );
		}
		if ( blockSize < 0 ) {
			ret = false;
			if ( errors != null )
				errors.append( "blockSize must be positive: " ).append( blockSize ).append( '\n' );
		}
		if ( metadata == null ) {
			ret = false;
			if ( errors != null )
				errors.append( "metadata must be specified" ).append( '\n' );
		}
		//TODO: beef this up by checking the fileSize to the blockSize
//		int numBlocks = fileSize / blockSize;

		return ret;
	}

	/**
	 * Creates the BlockFile that would be used to create the Database in
	 * {@link #createDatabase(File)}. The use for this method is to be able to create a wrapped
	 * BlockFile and give that to {@link #createDatabase(BlockFile)}.
	 * <p/>
	 * By default, this method tries to overwrite the given file with a new one. If the file is
	 * detected to be a "special" file, no attempt is made to overwrite it, and instead uses the
	 * file that already exists. The goal of a "special" file is to try to match with a Linux device
	 * path (such as /dev/sdc). If Java terms, a path that is not a directory, or a file, but does
	 * exist is considered "special."
	 * <p/>
	 * The file size set in the builder is the size used for the database, not the "special" file's
	 * size. Therefore the DB size must be less than the "special" file size.
	 *
	 * @param f non-null File location
	 *
	 * @return non-null BlockFile
	 */
	public BlockFile createBlockFile( File f ) throws IOException {
		RandomAccessFile raf = new RandomAccessFile( f, "rwd" );
		if ( !f.exists() || f.isDirectory() || f.isFile() ) {
			raf.setLength( fileSize );
		}

		return new RandomAccessBlockFile( raf, blockSize, fileSize / blockSize );
	}

	/**
	 * Creates a Database by constructing a new block file at the given location.
	 *
	 * @param f non-null File location
	 *
	 * @return a Database opened to this file
	 */
	public Database createDatabase( File f ) throws IOException {
		CheckUtils.checkNull( f, "f" );
		ensureValid();

		return createDatabase( createBlockFile( f ) );
	}

	/**
	 * Creates a Database by writing one to the given BlockFile. The BlockFile's block size must
	 * match the blockSize set for this builder. The fileSize setting is ignored (for now).
	 *
	 * @param bf non-null BlockFile
	 *
	 * @return a Database that uses this BlockFile
	 */
	public Database createDatabase( BlockFile bf ) throws IOException {
		CheckUtils.checkNull( bf, "bf" );
		if ( blockSize != bf.getBlockSize() ) {
			throw new IllegalArgumentException( "Given BlockFile size is " + bf.getBlockSize() +
			                                    ", but expected " + blockSize );
		}

		return Database.createNew( bf, metadata );
	}

	/**
	 * Parses a value from the options Map. The given value should already exist.
	 */
	private static int parseRDLInt( String name, Map<String, String> options )
			throws BadFileException {
		String value = options.get( name );
		try {
			return Integer.parseInt( value );
		} catch ( NumberFormatException e ) {
			throw new BadFileException( "Option " + name + " is not a valid number: " + value );
		}
	}

	/**
	 * Ensures that this builder is valid, and if it isn't, throws IllegalStateException.
	 */
	private void ensureValid() throws IllegalStateException {
		StringBuilder errors = new StringBuilder();
		if ( !validate( errors ) )
			throw new IllegalStateException( "DatabaseBuilder is not valid: " + errors );
	}
}
