/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.util;

/**
 * ConsoleProgressMeter prints progress out to the console as a single line. A sequence of dots and
 * percents are printed, such as:
 * <p/>
 * .....25%.....50%.....75%.....100%
 *
 * @author Jason Winnebeck
 */
public class ConsoleProgressMeter {
	private final float maxValue;
	private final int dotStep;
	private final int labelStep;

	private int nextStep = 0;

	/**
	 * Constructs a new ConsoleProgressMeter.
	 *
	 * @param maxValue  Value for 100%
	 * @param dotStep   each dot represents this percent (0-100) of progress
	 * @param labelStep labels will be printed every 'labelStep' percent; must be a multiple of
	 *                  dotStep.
	 */
	public ConsoleProgressMeter( float maxValue, int dotStep, int labelStep ) {
		if ( labelStep % dotStep != 0 )
			throw new IllegalArgumentException( "labelStep must be a multiple of dotStep" );

		this.maxValue = maxValue;
		this.dotStep = dotStep;
		this.labelStep = labelStep;
	}

	/**
	 * Updates the value, which may print dots and labels.
	 */
	public void updateValue( float val ) {
		int percent = (int) ( val / maxValue * 100.0f );
		while ( percent >= nextStep ) {
			System.out.print( '.' );
			if ( nextStep != 0 && nextStep % labelStep == 0 )
				System.out.print( nextStep + "%" );
			nextStep += dotStep;
		}
	}

	/**
	 * Ensures that progress up to and including 100% is printed, then calls System.out.println().
	 */
	public void finish() {
		updateValue( maxValue ); //I think that is guaranteed to calculate to 100%
		System.out.println();
	}

	/**
	 * Resets the class as if it was newly constructed.
	 */
	public void reset() {
		nextStep = 0;
	}
}
