/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.integration;

import org.gillius.realdb.BadFileException;
import org.gillius.realdb.io.BlockFile;
import org.gillius.realdb.io.ByteArrayBlockFile;
import org.gillius.realdb.io.FailingBlockFile;
import org.gillius.realdb.io.ProfilingBlockFile;
import org.gillius.realdb.io.rdl.RDLFile;
import org.gillius.realdb.model.DataStream;
import org.gillius.realdb.model.Database;
import org.gillius.realdb.model.data.GenericRecord;
import org.gillius.realdb.model.data.Record;
import org.gillius.realdb.tools.DatabaseBuilder;
import org.gillius.realdb.util.DBIterator;
import org.junit.*;

import java.io.IOException;
import java.util.Arrays;
import java.util.Random;

import static org.hamcrest.CoreMatchers.*;
import static org.junit.Assert.*;
import static org.junit.Assume.*;

/**
 * TestCorruptingWriteDualStream is a version of {@link TestCorruptingWriteSingleStream} that
 * tests with 2 streams.
 * <p/>
 * TODO: refactor code to share with single and multi versions
 *
 * @author Jason Winnebeck
 */
public class TestCorruptingWriteDualStream {
	private static final String RDL_CONTENT =
			"SET blockSize     = 256 " +
			"SET fileSize      = 204800 " +
			"SET maxStreams    = 3 " +
			"SET dataBlockSize = 2 " +
			"CREATE STREAM Test0 WITH ID 0 {value sint32 WITH CODEC SampledAlgorithm}" +
			"CREATE STREAM Test1 WITH ID 1 {value sint32 WITH CODEC SampledAlgorithm}" ;

	private static final int FLUSH_INTERVAL =   500; //flush every 500 records
	private static final int ITERATIONS     = 15000;
	private static final int NUM_STREAMS    =     2;
	/**
	 * An array the size of NUM_STREAMS. If non-zero, that stream skips writing every ith records.
	 */
	private static final int[] SKIPS = new int[] { 0, 2 };

	private RDLFile rdl;

	@Before
	public void setUp() throws Exception {
		rdl = RDLFile.load( RDL_CONTENT );
	}

	@Test
	public void doTest() throws Exception {
		assumeThat( System.getProperty( "realdb.testtype.integration" ), equalTo( "true" ) );

		int blockSize = Integer.parseInt( rdl.getOptionsMap().get( "blockSize" ) );
		int fileSize  = Integer.parseInt( rdl.getOptionsMap().get( "fileSize" ) );
		assertEquals( 0, fileSize % blockSize );
		ByteArrayBlockFile rawFile = new ByteArrayBlockFile( blockSize, fileSize / blockSize );
		ProfilingBlockFile file    = new ProfilingBlockFile( rawFile );

		RunData run = writeData( file );
		assertNull( run.lastError );
		for ( int stream = 0; stream < NUM_STREAMS; ++stream ) {
			int expected = ( SKIPS[stream] != 0 && ITERATIONS % SKIPS[stream] == 0 ) ? ITERATIONS - 1 :
			               ITERATIONS;
			assertEquals( "stream=" + stream, expected, run.lastFlushed[stream] );
			assertEquals( "stream=" + stream, expected, run.lastWritten[stream] );
		}
		assertNull( run.lastError );
		int totalOps = file.getReadCalls() + file.getWriteCalls() + 1; //+1 for close call
		System.out.println( "Total operations performed: " + totalOps );
		verifyData( rawFile, run, true );
		System.out.println( "Now performing failure tests on all operations" );

		boolean printStats = true;
		//Now, hold on tight! We're about to run this test a few thousand times.
		for ( int i = 1; i <= totalOps; ++i ) {
			if ( i % 250 == 0 ) {
				System.out.println( "Beginning iteration " + i );
				printStats = true;
			}
			try {
				rawFile = new ByteArrayBlockFile( blockSize, fileSize / blockSize );
				FailingBlockFile blockFile = new FailingBlockFile( rawFile, new Random( 500 ) );
				blockFile.setFailureCount( i );
				run = writeData( blockFile );
				//TODO: see if we can check anything in the lastFlushed == -1 && lastWritten >= 0 case
				if ( !Arrays.equals( new int[NUM_STREAMS], run.lastFlushed ) ) {
					//If any records were actually flushed (any element != 0)
					verifyData( rawFile, run, printStats );
					printStats = false;
				}
			} catch ( Throwable t ) {
				if ( run.lastError != null )
					run.lastError.printStackTrace();
				else
					System.err.println( "No write exception given!" );
				throw new Exception( "Failed on iteration " + i, t );
			}
		}
	}

	private RunData writeData( BlockFile file ) {
		RunData ret = new RunData();
		try {
			Database db = new DatabaseBuilder( rdl ).createDatabase( file );

			assertEquals( NUM_STREAMS, db.getStreams().size() );
			DataStream[] streams = new DataStream[ NUM_STREAMS ];
			for ( int i = 0; i < NUM_STREAMS; ++i ) {
				streams[i] = db.getStreamsMap().get( i );
			}

			//This starves the other stream and deletes all of its blocks.
			//Perhaps we should implement a "skip" instead of "multiplier"
			for ( int i = 1; i <= ITERATIONS; ++i ) {
				//Since we are using sampled codec, this should always return true
				for ( int j = 0; j < streams.length; ++j ) {
					DataStream stream = streams[j];
					//If "i" isn't a skipped iteration, write a record
					if ( SKIPS[j] == 0 || i % SKIPS[j] != 0 ) {
						assertTrue( stream.writeRecord( makeRecord( i ) ) );
						ret.lastWritten[j] = i;
						if ( i % FLUSH_INTERVAL == 0 ) {
							db.flush();
							ret.lastFlushed[j] = ret.lastWritten[j];
						}
					}
				}
			}

			db.close();
			System.arraycopy( ret.lastWritten, 0, ret.lastFlushed, 0, ret.lastWritten.length );

		} catch ( IOException e ) {
			ret.lastError = e;

		} catch ( BadFileException e ) {
			ret.lastError = e;
		}

		return ret;
	}

	private static void verifyData( BlockFile file, RunData run, boolean printStats )
			throws IOException, BadFileException {
		Database db = Database.load( file );
		assertEquals( NUM_STREAMS, db.getStreams().size() );

		for ( int i = 0; i < NUM_STREAMS; ++i ) {
			DataStream stream = db.getStreamsMap().get( i );
			verifyStream( stream, run.lastFlushed[i], run.lastWritten[i], SKIPS[i], printStats );
		}
	}

	/**
	 * Test that the database contains some subset of records beyond 0 and until lastFlushed, and
	 * possibly as far as lastWritten.
	 */
	private static void verifyStream( DataStream stream, int lastFlushed, int lastWritten, int skip,
	                                  boolean printStats ) throws IOException, BadFileException {
		DBIterator<Record> recordsIter = stream.readRecords();
		assertTrue( recordsIter.hasNext() );
		int firstRecord = (int) recordsIter.next().getTime();
		int lastRecord  = firstRecord;
		int numRecords  = 0;
		assertTrue( firstRecord > 0 );

		int i;
		for ( i = firstRecord + 1; i <= lastFlushed; ++i ) {
			if ( skip == 0 || i % skip != 0 ) {
				assertTrue( recordsIter.hasNext() );
				assertEquals( makeRecord( i ), recordsIter.next() );
				lastRecord = i;
				++numRecords;
			}
		}
		for ( ; i <= lastWritten; ++i ) {
			if ( ( skip == 0 || i % skip != 0 ) && recordsIter.hasNext() ) {
				assertEquals( makeRecord( i ), recordsIter.next() );
				lastRecord = i;
				++numRecords;
			}
		}
		assertFalse( recordsIter.hasNext() );

		if ( printStats ) {
			System.out.format( "Stream %s has %d records from %d to %d%n",
			                   stream, numRecords, firstRecord, lastRecord );
		}
	}

	private static GenericRecord makeRecord( int i ) {
		return new GenericRecord( i, i );
	}

	private static class RunData {
		public Exception lastError   = null;
		public int[]     lastFlushed = new int[ NUM_STREAMS ];
		public int[]     lastWritten = new int[ NUM_STREAMS ];
	}
}