/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.integration;

import org.gillius.realdb.BadFileException;
import org.gillius.realdb.io.BlockFile;
import org.gillius.realdb.io.ByteArrayBlockFile;
import org.gillius.realdb.io.FailingBlockFile;
import org.gillius.realdb.io.ProfilingBlockFile;
import org.gillius.realdb.io.rdl.RDLFile;
import org.gillius.realdb.model.DataStream;
import org.gillius.realdb.model.Database;
import org.gillius.realdb.model.data.GenericRecord;
import org.gillius.realdb.model.data.Record;
import org.gillius.realdb.tools.DatabaseBuilder;
import org.gillius.realdb.util.DBIterator;
import org.junit.*;

import java.io.IOException;
import java.util.Random;

import static org.hamcrest.CoreMatchers.*;
import static org.junit.Assert.*;
import static org.junit.Assume.*;

/**
 * TestCorruptingWriteSingleStream is an integration test to test that the guarantees made by
 * RealDB hold even if the file fails. This test is done as "black-box" as possible using the
 * APIs intended for the user. Of course, the BlockFile wrappers are used to exercise the test and
 * to force faults.
 * <p/>
 * In general, RealDB guarantees that all data up to the last flush is preserved. It is very
 * possible that even more data is preserved, but this test does not check for that. It only checks
 * the bare minimum assumptions.
 * <p/>
 * The purpose of this integration test is to provide an evaluation to pass milestone 3.
 * <p/>
 * TODO: refactor code to share with dual and multi versions
 *
 * @author Jason Winnebeck
 */
public class TestCorruptingWriteSingleStream {
	private static final String RDL_CONTENT =
			"SET blockSize     = 2048 " +
			"SET fileSize      = 204800 " +
			"SET maxStreams    = 3 " +
			"SET dataBlockSize = 2 " +
			"CREATE STREAM Test WITH ID 1 {value sint32 WITH CODEC SampledAlgorithm}";

	private static final int FLUSH_INTERVAL =    500; //flush every 500 records
	private static final int MAX_RECORD     = 200000;

	private RDLFile rdl;

	@Before
	public void setUp() throws Exception {
		rdl = RDLFile.load( RDL_CONTENT );
	}

	@Test
	public void doTest() throws Exception {
		assumeThat( System.getProperty( "realdb.testtype.integration" ), equalTo( "true" ) );

		ByteArrayBlockFile rawFile = new ByteArrayBlockFile( 2048, 100 );
		ProfilingBlockFile file    = new ProfilingBlockFile( rawFile );

		RunData run = writeData( file );
		assertNull( run.lastError );
		assertEquals( MAX_RECORD, run.lastFlushed );
		assertEquals( MAX_RECORD, run.lastWritten );
		assertNull( run.lastError );
		int totalOps = file.getReadCalls() + file.getWriteCalls() + 1; //+1 for close call
		System.out.println( "Total operations performed: " + totalOps );
		verifyData( rawFile, run );

		//Now, hold on tight! We're about to run this test a few thousand times.
		for ( int i = 1; i <= totalOps; ++i ) {
			if ( i % 500 == 0 )
				System.out.println( "Beginning iteration " + i );
			try {
				rawFile = new ByteArrayBlockFile( 2048, 100 );
				FailingBlockFile blockFile = new FailingBlockFile( rawFile, new Random( 500 ) );
				blockFile.setFailureCount( i );
				run = writeData( blockFile );
				//TODO: see if we can check anything in the lastFlushed == -1 && lastWritten >= 0 case
				if ( run.lastFlushed >= 0 ) { //If any records were actually flushed
					verifyData( rawFile, run );
				}
			} catch ( Throwable t ) {
				if ( run.lastError != null )
					run.lastError.printStackTrace();
				else
					System.err.println( "No write exception given!" );
				throw new Exception( "Failed on iteration " + i, t );
			}
		}
	}

	private RunData writeData( BlockFile file ) {
		RunData ret = new RunData();
		try {
			Database db = new DatabaseBuilder( rdl ).createDatabase( file );

			assertEquals( 1, db.getStreams().size() );
			DataStream stream = db.getStreams().iterator().next();

			for ( int i = 1; i <= MAX_RECORD; ++i ) {
				//Since we are using sampled codec, this should always return true
				assertTrue( stream.writeRecord( makeRecord( i ) ) );
				ret.lastWritten = i;
				if ( i % FLUSH_INTERVAL == 0 ) {
					db.flush();
					ret.lastFlushed = i;
				}
			}

			db.close();
			ret.lastFlushed = MAX_RECORD;

		} catch ( IOException e ) {
			ret.lastError = e;

		} catch ( BadFileException e ) {
			ret.lastError = e;
		}

		return ret;
	}

	private static void verifyData( BlockFile file, RunData run )
			throws IOException, BadFileException {
		Database db = Database.load( file );
		assertEquals( 1, db.getStreams().size() );
		DataStream stream = db.getStreams().iterator().next();

		//Test that the database contains some subset of records beyond 0 and until lastFlushed, and
		//possibly as far as lastWritten.
		DBIterator<Record> recordsIter = stream.readRecords();
		assertTrue( recordsIter.hasNext() );
		int firstRecord = (int) recordsIter.next().getTime();
		assertTrue( firstRecord > 0 );

		int i;
		for ( i = firstRecord + 1; i <= run.lastFlushed; ++i ) {
			assertTrue( recordsIter.hasNext() );
			assertEquals( makeRecord( i ), recordsIter.next() );
		}
		for ( ; i <= run.lastWritten; ++i ) {
			if ( recordsIter.hasNext() )
				assertEquals( makeRecord( i ), recordsIter.next() );
		}
		assertFalse( recordsIter.hasNext() );
	}

	private static GenericRecord makeRecord( int i ) {
		return new GenericRecord( i, i );
	}

	private static class RunData {
		public Exception lastError   = null;
		public int       lastFlushed = -1;
		public int       lastWritten = -1;
	}
}
