/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.integration;

import org.gillius.realdb.BadFileException;
import org.gillius.realdb.io.rdl.RDLFile;
import org.gillius.realdb.model.DataStream;
import org.gillius.realdb.model.Database;
import org.gillius.realdb.model.data.GenericRecord;
import org.gillius.realdb.model.data.Record;
import org.gillius.realdb.model.data.StreamInterval;
import org.gillius.realdb.tools.DatabaseBuilder;
import org.gillius.realdb.util.DBIterator;
import org.junit.*;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import static java.util.Arrays.*;
import static org.junit.Assert.*;

/**
 * TestReadingIntervals tests writing some data to a DataStream then reading intervals out of it.
 *
 * @author Jason Winnebeck
 */
public class TestReadingIntervals {
	private static final String RDL =
			"SET blockSize     = 512\n" +
			"SET fileSize      = 512000\n" +
			"SET maxStreams    = 3\n" +
			"SET dataBlockSize = 2\n" +
			"CREATE STREAM Test1 WITH ID 1 {value double WITH CODEC StepAlgorithm}";

	private DataStream stream;
	private List<? extends Record> result;

	@Before
	public void setUp() throws Exception {
		Database db = DatabaseBuilder.createMemoryDatabase( RDLFile.load( RDL ) );

		stream = db.getStreamsMap().get( 1 );

		List<? extends Record> rawData = asList(
				new GenericRecord( 10, 10.0 ),
				new GenericRecord( 20, 12.0 ),
				new GenericRecord( 25, 12.0 ),
				new GenericRecord( 30, 13.0 ),
				new GenericRecord( 35, 13.0 ),
				new GenericRecord( 40, 18.0 ) );

		result = asList(
				rawData.get( 0 ),
				rawData.get( 1 ),
				//2 should be removed
				rawData.get( 3 ),
				//4 should be removed
				rawData.get( 5 ) );

		for ( Record record : rawData ) {
			assertEquals( record.toString() + " not handled as expected",
			              result.contains( record ), stream.writeRecord( record ) );
		}
	}

	@Test
	public void testExpectedAreWritten() throws Exception {
		//Read back all of the raw records and make sure they match the expected result list.
		List<Record> actual = new ArrayList<Record>();
		DBIterator<Record> iter = stream.readRecords();
		while ( iter.hasNext() ) {
			actual.add( iter.next() );
		}
		assertEquals( result, actual );
	}

	@Test
	public void testReadAll() throws Exception {
		//Read back all of the intervals
		DBIterator<StreamInterval> iIter = stream.readIntervals();
		checkAllRead( iIter );
	}

	@Test
	public void testReadAllRange() throws Exception {
		DBIterator<StreamInterval> iIter;//Read back a range of the intervals
		iIter = stream.readIntervals( 0, Long.MAX_VALUE );
		checkAllRead( iIter );
	}

	@Test
	public void testReadAllSmallestRange() throws Exception {
		DBIterator<StreamInterval> iIter;
		iIter = stream.readIntervals( 10, 40 );
		checkAllRead( iIter );
	}

	@Test
	public void testReadSingleRange() throws IOException, BadFileException {
		DBIterator<StreamInterval> iIter;
		iIter = stream.readIntervals( 20, 20 );
		assertTrue( iIter.hasNext() );
		checkStreamInterval( iIter.next(), 12.0, 20, 30 );
		assertFalse( iIter.hasNext() );
	}

	@Test
	public void testReadMiddleRange() throws IOException, BadFileException {
		DBIterator<StreamInterval> iIter;
		iIter = stream.readIntervals( 27, 40 );
		assertTrue( iIter.hasNext() );
		checkStreamInterval( iIter.next(), 12.0, 20, 30 );
		assertTrue( iIter.hasNext() );
		checkStreamInterval( iIter.next(), 13.0, 30, 40 );
		assertTrue( iIter.hasNext() );
		checkStreamInterval( iIter.next(), 18.0, 40, 40 );
		assertFalse( iIter.hasNext() );
	}

	private static void checkAllRead( DBIterator<StreamInterval> iIter ) throws Exception {
		assertTrue( iIter.hasNext() );
		checkStreamInterval( iIter.next(), 10.0, 10, 20 );
		assertTrue( iIter.hasNext() );
		checkStreamInterval( iIter.next(), 12.0, 20, 30 );
		assertTrue( iIter.hasNext() );
		checkStreamInterval( iIter.next(), 13.0, 30, 40 );
		assertTrue( iIter.hasNext() );
		checkStreamInterval( iIter.next(), 18.0, 40, 40 );

		assertFalse( iIter.hasNext() );
	}

	private static void checkStreamInterval( StreamInterval i, double val, long start, long end ) {
		long timespan = end - start;

		assertEquals( 1, i.getNumElements() );
		assertEquals( val, i.getAverageValue( 0 ) );
		assertEquals( val, i.getMaximumValue( 0 ) );
		assertEquals( val, i.getMinimumValue( 0 ) );
		assertEquals( start, i.getStartTime() );
		assertEquals( end, i.getEndTime() );
		assertEquals( timespan, i.getTimespan() );
		assertEquals( timespan * val, (Double) i.getIntegral( 0 ), 0.00001 );
	}
}
