/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.integration;

import org.gillius.realdb.BadFileException;
import org.gillius.realdb.io.BlockFile;
import org.gillius.realdb.io.ByteArrayBlockFile;
import org.gillius.realdb.io.FailingBlockFile;
import org.gillius.realdb.io.rdl.RDLFile;
import org.gillius.realdb.model.DataStream;
import org.gillius.realdb.model.Database;
import org.gillius.realdb.model.data.GenericRecord;
import org.gillius.realdb.model.data.Record;
import org.gillius.realdb.tools.DatabaseBuilder;
import org.gillius.realdb.util.DBIterator;
import org.junit.*;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Random;

import static org.hamcrest.CoreMatchers.*;
import static org.junit.Assert.*;
import static org.junit.Assume.*;

/**
 * TestResumingWriteDualStream checks that we can resume writing after failure. This test was
 * created in response to the RealDBDemo not being able to resume from a failure sometimes.
 *
 * @author Jason Winnebeck
 */
public class TestResumingWriteDualStream {
	private static final String RDL_CONTENT =
			"SET blockSize     = 2048 " +
			"SET fileSize      = 204800 " +
			"SET maxStreams    = 3 " +
			"SET dataBlockSize = 2 " +
			"CREATE STREAM Test1 WITH ID 1 {value float WITH CODEC SampledAlgorithm}" +
			"CREATE STREAM Test2 WITH ID 2 {value float WITH CODEC SampledAlgorithm}";

	/**
	 * Runs the actual test failing at 1 IO operation later each time until the entire test proceeds
	 * with no IO failure, to test every possible failure point.
	 */
	@Test
	public void runAllTests() throws Exception {
		assumeThat( System.getProperty( "realdb.testtype.integration" ), equalTo( "true" ) );

		int failureCount = 1;
		try {
			while ( doTest( failureCount ) ) {
				++failureCount;
			}
		} catch ( Exception e ) {
			System.err.println( "Failure when failureCount = " + failureCount );
			throw e;
		}
	}

	private static boolean doTest( int failureCount ) throws Exception {
		RDLFile rdl = RDLFile.load( RDL_CONTENT );
		DatabaseBuilder builder = new DatabaseBuilder( rdl );
		BlockFile file = new ByteArrayBlockFile( builder.getBlockSize(),
		                                         builder.getFileSize() / builder.getBlockSize() );
		FailingBlockFile fbf = new FailingBlockFile( file, new Random() );
		Database db = builder.createDatabase( fbf );

		fbf.setFailureCount( failureCount );

		int numStreams = rdl.getStreamInfos().size();
		assertEquals( numStreams, db.getStreams().size() );
		List<DataStream> streams = new ArrayList<DataStream>( db.getStreams() );

		boolean failed = false;
		final int START_RECORD =     1;
		final int MAX_RECORD   = 50000;
		int[] curr = new int[ numStreams ];
		Arrays.fill( curr, START_RECORD );
		int minStreamIdx = findMin( curr );
		while ( curr[ minStreamIdx ] < MAX_RECORD ) {
			DataStream stream = streams.get( minStreamIdx );
			try {
				//Since we are using sampled codec, this should always return true
				assertTrue( stream.writeRecord( makeRecord( curr[ minStreamIdx ] ) ) );
				++curr[ minStreamIdx ];
				//If we just wrote the last record, close the DB. This can fail so we need to put it in
				//the loop.
				if ( curr[ findMin( curr ) ] == MAX_RECORD )
					db.close();

			} catch ( IOException e ) {
				//Reload the database; this should happen at most once
				assertFalse( failed );
				failed = true;
//				System.err.println( "Failed at " + Arrays.toString( curr ) );

				db = Database.load( fbf );
				streams = new ArrayList<DataStream>( db.getStreams() );
				for ( int i = 0; i < curr.length; ++i ) {
					long lastRecord = streams.get( i ).getEndTime();
					assertTrue( lastRecord <= curr[i] );
					curr[i] = ( lastRecord < 0 ) ? START_RECORD : (int) lastRecord + 1;
				}
//				System.err.println( "Resuming at " + Arrays.toString( curr ) );
			}

			minStreamIdx = findMin( curr );
		}

		//Disable failures for the read phase
		fbf.setFailureCount( 0 );
		//Reload the database and verify the data
		verifyData( file, MAX_RECORD, numStreams );

		return failed;
	}

	private static int findMin( int[] arr ) {
		int min = Integer.MAX_VALUE;
		int minIdx = -1;
		for ( int i = 0; i < arr.length; ++i ) {
			if ( arr[i] < min ) {
				min = arr[i];
				minIdx = i;
			}
		}

		return minIdx;
	}

	/**
	 * Reloads the database from the given BlockFile and verifies that there are records written up
	 * to and including lastRecord.
	 */
	private static void verifyData( BlockFile file, int lastRecord, int numStreams )
			throws IOException, BadFileException {
		Database db = Database.load( file );
		assertEquals( numStreams, db.getStreams().size() );
		List<DataStream> streams = new ArrayList<DataStream>( db.getStreams() );

		//Test that the database contains some subset of records beyond 0 and until lastRecord
		for ( DataStream stream : streams ) {
			DBIterator<Record> recordsIter = stream.readRecords();
			int firstRecord = (int) recordsIter.next().getTime();
			assertTrue( firstRecord > 0 );
			for ( int i = firstRecord + 1; i < lastRecord; ++i ) {
				assertTrue( stream.getInfo().getName(), recordsIter.hasNext() );
				assertEquals( stream.getInfo().getName(), makeRecord( i ), recordsIter.next() );
			}
			assertFalse( stream.getInfo().getName(), recordsIter.hasNext() );

			//We can't 100% guarantee the length of the stream, so we have to human-check it, kinda
//			System.out.println( "Read records " + firstRecord + " to " + lastRecord );
		}
	}

	private static GenericRecord makeRecord( int i ) {
		return new GenericRecord( i, i * 2.5f );
	}
}