/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.integration;

import org.gillius.realdb.BadFileException;
import org.gillius.realdb.io.BlockFile;
import org.gillius.realdb.io.ByteArrayBlockFile;
import org.gillius.realdb.io.FailingBlockFile;
import org.gillius.realdb.io.rdl.RDLFile;
import org.gillius.realdb.model.DataStream;
import org.gillius.realdb.model.Database;
import org.gillius.realdb.model.data.GenericRecord;
import org.gillius.realdb.model.data.Record;
import org.gillius.realdb.tools.DatabaseBuilder;
import org.gillius.realdb.util.DBIterator;
import org.junit.*;

import java.io.IOException;
import java.util.Random;

import static org.hamcrest.CoreMatchers.*;
import static org.junit.Assert.*;
import static org.junit.Assume.*;

/**
 * TestResumingWriteSingleStream checks that we can resume writing after failure. This test was
 * created in response to a bug in DefaultBlockPool that was never uncovered because the other tests
 * never ensured that you can resume from a corrupted database -- only that the data contained
 * within was correct.
 *
 * @author Jason Winnebeck
 */
public class TestResumingWriteSingleStream {
	private static final String RDL_CONTENT =
			"SET blockSize     = 2048 " +
			"SET fileSize      = 204800 " +
			"SET maxStreams    = 3 " +
			"SET dataBlockSize = 2 " +
			"CREATE STREAM Test WITH ID 1 {value float WITH CODEC SampledAlgorithm}";

	/**
	 * Runs the actual test failing at 1 IO operation later each time until the entire test proceeds
	 * with no IO failure, to test every possible failure point.
	 */
	@Test
	public void runAllTests() throws Exception {
		assumeThat( System.getProperty( "realdb.testtype.integration" ), equalTo( "true" ) );

		int failureCount = 1;
		try {
			while ( doTest( failureCount ) ) {
				++failureCount;
			}
		} catch ( Exception e ) {
			System.err.println( "Failure when failureCount = " + failureCount );
			throw e;
		}
	}

	private static boolean doTest( int failureCount ) throws Exception {
		RDLFile rdl = RDLFile.load( RDL_CONTENT );
		DatabaseBuilder builder = new DatabaseBuilder( rdl );
		BlockFile file = new ByteArrayBlockFile( builder.getBlockSize(),
		                                         builder.getFileSize() / builder.getBlockSize() );
		FailingBlockFile fbf = new FailingBlockFile( file, new Random() );
		Database db = builder.createDatabase( fbf );

		fbf.setFailureCount( failureCount );

		assertEquals( 1, db.getStreams().size() );
		DataStream stream = db.getStreams().iterator().next();

		boolean failed = false;
		final int MAX_RECORD = 50000;
		int curr = 0;
		while ( curr < MAX_RECORD ) {
			try {
				//Since we are using sampled codec, this should always return true
				assertTrue( stream.writeRecord( makeRecord( curr ) ) );
				++curr;
				//If we just wrote the last record, close the DB. This can fail so we need to put it in
				//the loop.
				if ( curr == MAX_RECORD )
					db.close();

			} catch ( IOException e ) {
				//Reload the database; this should happen at most once
				assertFalse( failed );
				failed = true;
//				System.err.println( "Failed at " + curr );

				db = Database.load( fbf );
				stream = db.getStreams().iterator().next();
				int lastRecord = (int) stream.getEndTime();
				assertTrue( lastRecord <= curr );
				curr = ( lastRecord < 0 ) ? 0 : lastRecord + 1;
//				System.err.println( "Resuming at " + curr );
			}
		}

		//Disable failures for the read phase
		fbf.setFailureCount( 0 );
		//Reload the database and verify the data
		verifyData( file, MAX_RECORD );

		return failed;
	}

	/**
	 * Reloads the database from the given BlockFile and verifies that there are records written up
	 * to and including lastRecord.
	 */
	private static void verifyData( BlockFile file, int lastRecord )
			throws IOException, BadFileException {
		Database db = Database.load( file );
		assertEquals( 1, db.getStreams().size() );
		DataStream stream = db.getStreams().iterator().next();

		//Test that the database contains some subset of records beyond 0 and until lastRecord
		DBIterator<Record> recordsIter = stream.readRecords();
		int firstRecord = (int) recordsIter.next().getTime();
		assertTrue( firstRecord > 0 );
		for ( int i = firstRecord + 1; i < lastRecord; ++i ) {
			assertTrue( recordsIter.hasNext() );
			assertEquals( makeRecord( i ), recordsIter.next() );
		}
		assertFalse( recordsIter.hasNext() );

		//We can't 100% guarantee the length of the stream, so we have to human-check it, kinda
//		System.out.println( "Read records " + firstRecord + " to " + lastRecord );
	}

	private static GenericRecord makeRecord( int i ) {
		return new GenericRecord( i, i * 2.5f );
	}
}
