/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.integration;

import org.gillius.realdb.BadFileException;
import org.gillius.realdb.io.BlockFile;
import org.gillius.realdb.io.ByteArrayBlockFile;
import org.gillius.realdb.io.rdl.RDLFile;
import org.gillius.realdb.model.DataStream;
import org.gillius.realdb.model.Database;
import org.gillius.realdb.model.data.GenericRecord;
import org.gillius.realdb.model.data.Record;
import org.gillius.realdb.tools.DatabaseBuilder;
import org.gillius.realdb.util.DBIterator;
import org.junit.*;

import java.io.IOException;

import static org.junit.Assert.*;

/**
 * TestSingleStreamSpaceManagement tests that we can write more data into a single stream than can
 * fit into the entire database, implying that single stream space management works.
 *
 * @author Jason Winnebeck
 */
public class TestSingleStreamSpaceManagement {
	private static final String RDL_CONTENT =
			"SET blockSize     = 2048 " +
			"SET fileSize      = 204800 " +
			"SET maxStreams    = 3 " +
			"SET dataBlockSize = 2 " +
			"CREATE STREAM Test WITH ID 1 {value float WITH CODEC SampledAlgorithm}";

	@Test
	public void doTest() throws Exception {
		RDLFile rdl = RDLFile.load( RDL_CONTENT );
		DatabaseBuilder builder = new DatabaseBuilder( rdl );
		BlockFile file = new ByteArrayBlockFile( builder.getBlockSize(),
		                                         builder.getFileSize() / builder.getBlockSize() );
		Database db = new DatabaseBuilder( rdl ).createDatabase( file );

		assertEquals( 1, db.getStreams().size() );
		DataStream stream = db.getStreams().iterator().next();

		final int MAX_RECORD = 200000;
		for ( int i = 0; i < MAX_RECORD; ++i ) {
			//Since we are using sampled codec, this should always return true
			assertTrue( stream.writeRecord( makeRecord( i ) ) );
		}

		//Reload the database
		db.close();
		verifyData( file, MAX_RECORD );
	}

	private static void verifyData( BlockFile file, int lastRecord )
			throws IOException, BadFileException {
		Database db = Database.load( file );
		assertEquals( 1, db.getStreams().size() );
		DataStream stream = db.getStreams().iterator().next();

		//Test that the database contains some subset of records beyond 0 and until lastRecord
		DBIterator<Record> recordsIter = stream.readRecords();
		int firstRecord = (int) recordsIter.next().getTime();
		assertTrue( firstRecord > 0 );
		for ( int i = firstRecord + 1; i < lastRecord; ++i ) {
			assertTrue( recordsIter.hasNext() );
			assertEquals( makeRecord( i ), recordsIter.next() );
		}
		assertFalse( recordsIter.hasNext() );

		//We can't 100% guarantee the length of the stream, so we have to human-check it, kinda
		System.out.println( "Read records " + firstRecord + " to " + lastRecord );
	}

	private static GenericRecord makeRecord( int i ) {
		return new GenericRecord( i, i * 2.5f );
	}
}
