/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.integration;

import org.gillius.realdb.io.ByteArrayBlockFile;
import org.gillius.realdb.io.rdl.RDLFile;
import org.gillius.realdb.metadata.ElementType;
import org.gillius.realdb.metadata.RecordElement;
import org.gillius.realdb.model.DataStream;
import org.gillius.realdb.model.Database;
import org.gillius.realdb.model.algorithms.SampledAlgorithm;
import org.gillius.realdb.model.data.GenericRecord;
import org.gillius.realdb.model.data.Record;
import org.gillius.realdb.tools.DatabaseBuilder;
import org.gillius.realdb.util.DBIterator;
import org.junit.*;

import static org.junit.Assert.*;

/**
 * TestWriteSingleStreamDatabase is an automated integration test for writing a single stream to a
 * database. This test should verify the data storage requirement of milestone 2.
 *
 * @author Jason Winnebeck
 */
public class TestWriteSingleStreamDatabase {
	private static final String RDL_CONTENT =
			"SET blockSize     = 2048 " +
			"SET fileSize      = 204800 " +
			"SET maxStreams    = 3 " +
			"SET dataBlockSize = 2 " +
			"CREATE STREAM Test WITH ID 1 {value float WITH CODEC SampledAlgorithm}";

	@Test
	public void doTest() throws Exception {
		ByteArrayBlockFile file = new ByteArrayBlockFile( 2048, 100 );
		RDLFile rdl = RDLFile.load( RDL_CONTENT );
		Database db = new DatabaseBuilder( rdl ).createDatabase( file );

		assertEquals( 1, db.getStreams().size() );
		DataStream stream = db.getStreams().iterator().next();

		//Check some basic info about this stream that we expect to be true
		assertEquals( "Test", stream.getInfo().getName() );
		assertEquals( 1, stream.getInfo().getUserId() );
		RecordElement[] recordElements = stream.getInfo().getElements();
		assertEquals( 1, recordElements.length );
		assertEquals( "value", recordElements[0].getName() );
		assertSame( SampledAlgorithm.class, recordElements[0].getCodec().getClass() );
		assertEquals( ElementType.Float, recordElements[0].getType() );

		for ( int i = 0; i < 2000; ++i ) {
			Record r = new GenericRecord( i, i * 2.5f );
			//Since we are using sampled codec, this should always return true
			assertTrue( stream.writeRecord( r ) );
			assertEquals( 0, stream.getStartTime() );
			assertEquals( i, stream.getEndTime() );
		}

		//Reload the database
		db.close();
		db = Database.load( file );

		DBIterator<Record> recordsIter = stream.readRecords();
		for ( int i = 0; i < 2000; ++i ) {
			assertTrue( recordsIter.hasNext() );
			Record r = new GenericRecord( i, i * 2.5f );
			assertEquals( r, recordsIter.next() );
		}
		assertFalse( recordsIter.hasNext() );
	}
}
