/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.io;

import org.junit.*;

import java.io.EOFException;
import java.io.IOException;
import java.util.Arrays;

import static org.junit.Assert.*;

/**
 * TestByteArrayBlockFile tests the {@link ByteArrayBlockFile} class.
 *
 * @author Jason Winnebeck
 */
public class TestByteArrayBlockFile {
	/**
	 * BlockFile initially with all 0 data.
	 */
	private BlockFile file;

	@Before
	public void setUp() {
		file = new ByteArrayBlockFile( 512, 10 );
	}

	@Test
	public void testGetters() {
		assertEquals( 512, file.getBlockSize() );
		assertEquals( 10,  file.getSize() );
	}

	/**
	 * Tests that we can read the entire (starting) file -- should be all zeros.
	 */
	@Test
	public void emptyRead() throws IOException {
		byte[] b        = new byte[512];
		byte[] expected = new byte[512];

		for ( int i = 0; i < file.getSize(); ++i ) {
			file.readBlock( i, b );
			assertArrayEquals( expected, b );
		}
	}

	@Test
	public void readWriteBlocks() throws IOException {
		byte[] data = new byte[ 700 ];
		Arrays.fill( data, (byte) 5 );

		byte[] expected = new byte[ 1024 ];
		System.arraycopy( data, 0, expected, 0, data.length ); //rest of the array is 0

		file.writeBlocks( 2, data );
		byte[] read = new byte[ 700 ];
		file.readBlocks( 2, read );
		assertArrayEquals( data, read );

		read = new byte[ 1024 ];
		file.readBlocks( 2, read );
		assertArrayEquals( expected, read );
	}

	@Test
	public void readWriteBlocksLongForm() throws IOException {
		byte[] data = new byte[ 700 ];
		Arrays.fill( data, 10, 20, (byte) 5 );

		byte[] expected = new byte[ 1024 ];
		System.arraycopy( data, 10, expected, 0, 10 ); //rest of the array is 0

		file.writeBlocks( 2, data, 10, 10 );
		byte[] read = new byte[ 10 ];
		file.readBlocks( 2, read );
		assertArrayEquals( Arrays.copyOfRange( data, 10, 20 ), read );

		read = new byte[ 1024 ];
		file.readBlocks( 2, read );
		assertArrayEquals( expected, read );
	}

	@Test(expected = IOException.class)
	public void testOverwriteMultiBlock() throws IOException {
		file.writeBlocks( file.getSize() - 1, new byte[ file.getBlockSize() + 1 ] );
	}

	@Test(expected = IOException.class)
	public void badReadBlocks() throws IOException {
		file.readBlocks( file.getSize() - 1, new byte[ file.getBlockSize() + 1 ] );
	}

	@Test(expected = IllegalArgumentException.class)
	public void testNegativeRead() throws IOException {
		file.readBlock( -1, new byte[ file.getBlockSize() ] );
	}

	@Test(expected = IllegalArgumentException.class)
	public void testNegativeWrite() throws IOException {
		file.writeBlock( -1, new byte[ file.getBlockSize() ] );
	}

	@Test(expected = EOFException.class)
	public void testOverwriteSingleBlock() throws Exception {
		file.writeBlocks( file.getSize(), new byte[1] );
	}

	@Test
	public void testReadLastBlock() throws Exception {
		file.readBlocks( file.getSize() - 1, new byte[ file.getBlockSize() ] );
	}

	/**
	 * Regression test for bug found where if we write an exact-sized block with writeBlocks, we
	 * overwrite the entire next block with zeros.
	 */
	@Test
	public void testWriteLastBlock() throws IOException {
		file.writeBlocks( file.getSize() - 1, new byte[ file.getBlockSize() ] );
	}

	/**
	 * Regression test for bug found where if we write an exact-sized block with writeBlocks, we
	 * overwrite the entire next block with zeros.
	 */
	@Test
	public void testWriteBlocksFullBlock() throws IOException {
		byte[] data = new byte[ file.getBlockSize() ];
		Arrays.fill( data, (byte)1 );
		file.writeBlocks( 1, data );
		file.writeBlocks( 0, data );

		byte[] result = new byte[ file.getBlockSize() ];
		file.readBlocks( 0, result );
		assertArrayEquals( data, result );
		file.readBlocks( 1, result );
		assertArrayEquals( data, result );
	}
}
