/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.io;

import org.easymock.*;
import org.gillius.realdb.metadata.ElementType;
import org.gillius.realdb.metadata.OptionalElementType;
import org.gillius.realdb.model.data.Discontinuity;
import org.gillius.realdb.model.data.GenericRecord;
import org.gillius.realdb.model.data.Record;
import org.junit.*;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInput;
import java.io.DataInputStream;
import java.io.DataOutput;
import java.io.DataOutputStream;
import java.io.IOException;

import static org.easymock.EasyMock.*;
import static org.junit.Assert.*;

/**
 * TestGenericRecordSerializer tests the {@link GenericRecordSerializer} class.
 *
 * @author Jason Winnebeck
 */
public class TestGenericRecordSerializer {

	private static final OptionalElementType[] types = {
	    new MyOET( ElementType.SInt32, false ),
	    new MyOET( ElementType.Float,  true ),
	    new MyOET( ElementType.Double, false )
	};

	private GenericRecordSerializer grs;
	private Record rNormal;
	private Record rWithNulls;
	private Record rDiscontinuity;

	@Before
	public void setUp() throws Exception {
		grs = new GenericRecordSerializer( types );
		rNormal = new GenericRecord( 10, 16, 34.4f, 923.3 );
		rWithNulls = new GenericRecord( rNormal.getTime(),
		                                rNormal.getElement( 0 ),
		                                null,
		                                rNormal.getElement( 2 ) );
		rDiscontinuity = new Discontinuity( 100 );
	}

	@Test
	public void testGetWrittenSize() throws Exception {
		assertEquals( 8 + 1 + 4 + 4 + 8, grs.getWrittenSize() );
	}

	@Test
	public void testWriteRecordNoNulls() throws Exception {
		DataOutput out = createStrictMock( DataOutput.class );

		out.writeLong( rNormal.getTime() );
		out.write( aryEq( new byte[1] ) ); //no bits set = no null fields null and no discontinuity
		out.writeInt( (Integer) rNormal.getElement( 0 ) );
		out.writeFloat( (Float) rNormal.getElement( 1 ) );
		out.writeDouble( (Double) rNormal.getElement( 2 ) );

		replay( out );

		grs.writeRecord( rNormal, out );

		verify( out );
	}

	@Test
	public void testWriteRecordWithNulls() throws Exception {
		DataOutput out = createStrictMock( DataOutput.class );

		out.writeLong( rWithNulls.getTime() );
		out.write( aryEq( new byte[] { 2 } ) ); //second bit set = first null field is null
		out.writeInt( (Integer) rWithNulls.getElement( 0 ) );
		out.writeFloat( 0.0f );
		out.writeDouble( (Double) rWithNulls.getElement( 2 ) );

		replay( out );

		grs.writeRecord( rWithNulls, out );

		verify( out );
	}

	@Test
	public void testWriteDiscontinuity() throws Exception {
		ByteArrayOutputStream baos = new ByteArrayOutputStream( grs.getWrittenSize() );
		grs.writeRecord( rDiscontinuity, new DataOutputStream( baos ) );

		assertArrayEquals( new byte[] { 0, 0, 0, 0, 0, 0, 0, 100, //timestamp = 100
		                                1,                        //first bit = discontinuity
		                                0,0,0,0,                  //placeholder int
		                                0,0,0,0,                  //placeholder float
		                                0,0,0,0,0,0,0,0 },        //placeholder double
		                   baos.toByteArray() );
	}

	@Test
	public void testReadRecord() throws Exception {
		DataInput in = createStrictMock( DataInput.class );

		expect( in.readLong() ).andReturn( rNormal.getTime() );
		in.readFully( (byte[]) anyObject() );
		expectLastCall().andAnswer( new SetByteArray( new byte[1] ) );
		expect( in.readInt() ).andReturn( (Integer) rNormal.getElement( 0 ) );
		expect( in.readFloat() ).andReturn( (Float) rNormal.getElement( 1 ) );
		expect( in.readDouble() ).andReturn( (Double) rNormal.getElement( 2 ) );

		replay( in );

		assertEquals( rNormal, grs.readRecord( in ) );

		verify( in );
	}

	@Test
	public void testRoundTrip() throws Exception {
		doRoundTrip( rNormal );
	}

	@Test
	public void testRoundTripDiscontinuity() throws Exception {
		doRoundTrip( rDiscontinuity );
	}

	@Test
	public void testRoundTripWithNulls() throws Exception {
		doRoundTrip( rWithNulls );
	}

	private void doRoundTrip( Record r ) throws IOException {
		ByteArrayOutputStream baos = new ByteArrayOutputStream( grs.getWrittenSize() );

		grs.writeRecord( r, new DataOutputStream( baos ) );

		ByteArrayInputStream bais = new ByteArrayInputStream( baos.toByteArray() );

		assertEquals( r, grs.readRecord( new DataInputStream( bais ) ) );
	}

	private static class MyOET implements OptionalElementType {
		private final ElementType type;
		private final boolean optional;

		private MyOET( ElementType type, boolean optional ) {
			this.type = type;
			this.optional = optional;
		}

		public ElementType getType() {
			return type;
		}

		public boolean isOptional() {
			return optional;
		}
	}

	private static class SetByteArray implements IAnswer<Void> {
		private final byte[] data;

		private SetByteArray( byte[] data ) {
			this.data = data;
		}

		public Void answer() throws Throwable {
			//noinspection SuspiciousSystemArraycopy
			System.arraycopy( data, 0, getCurrentArguments()[0], 0, data.length );
			return null;
		}
	}
}
