/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.model;

import org.gillius.realdb.BadFileException;
import org.gillius.realdb.io.BlockFile;
import org.gillius.realdb.io.ByteArrayBlockFile;
import org.gillius.realdb.io.DataInputReader;
import org.gillius.realdb.io.DataOutputWritable;
import org.gillius.realdb.io.FailingBlockFile;
import org.gillius.realdb.util.DBIterator;
import org.junit.*;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.NoSuchElementException;
import java.util.Random;

import static org.easymock.EasyMock.*;
import static org.junit.Assert.*;

/**
 * TestCircularQueue
 *
 * @author Jason Winnebeck
 */
public class TestCircularQueue {

	private BlockFile bf;
	private DataInputReader<Integer> reader;

	@Before
	public void setUp() {
		bf = new ByteArrayBlockFile( 512, 4 );

		reader = new DataInputReader<Integer>() {
			public Integer readValue( DataInput in ) throws IOException {
				return in.readInt();
			}
		};
	}

	@Test
	public void testGetMinSize() throws Exception {
		assertEquals( 3, CircularQueue.getMinSize( 100, 0, 4, 512 ) );
		assertEquals( 4, CircularQueue.getMinSize( 512, 0, 4, 512 ) );
	}

	/**
	 * An empty file (all zeros) should never be considered a valid file.
	 */
	@Test( expected = BadFileException.class )
	public void testEmptyFile() throws Exception {
		CircularQueue.load( bf, 0, 4, 10 );
	}

	@Test( expected = IllegalArgumentException.class )
	public void testTooSmallFile() throws Exception {
		CircularQueue.createNew( bf, 0, 1, 10 );
	}

	@Test( expected = IllegalArgumentException.class )
	public void testTooLargeMaxDow() throws Exception {
		CircularQueue.createNew( bf, 0, 1, bf.getBlockSize() + 1 );
	}

	@Test
	public void testLoadFreshlyFormatted() throws Exception {
		CircularQueue.createNew( bf, 0, 4, 10 );

		CircularQueue q = CircularQueue.load( bf, 0, 4, 10 );

		assertFalse( q.iterator( reader ).hasNext() );
	}

	@Test( expected = NoSuchElementException.class )
	public void testReadNothing() throws Exception {
		CircularQueue q = CircularQueue.createNew( bf, 0, 4, 4 );
		q.iterator( reader ).next();
	}

	/**
	 * Tests that the format command writes to all blocks.
	 */
	@Test
	public void testFormatTouchesAllBlocks() throws Exception {
		BlockFile bf = createMock( "bf", BlockFile.class );

		expect( bf.getBlockSize() ).andStubReturn( 512 );
		expect( bf.getSize() ).andStubReturn( 4 );

		bf.writeBlock( eq( 0 ), (byte[]) anyObject() );
		expectLastCall().atLeastOnce();
		bf.writeBlock( eq( 1 ), (byte[]) anyObject() );
		expectLastCall().atLeastOnce();
		bf.writeBlock( eq( 2 ), (byte[]) anyObject() );
		expectLastCall().atLeastOnce();
		bf.writeBlock( eq( 3 ), (byte[]) anyObject() );
		expectLastCall().atLeastOnce();

		replay( bf );
		CircularQueue.createNew( bf, 0, 4, 4 );
		verify( bf );
	}

	@Test
	public void testWriteOneReadOne() throws Exception {
		CircularQueue q = CircularQueue.createNew( bf, 0, 4, 4 );

		q.add( dow( 1 ) );
		assertEquals( seq( 1, 2 ), iterToList( q.iterator( reader ) ) );

		q.flush();

		//Can we do the same after reloading?
		q = CircularQueue.load( bf, 0, 4, 4 );
		assertEquals( seq( 1, 2 ), iterToList( q.iterator( reader ) ) );
	}

	@Test
	public void testWriteMoreThanOneButNotRotatingBlocksAndRead() throws Exception {
		CircularQueue q = CircularQueue.createNew( bf, 0, 4, 4 );

		int toWrite = q.getUsableBytesPerBlock() / 4 * 3 / 2;
		List<Integer> expected = seq( 1, toWrite );
		for( int i : expected ) {
			q.add( dow( i ) );
		}

		assertEquals( expected, iterToList( q.iterator( reader ) ) );
		q.flush();

		//Can we do the same after reloading?
		q = CircularQueue.load( bf, 0, 4, 4 );
		assertEquals( expected, iterToList( q.iterator( reader ) ) );
	}

	@Test
	public void testRotation() throws Exception {
		CircularQueue q = CircularQueue.createNew( bf, 0, 4, 4 );
		int blockCapacity = q.getUsableBytesPerBlock() / 4;
		int maxCapacity = q.getTotalUsableBytes() / 4;
		int highest = maxCapacity * 3;

		for ( int i = 0; i <= highest; ++i ) {
			q.add( dow( i ) );
		}

		List<Integer> result = iterToList( q.iterator( reader ) );
		assertEquals( highest, (int) result.get( result.size() - 1 ) );
		int lowest = result.get( 0 );
		assertTrue( String.valueOf( lowest ), lowest >= highest - maxCapacity &&
		                                      lowest <= highest - maxCapacity + blockCapacity );
		List<Integer> expected = seq( lowest, highest + 1 );
		assertEquals( "Solid sequence", expected, result );

		q.flush();
		//Can we do the same after reloading?
		q = CircularQueue.load( bf, 0, 4, 4 );
		assertEquals( expected, iterToList( q.iterator( reader ) ) );
	}

	@Test
	public void testFault() throws Exception {
		FailingBlockFile f = new FailingBlockFile( bf, new Random( 140934324099L ) );
		CircularQueue q = CircularQueue.createNew( f, 0, 4, 4 );

		int blockCapacity = q.getUsableBytesPerBlock() / 4;
		int highest = blockCapacity / 2;

		for ( int i = 0; i <= highest; ++i ) {
			q.add( dow( i ) );
		}

		List<Integer> expected = seq( 0, highest + 1 );
		assertEquals( expected, iterToList( q.iterator( reader ) ) );
		q.flush();
		assertEquals( expected, iterToList( q.iterator( reader ) ) );

		q.add( dow( highest + 1 ) );
		f.setFailureCount( 1 );
		try {
			q.flush();
			fail( "Expected IOException" );
		} catch ( IOException e ) {
		}

		q = CircularQueue.load( bf, 0, 4, 4 );
		assertEquals( expected, iterToList( q.iterator( reader ) ) );
	}

	@Test
	public void testWriteHeader() throws Exception {
		CircularQueue q = CircularQueue.createNew( bf, 0, 4, 4 );
		int blockCapacity = q.getUsableBytesPerBlock() / 4;

		DataOutputWritable header = createMock( DataOutputWritable.class );

		header.write( (DataOutput) anyObject() );
		replay( header );
		q.add( dow( 0 ), header );
		verify( header );

		reset( header );
		replay( header );
		for ( int i : seq( 1, blockCapacity ) ) {
			q.add( dow( i ), header ); //no write to header
		}

		//Now we will start a new block, so make sure header is written
		reset( header );
		header.write( (DataOutput) anyObject() );
		replay( header );

		q.add( dow( 0 ), header );
		verify( header );
	}

	private static <T> List<T> iterToList( DBIterator<T> iter ) throws IOException, BadFileException {
		List<T> ret = new ArrayList<T>();

		while( iter.hasNext() ) {
			ret.add( iter.next() );
		}

		return ret;
	}

	private static List<Integer> seq( int start, int end ) {
		ArrayList<Integer> ret = new ArrayList<Integer>( end - start );
		for ( int i = start; i < end; ++i ) {
			ret.add( i );
		}
		return ret;
	}

	private static IntegerDataOutputWritable dow( int i ) {
		return new IntegerDataOutputWritable( i );
	}

	private static class IntegerDataOutputWritable implements DataOutputWritable {
		private final int value;

		private IntegerDataOutputWritable( int value ) {
			this.value = value;
		}

		public void write( DataOutput output ) throws IOException {
			output.writeInt( value );
		}
	}
}
