/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.model;

import org.gillius.realdb.io.RecordReader;
import org.gillius.realdb.io.RecordWriter;
import org.gillius.realdb.model.data.GenericRecord;
import org.gillius.realdb.model.data.Record;
import org.junit.*;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;
import java.util.Arrays;

import static org.easymock.EasyMock.*;
import static org.junit.Assert.*;

/**
 * TestDataBlock tests the {@link DataBlock} class.
 *
 * @author Jason Winnebeck
 */
public class TestDataBlock {
	private DataBlock    block;
	private MyWriter     writer;
	private RecordReader reader;

	@Before
	public void setUp() throws Exception {
		block  = new DataBlock( 0, MyWriter.SIZE, 100 );
		writer = new MyWriter();
		reader = createMock( RecordReader.class );
		replay( reader );
	}

	@Test( expected = IllegalArgumentException.class )
	public void testCtorRecordTooBig() {
		new DataBlock( 0, 50, 5 );
	}

	@Test( expected = IllegalArgumentException.class )
	public void testCtorSizeTooSmall() {
		new DataBlock( 0, 5, 5 + DataBlock.getHeaderSize() - 1 );
	}

	@Test
	public void testCtor() {
		int totalSize = 5 + DataBlock.getHeaderSize();
		DataBlock b = new DataBlock( 0, 5, totalSize );
		assertEquals( 0, b.getBlockNum() );
		assertEquals( 5, b.getRecordSize() );
		assertEquals( totalSize, b.getTotalSize() );
		assertTrue( b.canWriteRecord() );

		assertEquals( -1L, b.getStartTime() );
		assertEquals( -1L, b.getEndTime() );
	}

	@Test
	public void testRoundTrip() throws Exception {
		block.writeRecord( new GenericRecord( 0 ), writer );
		block.writeRecord( new GenericRecord( 1 ), writer );

		DataBlock reloaded = DataBlock.load( block.getRawData(), block.getBlockNum(),
		                                     block.getRecordSize(), block.getTotalSize() );

		assertEquals( block, reloaded );
	}

	/**
	 * Tries to simulate a test like would happen when loading from a real BlockFile.
	 */
	@Test
	public void testRoundTripPadded() throws Exception {
		block.writeRecord( new GenericRecord( 0 ), writer );
		block.writeRecord( new GenericRecord( 1 ), writer );

		byte[] data = block.getRawData();
		data = Arrays.copyOf( data, data.length + 100 );

		DataBlock reloaded = DataBlock.load( data, block.getBlockNum(),
		                                     block.getRecordSize(), block.getTotalSize() );

		assertEquals( block, reloaded );
	}

	@Test
	public void testCanWriteRecord() throws IOException {
		block = new DataBlock( 0, writer.getWrittenSize(),
		                       writer.getWrittenSize() + DataBlock.getHeaderSize() );
		assertTrue( block.canWriteRecord() );
		block.writeRecord( new GenericRecord( 26 ), writer );
		assertFalse( block.canWriteRecord() );
	}

	@Test
	public void testWriteRecord() throws IOException {
		final long startTime = 100;
		assertTrue( block.canWriteRecord() );
		block.writeRecord( new GenericRecord( startTime ), writer );
		assertEquals( startTime, block.getStartTime() );
		assertEquals( startTime, block.getEndTime() );
		assertTrue( block.canWriteRecord() );

		block.writeRecord( new GenericRecord( startTime + 100 ), writer );
		assertEquals( startTime, block.getStartTime() );
		assertEquals( startTime + 100, block.getEndTime() );

		byte[] expected = new byte[] { 0,0,0,0,0,0,0, (byte) startTime,
		                               0,0,0,0,0,0,0, (byte) (startTime+100) };
		byte[] data     = block.getRawData();
		assertArrayEquals( expected, Arrays.copyOfRange( data, DataBlock.getHeaderSize(), data.length ) );
	}

	@Test
	public void testReadRecord() throws Exception {
		Record r = new GenericRecord( 26 );
		block.writeRecord( r, writer );

		reset( reader );
		expect( reader.getWrittenSize() ).andStubReturn( writer.getWrittenSize() );
		expect( reader.readRecord( (DataInput) anyObject() ) ).andReturn( r );
		replay( reader );

		assertSame( r, block.readRecord( 0, reader ) );

		verify( reader );
	}

	@Test( expected = IllegalArgumentException.class )
	public void testReadRecordNotYetWritten() throws Exception {
		Record r = new GenericRecord( 26 );
		block.writeRecord( r, writer );

		reset( reader );
		expect( reader.getWrittenSize() ).andStubReturn( writer.getWrittenSize() );
		replay( reader );
		block.readRecord( 1, reader );
	}

	@Test
	public void testFindRecord3() throws Exception {
		block.writeRecord( new GenericRecord( 25 ), writer );
		block.writeRecord( new GenericRecord( 50 ), writer );
		block.writeRecord( new GenericRecord( 100 ), writer );

		assertEquals( 0, block.findRecord( 0 ) );
		assertEquals( 0, block.findRecord( 24 ) );
		assertEquals( 0, block.findRecord( 25 ) );
		assertEquals( 0, block.findRecord( 49 ) );
		assertEquals( 1, block.findRecord( 50 ) );
		assertEquals( 1, block.findRecord( 99 ) );
		assertEquals( 2, block.findRecord( 100 ) );
		assertEquals( 2, block.findRecord( 125 ) );
	}

	@Test( expected = IllegalStateException.class )
	public void testFindRecordEmpty() throws Exception {
		block.findRecord( 5 );
	}

	@Test
	public void testFindRecord1() throws Exception {
		block.writeRecord( new GenericRecord( 25 ), writer );

		assertEquals( 0, block.findRecord( 0 ) );
		assertEquals( 0, block.findRecord( 24 ) );
		assertEquals( 0, block.findRecord( 25 ) );
		assertEquals( 0, block.findRecord( 49 ) );
	}

	private static class MyWriter implements RecordWriter {
		private static final int SIZE = 8;

		public int getWrittenSize() {
			return SIZE;
		}

		public void writeRecord( Record r, DataOutput out ) throws IOException {
			out.writeLong( r.getTime() );
		}
	}
}
