/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.model;

import org.easymock.*;
import org.gillius.realdb.BadFileException;
import org.gillius.realdb.io.ByteArrayBlockFile;
import org.gillius.realdb.util.DBIterator;
import org.junit.*;

import java.io.Flushable;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import static java.util.Collections.*;
import static org.easymock.EasyMock.*;
import static org.junit.Assert.*;

/**
 * TestDataIndex tests the {@link DataIndex} class.
 *
 * @author Jason Winnebeck
 */
public class TestDataIndex {
	private IMocksControl      ctrl;
	private Transaction        tx;
	private Capture<Flushable> capturedFlushable;

	@Before
	public void setUp() throws Exception {
		ctrl              = createStrictControl();
		tx                = ctrl.createMock( Transaction.class );
		capturedFlushable = new Capture<Flushable>();
		Flushable f = ctrl.createMock( Flushable.class );

		expect( tx.addAction( (Flushable) and( notNull(), capture( capturedFlushable ) ) ) ).andReturn( f );
		f.flush();

		ctrl.replay();
	}

	/**
	 * Tests that we can create an empty DataIndex then load it back, and also tests minimum sizes.
	 */
	@Test
	public void testEmpty() throws Exception {
		ByteArrayBlockFile file = new ByteArrayBlockFile( DataIndex.MIN_BLOCK_SIZE,
		                                                  DataIndex.MIN_SIZE );
		DataIndex idx = DataIndex.createNew( file, 0, DataIndex.MIN_SIZE );
		assertEquals( -1, idx.getFirstBlock() );
		assertEquals( -1, idx.getSecondBlock() );
		assertEquals( -1, idx.getStartTime() );
		assertEquals( -1, idx.getEndTime() );

		DataIndex reloaded = DataIndex.load( file, 0, DataIndex.MIN_SIZE );
		assertEquals( idx, reloaded );

		assertFalse( reloaded.iterator().hasNext() );
	}

	@Test( expected = BadFileException.class )
	public void testBadLoad() throws Exception {
		ByteArrayBlockFile file = new ByteArrayBlockFile( DataIndex.MIN_BLOCK_SIZE,
		                                                  DataIndex.MIN_SIZE );
		DataIndex.load( file, 0, DataIndex.MIN_SIZE );
	}

	/**
	 * In this test we try to create the condition where the head block only has 1 index, and the
	 * second block needs to be read to get the second index.
	 */
	@Test
	public void testGetSecondBlock() throws Exception {
		int size = DataIndex.getIndexSize( 2, DataIndex.MIN_BLOCK_SIZE );
		ByteArrayBlockFile file = new ByteArrayBlockFile( DataIndex.MIN_BLOCK_SIZE, size );
		DataIndex idx = DataIndex.createNew( file, 0, size );
		addBlocks( idx, 1, 3 );
		assertEquals( 1, idx.getFirstBlock() );
		assertEquals( 2, idx.getSecondBlock() );

		idx.removeFirstBlock( 2, null );
		assertEquals( 2, idx.getFirstBlock() );
	}

	@Test
	public void testGetLastBlock() throws Exception {
		DataIndex idx = makeIndex( 1000, 256 );
		idx.addDataBlock( 1, 2, 3, null );
		idx.addDataBlock( 2, 4, 5, null );
		idx.addDataBlock( 3, 6, 7, null );

		assertEquals( 3, idx.getLastBlock() );
	}

	@Test
	public void testGetIndexSizeEmpty() {
		assertEquals( DataIndex.MIN_SIZE, DataIndex.getIndexSize( 1, DataIndex.MIN_BLOCK_SIZE ) );
	}

	/**
	 * Tests that for an advertised index size, we can actually follow through.
	 */
	@Test
	public void testGetIndexSize() throws Exception {
		DataIndex idx = makeIndex( 1000, 256 );

		for ( int i = 0; i < 1000; ++i ) {
			idx.addDataBlock( i, i, i, null );
		}

		idx.flush();
		assertEquals( 0, idx.getFirstBlock() );
		assertEquals( 1, idx.getSecondBlock() );

		//Read back all of the blocks
		int i = 0;
		DBIterator<Integer> iter = idx.iterator();
		while ( iter.hasNext() ) {
			int dataBlockIndex = iter.next();
			assertEquals( i, dataBlockIndex );
			++i;
		}
	}

	/**
	 * Tests that we can perform some index operations and reload and get the expected results.
	 */
	@Test
	public void testLoad() throws Exception {
		int size = DataIndex.getIndexSize( 1000, 256 );
		ByteArrayBlockFile file = new ByteArrayBlockFile( 256, size );
		DataIndex idx = DataIndex.createNew( file, 0, size );

		addBlocks( idx, 0, 5 );
		removeBlocks( idx, 0, 2 );
		addBlocks( idx, 5, 200 );
		removeBlocks( idx, 2, 50 );
		addBlocks( idx, 200, 800 );

		assertEquals( seq( 50, 800 ), idx.getAllIndicies() );

		idx.flush();
		idx = DataIndex.load( file, 0, size );
		assertEquals( seq( 50, 800 ), idx.getAllIndicies() );

		removeBlocks( idx, 50, 300 );
		addBlocks( idx, 800, 900 );
		assertEquals( seq( 300, 900 ), idx.getAllIndicies() );

		assertEquals( 300, idx.getFirstBlock() );
		assertEquals( 301, idx.getSecondBlock() );
	}

	/**
	 * This is actually a regression test for when headAdvanced was true, when it should have been
	 * false. What happened is that we write true for headAdvanced, which causes load to ignore the
	 * ALT_HEAD. The problem was that we never set it back to false after writing to ALT_HEAD after
	 * advancing.
	 * <p/>
	 * Unfortunately, I don't have as much direct control as I'd like; this probably means DataIndex
	 * does too much. I try to "provoke" the bug by deleting more than 1 block's worth of records,
	 * then flushing.
	 */
	@Test
	public void testLoadAfterAdvance() throws Exception {
		int size = DataIndex.getIndexSize( 1000, 256 );
		ByteArrayBlockFile file = new ByteArrayBlockFile( 256, size );
		DataIndex idx = DataIndex.createNew( file, 0, size );

		addBlocks( idx, 0, 200 );
		idx.flush();
		removeBlocks( idx, 0, 80 );
		idx.flush();

		assertEquals( seq( 80, 200 ), idx.getAllIndicies() );

		idx = DataIndex.load( file, 0, size );
		assertEquals( seq( 80, 200 ), idx.getAllIndicies() );
	}

	@Test
	public void testRemoveFirstBlockSmall() throws Exception {
		//Pick a very large blockSize so we can ensure testing of head==tail case
		DataIndex idx = makeIndex( 15, 4096 );

		for ( int i = 0; i < 5; ++ i ) {
			idx.addDataBlock( i, i, i, null );
		}

		assertEquals( seq( 0, 5 ), idx.getAllIndicies() );

		idx.removeFirstBlock( 1, null );
		assertEquals( seq( 1, 5 ), idx.getAllIndicies() );

		idx.removeFirstBlock( 2, null );
		idx.removeFirstBlock( 3, null );
		idx.removeFirstBlock( 4, null );
		assertEquals( singletonList( 4 ), idx.getAllIndicies() );
		assertEquals(  4, idx.getFirstBlock() );
		assertEquals( -1, idx.getSecondBlock() );

		idx.removeFirstBlock( -1, null ); //param should be ignored
		assertEquals( Collections.EMPTY_LIST, idx.getAllIndicies() );
	}

	@Test
	public void testRemoveFirstBlockCycling1() throws Exception {
		DataIndex idx = makeIndex( 1000, 256 );

		for ( int i = 1; i < 10000; ++i ) {
			idx.addDataBlock( i, i, i, null );
			idx.removeFirstBlock( -1, null ); //param should be ignored
		}
	}

	@Test
	public void testRemoveFirstBlockFill() throws Exception {
		DataIndex idx = makeIndex( 1000, 256 );

		addBlocks( idx, 0, 1000 );
		removeBlocks( idx, 0, 1000 );

		assertEquals( Collections.EMPTY_LIST, idx.getAllIndicies() );
	}

	/**
	 * Tests the case where we write to the tail, then flush (to alt tail) then advance the tail
	 * block. When we read back, do we read the old or the new data?
	 */
	@Test
	public void testFlushThenAdvanceTail() throws Exception {
		//guesstimate 20-25 blocks per file block. If the structure changes, this test won't work
		//anymore.
		DataIndex idx = makeIndex( 1000, 25 * 4 );

		//Get into the second block
		addBlocks( idx, 0, 31 );
		idx.flush();
		addBlocks( idx, 31, 32 );
		idx.flush();
		addBlocks( idx, 32, 200 );

		assertEquals( seq( 0, 200 ), idx.getAllIndicies() );
	}

	@Test
	public void testStartEndTime() throws Exception {
		DataIndex idx = makeIndex( 1000, 256 );

		addBlocks( idx, 0, 1000 );
		assertEquals( 0, idx.getStartTime() );
		assertEquals( 999, idx.getEndTime() );

		removeBlocks( idx, 0, 200 );
		assertEquals( 200, idx.getStartTime() );
		assertEquals( 999, idx.getEndTime() );
	}

	@Test
	public void testTailTransactionHandlingExplicitFlush() throws Exception {
		DataIndex idx = makeIndex( 1000, 256 );

		idx.addDataBlock( 1, 10, 20, tx );
		idx.flush();

		ctrl.verify();
	}

	@Test
	public void testTailTransactionHandlingImplicitFlush() throws Exception {
		DataIndex idx = makeIndex( 1000, DataIndex.MIN_BLOCK_SIZE );

		idx.addDataBlock( 1, 1, 1, tx );
		addBlocks( idx, 2, 900 ); //This should force the tail to flush implicitly

		ctrl.verify();
	}

	@Test
	public void testHeadTransactionHandlingExplicitFlush() throws Exception {
		DataIndex idx = makeIndex( 1000, 256 );

		addBlocks( idx, 1, 500 );
		idx.removeFirstBlock( 2, tx );
		idx.flush();

		ctrl.verify();
	}

	@Test
	public void testHeadTransactionHandlingImplicitFlush() throws Exception {
		DataIndex idx = makeIndex( 1000, DataIndex.MIN_BLOCK_SIZE );

		addBlocks( idx, 1, 30 );
		idx.removeFirstBlock( 2, tx );
		addBlocks( idx, 30, 500 );

		ctrl.verify();
	}

	@Test
	public void testTailTransactionFlushThroughFlushable() throws Exception {
		DataIndex idx = makeIndex( 1000, 256 );

		idx.addDataBlock( 1, 10, 20, tx );
		capturedFlushable.getValue().flush(); //simulate a commit; this should cause a flush on our mock

		ctrl.verify();
	}

	@Test
	public void testHeadTransactionFlushThroughFlushable() throws Exception {
		DataIndex idx = makeIndex( 1000, 256 );

		addBlocks( idx, 1, 30 );
		idx.removeFirstBlock( 2, tx );
		capturedFlushable.getValue().flush();

		ctrl.verify();
	}

	@Test
	public void testIteratorOnEmpty() throws Exception {
		DataIndex idx = makeIndex( 1000, 256 );

		assertFalse( idx.iterator().hasNext() );
	}

	@Test
	public void testRangedIteratorOnEmpty() throws Exception {
		DataIndex idx = makeIndex( 1000, 256 );

		DataBlockSummaryReader reader = createMock( DataBlockSummaryReader.class );
		replay( reader );

		assertFalse( idx.iterator( reader, 50, 100 ).hasNext() );
	}

	private static void removeBlocks( DataIndex idx, int start, int end )
			throws IOException, BadFileException {
		for ( int i = start; i < end; ++i ) {
			assertEquals( i, idx.getFirstBlock() );
			idx.removeFirstBlock( i + 1, null );
		}
	}

	private static void addBlocks( DataIndex idx, int start, int end ) throws IOException {
		for ( int i = start; i < end; ++i ) {
			idx.addDataBlock( i, i, i, null );
		}
	}

	private static List<Integer> seq( int start, int end ) {
		ArrayList<Integer> ret = new ArrayList<Integer>( end - start );
		for ( int i = start; i < end; ++i ) {
			ret.add( i );
		}
		return ret;
	}

	private static DataIndex makeIndex( int numDataBlocks, int blockSize ) throws IOException {
		int size = DataIndex.getIndexSize( numDataBlocks, blockSize );
		ByteArrayBlockFile file = new ByteArrayBlockFile( blockSize, size );
		DataIndex idx = DataIndex.createNew( file, 0, size );
		return idx;
	}
}
