/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.model;

import org.gillius.realdb.BadFileException;
import org.gillius.realdb.util.DBIterator;
import org.junit.*;

import java.io.IOException;
import java.util.Arrays;

import static org.junit.Assert.*;

/**
 * TestDataIndexIterator tests the {@link DataIndexIterator} class.
 *
 * @author Jason Winnebeck
 */
public class TestDataIndexIterator {
	private DataStreamIndexReader reader;
	private IndexBlock[] blocks;

	@Before
	public void setUp() throws Exception {
		blocks = makeBlocks(
				seq( 1, 5 ),
				seq( 6, 30 ),
				seq( 31, 40 )
		);

		reader = new ArrayDataStreamIndexReader( blocks );
	}

	@Test
	public void testIterateAll() throws Exception {
		DataIndexIterator iter = new DataIndexIterator( reader );

		assertArrayEquals( seq( 1, 40 ), iterateAll( iter ) );
	}

	@Test
	public void testIterateAllExplicit() throws Exception {
		DataIndexIterator iter = new DataIndexIterator( reader, 0, 0,
		                                                Integer.MAX_VALUE, Integer.MAX_VALUE );

		assertArrayEquals( seq( 1, 40 ), iterateAll( iter ) );
	}

	@Test
	public void testIterateFirstBlock() throws Exception {
		DataIndexIterator iter = new DataIndexIterator( reader, 0, 0, 0, 5 );

		assertArrayEquals( blocks[0].getDataBlocks(), iterateAll( iter ) );
	}

	@Test
	public void testIterateSecondBlockLargeEndIndex() throws Exception {
		DataIndexIterator iter = new DataIndexIterator( reader, 1, 0, 1, 500 );

		assertArrayEquals( blocks[1].getDataBlocks(), iterateAll( iter ) );
	}

	@Test
	public void testIterateMiddleToEnd() throws Exception {
		DataIndexIterator iter = new DataIndexIterator( reader, 1, 2,
		                                                Integer.MAX_VALUE, Integer.MAX_VALUE );

		assertArrayEquals( seq( 8, 40 ), iterateAll( iter ) );
	}

	@Test( expected = IllegalArgumentException.class )
	public void testIterateStartAfterTheEnd() throws Exception {
		new DataIndexIterator( reader, 50, 0, Integer.MAX_VALUE, Integer.MAX_VALUE );
	}

	@Test
	public void testIterateOverNothing() throws Exception {
		reader = new ArrayDataStreamIndexReader( new IndexBlock[0] );
		DataIndexIterator iter = new DataIndexIterator( reader );

		assertArrayEquals( new int[0], iterateAll( iter ) );
	}

	private static int[] iterateAll( DBIterator<Integer> iter ) throws Exception {
		int[] ret = new int[ 1000 ];
		int length = 0;

		while ( iter.hasNext() ) {
			ret[length] = iter.next();
			++length;
		}

		return Arrays.copyOf( ret, length );
	}

	private static class ArrayDataStreamIndexReader implements DataStreamIndexReader {
		private final IndexBlock[] blocks;

		private ArrayDataStreamIndexReader( IndexBlock[] blocks ) {
			this.blocks = blocks;
		}

		public int getNumRanges() {
			return blocks.length;
		}

		public IndexBlock getIndexBlock( int idx ) throws IOException, BadFileException {
			return blocks[idx];
		}

		public Range getRange( int idx ) throws IOException, BadFileException {
			return blocks[idx];
		}
	}

	private static int[] seq( int start, int end ) {
		int[] ret = new int[ end - start + 1 ];

		for ( int i = 0; i < ret.length; ++i ) {
			ret[i] = start + i;
		}
		return ret;
	}

	private static IndexBlock[] makeBlocks( int[]... indices ) {
		IndexBlock[] ret = new IndexBlock[ indices.length ];

		for ( int i = 0; i < indices.length; i++ ) {
			ret[i] = new IndexBlock( i, i, i, indices[i] );
		}
		return ret;
	}
}
