/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.model;

import org.gillius.realdb.metadata.DataStreamInfo;
import org.gillius.realdb.metadata.ElementType;
import org.gillius.realdb.metadata.RecordElement;
import org.gillius.realdb.model.algorithms.SampledAlgorithm;
import org.gillius.realdb.model.data.GenericRecord;
import org.gillius.realdb.model.data.Record;
import org.gillius.realdb.util.DBIterator;
import org.gillius.realdb.util.DBIteratorAdaptor;
import org.junit.*;

import java.io.IOException;
import java.util.List;

import static java.util.Arrays.*;
import static java.util.Collections.*;
import static org.easymock.EasyMock.*;
import static org.junit.Assert.*;

/**
 * TestDataStream
 *
 * @author Jason Winnebeck
 */
public class TestDataStream {
	private DataBlockManager manager;
	private DataStreamIndex  index;
	private DataStream       stream;

	private static final GenericRecord record = new GenericRecord( 0, 32.0f );
	private static final int RECORD_SIZE = 13;

	@Before
	public void setUp() {
		RecordElement[] elements = new RecordElement[] {
				new RecordElement( 0, "y", new SampledAlgorithm(), ElementType.Float, false )
		};
		DataStreamInfo info = new DataStreamInfo( 10, "Test Stream", elements );

		manager = createStrictMock( DataBlockManager.class );
		replay( manager );
		index = createMock( DataStreamIndex.class );
		expect( index.getStartTime() ).andReturn( -1L );
		expect( index.getEndTime() ).andReturn( -1L );
		replay( index );

		stream = new DataStream( info, index, manager );
		assertSame( info, stream.getInfo() );
	}

	@Test
	public void testWriteRecord() throws Exception {
		writeSingleRecord();
	}

	@Test
	public void testWritesFlush() throws Exception {
		writeSingleRecord();
		reset( manager );
		manager.closeBlock();
		replay( manager );

		stream.flush();
		verify( manager );
	}

	@Test
	public void testNoWritesFlush() throws IOException {
		stream.flush();
	}

	@Test
	public void testReadEmpty() throws Exception {
		reset( index );
		List<Integer> list = emptyList();
		expect( index.iterator() ).andReturn( DBIteratorAdaptor.make( list.iterator() ) );
		replay( index );

		assertFalse( stream.readRecords().hasNext() );
		verify( index );
	}

	@Test
	public void testRead() throws Exception {
		DataBlock block = writeSingleRecord();

		reset( manager, index );
		expect( index.iterator() ).andReturn( DBIteratorAdaptor.make( asList( 0 ).iterator() ) );
		expect( manager.readBlock( 0, RECORD_SIZE ) ).andReturn( block );
		replay( manager, index );

		DBIterator<Record> iter = stream.readRecords();
		assertTrue( iter.hasNext() );
		assertEquals( record, iter.next() );
		assertFalse( iter.hasNext() );
		assertFalse( iter.hasNext() );

		verify( manager, index );
	}

	private DataBlock writeSingleRecord() throws Exception {
		reset( manager );
		DataBlock block = new DataBlock( 0, RECORD_SIZE, 512 );
		expect( manager.getNextBlock( RECORD_SIZE ) ).andReturn( block );
		replay( manager );

		assertTrue( stream.writeRecord( record ) );
		assertEquals( 1, block.getNumRecords() );
		verify( manager );

		return block;
	}
}
