/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.model;

import org.junit.*;

import java.util.NoSuchElementException;

import static org.junit.Assert.*;

/**
 * TestIndexBlockFactory tests the {@link IndexBlockFactory} class.
 *
 * @author Jason Winnebeck
 */
public class TestIndexBlockFactory {
	private IndexBlockFactory factory;

	@Before
	public void setUp() {
		factory = new IndexBlockFactory( 50, 100 );
	}

	@Test
	public void testEquals() {
		assertEquals( factory, factory );
		assertEquals( new IndexBlockFactory( 50, 100 ), new IndexBlockFactory( 50, 100 ) );
		assertFalse( new IndexBlockFactory( 50, 100 ).equals( new IndexBlockFactory( 20, 100 ) ) );
		assertFalse( new IndexBlockFactory( 50, 100 ).equals( new IndexBlockFactory( 50, 300 ) ) );
	}

	@Test
	public void testStartingNewCtor() {
		assertTrue( factory.hasSpace() );
		assertEquals(  50, factory.getMaxIndices() );
		assertEquals(   0, factory.getNumIndices() );
		assertEquals( 100, factory.getEndTime()   );
	}

	@Test
	public void testFullCtor() {
		int[] data = { 1, 2, 3, 4 };
		factory = new IndexBlockFactory( 50, 100, 500, data );
		assertEquals( 50, factory.getMaxIndices() );
		assertTrue( factory.hasSpace() );
		assertEquals( 4, factory.getNumIndices() );

		IndexBlock expected = new IndexBlock( 3, 100, 500, data );
		assertEquals( expected, factory.makeBlock( 3 ) );
	}

	@Test
	public void testResume() {
		int[] data = { 1, 2, 3, 4 };
		IndexBlock expected = new IndexBlock( 3, 100, 500, data );
		factory = new IndexBlockFactory( 50, expected );
		assertTrue( factory.hasSpace() );
		assertEquals( 50,       factory.getMaxIndices() );
		assertEquals(  4,       factory.getNumIndices() );
		assertEquals( expected, factory.makeBlock( 3 ) );
	}

	@Test
	public void testResumeEmptyIndexBlock() {
		IndexBlock empty = new IndexBlock( 0 );
		factory = new IndexBlockFactory( 50, empty );

		factory.addDataBlock( 1, 0, 500 );
		IndexBlock expected = new IndexBlock( 1, 0, 500, new int[] { 1 } );
		assertEquals( expected, factory.makeBlock( 1 ) );
	}

	@Test
	public void testCtorMakeBlock() {
		assertEquals( new IndexBlock( 0 ), factory.makeBlock( 0 ) );
	}

	@Test
	public void testSimple() {
		assertTrue( factory.hasSpace() );

		factory.addDataBlock( 10, 120, 500 );
		assertEquals( 500, factory.getEndTime() );
		assertEquals( 10, factory.getFirstIndex() );
		assertEquals( 10, factory.getLastIndex() );

		IndexBlock expected;
		expected = new IndexBlock( 3, 120, 500, new int[] { 10 } );
		assertEquals( expected, factory.makeBlock( 3 ) );

		//only difference -- 1 more seq num
		expected = new IndexBlock( 4, 120, 500, new int[] { 10 } );
		assertEquals( expected, factory.makeBlock( 4 ) );
	}

	@Test
	public void testGetSecondIndex() {
		factory.addDataBlock( 1, 120, 130 );
		factory.addDataBlock( 2, 140, 150 );

		assertEquals( 2, factory.getSecondIndex() );
	}

	@Test
	public void testUpdate() {
		factory.addDataBlock( 10, 120, 500 );
		factory.updateDataBlock( 10, 120, 510 );

		IndexBlock expected;
		expected = new IndexBlock( 3, 120, 510, new int[] { 10 } );
		assertEquals( expected, factory.makeBlock( 3 ) );
	}

	@Test( expected = IllegalStateException.class )
	public void testBadUpdate() {
		factory.updateDataBlock( 1, 2, 3 );
	}

	@Test( expected = IllegalArgumentException.class )
	public void testBadUpdateEndTime() {
		factory.addDataBlock( 10, 120, 500 );
		factory.updateDataBlock( 10, 120, 490 );
	}

	@Test
	public void testStartNextBlock() {
		factory.addDataBlock( 10, 120, 500 );

		factory.startNextBlock();
		assertEquals( 0, factory.getNumIndices() );

		factory.addDataBlock( 11, 501, 600 );
		IndexBlock expected = new IndexBlock( 3, 501, 600, new int[] { 11 } );
		assertEquals( expected, factory.makeBlock( 3 ) );
	}

	@Test
	public void testMaxBlocks() {
		for ( int i = 0; i < factory.getMaxIndices(); ++i ) {
			assertTrue( factory.hasSpace() );
			int startTime = 101 + ( i * 2 );
			factory.addDataBlock( i, startTime, startTime + 1 );
		}

		assertFalse( factory.hasSpace() );
	}

	@Test
	public void testRemoveFirstBlock() {
		factory.addDataBlock( 1, 101, 200 );
		factory.addDataBlock( 2, 201, 300 );
		factory.addDataBlock( 3, 301, 400 );

		IndexBlock expected;
		expected = new IndexBlock( 0, 101, 400, new int[] { 1, 2, 3 } );
		assertEquals( expected, factory.makeBlock( 0 ) );

		factory.removeFirstDataBlock( 201 );
		expected = new IndexBlock( 0, 201, 400, new int[] { 2, 3 } );
		assertEquals( expected, factory.makeBlock( 0 ) );

		factory.removeFirstDataBlock( 301 );
		expected = new IndexBlock( 0, 301, 400, new int[] { 3 } );
		assertEquals( expected, factory.makeBlock( 0 ) );

		factory.removeFirstDataBlock( 0 );
		expected = new IndexBlock( 0 );
		assertEquals( 0, factory.getNumIndices() );
	}

	@Test( expected = IllegalArgumentException.class )
	public void testRemoveFirstBlockBadStartTime() {
		factory.addDataBlock( 1, 101, 200 );
		factory.addDataBlock( 2, 201, 300 );
		factory.removeFirstDataBlock( 400 );
	}

	@Test( expected = NoSuchElementException.class )
	public void testRemoveFirstBlockOnEmpty() {
		factory.removeFirstDataBlock( 0 );
	}

	/**
	 * Tests if we try to add too many blocks, that we are stopped.
	 */
	@Test( expected = IllegalStateException.class )
	public void testOverwrite() {
		testMaxBlocks();
		factory.addDataBlock( 15, 100000, 100001 );
	}

	/**
	 * Tests that we can't perform an overlap, even of 1 millisecond.
	 */
	@Test( expected = IllegalArgumentException.class )
	public void testBadOrder() {
		factory.addDataBlock( 10, 300, 400 );
		factory.addDataBlock( 11, 400, 500 );
	}

	/**
	 * Tests that we can't perform an overlap.
	 */
	@Test( expected = IllegalArgumentException.class )
	public void testBadOrder2() {
		factory.addDataBlock( 10, 300, 400 );
		factory.addDataBlock( 11, 350, 500 );
	}

	/**
	 * Tests the effect of the minStartTime parameter from the constructor.
	 */
	@Test( expected = IllegalArgumentException.class )
	public void testOrderInitial() {
		factory.addDataBlock( 0, 95, 200 );
	}
}
