/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.model;

import org.gillius.realdb.model.data.GenericRecord;
import org.gillius.realdb.model.data.Record;
import org.gillius.realdb.model.data.StreamInterval;
import org.gillius.realdb.util.DBIterator;
import org.gillius.realdb.util.DBIteratorAdaptor;
import org.junit.*;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.NoSuchElementException;

import static java.util.Arrays.*;
import static java.util.Collections.*;
import static org.easymock.EasyMock.*;
import static org.junit.Assert.*;

/**
 * TestIntervalIterator tests the {@link IntervalIterator} class.
 *
 * @author Jason Winnebeck
 */
public class TestIntervalIterator {
	private StreamReconstructor recon;

	private List<GenericRecord> records;
	private List<StreamInterval> intervals;
	private DBIterator<Record> iter;

	private IntervalIterator ii;

	@Before
	public void setUp() throws Exception {
		recon = createStrictMock( "Recon", StreamReconstructor.class );
		records = asList(
				new GenericRecord( 10 ),
				new GenericRecord( 20 ),
				new GenericRecord( 30 ) );

		intervals = asList(
				makeInterval( 10, 20 ),
				makeInterval( 20, 30 ),
				makeInterval( 30, 30 ) );

		iter  = new DBIteratorAdaptor<Record>( records.iterator() );
		ii    = new IntervalIterator( iter, recon );

		replay( recon );
	}

	@Test( expected = NoSuchElementException.class )
	public void testEmptyNext() throws Exception {
		iter = new DBIteratorAdaptor<Record>( Collections.<Record>emptyList().iterator() );
		ii   = new IntervalIterator( iter, recon );

		iter.next();
	}

	@Test
	public void testEmptyHasNext() throws Exception {
		iter = new DBIteratorAdaptor<Record>( Collections.<Record>emptyList().iterator() );
		ii   = new IntervalIterator( iter, recon );

		assertFalse( ii.hasNext() );
	}

	@Test
	public void testSingle() throws Exception {
		iter = new DBIteratorAdaptor<Record>( records.subList( 0, 1 ).iterator() );
		ii   = new IntervalIterator( iter, recon );

		doTestSingle();
	}

	private void doTestSingle() throws Exception {
		StreamInterval inter = makeInterval( 10, 10 );

		reset( recon );
		recon.startRange( records.get( 0 ) );
		expect( recon.finishRecords( records.get( 0 ).getTime() ) ).andReturn( inter );
		expect( recon.finishRecords( records.get( 0 ).getTime() ) ).andReturn( null );
		replay( recon );

		assertEquals( singletonList( inter ), iterAsList( ii ) );

		verify( recon );
	}

	@Test
	public void testSingleWithExactEndTime() throws Exception {
		iter = new DBIteratorAdaptor<Record>( records.subList( 0, 1 ).iterator() );
		ii   = new IntervalIterator( iter, recon, 10 );

		doTestSingle();
	}

	@Test
	public void testReadAll() throws Exception {
		reset( recon );
		recon.startRange( records.get( 0 ) );
		expect( recon.handleRecord( records.get( 1 ) ) ).andReturn( intervals.get( 0 ) );
		expect( recon.handleRecord( records.get( 2 ) ) ).andReturn( intervals.get( 1 ) );
		expect( recon.finishRecords( records.get( 2 ).getTime() ) ).andReturn( intervals.get( 2 ) );
		expect( recon.finishRecords( records.get( 2 ).getTime() ) ).andReturn( null );
		replay( recon );

		assertEquals( intervals, iterAsList( ii ) );

		verify( recon );
	}

	@Test
	public void testReadAllWithExactEndTime() throws Exception {
		ii = new IntervalIterator( iter, recon, records.get( 2 ).getTime() );
		testReadAll();
	}

	@Test
	public void testReadRange() throws Exception {
		ii = new IntervalIterator( iter, recon, 25 );

		reset( recon );
		recon.startRange( records.get( 0 ) );
		expect( recon.handleRecord( records.get( 1 ) ) ).andReturn( intervals.get( 0 ) );
		expect( recon.handleRecord( records.get( 2 ) ) ).andReturn( intervals.get( 1 ) );
		replay( recon );

		assertEquals( intervals.subList( 0, 2 ), iterAsList( ii ) );

		verify( recon );
	}

	@Test
	public void testReadRangeWithExactEndTime() throws Exception {
		ii = new IntervalIterator( iter, recon, 20 );

		reset( recon );
		recon.startRange( records.get( 0 ) );
		expect( recon.handleRecord( records.get( 1 ) ) ).andReturn( intervals.get( 0 ) );
		expect( recon.handleRecord( records.get( 2 ) ) ).andReturn( intervals.get( 1 ) );
		replay( recon );

		assertEquals( intervals.subList( 0, 2 ), iterAsList( ii ) );

		verify( recon );
	}

	@Test
	public void testReadAllAfterEndTime() throws Exception {
		ii = new IntervalIterator( iter, recon, records.get( 0 ).getTime() - 1 );
		assertEquals( 0, iterAsList( ii ).size() );
	}

	private static <T> List<T> iterAsList( DBIterator<T> iter ) throws Exception {
		List<T> ret = new ArrayList<T>();
		while ( iter.hasNext() ) {
			ret.add( iter.next() );
		}
		return ret;
	}

	private static StreamInterval makeInterval( long startTime, long endTime ) {
		StreamInterval ret = createNiceMock( StreamInterval.class );
		expect( ret.getStartTime() ).andStubReturn( startTime );
		expect( ret.getEndTime() ).andStubReturn( endTime );
		replay( ret );
		return ret;
	}
}
