/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.model;

import org.gillius.realdb.BadFileException;
import org.gillius.realdb.io.ByteArrayBlockFile;
import org.gillius.realdb.io.FailingBlockFile;
import org.junit.*;

import java.io.DataOutput;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.util.Random;

import static org.junit.Assert.*;

/**
 * TestReliableSequencedBlock tests the {@link ReliableSequencedBlock} class.
 *
 * @author Jason Winnebeck
 */
public class TestReliableSequencedBlock {
	private FailingBlockFile file;

	@Before
	public void setUp() throws Exception {
		file = new FailingBlockFile( new ByteArrayBlockFile( 512, 4 ),
		                             new Random( 12309841L ) );
	}

	@Test( expected = BadFileException.class )
	public void testFailedLoad() throws Exception {
		ReliableSequencedBlock.load( file, 0, 1, 100, SimpleBlock.loader );
	}

	@Test
	public void testRoundTrip() throws Exception {
		ReliableSequencedBlock<SimpleBlock> rsb = ReliableSequencedBlock.createNew( file, 0, 1, 100 );
		assertNull( rsb.getBlock() );

		SimpleBlock sb = new SimpleBlock( rsb.getNextSeqNum() );
		rsb.writeBlock( sb );
		assertSame( sb, rsb.getBlock() );

		rsb = ReliableSequencedBlock.load( file, 0, 1, 100, SimpleBlock.loader );
		assertNotSame( sb, rsb.getBlock() );
		assertEquals( sb, rsb.getBlock() );
	}

	@Test
	public void testMultiWrite() throws Exception {
		ReliableSequencedBlock<SimpleBlock> rsb = ReliableSequencedBlock.createNew( file, 0, 1, 100 );
		SimpleBlock sb = null;
		for ( int i = 0; i < 100; ++i ) {
			sb = new SimpleBlock( rsb.getNextSeqNum() );
			rsb.writeBlock( sb );
			assertTrue( rsb.getNextSeqNum() != sb.getSeqNum() );
		}

		rsb = ReliableSequencedBlock.load( file, 0, 1, 100, SimpleBlock.loader );
		assertNotSame( sb, rsb.getBlock() );
		assertEquals( sb, rsb.getBlock() );
	}

	@Test( expected = IOException.class )
	public void testBadSize() throws Exception {
		ReliableSequencedBlock<SimpleBlock> rsb = ReliableSequencedBlock.createNew( file, 0, 1, 15 );
		SimpleBlock sb = new SimpleBlock( rsb.getNextSeqNum() );
		rsb.writeBlock( sb );
	}

	@Test
	public void testCorruption() throws Exception {
		for ( int i = 1; i < 100; ++i ) {
			runCorruptionTest( i );
		}
	}

	private void runCorruptionTest( int numBlocks ) throws Exception {
		ReliableSequencedBlock<SimpleBlock> rsb = ReliableSequencedBlock.createNew( file, 0, 1, 100 );
		SimpleBlock sb = null;
		for ( int i = 0; i < numBlocks; ++i ) {
			sb = new SimpleBlock( rsb.getNextSeqNum() );
			rsb.writeBlock( sb );
		}

		file.setFailureCount( 1 );
		try {
			SimpleBlock next = new SimpleBlock( rsb.getNextSeqNum() );
			assertFalse( next.equals( sb ) );
			rsb.writeBlock( next );
			fail( "IOException expected" );
		} catch ( IOException e ) {
			//expected
		}

		assertSame( sb, rsb.getBlock() );

		rsb = ReliableSequencedBlock.load( file, 0, 1, 100, SimpleBlock.loader );
		assertNotSame( sb, rsb.getBlock() );
		assertEquals( sb, rsb.getBlock() );
	}

	private static class SimpleBlock implements SequencedBlock {
		public static final SequencedBlockLoader<SimpleBlock> loader = new SequencedBlockLoader<SimpleBlock>() {
			public SimpleBlock load( byte[] serialized ) throws BadFileException {
				return new SimpleBlock( serialized );
			}
		};

		private final int seqNum;

		private SimpleBlock( int seqNum ) {
			this.seqNum = seqNum;
		}

		private SimpleBlock( byte[] data ) throws BadFileException {
			ByteBuffer buf = ByteBuffer.wrap( data );
			seqNum = buf.getInt();

			if ( buf.getInt() != 1 )
				throw new BadFileException( "Corrupted" );
			if ( buf.getInt() != 2 )
				throw new BadFileException( "Corrupted" );
			if ( buf.getInt() != 0xDEADBEEF )
				throw new BadFileException( "Corrupted" );
		}

		public int getSeqNum() {
			return seqNum;
		}

		public void write( DataOutput output ) throws IOException {
			output.writeInt( seqNum );
			output.writeInt( 1 );
			output.writeInt( 2 );
			output.writeInt( 0xDEADBEEF );
		}

		public boolean equals( Object o ) {
			if ( this == o ) return true;
			if ( o == null || getClass() != o.getClass() ) return false;

			SimpleBlock that = (SimpleBlock) o;

			if ( seqNum != that.seqNum ) return false;

			return true;
		}

		public int hashCode() {
			return seqNum;
		}
	}
}
