/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.model.data;

import org.gillius.realdb.model.ElementReconstructor;
import org.junit.*;

import static org.easymock.EasyMock.*;
import static org.junit.Assert.*;

/**
 * TestDefaultStreamReconstructor tests the {@link DefaultStreamReconstructor} class.
 *
 * @author Jason Winnebeck
 */
public class TestDefaultStreamReconstructor {
	private DefaultStreamReconstructor recon;
	private ElementReconstructor[] recons;

	private static final int pivot = 1;
	private static final int[][] looks = {
	    {pivot, 0},
	    {0,     1}
	};

	@Before
	public void setUp() throws Exception {
		recons = new ElementReconstructor[] { makeRecon(), makeRecon() };

		recon = new DefaultStreamReconstructor( recons, pivot, pivot + 2 + looks[1][1] );
	}

	@Test
	public void testHandleRecord() throws Exception {
		Record[] r = {
				new GenericRecord( 10, 11.0f, 21.0 ),
				new GenericRecord( 20, 12.0f, 22.0 ),
				new GenericRecord( 30, 13.0f, 23.0 ),
				new GenericRecord( 40, 14.0f, 24.0 )
		};

		//We need 3 records to start doing anything
		recon.startRange( r[0] );
		assertNull( recon.handleRecord( r[1] ) );

		//Now the lookahead is full, so we are asked to start generating real data.
		setupERs( new long[] { -1, 10, 20, 30 },
		          new Object[][] {
				          new Object[] { null, 11.0f, 12.0f, 13.0f },
				          new Object[] { null, 21.0, 22.0, 23.0 }
		          });
		StreamInterval si = recon.handleRecord( r[2] );
		assertEquals( 10,    si.getStartTime()       );
		assertEquals( 20,    si.getEndTime()         );
		assertEquals( 11.0f, si.getAverageValue( 0 ) );
		assertEquals( 21.0,  si.getAverageValue( 1 ) );
		verify( (Object[]) recons );

		//Now there is 1 rotation, now the lookbehind is full
		setupERs( new long[] { 10, 20, 30, 40 },
		          new Object[][] {
				          new Object[] { 11.0f, 12.0f, 13.0f, 14.0f },
				          new Object[] { 21.0, 22.0, 23.0, 24.0 }
		          });
		si = recon.handleRecord( r[3] );
		assertEquals( 20,    si.getStartTime()       );
		assertEquals( 30,    si.getEndTime()         );
		assertEquals( 12.0f, si.getAverageValue( 0 ) );
		assertEquals( 22.0,  si.getAverageValue( 1 ) );
		verify( (Object[]) recons );

		//Now we're out of data in the stream, so the lookahead isn't filled
		setupERs( new long[] { 20, 30, 40, 50 },
		          new Object[][] {
				          new Object[] { 12.0f, 13.0f, 14.0f, null },
				          new Object[] { 22.0, 23.0, 24.0, null }
		          });
		si = recon.finishRecords( 50 );
		assertEquals( 30,    si.getStartTime()       );
		assertEquals( 40,    si.getEndTime()         );
		assertEquals( 13.0f, si.getAverageValue( 0 ) );
		assertEquals( 23.0,  si.getAverageValue( 1 ) );
		verify( (Object[]) recons );

		//last interval
		setupERs( new long[] { 30, 40, 50, -1 },
		          new Object[][] {
				          new Object[] { 13.0f, 14.0f, null, null },
				          new Object[] { 23.0, 24.0, null, null }
		          });
		si = recon.finishRecords( 50 );
		assertEquals( 40,    si.getStartTime()       );
		assertEquals( 50,    si.getEndTime()         );
		assertEquals( 14.0f, si.getAverageValue( 0 ) );
		assertEquals( 24.0,  si.getAverageValue( 1 ) );
		verify( (Object[]) recons );

		//shouldn't matter how many times we call it, should still stay null, now that data is gone.
		assertNull( recon.finishRecords( 50 ) );
		assertNull( recon.finishRecords( 50 ) );
		assertNull( recon.finishRecords( 50 ) );
		assertNull( recon.finishRecords( 50 ) );
		assertNull( recon.finishRecords( 50 ) );
	}

	@Test( expected = IllegalStateException.class )
	public void testCallHandleBeforeStart() {
		recon.handleRecord( new GenericRecord( 10, 11.0f, 21.0 ) );
	}

	@Test
	public void testFinishBeforeStart() {
		assertNull( recon.finishRecords( 50 ) );
	}

	@Test
	public void testDiscontinuities() {
		Record[] r = new Record[] {
				new Discontinuity( 10 ),
				new Discontinuity( 20 ),
				new GenericRecord( 30, 14.0f, 24.0 )
		};

		//We need 3 records to start doing anything
		recon.startRange( r[0] );
		assertNull( recon.handleRecord( r[1] ) );

		StreamInterval si = recon.handleRecord( r[2] );
		assertTrue( si.isDiscontinuity() );
		assertEquals( 10, si.getStartTime() );
		assertEquals( 20, si.getEndTime()   );

		si = recon.finishRecords( 50 );
		assertTrue( si.isDiscontinuity() );
		assertEquals( 20, si.getStartTime() );
		assertEquals( 30, si.getEndTime()   );

		//Now we pull out a real range, 30-50
		setupERs( new long[] { 20, 30, 50, -1 },
		          new Object[][] {
				          new Object[] { null, 14.0f, null, null },
				          new Object[] { null, 24.0, null, null }
		          });
		si = recon.finishRecords( 50 );
		assertFalse( si.isDiscontinuity() );
		assertEquals( 30, si.getStartTime() );
		assertEquals( 50, si.getEndTime()   );

		//no more...
		assertNull( recon.finishRecords( 50 ) );
	}

	private void setupERs( long[] expectedTimes, Object[][] expectedData ) {
		for ( int i = 0; i < recons.length; ++i ) {
			ElementReconstructor recon = recons[i];
			reset( recon );
			expect( recon.getLookBehindDesired() ).andStubReturn( looks[i][0] );
			expect( recon.getLookAheadDesired() ).andStubReturn( looks[i][1] );
			expect( recon.getInterval( aryEq(expectedTimes), aryEq(expectedData[i]), eq(pivot) ) ).andReturn(
					makeI( expectedData[i][pivot] ) );
			replay( recon );
		}
	}

	private static ElementInterval makeI( Object value ) {
		ElementInterval ret = createMock( ElementInterval.class );
		expect( ret.getAverageValue() ).andStubReturn( value );
		replay( ret );
		return ret;
	}

	private static ElementReconstructor makeRecon() {
		ElementReconstructor ret = createMock( ElementReconstructor.class );
		replay( ret );
		return ret;
	}
}
