/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.model.data;

import org.junit.*;

import java.util.Arrays;

import static org.junit.Assert.*;

/**
 * TestRecordBuffer tests the {@link RecordBuffer} class.
 *
 * @author Jason Winnebeck
 */
public class TestRecordBuffer {
	private RecordBuffer buf = null;

	@Before
	public void setUp() {
		buf = new RecordBuffer( 3, 2, 1 );
	}

	@Test
	public void testInitialState() throws Exception {
		assertArrayEquals( new long[] { -1, -1, -1 },
		                   buf.getTimes() );

		assertArrayEquals(
				new Object[][] {
						new Object[] { null, null, null },
						new Object[] { null, null, null }
				},
				buf.getData() );

		assertTrue( Arrays.equals( new boolean[3], buf.getDiscontinuities() ) );
	}

	@Test
	public void testRotateEmpty() throws Exception {
		buf.fillLast( r( 1 ) );
		buf.rotateEmpty( 10 );

		assertArrayEquals( new long[] { -1, 1, 10 },
		                   buf.getTimes() );

		assertArrayEquals(
				new Object[][] {
						new Object[] { null, 1, null },
						new Object[] { null, 1.0, null }
				},
				buf.getData() );

		assertTrue( Arrays.equals( new boolean[3], buf.getDiscontinuities() ) );
	}

	@Test
	public void testRotate() throws Exception {
		buf.rotate( r( 1 ) );
		buf.rotate( r( 2 ) );
		buf.rotate( r( 3 ) );

		assertArrayEquals( new long[] { 1, 2, 3 },
		                   buf.getTimes() );

		assertArrayEquals(
				new Object[][] {
						new Object[] { 1, 2, 3 },
						new Object[] { 1.0, 2.0, 3.0 }
				},
				buf.getData() );

		assertTrue( Arrays.equals( new boolean[3], buf.getDiscontinuities() ) );
	}

	@Test
	public void testRotateDiscontinuity() throws Exception {
		buf.rotate( new Discontinuity( 100 ) );
		buf.rotate( new Discontinuity( 200 ) );

		assertArrayEquals( new long[] { -1, 100, 200 },
		                   buf.getTimes() );

		assertArrayEquals(
				new Object[][] {
						new Object[] { null, null, null },
						new Object[] { null, null, null }
				},
				buf.getData() );

		assertTrue( Arrays.equals( new boolean[] { false, true, true },
		                           buf.getDiscontinuities() ) );
	}

	@Test
	public void testDropPivot() throws Exception {
		buf.rotate( r( 1 ) );
		buf.rotate( r( 2 ) );
		buf.rotate( r( 3 ) );
		buf.dropPivot();

		assertArrayEquals( new long[] { 1, 3, -1 },
		                   buf.getTimes() );

		assertArrayEquals(
				new Object[][] {
						new Object[] { 1, 3, null },
						new Object[] { 1.0, 3.0, null }
				},
				buf.getData() );

		assertTrue( Arrays.equals( new boolean[3], buf.getDiscontinuities() ) );
	}

	@Test
	public void testFillLast() throws Exception {
		buf.fillLast( r( 1 ) );
		buf.fillLast( r( 2 ) );

		assertArrayEquals( new long[] { -1, -1, 2 },
		                   buf.getTimes() );

		assertArrayEquals(
				new Object[][] {
						new Object[] { null, null, 2 },
						new Object[] { null, null, 2.0 }
				},
				buf.getData() );

		assertTrue( Arrays.equals( new boolean[3], buf.getDiscontinuities() ) );
	}

	@Test
	public void testFillLastDiscontinuity() throws Exception {
		buf.fillLast( new Discontinuity( 100 ) );

		assertArrayEquals( new long[] { -1, -1, 100 },
		                   buf.getTimes() );

		assertArrayEquals(
				new Object[][] {
						new Object[] { null, null, null },
						new Object[] { null, null, null }
				},
				buf.getData() );

		assertTrue( Arrays.equals( new boolean[] { false, false, true },
		                           buf.getDiscontinuities() ) );
	}

	private static Record r( long time ) {
		return new GenericRecord( time, (int) time, (double) time );
	}
}
