/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.model.data;

import junit.framework.TestCase;
import org.gillius.realdb.metadata.ElementCodec;
import org.gillius.realdb.metadata.ElementType;
import org.gillius.realdb.metadata.RecordElement;
import org.gillius.realdb.model.ElementCompressor;

import static org.easymock.EasyMock.*;

/**
 * TestRecordCompressor tests the {@link RecordCompressor} class.
 *
 * @author Jason Winnebeck
 */
public class TestRecordCompressor extends TestCase {
	private ElementCompressor[] compressors;
	private RecordCompressor    compressor;
	private Record[]            records;

	private static final int pivot = 1;

	@Override
	protected void setUp() throws Exception {
		super.setUp();

		compressors = new ElementCompressor[] {
				createMock( "comp1", ElementCompressor.class ),
				createMock( "comp2", ElementCompressor.class ),
				createMock( "comp3", ElementCompressor.class )
		};
		for ( ElementCompressor c : compressors ) {
			expect( c.getLookBehindDesired() ).andReturn( 1 );
			expect( c.getLookAheadDesired() ).andReturn( 1 );
		}

		ElementCodec[] codecs = new ElementCodec[] {
				createMock( ElementCodec.class ),
				createMock( ElementCodec.class ),
				createMock( ElementCodec.class )
		};
		RecordElement[] elements = new RecordElement[3];

		for ( int i = 0; i < elements.length; ++i ) {
			expect( codecs[i].getCompressor( ElementType.Float ) ).andReturn( compressors[i] );
			elements[i] = new RecordElement( i, "", codecs[i], ElementType.Float, false );
		}

		replay( (Object[]) compressors );
		replay( (Object[]) codecs );
		compressor = new RecordCompressor( elements );
		verify( (Object[]) compressors );
		verify( (Object[]) codecs );
		reset( (Object[]) compressors );
		records = new GenericRecord[] {
				new GenericRecord( 10, 1, 1.0f, 1.0 ),
				new GenericRecord( 20, 2, 2.0f, 2.0 ),
				new GenericRecord( 30, 3, 3.0f, 3.0 ),
				new GenericRecord( 40, 4, 4.0f, 4.0 ),
				new GenericRecord( 50, 5, 5.0f, 5.0 )
		};
	}

	public void testHandleRecord() {
		doTestHandleRecord();
		compressor.reset();
		//Check that we can run the same test again, after the reset
		reset( (Object[]) compressors );
		doTestHandleRecord();
	}

	private void doTestHandleRecord() {
		replay( (Object[]) compressors );
		assertNull( compressor.handleRecord( records[0] ) );

		//Even though all compressors report false, forceNext should make a record come out.
		setUpCompressors( new boolean[] { false, false, false, false },
		                  new long[] { -1, rt(0), rt(1) },
		                  new Object[][] {
				                  new Object[] { null, rv(0,0), rv(1,0) },
				                  new Object[] { null, rv(0,1), rv(1,1) },
				                  new Object[] { null, rv(0,2), rv(1,2) }
		                  },
		                  pivot);
		assertSame( records[0], compressor.handleRecord( records[1] ) );
		verify( (Object[]) compressors );

		//Since at least one compressor reports true, the record should come out
		setUpCompressors( new boolean[] { false, true, false, true },
		                  new long[] { rt(0), rt(1), rt(2) },
		                  new Object[][] {
				                  new Object[] { rv(0,0), rv(1,0), rv(2,0) },
				                  new Object[] { rv(0,1), rv(1,1), rv(2,1) },
				                  new Object[] { rv(0,2), rv(1,2), rv(2,2) }
		                  },
		                  pivot );
		assertSame( records[1], compressor.handleRecord( records[2] ) );
		verify( (Object[]) compressors );

		//Since they all report false, this record is dropped
		setUpCompressors( new boolean[] { false, false, false, false },
		                  new long[] { rt(1), rt(2), rt(3) },
		                  new Object[][] {
				                  new Object[] { rv(1,0), rv(2,0), rv(3,0) },
				                  new Object[] { rv(1,1), rv(2,1), rv(3,1) },
				                  new Object[] { rv(1,2), rv(2,2), rv(3,2) }
		                  },
		                  null );
		assertNull( compressor.handleRecord( records[3] ) );
		verify( (Object[]) compressors );

		//Now we need to check that the proper rotation takes place (r[1] stays where it is)
		setUpCompressors( new boolean[] { false, false, false, false },
		                  new long[] { rt(1), rt(3), rt(4) },
		                  new Object[][] {
				                  new Object[] { rv(1,0), rv(3,0), rv(4,0) },
				                  new Object[] { rv(1,1), rv(3,1), rv(4,1) },
				                  new Object[] { rv(1,2), rv(3,2), rv(4,2) }
		                  },
		                  null );
		assertNull( compressor.handleRecord( records[4] ) );
		verify( (Object[]) compressors );

		//Check the flushing now
		setUpCompressors( new boolean[] { false, false, true, false },
		                  new long[] { rt(1), rt(4), -1 },
		                  new Object[][] {
				                  new Object[] { rv(1,0), rv(4,0), null },
				                  new Object[] { rv(1,1), rv(4,1), null },
				                  new Object[] { rv(1,2), rv(4,2), null }
		                  },
		                  pivot );
		assertSame( records[4], compressor.flushRecords() );
		verify( (Object[]) compressors );

		reset( (Object[]) compressors );
		replay( (Object[]) compressors );
		assertNull( compressor.flushRecords() );
	}

	public void testNullToDiscontinuity() throws Exception {
		records = new Record[] {
				new GenericRecord( 10, null, null, null ),
				new Discontinuity( 20 ),
				new Discontinuity( 30 ),
				new Discontinuity( 40 ),
				new GenericRecord( 50, null, null, null )
		};

		replay( (Object[]) compressors );
		assertNull( compressor.handleRecord( records[0] ) );

		//Even though all compressors report false, forceNext should make a record come out.
		setUpCompressors( new boolean[] { false, false, false, false },
		                  new long[] { -1, rt(0), rt(1) },
		                  new Object[][] {
				                  new Object[] { null, rv(0,0), null },
				                  new Object[] { null, rv(0,1), null },
				                  new Object[] { null, rv(0,2), null }
		                  },
		                  pivot);
		assertSame( records[0], compressor.handleRecord( records[1] ) );
		verify( (Object[]) compressors );

		//Since we transition from known to discontinuity, we still expect a record and no handle calls
		//This is because the compressor short circuits with its own "true"
		setUpCompressors( null,
		                  new long[] { rt(0), rt(1), rt(2) },
		                  new Object[][] {
				                  new Object[] { rv(0,0), null, null },
				                  new Object[] { rv(0,1), null, null },
				                  new Object[] { rv(0,2), null, null }
		                  },
		                  pivot );
		assertSame( records[1], compressor.handleRecord( records[2] ) );
		verify( (Object[]) compressors );

		//Discontinuity to discontinuity -- now we ask the compressors (they say not to output
		setUpCompressors( new boolean[] { false, false, false, false },
		                  new long[] { rt(1), rt(2), rt(3) },
		                  new Object[][] {
				                  new Object[] { null, null, null },
				                  new Object[] { null, null, null },
				                  new Object[] { null, null, null }
		                  },
		                  null );
		assertNull( compressor.handleRecord( records[3] ) );
		verify( (Object[]) compressors );

		//Discontinuity to discontinuity -- still ask the compressors, but one says to output
		setUpCompressors( new boolean[] { false, true, false, false },
		                  new long[] { rt(1), rt(3), rt(4) },
		                  new Object[][] {
				                  new Object[] { null, null, rv(4,0) },
				                  new Object[] { null, null, rv(4,1) },
				                  new Object[] { null, null, rv(4,2) }
		                  },
		                  pivot );
		assertSame( records[3], compressor.handleRecord( records[4] ) );
		verify( (Object[]) compressors );

		//When we move back to known, then we don't call the handles
		//This is because the compressor short circuits with its own "true"
		setUpCompressors( null,
		                  new long[] { rt(1), rt(4), -1 },
		                  new Object[][] {
				                  new Object[] { null, rv(4,0), null },
				                  new Object[] { null, rv(4,1), null },
				                  new Object[] { null, rv(4,2), null }
		                  },
		                  pivot );
		assertSame( records[4], compressor.flushRecords() );
		verify( (Object[]) compressors );
	}

	private long rt( int i ) {
		return records[i].getTime();
	}

	private Object rv( int i, int v ) {
		return records[i].getElement( v );
	}

	private void setUpCompressors( boolean[] returns, long[] times, Object[][] data,
	                               Integer written ) {
		reset( (Object[]) compressors );

		for ( int i = 0; i < compressors.length; ++i ) {
			ElementCompressor c = compressors[i];
			if ( returns != null ) {
				expect( c.handleNewValue( aryEq( times ), aryEq( data[i] ), eq( pivot ) ) )
						.andReturn( returns[i] )
						.times( 0, 1 );
			}

			if ( written != null ) {
				c.valueWritten( times[ written ], data[i][written] );
			}
		}

		replay( (Object[]) compressors );
	}
}
