/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.model.pool;

import org.easymock.*;
import org.gillius.realdb.io.DataOutputWritable;
import org.gillius.realdb.model.CircularQueue;
import org.gillius.realdb.util.DBIterator;
import org.gillius.realdb.util.DBIteratorAdaptor;
import org.junit.*;

import java.util.List;

import static java.util.Arrays.*;
import static java.util.Collections.*;
import static org.easymock.EasyMock.*;
import static org.junit.Assert.*;

/**
 * TestTransactionLog tests the {@link TransactionLog} class.
 *
 * @author Jason Winnebeck
 */
public class TestTransactionLog {
	private IMocksControl ctl;
	private CircularQueue q;
	private DataOutputWritable dow;

	private List<TxEntry> tx;

	private TransactionLog log;

	@Before
	public void setUp() throws Exception {
		ctl = createControl();
		q = ctl.createMock( CircularQueue.class );
		dow = ctl.createMock( DataOutputWritable.class );
		ctl.replay();

		tx = emptyList();

		log = new TransactionLog( q, dow );
	}

	@Test
	public void testEmpty() throws Exception {
		ctl.reset();
		expect( q.iterator( isA( TxEntryReader.class ) ) ).andReturn( iter( tx ) );
		ctl.replay();

		log.loadOpenTransactions();

		ctl.verify();

		assertEquals( 0, log.getOpenTransactions().size() );
	}

	@Test
	public void testOpenAndFreeStatesOnLoad() throws Exception {
		ctl.reset();
		FreeStateTxEntry freeEntry = new FreeStateTxEntry( 1 );
		tx = asList( new FreeStateTxEntry( 0 ),
		             freeEntry,
		             new RemoveFreeTxEntry( 2 ) );
		expect( q.iterator( isA( TxEntryReader.class ) ) ).andReturn( iter( tx ) );
		ctl.replay();

		log.loadOpenTransactions();

		ctl.verify();

		assertEquals( singleton( 2 ), log.getOpenTransactions() );
	}

	@Test
	public void testOpensAndClosesOnLoad() throws Exception {
		ctl.reset();
		tx = asList( (TxEntry) new RemoveIndexTxEntry( 0, 0, 0, 0 ),
		             new AddIndexTxEntry( 0, 0, 0, 0 ) );
		expect( q.iterator( isA( TxEntryReader.class ) ) ).andReturn( iter( tx ) );
		ctl.replay();

		log.loadOpenTransactions();

		ctl.verify();

		assertEquals( 0, log.getOpenTransactions().size() );
	}

	@Test
	public void testAddTransactionWritesHeader() throws Exception {
		RemoveIndexTxEntry entry = new RemoveIndexTxEntry( 1, 2, 3, 4 );

		ctl.reset();
		q.add( entry, dow );
		ctl.replay();

		log.startTransaction( entry );
		ctl.verify();
	}

	@Test( expected = IllegalStateException.class )
	public void testAddOpenTx() throws Exception {
		RemoveFreeTxEntry entry = new RemoveFreeTxEntry( 1 );

		ctl.resetToNice();
		ctl.replay();

		log.startTransaction( entry );
		log.startTransaction( new RemoveIndexTxEntry( entry.getBlock(), 4, 5, 4 ) );
	}

	@Test( expected = IllegalStateException.class )
	public void testCloseUnopenedTx() throws Exception {
		log.endTransaction( new AddIndexTxEntry( 2, 4, 5, 6 ) );
	}

	@Test
	public void testOpenThenClose() throws Exception {
		RemoveIndexTxEntry open = new RemoveIndexTxEntry( 1, 2, 3, 4 );
		AddIndexTxEntry close = new AddIndexTxEntry( open.getBlock(), 3, 4, 5 );

		ctl.resetToStrict();
		q.add( open, dow );
		q.add( close, dow );
		ctl.replay();

		log.startTransaction( open );
		assertEquals( singleton( open.getBlock() ), log.getOpenTransactions() );
		log.endTransaction( close );
		assertEquals( 0, log.getOpenTransactions().size() );
		ctl.verify();
	}

	private static DBIterator<TxEntry> iter( Iterable<TxEntry> c ) {
		return new DBIteratorAdaptor<TxEntry>( c.iterator() );
	}
}
