/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.RealDBDemo;

import javax.swing.*;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.text.NumberFormat;

/**
 * ConfigPanel allows for the configuration of the various database parameters for the RealDB demo
 * application.
 *
 * @author Jason Winnebeck
 */
public class ConfigPanel extends JPanel {
	private JTextField         filePath;
	private SpinnerNumberModel blockSize;
	private SpinnerNumberModel fileSize;
	private SpinnerNumberModel maxStreams;
	private SpinnerNumberModel numStreams;
	private SpinnerNumberModel dataBlockSize;

	private JLabel totalSize;
	private NumberFormat sizeFormat;

	private JTextArea rdl;

	private static final String RDL_TEMPLATE =
			"SET blockSize     = %d\n" +
			"SET fileSize      = %s\n" +
			"SET maxStreams    = %d\n" +
			"SET dataBlockSize = %d\n";

	private static final String RDL_STREAM_TEMPLATE =
			"CREATE STREAM %s WITH ID %d {value double WITH CODEC SampledAlgorithm}\n";


	public ConfigPanel() {
		super( new GridBagLayout() );

		initModels();
		initView();
	}

	/**
	 * Creates a modal dialog against the window containing the given parent component and displays
	 * it. The values in the panel will have been modified after this method returns.
	 *
	 * @return true if the user clicked OK, false otherwise
	 */
	public boolean showDialog( Component parent, String title ) {
		int ret = JOptionPane.showConfirmDialog( parent,
		                                         this,
		                                         title,
		                                         JOptionPane.OK_CANCEL_OPTION,
		                                         JOptionPane.PLAIN_MESSAGE );

		return ret == JOptionPane.OK_OPTION;
	}

	public File getFile() {
		return new File( filePath.getText() );
	}

	public int getBlockSize() {
		return blockSize.getNumber().intValue();
	}

	public int getFileSize() {
		return fileSize.getNumber().intValue();
	}

	public int getMaxStreams() {
		return maxStreams.getNumber().intValue();
	}

	public int getNumStreams() {
		return numStreams.getNumber().intValue();
	}

	public int getDataBlockSize() {
		return dataBlockSize.getNumber().intValue();
	}

	public String getRDL() {
		StringBuilder ret = new StringBuilder();
		ret.append( String.format( RDL_TEMPLATE, getBlockSize(), getBlockSize() * getFileSize(),
		                           getMaxStreams(), getDataBlockSize() ) );

		for ( int i = 0; i < getNumStreams(); ++i ) {
			ret.append( String.format( RDL_STREAM_TEMPLATE, "Test" + i, i ) );
		}

		return ret.toString();
	}

	private void initModels() {
		filePath      = new JTextField( 40 );
		blockSize     = new SpinnerNumberModel(  512, 256, 32 * 1024, 1 );
		fileSize      = new SpinnerNumberModel( 1000,  16, 200000,    1 );
		maxStreams    = new SpinnerNumberModel(    3,   1,     10,    1 );
		numStreams    = new SpinnerNumberModel(    2,   1,     10,    1 );
		dataBlockSize = new SpinnerNumberModel(    2,   1,     16,    1 );

		totalSize = new JLabel( "", JLabel.RIGHT );
		sizeFormat = NumberFormat.getIntegerInstance();

		rdl = new JTextArea( getRDL() );
		rdl.setEditable( false );

		ChangeListener streamsConstraint = new ChangeListener() {
			public void stateChanged( ChangeEvent e ) {
				int max = getMaxStreams();

				if ( max < getNumStreams() )
					numStreams.setValue( max );
			}
		};

		maxStreams.addChangeListener( streamsConstraint );
		numStreams.addChangeListener( streamsConstraint );

		final ChangeListener rdlUpdater = new ChangeListener() {
			public void stateChanged( ChangeEvent e ) {
				rdl.setText( getRDL() );
			}
		};
		blockSize.addChangeListener( rdlUpdater );
		fileSize.addChangeListener( rdlUpdater );
		maxStreams.addChangeListener( rdlUpdater );
		numStreams.addChangeListener( rdlUpdater );
		dataBlockSize.addChangeListener( rdlUpdater );

		ChangeListener totalCalc = new ChangeListener() {
			public void stateChanged( ChangeEvent e ) {
				long total = getBlockSize() * getFileSize();
				totalSize.setText( sizeFormat.format( total ) );
			}
		};
		totalCalc.stateChanged( null );
		blockSize.addChangeListener( totalCalc );
		fileSize.addChangeListener( totalCalc );
	}

	private void initView() {
		Font monospaced = new Font( Font.MONOSPACED, Font.PLAIN, rdl.getFont().getSize() );
		rdl.setFont( monospaced );

		GridBagConstraints gbc = new GridBagConstraints();
		gbc.fill = GridBagConstraints.BOTH;
		gbc.insets = new Insets( 4, 4, 4, 4 );

		gbc.gridy = 0;
		add( new JLabel( "Database File", JLabel.RIGHT ), gbc );
		add( filePath, gbc );
		JButton browse = new JButton( "..." );
		browse.addActionListener( new ActionListener() {
			public void actionPerformed( ActionEvent e ) {
				chooseFile();
			}
		} );
		add( browse, gbc );

		gbc.gridy++;
		add( new JLabel( "Block Size (Bytes)", JLabel.RIGHT ), gbc );
		gbc.weightx = 1.0; //only need to do this once
		add( new JSpinner( blockSize ), gbc );
		gbc.weightx = 0.0;

		gbc.gridy++;
		add( new JLabel( "File Size (in blocks)", JLabel.RIGHT ), gbc );
		add( new JSpinner( fileSize ), gbc );

		gbc.gridy++;
		add( new JLabel( "File Size (bytes)", JLabel.RIGHT ), gbc );
		add( totalSize, gbc );

		gbc.gridy++;
		add( new JLabel( "Maximum Streams", JLabel.RIGHT ), gbc );
		add( new JSpinner( maxStreams ), gbc );

		gbc.gridy++;
		add( new JLabel( "Streams", JLabel.RIGHT ), gbc );
		add( new JSpinner( numStreams ), gbc );

		gbc.gridy++;
		add( new JLabel( "Data block Size (in blocks)", JLabel.RIGHT ), gbc );
		add( new JSpinner( dataBlockSize ), gbc );

		gbc.gridy++;
		gbc.weighty = 1.0;
		add( new JLabel( "RDL", JLabel.RIGHT ), gbc );
		add( new JScrollPane( rdl ), gbc );
	}

	private void chooseFile() {
		JFileChooser jfc = new JFileChooser( filePath.getText() );
		jfc.setFileSelectionMode( JFileChooser.FILES_ONLY );
		jfc.setDialogTitle( "Select Database File" );
		if ( jfc.showDialog( this, "Select DB" ) == JFileChooser.APPROVE_OPTION ) {
			File f = jfc.getSelectedFile();
			if ( f.exists() ) {
				int opt = JOptionPane.showConfirmDialog(
						this, "This will overwrite " + f.getName() + "; are you sure?",
						"Confirm DB File Selection", JOptionPane.YES_NO_OPTION );
				if ( opt == JOptionPane.YES_OPTION )
					filePath.setText( f.getAbsolutePath() );

			} else {
				filePath.setText( f.getAbsolutePath() );
			}
		}
	}

	public static void main( String[] args ) {
		ConfigPanel panel = new ConfigPanel();
		if ( panel.showDialog( null, "ConfigPanel Test" ) )
			System.out.println( "User clicked OK" );
		else
			System.out.println( "User did not click OK" );
	}
}
