/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.RealDBDemo;

import org.gillius.realdb.BadFileException;
import org.gillius.realdb.model.DataStream;
import org.gillius.realdb.model.Database;
import org.gillius.realdb.model.data.GenericRecord;
import org.gillius.realdb.model.data.Record;
import org.gillius.realdb.util.DBIterator;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.IOException;
import java.util.concurrent.CopyOnWriteArrayList;

import static org.gillius.RealDBDemo.DemoRunner.State.*;

/**
 * DemoRunner is the model in the Demo system and is observable through a
 * {@link PropertyChangeListener}. Generated events:
 * <table border="1">
 * <tr><th>Event</th><th>Type</th><th>Description</th></tr>
 * <tr><td>currIndex</td><td>Long</td><td></td></tr>
 * <tr><td>state</td><td>{@link State}</td><td></td></tr>
 * <tr><td>paused</td><td>Boolean</td><td></td></tr>
 * <tr><td>pausedOnTransition</td><td>Boolean</td><td></td></tr>
 * </table>
 *
 * @author Jason Winnebeck
 */
public class DemoRunner implements Runnable {

	public static enum State {
		Stopped, Running, ErrorWait, Validating, Repairing, Broken
	}

	private final DatabaseLoader loader;
	private final boolean        quiet;

	private Database db;
	private long     currIndex = 0;
	private State    state     = Stopped;

	private volatile boolean paused            = false;
	private volatile boolean pauseOnTransition = false;

	private Thread runThread = null;

	private final CopyOnWriteArrayList<PropertyChangeListener> listeners =
			new CopyOnWriteArrayList<PropertyChangeListener>();

	public DemoRunner( Database db, DatabaseLoader loader ) {
		this( db, loader, false );
	}

	public DemoRunner( Database db, DatabaseLoader loader, boolean quiet ) {
		this.db     = db;
		this.loader = loader;
		this.quiet  = quiet;
	}

	public void addListener( PropertyChangeListener listener ) {
		listeners.add( listener );
	}

	public void removeListener( PropertyChangeListener listener ) {
		listeners.remove( listener );
	}

	public void run() {
		synchronized ( this ) {
			//Don't allow re-entrant (even recursive) calls
			if ( runThread != null )
				throw new IllegalStateException( "DemoRunner already running" );

			runThread = Thread.currentThread();
		}
		setState( Running );

		while ( !Thread.interrupted() ) {
			synchronized ( this ) {
				while ( paused ) {
					try {
						wait();
					} catch ( InterruptedException e ) {
						Thread.currentThread().interrupt();
					}
				}
			}
			State old = state;
			switch ( state ) {
				case Running:
					running();
					break;

				case ErrorWait:
					try {
						errorWait();
					} catch ( InterruptedException e ) {
						Thread.currentThread().interrupt();
					}
					break;

				case Validating:
					validating();
					break;

				case Repairing:
					repairing();
					break;

				case Broken:
					return;

				default:
					throw new AssertionError( "Unknown state " + state );
			}

			if ( old != state && pauseOnTransition )
				setPaused( true );
		}

		synchronized ( this ) {
			runThread = null;
			Thread.interrupted(); //ensure interrupted flag is cleared as it is our stop signal
			setState( Stopped );
		}
	}

	public synchronized void stopRun() {
		if ( runThread != null )
			runThread.interrupt();
	}

	public long getCurrIndex() {
		return currIndex;
	}

	public State getState() {
		return state;
	}

	public boolean isPaused() {
		return paused;
	}

	public boolean isPauseOnTransition() {
		return pauseOnTransition;
	}

	public synchronized void setPaused( boolean paused ) {
		if ( paused != this.paused ) {
			boolean old = this.paused;
			this.paused = paused;
			fireEvent( "paused", old, paused );
			notifyAll();
		}
	}

	public void setPauseOnTransition( boolean pauseOnTransition ) {
		if ( pauseOnTransition != this.pauseOnTransition ) {
			boolean old = this.pauseOnTransition;
			this.pauseOnTransition = pauseOnTransition;
			fireEvent( "pausedOnTransition", old, pauseOnTransition );
		}
	}

	public double getReadsPerRecord() {
		return db.getProfilingStatistics().getReadBlocks() / (double) currIndex;
	}

	public double getWritesPerRecord() {
		return db.getProfilingStatistics().getWrittenBlocks() / (double) currIndex;
	}

	public Database getDatabase() {
		return db;
	}

	private void setDatabase( Database db ) {
		Database old = this.db;
		this.db = db;
		fireEvent( "database", old, db );
	}

	private void running() {
		try {
			writeNextRecord();

		} catch ( IOException e ) {
			writeError( "Database failed, waiting to reconnect" );
			if ( !quiet )
				e.printStackTrace();
			setState( ErrorWait );
			db.forceClose();
			setDatabase( null );
		} catch ( BadFileException e ) {
			writeError( "Database broken" );
			if ( !quiet )
				e.printStackTrace();
			setState( Broken );
		}
	}

	private void errorWait() throws InterruptedException {
		try {
			setDatabase( loader.reload() );
			setState( Validating );

		} catch ( IOException e ) {
			Thread.sleep( 1000 );

		} catch ( BadFileException e ) {
			writeError( "Database broken" );
			if ( !quiet )
				e.printStackTrace();
			setState( Broken );
		}
	}

	private void validating() {
		writeLog( "Validating the database" );

		writeLog( "Last index written: " + ( currIndex - 1 ) );

		for ( DataStream stream : db.getStreams() ) {
			writeLog( "For stream " + stream.getInfo().getName() + ":" );
			DBIterator<Record> records = stream.readRecords();
			int numRecords = 0;
			Record record;
			try {
				if ( records.hasNext() ) {
					record = records.next();
					++numRecords;
					long expected = record.getTime();
					writeLog( " - First index: " + expected );
					if ( expected != stream.getStartTime() ) {
						writeLog( " -- Does not match getStartTime: " + stream.getStartTime() );
						setState( Broken );
					}
					expected = validateRecord( record, expected, stream.getInfo().getUserId() );
					while ( records.hasNext() ) {
						expected = validateRecord( records.next(), expected, stream.getInfo().getUserId() );
						++numRecords;
					}
					writeLog( " - Last index: " + ( expected - 1 ) );
					if ( expected - 1 != stream.getEndTime() ) {
						writeLog( " -- Does not match getEndTime: " + stream.getEndTime() );
						setState( Broken );
					}
					writeLog( " - Total records read: " + numRecords );
				} else {
					writeLog( "Stream has no records" );
				}
			} catch ( Exception e ) {
				writeLog( "Error reading back the database" );
				e.printStackTrace( System.out );
			}
		}

		if ( state != Broken )
			setState( Repairing );
	}

	private void repairing() {
		writeLog( "Repairing the database, by writing records up to " + currIndex );

		for ( DataStream stream : db.getStreams() ) {
			for ( long t = stream.getEndTime() + 1; t < currIndex; ++t ) {
				try {
					final int sId = stream.getInfo().getUserId();
					stream.writeRecord( new GenericRecord( t, eval( t, sId ) ) );
				} catch ( IOException e ) {
					writeError( "Error while repairing; will resume later" );
					if ( !quiet )
						e.printStackTrace();
					setState( ErrorWait );
				} catch ( BadFileException e ) {
					writeError( "Corrupted database while repairing; cannot recover" );
					if ( !quiet )
						e.printStackTrace();
					setState( Broken );
				}
			}
		}

		setState( Running );
	}

	@SuppressWarnings( { "FloatingPointEquality" } )
	private long validateRecord( Record record, long expectedTime, int streamId ) {
		double value = (Double)record.getElement( 0 );
		double expected = eval( record.getTime(), streamId );

		if ( value != expected ) {
			writeError( " - Record %d does not have proper value", record.getTime() );
			setState( Broken );
		}
		if ( record.getTime() != expectedTime ) {
			writeError( " - Records jump from expected %d to %d%n", expectedTime, record.getTime() );
			setState( Broken );
			return record.getTime() + 1;
		} else {
			return expectedTime + 1;
		}
	}

	private void writeNextRecord() throws IOException, BadFileException {
		for ( DataStream stream : db.getStreams() ) {
			final int sId = stream.getInfo().getUserId();
			stream.writeRecord( new GenericRecord( currIndex, eval( currIndex, sId ) ) );
		}
		setCurrIndex( currIndex + 1 );
	}

	private static double eval( long index, int streamId ) {
		return index + streamId;
	}

	private void setCurrIndex( long currIndex ) {
		if ( this.currIndex != currIndex ) {
			long old = this.currIndex;
			this.currIndex = currIndex;
			fireEvent( "currIndex", old, currIndex );
		}
	}

	private void setState( State state ) {
		if ( this.state != state ) {
			State old = this.state;
			this.state = state;
			fireEvent( "state", old, state );
		}
	}

	private void fireEvent( String name, Object oldVal, Object newVal ) {
		PropertyChangeEvent ev = new PropertyChangeEvent( this, name, oldVal, newVal );
		for ( PropertyChangeListener listener : listeners ) {
			listener.propertyChange( ev );
		}
	}

	private void writeLog( String log ) {
		if ( !quiet )
			System.out.println( log );
	}

	private void writeError( String error ) {
		if ( !quiet )
			System.err.println( error );
	}

	private void writeError( String format, Object... args ) {
		if ( !quiet )
			System.err.format( format, args );
	}
}
