/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.RealDBDemo;

import javax.swing.*;
import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;

/**
 * DemoUI is the view and controller of the main demo, using DemoRunner as the model.
 *
 * @author Jason Winnebeck
 */
public class DemoUI extends JPanel {
	private final DemoRunner model;

	private final JLabel pausedLabel;
	private final JLabel stateLabel;
	private final JLabel currIndexLabel;
	private final JLabel readsPerRecordLabel;
	private final JLabel writesPerRecordLabel;

	private final JCheckBox potCheck;
	private final JButton   failureButton;
	private final JButton   resumeButton;

	public DemoUI( DemoRunner runner, final DemoDatabaseLoader loader ) {
		model = runner;

		pausedLabel    = new JLabel( model.isPaused() ? "Yes" : "No" );
		stateLabel     = new JLabel( model.getState().toString() );
		currIndexLabel = new JLabel( String.valueOf( model.getCurrIndex() ) );

		readsPerRecordLabel  = new JLabel( String.valueOf( model.getReadsPerRecord() ) );
		writesPerRecordLabel = new JLabel( String.valueOf( model.getWritesPerRecord() ) );

		potCheck       = new JCheckBox( "Auto-pause", model.isPauseOnTransition() );
		failureButton  = new JButton( "Induce failure" );
		resumeButton   = new JButton( "Unpause" );
		resumeButton.setEnabled( model.isPaused() );

		initComponents();

		runner.addListener( new PropertyChangeListener() {
			public void propertyChange( final PropertyChangeEvent evt ) {
				SwingUtilities.invokeLater( new Runnable() {
					public void run() {
						handlePropertyChange( evt );
					}
				} );
			}
		} );

		//Below is the "Controller" part
		potCheck.addItemListener( new ItemListener() {
			public void itemStateChanged( ItemEvent e ) {
				model.setPauseOnTransition( e.getStateChange() == ItemEvent.SELECTED );
			}
		} );
		failureButton.addActionListener( new ActionListener() {
			public void actionPerformed( ActionEvent e ) {
				loader.getBlockFile().setFailureFlag();
			}
		} );
		resumeButton.addActionListener( new ActionListener() {
			public void actionPerformed( ActionEvent e ) {
				model.setPaused( false );
			}
		} );
	}

	private void handlePropertyChange( PropertyChangeEvent evt ) {
		if ( "state".equals( evt.getPropertyName() ) ) {
			stateLabel.setText( evt.getNewValue().toString() );

		} else if ( "currIndex".equals( evt.getPropertyName() ) ) {
			long currIndex = (Long) evt.getNewValue();
			if ( currIndex % 1000 == 0 ) {
				currIndexLabel.setText( evt.getNewValue().toString() );
				updateStatistics();
			}

		} else if ( "pauseOnTransition".equals( evt.getPropertyName() ) ) {
			potCheck.setSelected( (Boolean) evt.getNewValue() );

		} else if ( "paused".equals( evt.getPropertyName() ) ) {
			boolean paused = (Boolean) evt.getNewValue();
			pausedLabel.setText( paused ? "Yes" : "No" );
			resumeButton.setEnabled( paused );
		}
	}

	private void updateStatistics() {
		readsPerRecordLabel.setText( String.valueOf( model.getReadsPerRecord() ) );
		writesPerRecordLabel.setText( String.valueOf( model.getWritesPerRecord() ) );
	}

	private void initComponents() {
		setLayout( new BorderLayout( 8, 8 ) );

		JPanel statusPanel = new JPanel( new GridBagLayout() );
		GridBagConstraints gbc = new GridBagConstraints();
		gbc.gridx = 0;
		gbc.insets = new Insets( 8, 8, 8, 8 );
		gbc.anchor = GridBagConstraints.EAST;
		gbc.fill = GridBagConstraints.BOTH;

		statusPanel.add( new JLabel( "Paused:"                  ), gbc );
		statusPanel.add( new JLabel( "Current state:"           ), gbc );
		statusPanel.add( new JLabel( "Current record index:"    ), gbc );
		statusPanel.add( new JLabel( "Write blocks per record:" ), gbc );
		statusPanel.add( new JLabel( "Read blocks per record:"  ), gbc );

		gbc.gridx = 1;
		statusPanel.add( pausedLabel,          gbc );
		statusPanel.add( stateLabel,           gbc );
		statusPanel.add( currIndexLabel,       gbc );
		statusPanel.add( readsPerRecordLabel,  gbc );
		statusPanel.add( writesPerRecordLabel, gbc );

		JPanel buttonPanel = new JPanel( new FlowLayout() );
		buttonPanel.add( potCheck );
		buttonPanel.add( failureButton );
		buttonPanel.add( resumeButton );

		add( statusPanel, BorderLayout.CENTER );
		add( buttonPanel, BorderLayout.SOUTH );
	}

	public void showFrame( String title, int defaultCloseAction ) {
		JFrame frame = new JFrame( title );
		frame.setContentPane( this );
		frame.setDefaultCloseOperation( defaultCloseAction );
		frame.pack();
		frame.setLocationByPlatform( true );
		frame.setVisible( true );
	}
}
