/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.RealDBDemo;

import javax.swing.*;
import java.awt.*;
import java.awt.image.BufferedImage;
import java.util.Arrays;

/**
 * RealTimeGraph is a component that can take real-time data and produce a graph with it.
 *
 * @author Jason Winnebeck
 */
public class RealTimeGraph extends JComponent {
	private final StreamInfo[] streams;
	private final BufferedImage buffer;
	private final int[] colBuf;
	private final int[] lineBuf;

	private final double scale;
	private final double translate;
	private final boolean drawTerminator;

	private int currSample;
	private static final int BG_CODE = 0x00FFFFFF;

	public RealTimeGraph( int sampleWindowSize, boolean drawTerminator,
	                      int height, double lower, double upper ) {
		this( new Color[] { Color.RED }, drawTerminator, sampleWindowSize, height, lower, upper );
	}

	/**
	 * Constructs a RealTimeGraph ready to display one or more real-time data streams.
	 *
	 * @param streamColors     non-null array of 1 or more Color for each data stream
	 * @param drawTerminator   if true, draws a black line at the end of the signal
	 * @param sampleWindowSize positive number of samples to show at one time (graph width)
	 * @param height           Height in pixels of the graph
	 * @param lower            lower limit of the graph has this y-axis value
	 * @param upper            upper limit of the graph has this y-axis value
	 */
	public RealTimeGraph( Color[] streamColors, boolean drawTerminator, int sampleWindowSize,
	                      int height,
	                      double lower, double upper ) {
		if ( streamColors == null )
			throw new IllegalArgumentException(
					"Parameter streamColors to RealTimeGraph.RealTimeGraph must not be null." );
		if ( streamColors.length <= 0 )
			throw new IllegalArgumentException(
					"Parameter streamColors must have at least one element in it." );
		if ( sampleWindowSize < 0 )
			throw new IllegalArgumentException(
					"Parameter sampleWindowSize to RealTimeGraph.RealTimeGraph must be positive or zero, but " + sampleWindowSize + " was given." );
		if ( height < 0 )
			throw new IllegalArgumentException(
					"Parameter height to RealTimeGraph.RealTimeGraph must be positive or zero, but " + height + " was given." );

		this.drawTerminator = drawTerminator;

		streams = new StreamInfo[ streamColors.length ];
		for ( int i = 0; i < streams.length; i++ )
			streams[i] = new StreamInfo( streamColors[i] );

		buffer = new BufferedImage( sampleWindowSize, height, BufferedImage.TYPE_INT_RGB );

		Graphics bufG = buffer.getGraphics();
		bufG.setColor( new Color( BG_CODE ) );
		bufG.fillRect( 0, 0, buffer.getWidth(), buffer.getHeight() );

		currSample = 0;
		scale = (double) height / ( lower - upper );
		translate = (double) height / ( -lower / upper + 1 );
		colBuf = new int[ height ];
		Arrays.fill( colBuf, BG_CODE );
		lineBuf = new int[ height ];
		Arrays.fill( lineBuf, 0 );

		final Dimension size = new Dimension( sampleWindowSize, height );
		setPreferredSize( size );
		setMaximumSize( size );
	}

	/**
	 * Adds a batch of samples on each data stream to the RealTimeGraph.
	 *
	 * @param samples non-null array of double of 1 or more non-null double[] of getNumStreams() length
	 *                that does not contain NaNs
	 */
	public void addSamples( double[][] samples ) {
		final int h = getBufferHeight();

		synchronized ( buffer ) {
			final int w = buffer.getWidth();
			int startRepaint = currSample;

			for ( double[] sample : samples ) {
				for ( int i = 0; i < streams.length; i++ ) {
					int pos = translateClamped( sample[i] );
					StreamInfo stream = streams[i];
					colBuf[pos] = stream.colorCode;
					stream.lastPos = pos;
				}

				buffer.setRGB( currSample, 0, 1, h, colBuf, 0, 1 );

				repaint( currSample, 0, 2, h );
				++currSample;
				if ( currSample >= w ) {
					repaint( startRepaint, 0, w - startRepaint, h );
					currSample = 0;
				}

				Arrays.fill( colBuf, BG_CODE );
			}

			if ( drawTerminator && currSample < w - 1 ) {
				buffer.setRGB( currSample + 1, 0, 1, h, lineBuf, 0, 1 );
				repaint( startRepaint, 0, w - startRepaint + 1, h );
			} else {
				repaint( startRepaint, 0, w - startRepaint, h );
			}
		}
	}

	/**
	 * Returns the number of data streams being displayed in this RTGraph.
	 */
	public int getNumStreams() {
		return streams.length;
	}

	/**
	 * Returns the height of the graph's buffer (logical size).
	 */
	public int getBufferHeight() {
		return colBuf.length;
	}

	protected void paintComponent( Graphics g ) {
		super.paintComponent( g );

		synchronized ( buffer ) {
			g.drawImage( buffer, 0, 0, null );
		}
	}

	private int translateClamped( double x ) {
		final int h = getBufferHeight();
		int ret = (int) ( x * scale + translate );

		if ( ret < 0 ) ret = 0;
		else if ( ret >= h ) ret = h - 1;

		return ret;
	}

	private static class StreamInfo {
		private final int colorCode;
		private int lastPos; //might eventually be used for line drawing

		public StreamInfo( Color color ) {
			colorCode = color.getRGB();
			lastPos = -1;
		}
	}
}
