/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.integration;

import org.gillius.RealDBDemo.DatabaseLoader;
import org.gillius.RealDBDemo.DemoRunner;
import org.gillius.realdb.BadFileException;
import org.gillius.realdb.io.BlockFile;
import org.gillius.realdb.io.ByteArrayBlockFile;
import org.gillius.realdb.io.ProfilingBlockFile;
import org.gillius.realdb.io.TriggeredFailingBlockFile;
import org.gillius.realdb.io.rdl.RDLFile;
import org.gillius.realdb.model.Database;
import org.gillius.realdb.model.DatabaseStatistics;
import org.gillius.realdb.tools.DatabaseBuilder;
import org.junit.*;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.IOException;
import java.util.Random;

import static org.junit.Assert.*;

/**
 * TestDemo runs the program in the {@link org.gillius.RealDBDemo} package and simulates faults at
 * various points to emulate what the user might do.
 *
 * @author Jason Winnebeck
 */
public class TestDemo {

	private static final String RDL =
			"SET blockSize     = 512\n" +
			"SET fileSize      = 512000\n" +
			"SET maxStreams    = 3\n" +
			"SET dataBlockSize = 2\n" +
			"CREATE STREAM Test0 WITH ID 0 {value double WITH CODEC SampledAlgorithm}\n" +
			"CREATE STREAM Test1 WITH ID 1 {value double WITH CODEC SampledAlgorithm}";

	//Below are fault induction indicies from a real run of the demo that failed.
//	private static final List<Long> faults =
//			asList( 6572L, 10540L, 14012L, 18166L, 25544L, 29822L, 32860L, 41292L );

	@Test
	public void runTest1() throws Exception {
		doTest( 2639, 19999, true );
	}

	@Test
	public void runTest2() throws Exception {
		doTest( 6000, 40000, true );
	}

	@Test
	public void runTest3() throws Exception {
		doTest( 2639, 40000, true );
	}

	@Test
	public void runTest4() throws Exception {
		doTest( 40001, 40000, true );
	}

	private static void doTest( final int failureStep, final int lastIndex, boolean quiet )
			throws BadFileException, IOException {
		Random rnd = new Random( 3149414L );

		RDLFile rdl = RDLFile.load( RDL );
		DatabaseBuilder builder = new DatabaseBuilder( rdl );
		BlockFile file;
		file = new ByteArrayBlockFile( builder.getBlockSize(),
		                               builder.getFileSize() / builder.getBlockSize() );
		new DatabaseBuilder( rdl ).createDatabase( file ).close();

		final MyDatabaseLoader loader = new MyDatabaseLoader( file, rnd );
		final DemoRunner runner = new DemoRunner( loader.reload(), loader, quiet );
		final DatabaseStatistics[] stats = { new DatabaseStatistics() };

		runner.addListener( new PropertyChangeListener() {
			public void propertyChange( PropertyChangeEvent evt ) {
				if ( "currIndex".equals( evt.getPropertyName() ) ) {
					long currIndex = (Long) evt.getNewValue();
					if ( currIndex > 0 && currIndex % failureStep == 0 ) {
						loader.getBlockFile().setFailureFlag();
					}
					if ( currIndex >= lastIndex ) {
						runner.stopRun();
					}

				} else if ( "database".equals( evt.getPropertyName() ) ) {
					Database db = (Database) evt.getOldValue();
					if ( db != null ) {
						stats[0] = stats[0].add( db.getProfilingStatistics() );
					}
				}
			}
		} );

		runner.run();
		loader.getBlockFile().close();

		assertEquals( DemoRunner.State.Stopped, runner.getState() );

		stats[0] = stats[0].add( runner.getDatabase().getProfilingStatistics() );

		System.out.println( stats[0] );
		System.out.println( "  Read blocks per record : " + stats[0].getReadBlocks() / (double) lastIndex );
		System.out.println( "  Write blocks per record: " + stats[0].getWrittenBlocks() / (double) lastIndex );
	}

	private static class MyDatabaseLoader implements DatabaseLoader {
		private TriggeredFailingBlockFile bf;
		private final BlockFile file;
		private final Random rnd;

		public MyDatabaseLoader( BlockFile file, Random rnd ) {
			this.file = file;
			this.rnd  = rnd;
		}

		public Database reload() throws IOException, BadFileException {
			bf = new TriggeredFailingBlockFile( rnd, file );
			return Database.load( new ProfilingBlockFile( bf ) );
		}

		public TriggeredFailingBlockFile getBlockFile() {
			return bf;
		}
	}
}
