/*
 * RealDB Project - Copyright (c) 2008-2010 by Jason Winnebeck
 *
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, see <http://www.gnu.org/licenses>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or combining it with JUnit (or a
 * modified version of that library), containing parts covered by the terms of Common Public
 * License, 1.0, the licensors of this Program grant you additional permission to convey the
 * resulting work.
 */
package org.gillius.realdb.concept;

import org.gillius.realdb.concept.data.BinaryVehicleDataParser;
import org.gillius.realdb.concept.data.DataHandlingException;
import org.gillius.realdb.concept.data.TextVehicleDataParser;
import org.gillius.realdb.concept.data.VehicleDataParser;
import org.gillius.realdb.io.rdl.RDLFile;
import org.gillius.realdb.model.Database;
import org.gillius.realdb.tools.DatabaseBuilder;

import java.io.BufferedInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.zip.GZIPInputStream;

/**
 * VehicleDataSimulation is the data processing component of the RealDB proof-of-concept. It
 * writes the vehicle data to a RealDB database as if it were an actual vehicle (writing the data
 * to all streams, in time order). The simulation runs as fast as it can; it doesn't run in
 * "real" time, since we could be simulating days, weeks, or months of data.
 * <p/>
 * VehicleDataSimulation supports 4 file types: .bin and .bin.gz - via
 * {@link BinaryVehicleDataParser} and .csv and .csv.gz via {@link TextVehicleDataParser}.
 *
 * @author Jason Winnebeck
 */
public class VehicleDataSimulation {
	public static void main( String[] args ) throws Exception {
		if ( args.length != 2 ) {
			System.err.println( "Program requires two arguments: binary data input file and database " +
			                    "file path" );
			System.exit( 1 );
		}

		System.out.println( "Running simulation on " + args[0] + " to create " + args[1] );

		Database db;

		File output = new File( args[1] );
		//read the RDL
		InputStream rdlData = VehicleDataSimulation.class.getResourceAsStream( "Definition.rdl" );
		ByteArrayOutputStream writer = new ByteArrayOutputStream( 5000 );
		IOUtils.copy( rdlData, writer );

		//create the DB
		RDLFile rdl = RDLFile.load( writer.toString( "UTF-8" ) );
		DatabaseBuilder builder = new DatabaseBuilder( rdl );

		db = builder.createDatabase( output );

		VehicleDataParser parser = getParser( new File( args[0] ), new ProgressReporter( "bytes" ) );

		runSimulation( parser, db );
		db.close();
		System.out.println( "Complete" );
	}

	public static void runSimulation( VehicleDataParser parser, Database db )
			throws IOException, DataHandlingException {
		RecordSinkWritingHandler writer = new RecordSinkWritingHandler( db );
		SimulationVehicleDataHandler handler = new SimulationVehicleDataHandler( writer );

		parser.setHandler( handler );
		parser.parse();
	}

	public static VehicleDataParser getParser( File f, ProgressReporter progressListener )
			throws IOException {
		String name = f.getName();
		progressListener.setTotal( f.length() );
		ProgressInputStream progress = new ProgressInputStream( new FileInputStream( f ),
		                                                        progressListener );

		BufferedInputStream bis = new BufferedInputStream( progress );
		if ( name.endsWith( ".bin" ) ) {
			return new BinaryVehicleDataParser( bis );

		} else if ( name.endsWith( ".bin.gz" ) ) {
			return new BinaryVehicleDataParser( new BufferedInputStream( new GZIPInputStream( progress ) ) );

		} else if ( name.endsWith( ".csv" ) ) {
			return new TextVehicleDataParser( new InputStreamReader( bis ), "," );

		} else if ( name.endsWith( ".csv.gz" ) ) {
			return new TextVehicleDataParser( new InputStreamReader( new GZIPInputStream( bis ) ), "," );

		} else {
			throw new IllegalArgumentException( "Unknown file type: " + f.getAbsolutePath() );
		}
	}
}
